//! this example demonstrates the type of four-core clocked inverter conceived
//! of in the https://uninsane.org/ blogpost.
//!
//! we're going for something like this:
//!
//! ```
//!          ____         ____         ____         ____
//!  + ___  /    \ +___  /    \ +___  /    \ +___  /    \  ___ +
//! IN ___ |o    o| ___ |x    o| ___ |x    o| ___ |x    o| ___ OUT
//!  -      \_o__/ -     \_o__/ -     \_o__/ -     \_o__/      -
//!           ||           ||           ||           ||
//!          +  -         +  -         +  -         +  -
//!          CTL0         CTL1         CTL2         CTL3
//! ```
//!
//! conventions:
//! - CW core polarization is logic '1'.
//! - differential signals are assumed to flow clockwise. i.e. left-to-right, with + on top and -
//!   on bottom.
//! - 'x' denotes a signal going "into" the page. 'o' for a signal going "out of" the page.
//! - the core driven by CTLn is named Sn, interchangeably used to represent the state (0 or 1) of
//!   that core
//!
//! hence:
//! - the input into core S0 has current flowing out of the page and tends to drive S0 to logic '0'.
//! - the input into S1, S2, S3 tends these cores to logic '1'.
//! - a positive current to CTLn will "clear" Sn to the logic '0' state.
//!
//! in this device, S0 is effectively an inverter, with S1, S2, S3 acting as buffers.
//! for the purpose of this simulation, all terminal wires are closed loops either explicitly
//! driven or measured.
//!
//! to run this, from toplevel directory:
//! ```
//! $ cargo run --release --bin multi_core_inverter
//! $ pushd crates/coremem; cargo run --release --bin viewer ../../out/applications/multi_core_inverter/0/ ; popd
//! ```
#![recursion_limit = "256"]

use coremem::geom::{Coord as _, Meters, Torus};
use coremem::mat::{Ferroxcube3R1MH, IsoConductorOr, IsomorphicConductor};
use coremem::meas;
#[allow(unused)]
use coremem::real::{self, Real as _};
use coremem::sim::spirv::{self, SpirvSim};
use coremem::sim::units::Seconds;
use coremem::stim::{CurlVectorField, Exp, Gated, ModulatedVectorField, Scaled, Shifted, TimeVaryingExt as _};
use coremem::Driver;

// type R = real::R32;
type R = f32;
type Mat = IsoConductorOr<R, Ferroxcube3R1MH>;
// type Backend = spirv::CpuBackend;
type Backend = spirv::WgpuBackend;
type Sim = SpirvSim::<R, Mat, Backend>;



enum DriveDirection {
    High,
    Low,
}
impl DriveDirection {
    fn as_signed_f32(&self) -> f32 {
        match self {
            DriveDirection::High => 1.0,
            DriveDirection::Low => -1.0,
        }
    }
}

/// different states each control signal can be in for some clock cycle.
#[derive(Copy, Clone)]
enum ClockState {
    HoldHigh,
    ReleaseHigh,
    HoldLow,
    ReleaseLow,
    Float,
}
type ClockSegment = Shifted<R, Gated<R, Scaled<Exp<R>, R>>>;

impl ClockState {
    fn time_stimulus(&self, params: &Params, cycle: u32)
        -> Option<ClockSegment>
    {
        use ClockState::*;
        match self {
            HoldHigh => Some(params.control_signal_hold(cycle, DriveDirection::High)),
            ReleaseHigh => Some(params.control_signal_release(cycle, DriveDirection::High)),
            HoldLow => Some(params.control_signal_hold(cycle, DriveDirection::Low)),
            ReleaseLow => Some(params.control_signal_release(cycle, DriveDirection::Low)),
            Float => None,
        }
    }
}

#[derive(Copy, Clone)]
struct Params {
    input_magnitude: f32,
    clock_phase_duration: f32,
    clock_decay: f32, // exp decay half-life
    ctl_conductivity: f32,
    coupling_conductivity: f32,
    // 's' = core (ferromagnetic part)
    s_major: f32,
    s_minor: f32,
    // 'io' = drive/control wire
    io_major: f32,
    io_minor: f32,
    coupling_major: f32,
    coupling_minor: f32,
    // coords for core 'n'
    sy: f32,
    sz: f32,
}
fn um(n: u32) -> f32 {
    n as f32 * 1e-6
}
fn ps(n: u32) -> f32 {
    n as f32 * 1e-12
}
impl Params {
    fn sx(&self, n: u32) -> f32 {
        um((n+1) * 400)
    }
    /// x-coord for the loop coupling core 'n' to core 'n+1'
    fn couplingx(&self, n: u32) -> f32 {
        self.sx(n) + um(200)
    }
    /// control loop for core n (alternately called "drive" loop)
    fn ctl(&self, n: u32) -> Torus {
        Torus::new_yz(Meters::new(self.sx(n), self.sy + self.s_major, self.sz), self.io_major, self.io_minor)
    }
    /// core '0' gets an external input in place of its coupling loop
    fn input0(&self) -> Torus {
        Torus::new_xz(Meters::new(self.sx(0) - self.s_major, self.sy, self.sz), self.io_major, self.io_minor)
    }
    /// the last core gets an external output in place of its coupling loop
    fn sense(&self, n: u32) -> Torus {
        Torus::new_xz(Meters::new(self.sx(n) + self.s_major, self.sy, self.sz), self.io_major, self.io_minor)
    }
    fn s(&self, n: u32) -> Torus {
        Torus::new_xy(Meters::new(self.sx(n), self.sy, self.sz), self.s_major, self.s_minor)
    }
    /// coupling(n) is the wire which couples core n into core n+1
    fn coupling(&self, n: u32) -> Torus {
        Torus::new_xz(Meters::new(self.couplingx(n), self.sy, self.sz), self.coupling_major, self.coupling_minor)
    }

    fn control_signal_hold(&self, cycle: u32, ty: DriveDirection) -> ClockSegment {
        // simple square wave:
        let amp = ty.as_signed_f32() * self.input_magnitude;
        let start = self.clock_phase_duration * cycle as f32;
        Exp::new(100f32.cast() /* very long decay */)
            .scaled(amp.cast())
            .gated(R::zero(), self.clock_phase_duration.cast())
            .shifted(start.cast())
    }

    fn control_signal_release(&self, cycle: u32, ty: DriveDirection) -> ClockSegment {
        // decaying exponential wave:
        let amp = ty.as_signed_f32() * self.input_magnitude;
        let start = self.clock_phase_duration * cycle as f32;
        Exp::new(self.clock_decay.cast() /* half life */)
            .scaled(amp.cast())
            .gated(R::zero(), self.clock_phase_duration.cast())
            .shifted(start.cast())
    }

    fn with_input_magnitude(mut self, p: f32) -> Self {
        self.input_magnitude = p;
        self
    }
    fn with_clock_phase_duration(mut self, p: f32) -> Self {
        self.clock_phase_duration = p;
        self
    }
    fn with_clock_decay(mut self, p: f32) -> Self {
        self.clock_decay = p;
        self
    }
    fn with_ctl_conductivity(mut self, p: f32) -> Self {
        self.ctl_conductivity = p;
        self
    }
    fn with_coupling_conductivity(mut self, p: f32) -> Self {
        self.coupling_conductivity = p;
        self
    }
}

/// 5 cores in sequence; analyze how they propagate a *specific* input signal
#[allow(unused)]
fn drive_map_5_core_inv(s0_init_state: ClockState, s0_rel_state: ClockState) -> [[ClockState; 5]; 14] {
    use ClockState::*;
    [
        // charge each device to '1'
        [s0_init_state,HoldHigh,     HoldHigh,     HoldHigh,     HoldHigh],
        // this is when we'd ordinarily open S0 for write
        [s0_rel_state, HoldHigh,     HoldHigh,     HoldHigh,     HoldHigh],

        // this is when S0 would ordinarily receive a write
        [Float,        HoldHigh,     HoldHigh,     HoldHigh,     HoldHigh],
        // open S1 for write
        [Float,        ReleaseHigh,  HoldHigh,     HoldHigh,     HoldHigh],

        // write S0' => S1
        [HoldLow,      Float,        HoldHigh,     HoldHigh,     HoldHigh],
        // open S2 for write
        [HoldLow,      Float,        ReleaseHigh,  HoldHigh,     HoldHigh],

        // write S1' => S2
        [HoldLow,      HoldLow,      Float,        HoldHigh,     HoldHigh],
        // open S3 for write
        [HoldLow,      HoldLow,      Float,        ReleaseHigh,  HoldHigh],

        // write S2' => S3; RESET S0
        [HoldHigh,     HoldLow,      HoldLow,      Float,        HoldHigh],
        // open S4 for write
        [HoldHigh,     HoldLow,      HoldLow,      Float,        ReleaseHigh],

        // write S3' => S4; RESET S1
        [HoldHigh,     HoldHigh,     HoldLow,      HoldLow,      Float],
        // open S0 for write
        [ReleaseHigh,  HoldHigh,     HoldLow,      HoldLow,      Float],

        // write S4' => output; RESET S2
        [Float,        HoldHigh,     HoldHigh,     HoldLow,      HoldLow],
        // open S1 for write
        [Float,        ReleaseHigh,  HoldHigh,     HoldLow,      HoldLow],
    ]
}

/// minimal 2-core inverter.
/// analyze how the inverter transfers a zero v.s. a one.
#[allow(unused)]
fn drive_map_isolated_inv() -> [[ClockState; 2]; 6] {
    use ClockState::*;
    [
        // charge each device to '1'
        [HoldHigh,   HoldHigh   ],
        // let the cores settle
        [ReleaseHigh,ReleaseHigh],
        // write S0 -> S1. S1 should be *cleared* to 0.
        [HoldLow,    Float      ],

        // charge S0=0, reset S1 for next write
        [HoldLow,    HoldHigh   ],
        // let the cores settle
        [ReleaseLow, ReleaseHigh],
        // write S0 -> S1. S1 should *keep its state* of 1.
        [HoldLow,    Float      ],
    ]
}

/// minimal 2-core inverter, but strictly square clock signals: no artificial decay
#[allow(unused)]
fn drive_map_isolated_inv_square() -> [[ClockState; 2]; 6] {
    use ClockState::*;
    [
        // charge each device to '1'
        [HoldHigh,   HoldHigh   ],
        // let the cores settle
        [Float,      Float      ],
        // write S0 -> S1. S1 should be *cleared* to 0.
        [HoldLow,    Float      ],

        // charge S0=0, reset S1 for next write
        [HoldLow,    HoldHigh   ],
        // let the cores settle
        [Float,      Float      ],
        // write S0 -> S1. S1 should *keep its state* of 1.
        [HoldLow,    Float      ],
    ]
}

fn main() {
    coremem::init_logging();
    // coremem::init_debug();

    let params = Params {
        input_magnitude: 0.0,
        clock_phase_duration: 0.0,
        clock_decay: 0.0,
        ctl_conductivity: 0.0,
        coupling_conductivity: 0.0,
        // 's' = core (ferromagnetic part)
        s_major: um(160),
        s_minor: um(30),
        // 'io' = drive/control wire
        io_major: um(80),
        io_minor: um(30),
        coupling_major: um(130),
        coupling_minor: um(30),
        // coords for core 'n'
        sy: um(400),
        sz: um(280),
    };
    // TODO: use a deque, with push_front and push_back
    let deferred = || {}; // add to this to schedule sims at a lower priority
    // let drive_map = drive_map_5_core_inv(ClockState::HoldLow, ClockState::ReleaseLow);
    // let drive_map = drive_map_isolated_inv();
    // let drive_map = drive_map_isolated_inv_square();
    // run_sim(
    //     "32-160ns-12000ps-2e7A",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(160000))
    //         .with_clock_decay(ps(12000))
    //         .with_input_magnitude(2.0e7)
    //         .with_ctl_conductivity(1e6)
    //         .with_coupling_conductivity(1e6)
    // );
    // run_sim(
    //     "33-160ns-12000ps-1e7A",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(160000))
    //         .with_clock_decay(ps(12000))
    //         .with_input_magnitude(1.0e7)
    //         .with_ctl_conductivity(1e6)
    //         .with_coupling_conductivity(1e6)
    // );
    // run_sim(
    //     "34-320ns-12000ps-1e7A",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(320000))
    //         .with_clock_decay(ps(12000))
    //         .with_input_magnitude(1.0e7)
    //         .with_ctl_conductivity(1e6)
    //         .with_coupling_conductivity(1e6)
    // );
    // run_sim(
    //     "35-320ns-24000ps-1e7A",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(320000))
    //         .with_clock_decay(ps(24000))
    //         .with_input_magnitude(1.0e7)
    //         .with_ctl_conductivity(1e6)
    //         .with_coupling_conductivity(1e6)
    // );
    // run_sim(
    //     "36-40ns-4000ps-1e7A-1e5cond",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(40000))
    //         .with_clock_decay(ps(4000))
    //         .with_input_magnitude(1.0e7)
    //         .with_ctl_conductivity(1e5)
    //         .with_coupling_conductivity(1e5)
    // );
    // run_sim(
    //     "37-40ns-4000ps-5e8A-1e5cond",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(40000))
    //         .with_clock_decay(ps(4000))
    //         .with_input_magnitude(5e8)
    //         .with_ctl_conductivity(1e5)
    //         .with_coupling_conductivity(1e5)
    // );
    // run_sim(
    //     "38-40ns-4000ps-2e8A-1e5cond",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(40000))
    //         .with_clock_decay(ps(4000))
    //         .with_input_magnitude(2e8)
    //         .with_ctl_conductivity(1e5)
    //         .with_coupling_conductivity(1e5)
    // );
    // run_sim(
    //     "39-40ns-4000ps-2e9A-1e4cond",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(40000))
    //         .with_clock_decay(ps(4000))
    //         .with_input_magnitude(2e9)
    //         .with_ctl_conductivity(1e4)
    //         .with_coupling_conductivity(1e4)
    // );
    // run_sim(
    //     "40-40ns-4000ps-2e10A-1e3cond",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(40000))
    //         .with_clock_decay(ps(4000))
    //         .with_input_magnitude(2e10)
    //         .with_ctl_conductivity(1e3)
    //         .with_coupling_conductivity(1e3)
    // );
    // run_sim(
    //     "41-40ns-4000ps-2e6A-1e7cond",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(40000))
    //         .with_clock_decay(ps(4000))
    //         .with_input_magnitude(2e6)
    //         .with_ctl_conductivity(1e7)
    //         .with_coupling_conductivity(1e7)
    // );
    // run_sim(
    //     "42-40ns-4000ps-2e5A-1e8cond",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(40000))
    //         .with_clock_decay(ps(4000))
    //         .with_input_magnitude(2e5)
    //         .with_ctl_conductivity(1e8)
    //         .with_coupling_conductivity(1e8)
    // );
    // run_sim(
    //     "43-40ns-4000ps-2e9A-1e4pctl-1e6pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(40000))
    //         .with_clock_decay(ps(4000))
    //         .with_input_magnitude(2e9)
    //         .with_ctl_conductivity(1e4)
    //         .with_coupling_conductivity(1e6)
    // );
    // run_sim(
    //     "44-40ns-4000ps-2e9A-1e6pctl-1e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(40000))
    //         .with_clock_decay(ps(4000))
    //         .with_input_magnitude(2e9)
    //         .with_ctl_conductivity(1e6)
    //         .with_coupling_conductivity(1e4)
    // );
    // run_sim(
    //     "45-40ns-4000ps-2e9A-1e4pctl-1e7pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(40000))
    //         .with_clock_decay(ps(4000))
    //         .with_input_magnitude(2e9)
    //         .with_ctl_conductivity(1e4)
    //         .with_coupling_conductivity(1e7)
    // );
    // run_sim(
    //     "46-80ns-4000ps-2e9A-1e4pctl-1e6pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(80000))
    //         .with_clock_decay(ps(4000))
    //         .with_input_magnitude(2e9)
    //         .with_ctl_conductivity(1e4)
    //         .with_coupling_conductivity(1e6)
    // );
    // run_sim(
    //     "47-80ns-4000ps-5e9A-1e4pctl-1e7pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(80000))
    //         .with_clock_decay(ps(4000))
    //         .with_input_magnitude(5e9)
    //         .with_ctl_conductivity(1e4)
    //         .with_coupling_conductivity(1e7)
    // );
    // run_sim(
    //     "48-40ns-4000ps-1e7A-1e6pctl-1e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(40000))
    //         .with_clock_decay(ps(4000))
    //         .with_input_magnitude(1e7)
    //         .with_ctl_conductivity(1e6)
    //         .with_coupling_conductivity(1e4)
    // );
    // run_sim(
    //     "49-80ns-4000ps-2e10A-1e3pctl-1e6pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(80000))
    //         .with_clock_decay(ps(4000))
    //         .with_input_magnitude(2e10)
    //         .with_ctl_conductivity(1e3)
    //         .with_coupling_conductivity(1e6)
    // );
    // run_sim(
    //     "50-80ns-4000ps-1e10A-1e3pctl-1e6pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(80000))
    //         .with_clock_decay(ps(4000))
    //         .with_input_magnitude(1e10)
    //         .with_ctl_conductivity(1e3)
    //         .with_coupling_conductivity(1e6)
    // );
    // run_sim(
    //     "51-80ns-4000ps-5e9A-1e4pctl-5e6pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(80000))
    //         .with_clock_decay(ps(4000))
    //         .with_input_magnitude(5e9)
    //         .with_ctl_conductivity(1e4)
    //         .with_coupling_conductivity(5e6)
    // );
    // run_sim(
    //     "52-80ns-4000ps-2e10A-1e3pctl-5e6pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(80000))
    //         .with_clock_decay(ps(4000))
    //         .with_input_magnitude(2e10)
    //         .with_ctl_conductivity(1e3)
    //         .with_coupling_conductivity(5e6)
    // );
    // run_sim(
    //     "53-80ns-4000ps-5e10A-1e3pctl-5e6pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(80000))
    //         .with_clock_decay(ps(4000))
    //         .with_input_magnitude(5e10)
    //         .with_ctl_conductivity(1e3)
    //         .with_coupling_conductivity(5e6)
    // );
    // run_sim(
    //     "54-320ns-4000ps-2e9A-1e4pctl-2e6pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(320000))
    //         .with_clock_decay(ps(4000))
    //         .with_input_magnitude(2e9)
    //         .with_ctl_conductivity(1e4)
    //         .with_coupling_conductivity(2e6)
    // );
    // run_sim(
    //     "55-80ns-4000ps-2e9A-1e4pctl-1e5pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(80000))
    //         .with_clock_decay(ps(4000))
    //         .with_input_magnitude(2e9)
    //         .with_ctl_conductivity(1e4)
    //         .with_coupling_conductivity(1e5)
    // );
    // run_sim(
    //     "56-80ns-4000ps-2e9A-1e4pctl-5e5pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(80000))
    //         .with_clock_decay(ps(4000))
    //         .with_input_magnitude(2e9)
    //         .with_ctl_conductivity(1e4)
    //         .with_coupling_conductivity(5e5)
    // );
    // run_sim(
    //     "57-320ns-4000ps-5e10A-1e3pctl-5e6pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(320000))
    //         .with_clock_decay(ps(4000))
    //         .with_input_magnitude(5e10)
    //         .with_ctl_conductivity(1e3)
    //         .with_coupling_conductivity(5e6)
    // );
    // run_sim(
    //     "58-20ns-1000ps-2e9A-1e4pctl-1e5pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(20000))
    //         .with_clock_decay(ps(1000))
    //         .with_input_magnitude(2e9)
    //         .with_ctl_conductivity(1e4)
    //         .with_coupling_conductivity(1e5)
    // );
    // run_sim(
    //     "59-20ns-500ps-2e9A-1e4pctl-1e5pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(20000))
    //         .with_clock_decay(ps(500))
    //         .with_input_magnitude(2e9)
    //         .with_ctl_conductivity(1e4)
    //         .with_coupling_conductivity(1e5)
    // );
    // run_sim(
    //     "60-20ns-1000ps-2e9A-5e3pctl-1e5pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(20000))
    //         .with_clock_decay(ps(1000))
    //         .with_input_magnitude(2e9)
    //         .with_ctl_conductivity(5e3)
    //         .with_coupling_conductivity(1e5)
    // );
    // run_sim(
    //     "61-20ns-1000ps-4e9A-5e3pctl-1e5pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(20000))
    //         .with_clock_decay(ps(1000))
    //         .with_input_magnitude(4e9)
    //         .with_ctl_conductivity(5e3)
    //         .with_coupling_conductivity(1e5)
    // );
    // run_sim(
    //     "62-20ns-500ps-2e9A-5e3pctl-5e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(20000))
    //         .with_clock_decay(ps(500))
    //         .with_input_magnitude(2e9)
    //         .with_ctl_conductivity(5e3)
    //         .with_coupling_conductivity(5e4)
    // );
    // run_sim(
    //     "63-10ns-400ps-4e9A-5e3pctl-1e5pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(10000))
    //         .with_clock_decay(ps(400))
    //         .with_input_magnitude(4e9)
    //         .with_ctl_conductivity(5e3)
    //         .with_coupling_conductivity(1e5)
    // );
    // run_sim(
    //     "64-10ns-400ps-4e9A-5e3pctl-5e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(10000))
    //         .with_clock_decay(ps(400))
    //         .with_input_magnitude(4e9)
    //         .with_ctl_conductivity(5e3)
    //         .with_coupling_conductivity(5e4)
    // );
    // run_sim(
    //     "65-10ns-400ps-4e9A-2e3pctl-2e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(10000))
    //         .with_clock_decay(ps(400))
    //         .with_input_magnitude(4e9)
    //         .with_ctl_conductivity(2e3)
    //         .with_coupling_conductivity(2e4)
    // );
    // run_sim(
    //     "66-10ns-200ps-4e9A-5e3pctl-5e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(10000))
    //         .with_clock_decay(ps(200))
    //         .with_input_magnitude(4e9)
    //         .with_ctl_conductivity(5e3)
    //         .with_coupling_conductivity(5e4)
    // );
    // run_sim(
    //     "67-5ns-200ps-4e9A-2e3pctl-2e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(5000))
    //         .with_clock_decay(ps(200))
    //         .with_input_magnitude(4e9)
    //         .with_ctl_conductivity(2e3)
    //         .with_coupling_conductivity(2e4)
    // );
    // run_sim(
    //     "68-5ns-200ps-8e9A-2e3pctl-2e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(5000))
    //         .with_clock_decay(ps(200))
    //         .with_input_magnitude(8e9)
    //         .with_ctl_conductivity(2e3)
    //         .with_coupling_conductivity(2e4)
    // );
    // run_sim(
    //     "69-2ns-80ps-2e10A-1e3pctl-1e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(2000))
    //         .with_clock_decay(ps(80))
    //         .with_input_magnitude(2e10)
    //         .with_ctl_conductivity(1e3)
    //         .with_coupling_conductivity(1e4)
    // );
    // run_sim(
    //     "70-2ns-80ps-1e10A-1e3pctl-1e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(2000))
    //         .with_clock_decay(ps(80))
    //         .with_input_magnitude(1e10)
    //         .with_ctl_conductivity(1e3)
    //         .with_coupling_conductivity(1e4)
    // );
    // run_sim(
    //     "71-2ns-100ps-5e9A-1e3pctl-1e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(2000))
    //         .with_clock_decay(ps(100))
    //         .with_input_magnitude(5e9)
    //         .with_ctl_conductivity(1e3)
    //         .with_coupling_conductivity(1e4)
    // );
    // run_sim(
    //     "72-160ns-8000ps-5e9A-1e3pctl-1e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(160000))
    //         .with_clock_decay(ps(8000))
    //         .with_input_magnitude(5e9)
    //         .with_ctl_conductivity(1e3)
    //         .with_coupling_conductivity(1e4)
    // );
    // run_sim(
    //     "73-160ns-8000ps-5e9A-1e3pctl-1e6pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(160000))
    //         .with_clock_decay(ps(8000))
    //         .with_input_magnitude(5e9)
    //         .with_ctl_conductivity(1e3)
    //         .with_coupling_conductivity(1e6)
    // );
    // run_sim(
    //     "74-20ns-1000ps-5e9A-1e3pctl-1e5pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(20000))
    //         .with_clock_decay(ps(1000))
    //         .with_input_magnitude(5e9)
    //         .with_ctl_conductivity(1e3)
    //         .with_coupling_conductivity(1e5)
    // );
    //
    // run_sim(
    //     "75-2ns-100ps-5e9A-1e3pctl-1e3pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(2000))
    //         .with_clock_decay(ps(100))
    //         .with_input_magnitude(5e9)
    //         .with_ctl_conductivity(1e3)
    //         .with_coupling_conductivity(1e3)
    // );
    // run_sim(
    //     "76-2ns-100ps-1e10A-1e3pctl-1e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(2000))
    //         .with_clock_decay(ps(100))
    //         .with_input_magnitude(1e10)
    //         .with_ctl_conductivity(1e3)
    //         .with_coupling_conductivity(1e4)
    // );
    // run_sim(
    //     "77-2ns-100ps-2e10A-1e3pctl-1e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(2000))
    //         .with_clock_decay(ps(100))
    //         .with_input_magnitude(2e10)
    //         .with_ctl_conductivity(1e3)
    //         .with_coupling_conductivity(1e4)
    // );
    // run_sim(
    //     "78-20ns-1000ps-1e10A-1e3pctl-1e5pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(20000))
    //         .with_clock_decay(ps(1000))
    //         .with_input_magnitude(1e10)
    //         .with_ctl_conductivity(1e3)
    //         .with_coupling_conductivity(1e5)
    // );
    // run_sim(
    //     "79-20ns-1000ps-2e10A-1e3pctl-1e5pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(20000))
    //         .with_clock_decay(ps(1000))
    //         .with_input_magnitude(2e10)
    //         .with_ctl_conductivity(1e3)
    //         .with_coupling_conductivity(1e5)
    // );
    // run_sim(
    //     "80-2ns-100ps-1e10A-1e3pctl-1e3pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(2000))
    //         .with_clock_decay(ps(100))
    //         .with_input_magnitude(1e10)
    //         .with_ctl_conductivity(1e3)
    //         .with_coupling_conductivity(1e3)
    // );
    // run_sim(
    //     "81-2ns-100ps-2e10A-1e3pctl-1e3pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(2000))
    //         .with_clock_decay(ps(100))
    //         .with_input_magnitude(2e10)
    //         .with_ctl_conductivity(1e3)
    //         .with_coupling_conductivity(1e3)
    // );
    // run_sim(
    //     "82-2ns-100ps-5e10A-1e3pctl-1e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(2000))
    //         .with_clock_decay(ps(100))
    //         .with_input_magnitude(5e10)
    //         .with_ctl_conductivity(1e3)
    //         .with_coupling_conductivity(1e4)
    // );
    // run_sim(
    //     "83-2ns-100ps-1e9A-1e4pctl-1e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(2000))
    //         .with_clock_decay(ps(100))
    //         .with_input_magnitude(1e9)
    //         .with_ctl_conductivity(1e4)
    //         .with_coupling_conductivity(1e4)
    // );
    // run_sim(
    //     "84-2ns-100ps-2e11A-1e2pctl-1e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(2000))
    //         .with_clock_decay(ps(100))
    //         .with_input_magnitude(2e11)
    //         .with_ctl_conductivity(1e2)
    //         .with_coupling_conductivity(1e4)
    // );
    // run_sim(
    //     "85-2ns-100ps-1e11A-1e2pctl-1e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(2000))
    //         .with_clock_decay(ps(100))
    //         .with_input_magnitude(1e11)
    //         .with_ctl_conductivity(1e2)
    //         .with_coupling_conductivity(1e4)
    // );
    // run_sim(
    //     "86-2ns-100ps-5e10A-1e2pctl-1e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(2000))
    //         .with_clock_decay(ps(100))
    //         .with_input_magnitude(5e10)
    //         .with_ctl_conductivity(1e2)
    //         .with_coupling_conductivity(1e4)
    // );
    // run_sim(
    //     "87-2ns-100ps-5e10A-1e2pctl-1e3pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(2000))
    //         .with_clock_decay(ps(100))
    //         .with_input_magnitude(5e10)
    //         .with_ctl_conductivity(1e2)
    //         .with_coupling_conductivity(1e3)
    // );
    // run_sim(
    //     "88-2ns-100ps-1e11A-1e3pctl-1e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(2000))
    //         .with_clock_decay(ps(100))
    //         .with_input_magnitude(1e11)
    //         .with_ctl_conductivity(1e3)
    //         .with_coupling_conductivity(1e4)
    // );
    // run_sim(
    //     "89-2ns-100ps-2e11A-1e3pctl-1e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(2000))
    //         .with_clock_decay(ps(100))
    //         .with_input_magnitude(2e11)
    //         .with_ctl_conductivity(1e3)
    //         .with_coupling_conductivity(1e4)
    // );
    // run_sim(
    //     "90-2ns-100ps-5e11A-1e3pctl-1e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(2000))
    //         .with_clock_decay(ps(100))
    //         .with_input_magnitude(5e11)
    //         .with_ctl_conductivity(1e3)
    //         .with_coupling_conductivity(1e4)
    // );
    // run_sim(
    //     "91-2ns-100ps-1e12A-1e3pctl-1e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(2000))
    //         .with_clock_decay(ps(100))
    //         .with_input_magnitude(1e12)
    //         .with_ctl_conductivity(1e3)
    //         .with_coupling_conductivity(1e4)
    // );
    // run_sim(
    //     "92-2ns-100ps-2e12A-1e3pctl-1e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(2000))
    //         .with_clock_decay(ps(100))
    //         .with_input_magnitude(2e12)
    //         .with_ctl_conductivity(1e3)
    //         .with_coupling_conductivity(1e4)
    // );
    // run_sim(
    //     "93-2ns-100ps-5e12A-1e3pctl-1e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(2000))
    //         .with_clock_decay(ps(100))
    //         .with_input_magnitude(5e12)
    //         .with_ctl_conductivity(1e3)
    //         .with_coupling_conductivity(1e4)
    // );
    // run_sim(
    //     "94-5ns-100ps-2e11A-1e2pctl-1e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(5000))
    //         .with_clock_decay(ps(200))
    //         .with_input_magnitude(2e11)
    //         .with_ctl_conductivity(1e2)
    //         .with_coupling_conductivity(1e4)
    // );
    // run_sim(
    //     "95-5ns-100ps-4e11A-1e2pctl-1e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(5000))
    //         .with_clock_decay(ps(200))
    //         .with_input_magnitude(4e11)
    //         .with_ctl_conductivity(1e2)
    //         .with_coupling_conductivity(1e4)
    // );
    // run_sim(
    //     "96-5ns-100ps-8e11A-1e2pctl-1e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(5000))
    //         .with_clock_decay(ps(200))
    //         .with_input_magnitude(8e11)
    //         .with_ctl_conductivity(1e2)
    //         .with_coupling_conductivity(1e4)
    // );
    // run_sim(
    //     "97-5ns-100ps-16e11A-1e2pctl-1e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(5000))
    //         .with_clock_decay(ps(200))
    //         .with_input_magnitude(16e11)
    //         .with_ctl_conductivity(1e2)
    //         .with_coupling_conductivity(1e4)
    // );
    // run_sim(
    //     "98-5ns-100ps-6e11A-1e2pctl-1e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(5000))
    //         .with_clock_decay(ps(200))
    //         .with_input_magnitude(6e11)
    //         .with_ctl_conductivity(1e2)
    //         .with_coupling_conductivity(1e4)
    // );
    // run_sim(
    //     "99-5ns-100ps-7e11A-1e2pctl-1e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(5000))
    //         .with_clock_decay(ps(200))
    //         .with_input_magnitude(7e11)
    //         .with_ctl_conductivity(1e2)
    //         .with_coupling_conductivity(1e4)
    // );
    // run_sim(
    //     "100-5ns-100ps-6.5e11A-1e2pctl-1e4pcpl",
    //     drive_map_isolated_inv(),
    //     params
    //         .with_clock_phase_duration(ps(5000))
    //         .with_clock_decay(ps(200))
    //         .with_input_magnitude(6.5e11)
    //         .with_ctl_conductivity(1e2)
    //         .with_coupling_conductivity(1e4)
    // );
    run_sim(
        "101-2ns-100ps-4e11A-1e3pctl-1e4pcpl",
        drive_map_isolated_inv(),
        params
            .with_clock_phase_duration(ps(2000))
            .with_clock_decay(ps(100))
            .with_input_magnitude(4e11)
            .with_ctl_conductivity(1e3)
            .with_coupling_conductivity(1e4)
    );

    deferred();
}


fn run_sim<const C: usize, const R: usize>(
    name: &str, drive_map: [[ClockState; C]; R], params: Params
) {
    // let ns = |n| n as f32 * 1e-9;
    let feat_size = um(10);

    let sim_bounds = |num_cores| Meters::new(params.sx(num_cores), params.sy * 2.0, params.sz * 2.0);
    let sim_padding = Meters::new(um(80), um(80), um(80));

    //////// define the control signals/transitions
    // each row N denotes the drive currents at clock cycle N.
    // each col M denotes the drive current at core M.
    let num_cycles = drive_map.len() as u32;
    let num_cores = drive_map[0].len() as u32;
    let mut core_drivers = vec![Vec::default(); num_cores as usize];
    for (cycle, cores) in drive_map.into_iter().enumerate() {
        for (core, clock) in cores.into_iter().enumerate() {
            core_drivers[core as usize].extend(clock.time_stimulus(&params, cycle as u32));
        }
    }
    let stim: Vec<_> = core_drivers.into_iter()
        .enumerate()
        .map(|(core, time_varying)| {
            let region = params.ctl(core as u32);
            let area = region.cross_section();
            let amp = 1.0 / area;
            let v_field = CurlVectorField::new(region.clone());
            ModulatedVectorField::new(v_field, time_varying.scaled(amp.cast::<R>()))
        })
        .collect();
    assert_eq!(stim.len(), num_cores as usize);


    let ctl_mat = IsomorphicConductor::new(params.ctl_conductivity.cast::<R>());
    let coupling_mat = IsomorphicConductor::new(params.coupling_conductivity.cast::<R>());
    let ferro_mat = Ferroxcube3R1MH::new();

    let last_core = num_cores - 1;

    let mut driver = Driver::new(Sim::new(
        sim_bounds(num_cores).to_index(feat_size), feat_size,
    ));
    driver.add_classical_boundary(sim_padding);

    //////// create the wires and toroids
    driver.fill_region(&params.input0(), coupling_mat);
    driver.fill_region(&params.sense(last_core), coupling_mat);
    for core in 0..num_cores {
        driver.fill_region(&params.s(core), ferro_mat);
        driver.fill_region(&params.ctl(core), ctl_mat);
        if core != last_core {
            driver.fill_region(&params.coupling(core), coupling_mat);
        }
    }

    //////// monitor some measurements
    // driver.add_measurement(meas::CurrentLoop::new("input0", input0.clone()));
    for core in 0..num_cores {
        driver.add_measurement(meas::CurrentLoop::new(
            &format!("drive{}", core),
            params.ctl(core),
        ));
    }
    for core in 0..num_cores {
        let name = format!("sense{}", core);
        if core != last_core {
            driver.add_measurement(meas::CurrentLoop::new(
                &name, params.coupling(core),
            ));
        } else {
            driver.add_measurement(meas::CurrentLoop::new(
                &name, params.sense(core),
            ));
        }
    }
    for core in 0..num_cores {
        driver.add_measurement(meas::MagneticLoop::new(
            &format!("state{}", core),
            params.s(core),
        ));
    }


    let duration = Seconds(params.clock_phase_duration * num_cycles as f32);

    // let stim = DynStimuli::from_vec(stim.map(MapIntoBoxStimulus).into_vec());
    let mut driver = driver.with_modulated_stimulus();
    for s in stim {
        driver.add_stimulus(s);
    }

    let prefix = format!("out/applications/multi_core_inverter/{}/", name);
    let _ = std::fs::create_dir_all(&prefix);
    driver.add_state_file(&*format!("{}state.bc", prefix), 25600);
    // driver.add_serializer_renderer(&*format!("{}frame-", prefix), 6400, None);
    // driver.add_csv_renderer(&*format!("{}meas-detailed.csv", prefix), 100, None);
    driver.add_csv_renderer(&*format!("{}meas.csv", prefix), 1600, None);
    driver.add_csv_renderer(&*format!("{}meas-sparse.csv", prefix), 12800, None);
    driver.set_steps_per_stimulus(200);

    driver.step_until(duration);
}
