import plotly.express as px
from pandas import DataFrame

from math import floor

def extrema(arr):
    arr = list(arr)
    return min(arr), max(arr)

class Line:
    def __init__(self, from_, to):
        self.from_ = from_
        self.to = to

    def slope(self) -> float:
        return (self.to[1] - self.from_[1]) / (self.to[0] - self.from_[0])

    def get(self, x: float) -> float:
        from_x, from_y = self.from_
        to_x, to_y = self.to
        if x == from_x:
            return from_y
        tween = (x - from_x) / (to_x - from_x)
        return tween * to_y + (1-tween) * from_y


class Piecewise:
    def __init__(self, xy: list):
        """ xy is a list of (x, y) pairs """
        self.xy = list(xy)

    @property
    def num_pieces(self) -> int:
        return len(self.xy) - 1

    def normalized(self, prev_max: float) -> 'Piecewise':
        """ map every coordinate from [-prev_max, prev_max] to [0, 1] """
        p = prev_max
        r = 2*prev_max
        s = 1.0/r
        return Piecewise([
            (s*(x + p), s*(y + p)) for (x, y) in self.xy
        ])

    def inverted_y(self) -> 'Piecewise':
        return Piecewise([
            (x, -y) for (x, y) in self.xy
        ])

    def inverted_x(self) -> 'Piecewise':
        return Piecewise([
            (-x, y) for (x, y) in self.xy
        ][::-1])

    def inverted_xy(self) -> 'Piecewise':
        return self.inverted_x().inverted_y()

    def logically_inverted(self) -> 'Piecewise':
        """ return a Piecewise that evaluates to 1-y """
        return self.inverted_y().shifted_y(1.0)

    def logically_inverted_x(self) -> 'Piecewise':
        return self.inverted_x().shifted_x(1.0)

    def logically_inverted_xy(self) -> 'Piecewise':
        return self.logically_inverted_x().logically_inverted()

    def shifted_x(self, shift: float) -> 'Piecewise':
        return Piecewise([
            (x + shift, y) for (x, y) in self.xy
        ])

    def shifted_y(self, shift: float) -> 'Piecewise':
        return Piecewise([
            (x, y + shift) for (x, y) in self.xy
        ])

    def scaled_y(self, scale: float) -> 'Piecewise':
        return Piecewise([
            (x, y * scale) for (x, y) in self.xy
        ])

    def clipped(self, min_y: float = 0.00, max_y: float = 1.0) -> 'Piecewise':
        return Piecewise([
            (x, min(max_y, max(min_y, y))) for (x, y) in self.xy
        ])

    def flat_extrapolation(self) -> 'Piecewise':
        """ make it so f(x) for x OOB returns the nearest in-bounds y """
        first = self.xy[0]
        last = self.xy[-1]
        new_first = (first[0] - 1.0, first[1])
        new_last = (last[0] + 1.0, last[1])
        return Piecewise([new_first] + self.xy + [new_last])

    def cascaded(self, second: 'Piecewise') -> 'Piecewise':
        """ return a function equivalent to y = second(self(x)) """
        b_min, b_max = extrema(x for (x, y) in self.xy)
        bounds = [ int(floor(b_min * 100)), 1 + int(floor(b_max * 100)) ]
        return Piecewise([
            (0.01 * x, second.get(self.get(0.01 * x))) for x in range(*bounds)
        ])

    def line_for(self, x: float) -> Line:
        for first_lower in self.xy[:-1][::-1]:
            if first_lower[0] < x: break
        for first_upper in self.xy[1:]:
            if not first_upper[0] < x: break

        return Line(first_lower, first_upper)

    def is_viable_inverter(self) -> bool:
        res = 1000
        for i in range(res+1):
            x = i/res
            n = (i+1)/res
            if self.get(self.get(x)) < x and self.get(self.get(n)) > n:
                return True

        return False

    def get(self, x: float) -> float:
        """
        evaluate the piecewise function at the provided x value.
        OOB points are just extrapolated from the nearest piece.
        """
        return self.line_for(x).get(x)

    def get_slope(self, x: float) -> float:
        return self.line_for(x).slope()

    def get_repeated(self, x: float, n: int = 255) -> float:
        for _ in range(n):
            x = self.get(x)
        return x

    def get_mean(self, x: float, n: int = 100) -> float:
        return 1/n * sum(self.get(x * i/(n-1)) for i in range(n))

    def get_integral(self, x: float, n: int = 100) -> float:
        return self.get_mean(x) * x

    def get_range(self, xmin: float = 0.0, xmax: float = 1.0) -> float:
        return abs(self.get(xmax) - self.get(xmin))

    def df_for(self, from_: float, to: float, points: int, f) -> DataFrame:
        x_step = (to - from_) / (points - 1)
        x = [from_ + x_step*x for x in range(points)]
        y = [f(xi) for xi in x]
        return DataFrame(data=dict(x=x, y=y))

    def df(self, from_: float = 0.0, to: float = 1.0, points: int = 101) -> DataFrame:
        return self.df_for(from_, to, points, self.get)

    def slope_df(self, from_: float = 0.0, to: float = 1.0, points: int = 101) -> DataFrame:
        return self.df_for(from_, to, points, self.get_slope)

    def min_max_slope(self):
        slope = [self.get_slope(0.01*x) for x in range(101)]
        return min(slope), max(slope)

    def max_abs_slope(self) -> float:
        return max(abs(s) for s in self.min_max_slope())

    def plot_for(self, from_: float, to: float, title: str, f):
        df = self.df_for(from_, to, points=101, f=f)
        fig = px.line(df, x="x", y="y", title=title)
        fig.show()

    def plot(self, from_: float = 0.0, to: float = 1.0, title: str = "Piecewise"):
        self.plot_for(from_, to, title, self.get)

    def plot_slope(self, from_: float = 0.0, to: float = 1.0, title: str = "Piecewise"):
        self.plot_for(from_, to, title, self.get_slope)

    def plot_equilibrium(self, from_: float = 0.0, to: float = 1.0, title: str = "Piecewise"):
        self.plot_for(from_, to, title, self.get_repeated)

    def plot_integral(self, from_: float = 0.0, to: float = 1.0, title: str = "Piecewise"):
        self.plot_for(from_, to, title, self.get_integral)

