#!/usr/bin/env python3
"""
invoke with the path to a meas.csv file for the stacked_core 39-xx demos
to extract higher-level info from them.
"""
import os
import sys
import re

from stacked_cores import load_csv, labeled_rows, last_row_before_t, extract_m

def extract_one(path: str, t_first: float, t_last: float, t_mid: float = None):
    header, raw_rows = load_csv(path)
    rows = labeled_rows(header, raw_rows)

    tx_init = last_row_before_t(rows, t_first)
    tx_fini = last_row_before_t(rows, t_last)
    tx_mid = last_row_before_t(rows, t_mid) if t_mid is not None else None

    if tx_fini and float(tx_fini["time"]) < 0.95 * t_last:
        tx_fini = None

    m_init = extract_m(tx_init) if tx_init is not None else [None]
    m_fini = extract_m(tx_fini) if tx_fini is not None else [None]
    m_mid = extract_m(tx_mid)[1:-1] if tx_mid is not None else []

    return m_init[0], m_fini[-1], m_mid

def extract_polarity(stem: str) -> float:
    s = None
    if re.search("-p\d\d\d\d?", stem):
        s = re.search("-p\d\d\d\d?", stem).group(0)
    if re.search("-n\d\d\d\d?", stem):
        s = re.search("-n\d\d\d\d?", stem).group(0)

    if s:
        sign = {'n': -1, 'p': 1}[s[1]]
        mag = int(s[2:])
        max_mag = 10**(len(s[2:]) - 1)
        return sign * mag / max_mag

    if "-000" in stem:
        return 0.00

def extract_39xx(base_path: str, t_first: str = "2e-9", t_last: str = "3e-9", t_mid: str = None):
    t_first = float(t_first)
    t_last = float(t_last)
    t_mid = float(t_mid) if t_mid is not None else None

    base_dir, prefix = os.path.split(base_path)
    mappings = {}
    for entry in os.listdir(base_dir):
        if entry.startswith(prefix):
            (input_, output, mid) = extract_one(os.path.join(base_dir, entry, "meas.csv"), t_first, t_last, t_mid)
            polarity = extract_polarity(entry)
            if input_ is not None and output is not None:
                mappings[int(round(input_))] = (int(round(output)), polarity, [round(m) for m in mid])

    if mappings:
        print("Piecewise(")
        print("    [")
        for i, (o, polarity, mid) in sorted(mappings.items()):
            comments = []
            if polarity is not None:
                comments += [f"{polarity:= 05.3f}"]
            for core, val in enumerate(mid):
                comments += [f"M{core+1}={val:5}"]

            comment = "  # " + ", ".join(comments) if comments else ""
            print(f"        [ {i:6}, {o:6} ],{comment}")
        print("    ]")
        print(")")


if __name__ == '__main__':
    extract_39xx(*sys.argv[1:])
