use std::sync::{Arc, Mutex};
use std::time::{Duration, Instant};

/// this is just a big dumb bag of perf-related metrics,
/// gathered with the intention of identifying areas for optimization
pub struct Diagnostics {
    frames_completed: u64,
    /// all known time spent in driver code, as measured from the toplevel
    time_in_driver: Duration,
    /// time during which driver had passed control to stimulus
    time_sim_step: Duration,
    time_prepping_stim: Duration,
    time_on_stimuli: Duration,
    /// time during which driver was preparing to render (e.g. cloning state)
    time_prepping_render: Duration,
    time_rendering: Duration,
    /// time during which driver was waiting for an async stimulus job
    time_blocked_on_stim: Duration,
    /// time during which driver was waiting for an async render job
    time_blocked_on_render: Duration,
    /// time during which CPU was waiting for GPU data
    time_reading_device: Duration,
    /// time during which CPU was transferring data to GPU
    time_writing_device: Duration,
    /// time during which the GPU was actively computing steps
    time_stepping_device: Duration,
    start_time: Instant,
}

#[derive(Clone, Default)]
pub struct SyncDiagnostics(Arc<Mutex<Diagnostics>>);

impl Default for Diagnostics {
    fn default() -> Self {
        Self::new()
    }
}

impl Diagnostics {
    pub fn new() -> Self {
        Self {
            frames_completed: 0,
            time_in_driver: Default::default(),
            time_sim_step: Default::default(),
            time_prepping_stim: Default::default(),
            time_on_stimuli: Default::default(),
            time_prepping_render: Default::default(),
            time_rendering: Default::default(),
            time_blocked_on_stim: Default::default(),
            time_blocked_on_render: Default::default(),
            time_reading_device: Default::default(),
            time_writing_device: Default::default(),
            time_stepping_device: Default::default(),
            start_time: Instant::now(),
        }
    }

    pub fn format(&self) -> String {
        let overall_time = self.start_time.elapsed().as_secs_f64();
        let driver_time = self.time_in_driver.as_secs_f64();
        let fps = (self.frames_completed as f64) / overall_time;
        let fps_line = format!("fps: {:6.2}", fps);

        let step_time = self.time_sim_step.as_secs_f64();
        let render_prep_time = self.time_prepping_render.as_secs_f64();
        let stim_block_time = self.time_blocked_on_stim.as_secs_f64();
        let render_block_time = self.time_blocked_on_render.as_secs_f64();
        let stim_prep_time = self.time_prepping_stim.as_secs_f64();

        let other_driver_time = driver_time - (
            step_time + stim_block_time + stim_prep_time + render_block_time + render_prep_time
        );
        let other_time = overall_time - driver_time;
        let toplevel_line = format!("toplevel\tstep: {:.1}s, stim_blocked: {:.1}s, render_blocked: {:.1}s, stim_prep: {:.1}s, render_prep: {:.1}s, driver_other: {:.1}s, unknown: {:.1}s",
            step_time,
            stim_block_time,
            render_block_time,
            stim_prep_time,
            render_prep_time,
            other_driver_time,
            other_time,
        );

        let device_step_time = self.time_stepping_device.as_secs_f64();
        let device_write_time = self.time_writing_device.as_secs_f64();
        let device_read_time = self.time_reading_device.as_secs_f64();
        let device_line = format!("> gpu\tstep: {:.1}s, write: {:.1}s, read: {:.1}s",
            device_step_time,
            device_write_time,
            device_read_time,
        );

        let stim_bg_time = self.time_on_stimuli.as_secs_f64();
        let render_bg_time = self.time_rendering.as_secs_f64();
        let bg_line = format!("> async\tstim: {:.1}s, render: {:.1}s",
            stim_bg_time,
            render_bg_time,
        );

        format!("{}\n  {}\n  {}\n  {}", fps_line, toplevel_line, device_line, bg_line)
    }
}

impl SyncDiagnostics {
    pub fn new() -> Self {
        Self(Arc::new(Mutex::new(Diagnostics::new())))
    }
    pub fn format(&self) -> String {
        self.0.lock().unwrap().format()
    }
    /// measure the duration of some arbitrary chunk of code.
    /// used internally.
    pub fn measure<R, F: FnOnce() -> R>(f: F) -> (Duration, R) {
        let start = Instant::now();
        let r = f();
        (start.elapsed(), r)
    }

    pub fn instrument_driver<R, F: FnOnce() -> R>(&self, f: F) -> R {
        let (elapsed, ret) = Self::measure(f);
        self.0.lock().unwrap().time_in_driver += elapsed;
        ret
    }

    /// record the duration of the sim step operation.
    pub fn instrument_step<R, F: FnOnce() -> R>(&self, frames: u64, f: F) -> R {
        let (elapsed, ret) = Self::measure(f);
        let mut me = self.0.lock().unwrap();
        me.time_sim_step += elapsed;
        me.frames_completed += frames as u64;
        ret
    }

    /// record the duration spent preparing for render (i.e. cloning stuff and moving it into a
    /// render pool).
    pub fn instrument_render_prep<R, F: FnOnce() -> R>(&self, f: F) -> R {
        let (elapsed, ret) = Self::measure(f);
        self.0.lock().unwrap().time_prepping_render += elapsed;
        ret
    }

    /// record the duration actually spent doing CPU render work
    pub fn instrument_render_cpu_side<R, F: FnOnce() -> R>(&self, f: F) -> R {
        let (elapsed, ret) = Self::measure(f);
        self.0.lock().unwrap().time_rendering += elapsed;
        ret
    }
    pub fn instrument_stimuli_prep<R, F: FnOnce() -> R>(&self, f: F) -> R {
        let (elapsed, ret) = Self::measure(f);
        self.0.lock().unwrap().time_prepping_stim += elapsed;
        ret
    }

    /// record the duration spent blocking the simulation because the stimulus queue is full.
    pub fn instrument_stimuli_blocked<R, F: FnOnce() -> R>(&self, f: F) -> R{
        let (elapsed, ret) = Self::measure(f);
        self.0.lock().unwrap().time_blocked_on_stim += elapsed;
        ret
    }

    /// record the duration spent blocking the simulation because the render queue is full.
    pub fn instrument_render_blocked<R, F: FnOnce() -> R>(&self, f: F) -> R{
        let (elapsed, ret) = Self::measure(f);
        self.0.lock().unwrap().time_blocked_on_render += elapsed;
        ret
    }

    pub fn instrument_stimuli<R, F: FnOnce() -> R>(&self, f: F) -> R {
        let (elapsed, ret) = Self::measure(f);
        self.0.lock().unwrap().time_on_stimuli += elapsed;
        ret
    }

    pub fn record_step_device(&self, t: Duration) {
        self.0.lock().unwrap().time_stepping_device += t;
    }
    pub fn instrument_read_device<R, F: FnOnce() -> R>(&self, f: F) -> R {
        let (elapsed, ret) = Self::measure(f);
        self.0.lock().unwrap().time_reading_device += elapsed;
        ret
    }
    pub fn instrument_write_device<R, F: FnOnce() -> R>(&self, f: F) -> R {
        let (elapsed, ret) = Self::measure(f);
        self.0.lock().unwrap().time_writing_device += elapsed;
        ret
    }
}

