use coremem_cross::real::Real;
use coremem_cross::vec::Vec2;

use std::ops::Add;

#[derive(Copy, Clone, Debug, Default, PartialEq)]
pub struct Line2d<R> {
    from: Vec2<R>,
    to: Vec2<R>,
}

impl<R: Real> Add for Line2d<R> {
    type Output = Self;
    fn add(self, other: Self) -> Self {
        Line2d {
            from: self.from + Vec2::new(R::zero(), other.y(self.from.x())),
            to: self.to + Vec2::new(R::zero(), other.y(self.to.x())),
        }
    }
}

impl<R> Line2d<R> {
    pub fn new(from: Vec2<R>, to: Vec2<R>) -> Self {
        Self {
            from,
            to,
        }
    }
}

impl<R: Real> Line2d<R> {
    pub fn from(&self) -> Vec2<R> {
        self.from
    }
    pub fn to(&self) -> Vec2<R> {
        self.to
    }

    pub fn x(&self, y: R) -> R {
        self.shifted(Vec2::new(R::zero(), -y)).x_intercept()
    }
    pub fn y(&self, x: R) -> R {
        (self.from.y + (x - self.from.x) * self.slope()).into()
    }
    pub fn at_x(&self, x: R) -> Vec2<R> {
        Vec2::new(x, self.y(x))
    }

    pub fn x_intercept(&self) -> R {
        let t = (-self.from.y) / (self.to.y - self.from.y);
        (self.from.x + t * (self.to.x - self.from.x)).into()
    }

    pub fn as_vector(&self) -> Vec2<R> {
        self.to + (-self.from)
    }

    pub fn length_sq(&self) -> R {
        let Vec2 { x, y } = self.as_vector();
        (x*x + y*y).into()
    }

    pub fn distance_sq(&self, pt: Vec2<R>) -> R {
        // source: https://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line#Line_defined_by_two_points
        let d = self.as_vector();
        let twice_area = d.y*pt.x - d.x*pt.y + self.to.x*self.from.y - self.to.y*self.from.x;
        let d2 = twice_area * twice_area / self.length_sq();
        d2.into()
    }

    pub fn is_ascending(&self) -> bool {
        self.to.y > self.from.y
    }

    pub fn is_ascending_x(&self) -> bool {
        self.to.x > self.from.x
    }

    pub fn contains_x(&self, x: R) -> bool {
        if self.is_ascending_x() {
            x >= self.from.x && x < self.to.x
        } else {
            x <= self.from.x && x > self.to.x
        }
    }

    pub fn contains_y(&self, y: R) -> bool {
        if self.is_ascending_x() {
            y >= self.from.y && y < self.to.y
        } else {
            y <= self.from.y && y > self.to.y
        }
    }

    pub fn min_x(&self) -> R {
        self.from.x.min_or_undefined(self.to.x)
    }

    pub fn max_x(&self) -> R {
        self.from.x.max_or_undefined(self.to.x)
    }

    pub fn clamp_x(&self, x: R) -> R {
        match x {
            v if v < self.min_x() => self.min_x(),
            v if v > self.max_x() => self.max_x(),
            v => v,
        }
    }

    pub fn clamp_by_x(&self, x: R) -> Vec2<R> {
        self.at_x(self.clamp_x(x))
    }

    pub fn slope(&self) -> R {
        (self.to.y - self.from.y) / (self.to.x - self.from.x)
    }

    /// Move the coordinate along the line toward some desired `y(x)` value.
    /// The returned value could be OOB: check `self.contains_x` to find out.
    pub fn move_toward_y_unclamped(&self, start_x: R, target_y: R) -> R {
        let start_y = self.y(start_x);
        let delta_x = (target_y - start_y) / self.slope();
        start_x + delta_x
    }

    pub fn shifted(&self, p: Vec2<R>) -> Self {
        Line2d {
            from: self.from + p,
            to: self.to + p
        }
    }
}

