use crate::geom::Meters;
use coremem_cross::real::Real as _;
use coremem_cross::vec::{Vec2, Vec3};

use serde::{Serialize, Deserialize};
use std::fmt::{self, Display};
use std::ops::Range;

use super::{HasCrossSection, Region};

#[derive(Copy, Clone, Default, Serialize, Deserialize)]
pub struct CylinderZ {
    center: Vec2<f32>,
    radius: f32,
}

impl CylinderZ {
    // TODO: should be unit-typed (Meters)
    pub fn new(center: Vec2<f32>, radius: f32) -> Self {
        Self {
            center,
            radius,
        }
    }
}

impl Region for CylinderZ {
    fn contains(&self, p: Meters) -> bool {
        p.xy().distance_sq(self.center) <= self.radius * self.radius
    }
}

impl Display for CylinderZ {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "d({:.1e}, {:.1e}) <= {:.1e}", self.center.x(), self.center.y(), self.radius)
    }
}

/// describes all 3d space which falls within a given angular space, relative to the Z axis.
#[derive(Copy, Clone, Default, Serialize, Deserialize)]
pub struct WedgeZ {
    arg_min: f32,
    arg_max: f32,
}

impl WedgeZ {
    pub fn new(arg_min: f32, arg_max: f32) -> Self {
        Self { arg_min, arg_max }
    }
}

impl Region for WedgeZ {
    fn contains(&self, p: Meters) -> bool {
        let arg = p.xy().arg();
        // arg is [-pi, pi).
        // if the user supplied some desired range where arg_max > pi, then we need to rotate
        // one revolution "into" that range.
        let arg_next = arg + f32::two_pi();
        (arg >= self.arg_min && arg <= self.arg_max) ||
        (arg_next >= self.arg_min && arg_next <= self.arg_max)
    }
}

#[derive(Copy, Clone, Debug, Default, Serialize, Deserialize)]
pub struct Torus {
    center: Meters,
    /// Unit-length vector describing the axis of the torus
    normal: Meters,
    /// Distance from origin to the "center" of the solid part of the torus
    major_rad: f32,
    /// Distance from a center-point of the torus to its surface
    minor_rad: f32,
}

impl Torus {
    pub fn new(center: Meters, normal: Meters, major_rad: f32, minor_rad: f32) -> Self {
        Self {
            center,
            normal,
            major_rad,
            minor_rad,
        }
    }
    pub fn new_xy(center: Meters, major_rad: f32, minor_rad: f32) -> Self {
        Self::new(center, Meters(Vec3::new(0.0, 0.0, 1.0)), major_rad, minor_rad)
    }
    pub fn new_xz(center: Meters, major_rad: f32, minor_rad: f32) -> Self {
        Self::new(center, Meters(Vec3::new(0.0, 1.0, 0.0)), major_rad, minor_rad)
    }
    pub fn new_yz(center: Meters, major_rad: f32, minor_rad: f32) -> Self {
        Self::new(center, Meters(Vec3::new(1.0, 0.0, 0.0)), major_rad, minor_rad)
    }
    pub fn center(&self) -> Meters {
        self.center
    }
    pub fn axis(&self) -> Meters {
        self.normal
    }
    pub fn cross_section(&self) -> f32 {
        self.minor_rad * self.minor_rad * f32::pi()
    }
}

impl Region for Torus {
    fn contains(&self, p: Meters) -> bool {
        // a torus is the set of all points < distance `r` from the circle of radius `R`,
        // where `r`= minor_rad, `R` = major_rad, and `center` and `normal` define the center and
        // plane of the circle.
        // 1. Project `p` onto the plane of the circle.
        // 2. Find the point `q` on the circle which is nearest to `p`.
        // 3. Consider the distance from `p` to `q`.
        let rel_p = *p - *self.center;
        let p_on_plane = rel_p - self.normal.with_mag(self.normal.dot(rel_p)).unwrap();
        let q = if p_on_plane == Vec3::zero() {
            // avoid division by zero.
            // The point is precisely on the axis of the torus.
            // We can choose any point on the major radius to compare it to
            // and they all give the same answer.
            // Such a point is given by rotating the normal axis by 90 degrees in ANY DIRECTION
            let off_axis = self.normal.arbitrary_orthogonal_vector();
            off_axis.with_mag(self.major_rad).unwrap()
        } else {
            p_on_plane.with_mag(self.major_rad).unwrap()
        };
        let distance_to_circle_sq = rel_p.distance_sq(q);
        distance_to_circle_sq < self.minor_rad * self.minor_rad
    }
}

impl HasCrossSection for Torus {
    fn cross_section_normal(&self, coord: Meters) -> Vec3<f32> {
        let axis = self.axis();
        let to_coord = *coord - *self.center();
        // this creates a normal which always points "counter-clockwise" along the shape
        axis.cross(to_coord).with_mag(self.cross_section()).unwrap_or_default()
    }
}

#[derive(Copy, Clone, Default, Serialize, Deserialize)]
pub struct Sphere {
    center: Meters,
    rad: f32,
}

impl Sphere {
    pub fn new(center: Meters, rad: f32) -> Self {
        Self {
            center,
            rad,
        }
    }
}

impl Region for Sphere {
    fn contains(&self, p: Meters) -> bool {
        p.distance_sq(*self.center) < self.rad * self.rad
    }
}

#[derive(Copy, Clone, Default, Serialize, Deserialize)]
pub struct Cube {
    /// (x0, y0, z0))
    small: Meters,
    /// (i.e. (x1, y1, z1)) where x1 >= x0, y1 >= y0, z1 >= z0
    large: Meters,
}

impl Cube {
    pub fn new(small: Meters, large: Meters) -> Self {
        Self { small, large }
    }
    pub fn new_centered(center: Meters, size: Meters) -> Self {
        Self::new(center - size*0.5, center + size*0.5)
    }
    /// 'new', but handle the case where pt1's coordinate is lower than pt0 by 'swapping' them
    pub fn new_including_negatives(pt0: Meters, pt1: Meters) -> Self {
        Self::new(
            Meters::new(pt0.x().min(pt1.x()), pt0.y().min(pt1.y()), pt0.z().min(pt1.z())),
            Meters::new(pt0.x().max(pt1.x()), pt0.y().max(pt1.y()), pt0.z().max(pt1.z())),
        )
    }
    pub fn x_range(&self) -> Range<f32> {
        self.small.x()..self.large.x()
    }
    pub fn y_range(&self) -> Range<f32> {
        self.small.y()..self.large.y()
    }
    pub fn z_range(&self) -> Range<f32> {
        self.small.z()..self.large.z()
    }
    pub fn small(&self) -> Meters {
        self.small
    }
    pub fn large(&self) -> Meters {
        self.large
    }

    pub fn center(&self) -> Meters {
        (self.small + self.large) * 0.5
    }

    pub fn left(&self) -> f32 {
        self.small.x()
    }
    pub fn right(&self) -> f32 {
        self.small.x()
    }
    pub fn top(&self) -> f32 {
        self.small.y()
    }
    pub fn bot(&self) -> f32 {
        self.large.y()
    }
    pub fn out(&self) -> f32 {
        self.small.z()
    }
    pub fn in_(&self) -> f32 {
        self.large.z()
    }

    pub fn bot_left_out(&self) -> Meters {
        Meters::new(self.left(), self.bot(), self.out())
    }
    pub fn bot_left_center(&self) -> Meters {
        Meters::new(self.left(), self.bot(), self.center().z())
    }
    pub fn bot_left_in(&self) -> Meters {
        Meters::new(self.left(), self.bot(), self.in_())
    }

    pub fn bot_right_out(&self) -> Meters {
        Meters::new(self.right(), self.bot(), self.out())
    }
    pub fn bot_right_center(&self) -> Meters {
        Meters::new(self.right(), self.bot(), self.center().z())
    }
    pub fn bot_right_in(&self) -> Meters {
        Meters::new(self.right(), self.bot(), self.in_())
    }

    pub fn top_left_out(&self) -> Meters {
        Meters::new(self.left(), self.top(), self.out())
    }
    pub fn top_left_center(&self) -> Meters {
        Meters::new(self.left(), self.top(), self.center().z())
    }
    pub fn top_left_in(&self) -> Meters {
        Meters::new(self.left(), self.top(), self.in_())
    }

    pub fn top_right_out(&self) -> Meters {
        Meters::new(self.right(), self.top(), self.out())
    }
    pub fn top_right_center(&self) -> Meters {
        Meters::new(self.right(), self.top(), self.center().z())
    }
    pub fn top_right_in(&self) -> Meters {
        Meters::new(self.right(), self.top(), self.in_())
    }
}

impl Region for Cube {
    fn contains(&self, p: Meters) -> bool {
        self.x_range().contains(&p.x()) &&
            self.y_range().contains(&p.y()) &&
            self.z_range().contains(&p.z())
    }
}

/// a Spiral traces out a circle on the xy plane as z increases.
#[derive(Copy, Clone, Default, Serialize, Deserialize)]
pub struct Spiral {
    /// radius of the spiral
    major: f32,
    /// effective "width" of the spiral.
    /// any point within this distance from the 1-dimensional spiral is considered to be in the
    /// region.
    minor: f32,
    /// determines how stretched the spiral is in z-space.
    /// the xy plane at z=0 looks identical again at t=period.
    period: f32,
}

impl Spiral {
    pub fn new(major: f32, minor: f32, period: f32) -> Self {
        Self { major, minor, period }
    }
}

impl Region for Spiral {
    fn contains(&self, p: Meters) -> bool {
        let revs = p.z() / self.period;
        let expected_angle = (revs % 1.0) * std::f32::consts::PI * 2.0;
        let expected = Meters::new(self.major * expected_angle.cos(), self.major * expected_angle.sin(), p.z());
        p.distance_sq(*expected) < self.minor * self.minor
    }
}

#[cfg(test)]
mod test {
    use super::*;
    #[test]
    fn torus_as_sphere() {
        let tor = Torus::new_xy(Meters::new(0.0, 0.0, 0.0), 0.0, 10.0);
        assert!(tor.contains(Meters::new(0.0, 0.0, 0.0)));
        assert!(tor.contains(Meters::new(5.0, 0.0, 0.0)));

        assert!(tor.contains(Meters::new(9.0, 0.0, 0.0)));
        assert!(tor.contains(Meters::new(-9.0, 0.0, 0.0)));
        assert!(tor.contains(Meters::new(0.0, 9.0, 0.0)));
        assert!(tor.contains(Meters::new(0.0, -9.0, 0.0)));
        assert!(tor.contains(Meters::new(0.0, 0.0, 9.0)));
        assert!(tor.contains(Meters::new(0.0, 0.0, -9.0)));

        assert!(tor.contains(Meters::new(5.0, 5.0, 5.0)));
        assert!(tor.contains(Meters::new(-5.0, -5.0, -5.0)));

        // boundary
        assert!(!tor.contains(Meters::new(10.0, 0.0, 0.0)));
        assert!(!tor.contains(Meters::new(-10.0, 0.0, 0.0)));
        assert!(!tor.contains(Meters::new(0.0, 10.0, 0.0)));
        assert!(!tor.contains(Meters::new(0.0, -10.0, 0.0)));
        assert!(!tor.contains(Meters::new(0.0, 0.0, 10.0)));
        assert!(!tor.contains(Meters::new(0.0, 0.0, -10.0)));

        assert!(!tor.contains(Meters::new(11.0, 0.0, 0.0)));
        assert!(!tor.contains(Meters::new(9.0, 9.0, 9.0)));
        assert!(!tor.contains(Meters::new(-9.0, -9.0, -9.0)));
        assert!(!tor.contains(Meters::new(-9.0, 9.0, 0.0)));
    }

    #[test]
    fn torus_xy() {
        let tor = Torus::new_xy(Meters::new(0.0, 0.0, 0.0), 10.0, 2.0);
        assert!(!tor.contains(Meters::new(0.0, 0.0, 0.0)));

        assert!(tor.contains(Meters::new(10.0, 0.0, 0.0)));
        assert!(tor.contains(Meters::new(-10.0, 0.0, 0.0)));
        assert!(tor.contains(Meters::new(0.0, 10.0, 0.0)));
        assert!(tor.contains(Meters::new(0.0, -10.0, 0.0)));
        assert!(!tor.contains(Meters::new(0.0, 0.0, 10.0)));
        assert!(!tor.contains(Meters::new(0.0, 0.0, -10.0)));

        assert!(tor.contains(Meters::new(11.0, 0.0, 0.0)));
        assert!(tor.contains(Meters::new(-11.0, 0.0, 0.0)));
        assert!(tor.contains(Meters::new(11.0, 2.0, 0.0)));
        assert!(tor.contains(Meters::new(11.0, 0.0, 1.0)));
        assert!(tor.contains(Meters::new(-11.0, 0.0, 1.0)));
        assert!(tor.contains(Meters::new(-7.0, 7.0, 1.0)));
        assert!(tor.contains(Meters::new(-7.0, -7.0, -1.0)));

        // boundary
        assert!(!tor.contains(Meters::new(12.0, 0.0, 0.0)));
        assert!(!tor.contains(Meters::new(10.0, 0.0, 2.0)));
        assert!(!tor.contains(Meters::new(-10.0, 0.0, 2.0)));
    }

    #[test]
    fn torus_xz() {
        let tor = Torus::new_xz(Meters::new(0.0, 0.0, 0.0), 10.0, 2.0);
        assert!(!tor.contains(Meters::new(0.0, 0.0, 0.0)));

        assert!(tor.contains(Meters::new(10.0, 0.0, 0.0)));
        assert!(tor.contains(Meters::new(-10.0, 0.0, 0.0)));
        assert!(!tor.contains(Meters::new(0.0, 10.0, 0.0)));
        assert!(!tor.contains(Meters::new(0.0, -10.0, 0.0)));
        assert!(tor.contains(Meters::new(0.0, 0.0, 10.0)));
        assert!(tor.contains(Meters::new(0.0, 0.0, -10.0)));

        assert!(tor.contains(Meters::new(11.0, 0.0, 0.0)));
        assert!(tor.contains(Meters::new(-11.0, 0.0, 0.0)));
        assert!(tor.contains(Meters::new(11.0, 1.0, 0.0)));
        assert!(tor.contains(Meters::new(11.0, 0.0, 2.0)));
        assert!(tor.contains(Meters::new(-11.0, 1.0, 0.0)));
        assert!(tor.contains(Meters::new(-7.0, 1.0, 7.0)));
        assert!(tor.contains(Meters::new(-7.0, -1.0, -7.0)));

        // boundary
        assert!(!tor.contains(Meters::new(12.0, 0.0, 0.0)));
        assert!(!tor.contains(Meters::new(10.0, 2.0, 0.0)));
        assert!(!tor.contains(Meters::new(-10.0, 2.0, 0.0)));
    }

    #[test]
    fn torus_yz() {
        let tor = Torus::new_yz(Meters::new(0.0, 0.0, 0.0), 10.0, 2.0);
        assert!(!tor.contains(Meters::new(0.0, 0.0, 0.0)));

        assert!(tor.contains(Meters::new(0.0, 10.0, 0.0)));
        assert!(tor.contains(Meters::new(0.0, -10.0, 0.0)));
        assert!(!tor.contains(Meters::new(10.0, 0.0, 0.0)));
        assert!(!tor.contains(Meters::new(-10.0, 0.0, 0.0)));
        assert!(tor.contains(Meters::new(0.0, 0.0, 10.0)));
        assert!(tor.contains(Meters::new(0.0, 0.0, -10.0)));

        assert!(tor.contains(Meters::new(0.0, 11.0, 0.0)));
        assert!(tor.contains(Meters::new(0.0, -11.0, 0.0)));
        assert!(tor.contains(Meters::new(1.0, 11.0, 0.0)));
        assert!(tor.contains(Meters::new(0.0, 11.0, 2.0)));
        assert!(tor.contains(Meters::new(1.0, -11.0, 0.0)));
        assert!(tor.contains(Meters::new(1.0, -7.0, 7.0)));
        assert!(tor.contains(Meters::new(-1.0, -7.0, -7.0)));

        // boundary
        assert!(!tor.contains(Meters::new(0.0, 12.0, 0.0)));
        assert!(!tor.contains(Meters::new(2.0, 10.0, 0.0)));
        assert!(!tor.contains(Meters::new(2.0, -10.0, 0.0)));
    }

    #[test]
    fn torus_with_center() {
        let tor = Torus::new_xy(Meters::new(1.0, 2.0, 3.0), 0.0, 10.0);
        assert!(tor.contains(Meters::new(1.0, 2.0, 3.0)));
        assert!(tor.contains(Meters::new(1.0, 2.0, 12.0)));
        assert!(!tor.contains(Meters::new(1.0, 2.0, 13.0)));

        let tor = Torus::new_xy(Meters::new(1.0, 2.0, 3.0), 10.0, 3.0);
        assert!(!tor.contains(Meters::new(1.0, 2.0, 3.0)));
        assert!(tor.contains(Meters::new(1.0, 14.0, 3.0)));
        assert!(!tor.contains(Meters::new(1.0, 15.0, 3.0)));

        let tor = Torus::new_xz(Meters::new(1.0, 2.0, 3.0), 5.0, 5.0);
        assert!(!tor.contains(Meters::new(1.0, 2.0, 3.0)));
        assert!(tor.contains(Meters::new(2.0, 2.0, 3.0)));
        assert!(tor.contains(Meters::new(1.0, 2.0, 12.0)));
        assert!(!tor.contains(Meters::new(1.0, 2.0, 13.0)));
    }

    #[test]
    fn test_spiral_doesnt_contain_zero() {
        let spiral = Spiral::new(10.0, 1.0, 4.0);
        assert!(!spiral.contains(Meters::new(0.0, 0.0, 0.0)));
        assert!(!spiral.contains(Meters::new(0.0, 0.0, 1.0)));
        assert!(!spiral.contains(Meters::new(0.0, 0.0, 2.0)));
        assert!(!spiral.contains(Meters::new(0.0, 0.0, 3.0)));
        assert!(!spiral.contains(Meters::new(0.0, 0.0, 4.0)));
        assert!(!spiral.contains(Meters::new(0.0, 0.0, 5.0)));
        assert!(!spiral.contains(Meters::new(0.0, 0.0, 6.0)));
        assert!(!spiral.contains(Meters::new(0.0, 0.0, 7.0)));
        assert!(!spiral.contains(Meters::new(0.0, 0.0, 8.0)));
        assert!(!spiral.contains(Meters::new(0.0, 0.0, 9.0)));
        assert!(!spiral.contains(Meters::new(0.0, 0.0, 10.0)));
        assert!(!spiral.contains(Meters::new(0.0, 0.0, 11.0)));
    }

    #[test]
    fn test_spiral_exactly_on_course() {
        let spiral = Spiral::new(10.0, 1.0, 4.0);
        assert!(spiral.contains(Meters::new(10.0, 0.0, 0.0)));
        assert!(spiral.contains(Meters::new(0.0, 10.0, 1.0)));
        assert!(spiral.contains(Meters::new(-10.0, 0.0, 2.0)));
        assert!(spiral.contains(Meters::new(0.0, -10.0, 3.0)));
        assert!(spiral.contains(Meters::new(10.0, 0.0, 4.0)));
        assert!(spiral.contains(Meters::new(0.0, 10.0, 5.0)));
    }

    #[test]
    fn test_spiral_exactly_on_course_negative() {
        let spiral = Spiral::new(10.0, 1.0, 4.0);
        assert!(spiral.contains(Meters::new(0.0, -10.0, -1.0)));
        assert!(spiral.contains(Meters::new(-10.0, 0.0, -2.0)));
        assert!(spiral.contains(Meters::new(0.0, 10.0, -3.0)));
        assert!(spiral.contains(Meters::new(10.0, 0.0, -4.0)));
        assert!(spiral.contains(Meters::new(0.0, -10.0, -5.0)));
    }

    #[test]
    fn test_spiral_within_minor_radius() {
        let spiral = Spiral::new(10.0, 1.0, 4.0);
        assert!(spiral.contains(Meters::new(10.5, 0.0, 0.0)));
        assert!(spiral.contains(Meters::new(10.5, -0.5, 0.0)));
        assert!(spiral.contains(Meters::new(9.1, 0.0, 0.0)));
        assert!(!spiral.contains(Meters::new(8.9, 0.0, 0.0)));
        assert!(!spiral.contains(Meters::new(9.1, 0.9, 0.0)));

        assert!(spiral.contains(Meters::new(0.5, 10.5, 1.0)));
        assert!(spiral.contains(Meters::new(-0.9, 10.0, 1.0)));
        assert!(!spiral.contains(Meters::new(-0.9, 10.9, 1.0)));

        assert!(spiral.contains(Meters::new(-10.9, 0.0, 2.0)));
        assert!(spiral.contains(Meters::new(-10.0, 0.6, 2.0)));
        assert!(!spiral.contains(Meters::new(-10.8, 0.8, 2.0)));
    }
}


