use futures::FutureExt as _;
use log::info;
use std::borrow::Cow;
use std::num::NonZeroU64;
use std::time::Duration;
use wgpu;
use wgpu::util::DeviceExt as _;

use crate::diagnostics::SyncDiagnostics;

use coremem_cross::vec::{Vec3, Vec3u};
use coremem_cross::step::SimMeta;

use spirv_backend::HasEntryPoints;

use super::SimBackend;

#[derive(Default)]
pub struct WgpuBackend {
    handles: Option<(&'static str /* step_h */, &'static str /* step_e */, WgpuHandles)>,
}

struct WgpuHandles {
    step_bind_group_layout: wgpu::BindGroupLayout,
    step_e_pipeline: wgpu::ComputePipeline,
    step_h_pipeline: wgpu::ComputePipeline,
    device: wgpu::Device,
    queue: wgpu::Queue,
}

impl WgpuHandles {
    fn open<R, M: HasEntryPoints<R>>(dim: Vec3u) -> Self {
        info!("WgpuHandles::open({})", dim);
        use std::mem::size_of;
        let volume = dim.product_sum_usize() as u64;
        let max_elem_size = size_of::<M>().max(size_of::<Vec3<R>>());
        let max_array_size = volume * max_elem_size as u64;
        let max_buf_size = max_array_size + 0x1000; // allow some overhead

        let (device, queue) = futures::executor::block_on(open_device(max_buf_size));
        let shader_binary = get_shader();
        let shader_module = unsafe { device.create_shader_module_spirv(&shader_binary) };
        let (step_bind_group_layout, step_h_pipeline, step_e_pipeline) = make_pipelines(
            &device, &shader_module, M::step_h(), M::step_e()
        );
        WgpuHandles {
            step_bind_group_layout,
            step_h_pipeline,
            step_e_pipeline,
            device,
            queue,
        }
    }
}

// TODO: these bounds aren't 100% right. we're sending R and M over to the GPU by a bitwise copy.
// that probably means the types should be Send + Copy
impl<R: Copy, M: Send + Sync + HasEntryPoints<R>> SimBackend<R, M> for WgpuBackend {
    fn step_n(
        &mut self,
        diag: &SyncDiagnostics,
        meta: SimMeta<R>,
        mat: &[M],
        stim_cpu_e: &[Vec3<R>],
        stim_cpu_h: &[Vec3<R>],
        e: &mut [Vec3<R>],
        h: &mut [Vec3<R>],
        m: &mut [Vec3<R>],
        num_steps: u32,
    ) {
        let dim = meta.dim();
        let field_bytes = dim.product_sum() as usize * std::mem::size_of::<Vec3<f32>>();

        let (step_h, step_e, handles) = self.handles.get_or_insert_with(|| (
            M::step_h(),
            M::step_e(),
            WgpuHandles::open::<R, M>(dim)
        ));
        // if device is opened, make sure we're open for the right types
        assert_eq!(*step_h, M::step_h());
        assert_eq!(*step_e, M::step_e());
        let device = &handles.device;
        let queue = &handles.queue;
        let step_bind_group_layout = &handles.step_bind_group_layout;
        let step_e_pipeline = &handles.step_e_pipeline;
        let step_h_pipeline = &handles.step_h_pipeline;
        let timestamp_buffer = device.create_buffer(&wgpu::BufferDescriptor {
            label: Some("timestamps"),
            // each timestamp is 8 bytes, and we do 4 per step
            size: 8 * 4 * num_steps as u64,
            usage: wgpu::BufferUsages::MAP_READ | wgpu::BufferUsages::COPY_DST,
            mapped_at_creation: true,
        });
        timestamp_buffer.unmap();

        let sim_meta_buffer = device.create_buffer_init(&wgpu::util::BufferInitDescriptor {
            label: Some("gpu-side simulation metadata"),
            contents: to_bytes(&[meta][..]),
            usage: wgpu::BufferUsages::STORAGE,
        });

        let stim_e_buffer = device.create_buffer_init(&wgpu::util::BufferInitDescriptor {
            label: Some("gpu-side stimulus e field"),
            contents: to_bytes(stim_cpu_e),
            usage: wgpu::BufferUsages::STORAGE
        });

        let stim_h_buffer = device.create_buffer_init(&wgpu::util::BufferInitDescriptor {
            label: Some("gpu-side stimulus h field"),
            contents: to_bytes(stim_cpu_h),
            usage: wgpu::BufferUsages::STORAGE
        });

        let mat_buffer = device.create_buffer_init(&wgpu::util::BufferInitDescriptor {
            label: Some("gpu-side materials matrix"),
            contents: to_bytes(mat),
            // Can be used by the GPU and copied back to the CPU
            usage: wgpu::BufferUsages::STORAGE,
        });

        let e_field_buffer = device.create_buffer_init(&wgpu::util::BufferInitDescriptor {
            label: Some("gpu-side in/out e field"),
            contents: to_bytes(e),
            usage: wgpu::BufferUsages::STORAGE.union(wgpu::BufferUsages::COPY_SRC),
        });

        let h_field_buffer = device.create_buffer_init(&wgpu::util::BufferInitDescriptor {
            label: Some("gpu-side in/out h field"),
            contents: to_bytes(h),
            // Can be used by the GPU and copied back to the CPU
            usage: wgpu::BufferUsages::STORAGE.union(wgpu::BufferUsages::COPY_SRC),
        });

        let m_field_buffer = device.create_buffer_init(&wgpu::util::BufferInitDescriptor {
            label: Some("gpu-side in/out m field"),
            contents: to_bytes(m),
            // Can be used by the GPU and copied back to the CPU
            usage: wgpu::BufferUsages::STORAGE.union(wgpu::BufferUsages::COPY_SRC),
        });

        let e_readback_buffer = device.create_buffer(&wgpu::BufferDescriptor {
            label: Some("cpu-side copy of e output buffer"),
            size: field_bytes as wgpu::BufferAddress,
            // Can be read to the CPU, and can be copied from the shader's storage buffer
            usage: wgpu::BufferUsages::MAP_READ | wgpu::BufferUsages::COPY_DST,
            mapped_at_creation: false,
        });

        let h_readback_buffer = device.create_buffer(&wgpu::BufferDescriptor {
            label: Some("cpu-side copy of h output buffer"),
            size: field_bytes as wgpu::BufferAddress,
            // Can be read to the CPU, and can be copied from the shader's storage buffer
            usage: wgpu::BufferUsages::MAP_READ | wgpu::BufferUsages::COPY_DST,
            mapped_at_creation: false,
        });

        let m_readback_buffer = device.create_buffer(&wgpu::BufferDescriptor {
            label: Some("cpu-side copy of m output buffer"),
            size: field_bytes as wgpu::BufferAddress,
            // Can be read to the CPU, and can be copied from the shader's storage buffer
            usage: wgpu::BufferUsages::MAP_READ | wgpu::BufferUsages::COPY_DST,
            mapped_at_creation: false,
        });

        let bind_group = device.create_bind_group(&wgpu::BindGroupDescriptor {
            label: None,
            layout: &step_bind_group_layout,
            entries: &[
                wgpu::BindGroupEntry {
                    binding: 0,
                    resource: sim_meta_buffer.as_entire_binding(),
                },
                wgpu::BindGroupEntry {
                    binding: 1,
                    resource: stim_e_buffer.as_entire_binding(),
                },
                wgpu::BindGroupEntry {
                    binding: 2,
                    resource: stim_h_buffer.as_entire_binding(),
                },
                wgpu::BindGroupEntry {
                    binding: 3,
                    resource: mat_buffer.as_entire_binding(),
                },
                wgpu::BindGroupEntry {
                    binding: 4,
                    resource: e_field_buffer.as_entire_binding(),
                },
                wgpu::BindGroupEntry {
                    binding: 5,
                    resource: h_field_buffer.as_entire_binding(),
                },
                wgpu::BindGroupEntry {
                    binding: 6,
                    resource: m_field_buffer.as_entire_binding(),
                },
            ],
        });

        let queries = device.create_query_set(&wgpu::QuerySetDescriptor {
            label: None,
            count: 4 * num_steps,
            ty: wgpu::QueryType::Timestamp,
        });

        let workgroups = ((dim.x()+3) / 4, (dim.y()+3) / 4, (dim.z()+3) / 4);

        let mut encoder =
            device.create_command_encoder(&wgpu::CommandEncoderDescriptor { label: None });

        for step in 0..num_steps {
            {
                let mut cpass = encoder.begin_compute_pass(&wgpu::ComputePassDescriptor { label: None });
                cpass.set_bind_group(0, &bind_group, &[]);
                cpass.set_pipeline(&step_e_pipeline);
                cpass.write_timestamp(&queries, 4*step);
                cpass.dispatch(workgroups.0, workgroups.1, workgroups.2);
                cpass.write_timestamp(&queries, 4*step + 1);
            }

            {
                let mut cpass = encoder.begin_compute_pass(&wgpu::ComputePassDescriptor { label: None });
                cpass.set_bind_group(0, &bind_group, &[]);
                cpass.set_pipeline(&step_h_pipeline);
                cpass.write_timestamp(&queries, 4*step + 2);
                cpass.dispatch(workgroups.0, workgroups.1, workgroups.2);
                cpass.write_timestamp(&queries, 4*step + 3);
            }
        }

        encoder.copy_buffer_to_buffer(
            &e_field_buffer,
            0,
            &e_readback_buffer,
            0,
            field_bytes as u64,
        );

        encoder.copy_buffer_to_buffer(
            &h_field_buffer,
            0,
            &h_readback_buffer,
            0,
            field_bytes as u64,
        );

        encoder.copy_buffer_to_buffer(
            &m_field_buffer,
            0,
            &m_readback_buffer,
            0,
            field_bytes as u64,
        );

        encoder.resolve_query_set(&queries, 0..4*num_steps, &timestamp_buffer, 0);

        diag.instrument_write_device(move || {
            queue.submit(Some(encoder.finish()));
        });

        let e_readback_slice = e_readback_buffer.slice(..);
        let e_readback_future = e_readback_slice.map_async(wgpu::MapMode::Read).then(|_| async {
            e.copy_from_slice(unsafe {
                from_bytes(e_readback_slice.get_mapped_range().as_ref())
            });
            e_readback_buffer.unmap();
        });

        let h_readback_slice = h_readback_buffer.slice(..);
        let h_readback_future = h_readback_slice.map_async(wgpu::MapMode::Read).then(|_| async {
            h.copy_from_slice(unsafe {
                from_bytes(h_readback_slice.get_mapped_range().as_ref())
            });
            h_readback_buffer.unmap();
        });

        let m_readback_slice = m_readback_buffer.slice(..);
        let m_readback_future = m_readback_slice.map_async(wgpu::MapMode::Read).then(|_| async {
            m.copy_from_slice(unsafe {
                from_bytes(m_readback_slice.get_mapped_range().as_ref())
            });
            m_readback_buffer.unmap();
        });


        let timestamp_period = queue.get_timestamp_period();
        let timestamp_readback_slice = timestamp_buffer.slice(..);
        let timestamp_readback_future = timestamp_readback_slice.map_async(wgpu::MapMode::Read).then(|_| async {
            let (e_time, h_time) = {
                let mapped = timestamp_readback_slice.get_mapped_range();
                let timings: &[u64] = unsafe {
                    from_bytes(mapped.as_ref())
                };

                let (mut e_time, mut h_time) = (0, 0);
                for frame in timings.chunks(4) {
                    e_time += frame[1] - frame[0];
                    h_time += frame[3] - frame[2];
                }

                (
                    Duration::from_nanos((e_time as f64 * timestamp_period as f64) as u64),
                    Duration::from_nanos((h_time as f64 * timestamp_period as f64) as u64),
                )
            };
            timestamp_buffer.unmap();
            diag.record_step_device(e_time + h_time);
        });

        // optimization note: it may be possible to use `WaitForSubmission`
        // and copy data to/from even as the GPU begins executing the next job.
        device.poll(wgpu::Maintain::Wait);

        diag.instrument_read_device(move || {
            futures::executor::block_on(futures::future::join(
                e_readback_future, futures::future::join(
                    h_readback_future, futures::future::join(
                        m_readback_future, timestamp_readback_future
                    )
                )
            ));
        });
    }
}

/// Convert an arbitrary slice into a byte slice
fn to_bytes<T>(slice: &[T]) -> &[u8] {
    unsafe {
        std::slice::from_raw_parts(slice.as_ptr() as *const u8, slice.len() * std::mem::size_of::<T>())
    }
}

/// Convert a byte slice into a T slice
unsafe fn from_bytes<T>(slice: &[u8]) -> &[T] {
    let elem_size = std::mem::size_of::<T>();
    let new_len = slice.len() / elem_size;
    assert_eq!(new_len * elem_size, slice.len());
    std::slice::from_raw_parts(slice.as_ptr() as *const T, new_len)
}

/// Loads the shader
fn get_shader() -> wgpu::ShaderModuleDescriptorSpirV<'static> {
    let data = spirv_backend_runner::spirv_module();
    let spirv = Cow::Owned(wgpu::util::make_spirv_raw(&data).into_owned());

    wgpu::ShaderModuleDescriptorSpirV {
        label: None,
        source: spirv,
    }
}

async fn open_device(max_buf_size: u64) -> (wgpu::Device, wgpu::Queue) {
    // based on rust-gpu/examples/runners/wgpu/src/compute.rs:start_internal
    let instance = wgpu::Instance::new(wgpu::Backends::PRIMARY);
    info!("open_device: got instance");
    let adapter = instance
        .request_adapter(&wgpu::RequestAdapterOptions {
            power_preference: wgpu::PowerPreference::HighPerformance,
            force_fallback_adapter: false,
            compatible_surface: None,
        })
        .await
        .expect("Failed to find an appropriate adapter");
    info!("open_device: got adapter");

    // XXX not all adapters will support non-default limits, and it could
    // cause perf degradations even on the ones that do. May want to consider
    // folding some buffers together to avoid this.
    let mut limits = wgpu::Limits::default();
    //limits.max_bind_groups = 5;
    //limits.max_dynamic_storage_buffers_per_pipeline_layout = 5;
    limits.max_storage_buffers_per_shader_stage = 7;
    //limits.max_storage_buffer_binding_size = 128 MiB.
    //limits.max_storage_buffer_binding_size = 1024 * (1 << 20);
    limits.max_storage_buffer_binding_size = max_buf_size.try_into().unwrap();

    let (device, queue) = adapter
        .request_device(
            &wgpu::DeviceDescriptor {
                label: None,
                features: (
                    wgpu::Features::empty()
                        .union(wgpu::Features::SPIRV_SHADER_PASSTHROUGH)
                        .union(wgpu::Features::TIMESTAMP_QUERY)
                ),
                limits,
            },
            None,
        )
        .await
        .expect("Failed to create device");
    info!("open_device: got device");
    (device, queue)
}

fn make_pipelines(
    device: &wgpu::Device,
    shader_module: &wgpu::ShaderModule,
    entry_step_h: &'static str,
    entry_step_e: &'static str
) -> (
    wgpu::BindGroupLayout, wgpu::ComputePipeline, wgpu::ComputePipeline
) {
    let bind_group_layout = device.create_bind_group_layout(&wgpu::BindGroupLayoutDescriptor {
        label: None,
        entries: &[
            wgpu::BindGroupLayoutEntry {
                // meta
                binding: 0,
                count: None,
                visibility: wgpu::ShaderStages::COMPUTE,
                ty: wgpu::BindingType::Buffer {
                    has_dynamic_offset: false,
                    min_binding_size: Some(NonZeroU64::new(1).unwrap()),
                    ty: wgpu::BufferBindingType::Storage { read_only: true },
                },
            },
            wgpu::BindGroupLayoutEntry {
                // stimulus(e)
                binding: 1,
                count: None,
                visibility: wgpu::ShaderStages::COMPUTE,
                ty: wgpu::BindingType::Buffer {
                    has_dynamic_offset: false,
                    min_binding_size: Some(NonZeroU64::new(1).unwrap()),
                    ty: wgpu::BufferBindingType::Storage { read_only: true },
                },
            },
            wgpu::BindGroupLayoutEntry {
                // stimulus(h)
                binding: 2,
                count: None,
                visibility: wgpu::ShaderStages::COMPUTE,
                ty: wgpu::BindingType::Buffer {
                    has_dynamic_offset: false,
                    min_binding_size: Some(NonZeroU64::new(1).unwrap()),
                    ty: wgpu::BufferBindingType::Storage { read_only: true },
                },
            },
            wgpu::BindGroupLayoutEntry {
                // materials
                binding: 3,
                count: None,
                visibility: wgpu::ShaderStages::COMPUTE,
                ty: wgpu::BindingType::Buffer {
                    has_dynamic_offset: false,
                    min_binding_size: Some(NonZeroU64::new(1).unwrap()),
                    ty: wgpu::BufferBindingType::Storage { read_only: true },
                },
            },
            wgpu::BindGroupLayoutEntry {
                // e field
                binding: 4,
                count: None,
                visibility: wgpu::ShaderStages::COMPUTE,
                ty: wgpu::BindingType::Buffer {
                    has_dynamic_offset: false,
                    min_binding_size: Some(NonZeroU64::new(1).unwrap()),
                    ty: wgpu::BufferBindingType::Storage { read_only: false },
                },
            },
            wgpu::BindGroupLayoutEntry {
                // h field
                binding: 5,
                count: None,
                visibility: wgpu::ShaderStages::COMPUTE,
                ty: wgpu::BindingType::Buffer {
                    has_dynamic_offset: false,
                    min_binding_size: Some(NonZeroU64::new(1).unwrap()),
                    ty: wgpu::BufferBindingType::Storage { read_only: false },
                },
            },
            wgpu::BindGroupLayoutEntry {
                // m field
                binding: 6,
                count: None,
                visibility: wgpu::ShaderStages::COMPUTE,
                ty: wgpu::BindingType::Buffer {
                    has_dynamic_offset: false,
                    min_binding_size: Some(NonZeroU64::new(1).unwrap()),
                    ty: wgpu::BufferBindingType::Storage { read_only: false },
                },
            },
        ],
    });

    let pipeline_layout = device.create_pipeline_layout(&wgpu::PipelineLayoutDescriptor {
        label: None,
        bind_group_layouts: &[&bind_group_layout],
        push_constant_ranges: &[],
    });

    let compute_step_h_pipeline = device.create_compute_pipeline(&wgpu::ComputePipelineDescriptor {
        label: None,
        layout: Some(&pipeline_layout),
        module: shader_module,
        entry_point: entry_step_h,
    });

    let compute_step_e_pipeline = device.create_compute_pipeline(&wgpu::ComputePipelineDescriptor {
        label: None,
        layout: Some(&pipeline_layout),
        module: shader_module,
        entry_point: entry_step_e,
    });

    (bind_group_layout, compute_step_h_pipeline, compute_step_e_pipeline)
}

