use std::borrow::Cow;

use serde::{Deserialize, Serialize};
use log::{info, warn};

use crate::diagnostics::SyncDiagnostics;
use crate::geom::Index;
use crate::real::Real;
use crate::sim::{AbstractSim, Fields, GenericSim};
use crate::stim::{RenderedStimulus, Stimulus};
use coremem_cross::mat::{GenericMaterial, Material};
use coremem_cross::step::SimMeta;
use coremem_cross::vec::Vec3;

mod cpu;
mod gpu;
pub use cpu::CpuBackend;
pub use gpu::WgpuBackend;

/// while the SpirvSim type owns all the stateful data,
/// it dispatches to a pluggable backend for advancing that state.
///
/// this backend type abstracts over the specific compute used to advance the state.
/// it could advance the state locally, on the CPU,
/// or it could dispatch to a SPIR-V device, like a GPU.
pub trait SimBackend<R, M> {
    fn step_n(
        &mut self,
        diag: &SyncDiagnostics,
        meta: SimMeta<R>,
        mat: &[M],
        stim_e: &[Vec3<R>],
        stim_h: &[Vec3<R>],
        e: &mut [Vec3<R>],
        h: &mut [Vec3<R>],
        m: &mut [Vec3<R>],
        num_steps: u32,
    );
}

/// Wrapper around an inner state object which offloads stepping onto a spirv backend (e.g. GPU).
#[derive(Default, Serialize, Deserialize)]
// TODO: remove default R/M
pub struct SpirvSim<R=f32, M=GenericMaterial<R>, B=WgpuBackend>
where M: 'static
{
    meta: SimMeta<R>,
    e: Vec<Vec3<R>>,
    h: Vec<Vec3<R>>,
    m: Vec<Vec3<R>>,
    mat: Vec<M>,
    step_no: u64,
    #[serde(skip)]
    backend: B,
    #[serde(skip)]
    diag: SyncDiagnostics,
}

// B isn't always clonable (e.g. gpu backend) so rust can't auto-derive this.
impl<R: Clone, M: Clone, B: Default> Clone for SpirvSim<R, M, B> {
    fn clone(&self) -> Self {
        Self {
            meta: self.meta.clone(),
            e: self.e.clone(),
            h: self.h.clone(),
            m: self.m.clone(),
            mat: self.mat.clone(),
            step_no: self.step_no.clone(),
            // backends can be expensive to clone.
            // instead we default-init them and they're expected to acquire their resources on
            // first query.
            backend: Default::default(),
            diag: self.diag.clone(),
        }
    }
}

impl<R, M, B> AbstractSim for SpirvSim<R, M, B>
where
    R: Real,
    M: Send + Sync + Material<R> + Clone + Into<GenericMaterial<R>>,
    B: Send + Sync + SimBackend<R, M>,
{
    type Real = R;
    type Material = M;

    fn meta(&self) -> SimMeta<f32> {
        self.meta.cast()
    }
    fn step_no(&self) -> u64 {
        self.step_no
    }

    fn put_material_index(&mut self, pos: Index, mat: M) {
        let flat_idx = self.flat_index(pos).unwrap();
        self.mat[flat_idx] = mat;
    }

    fn get_material_index(&self, pos: Index) -> &M {
        let flat_idx = self.flat_index(pos).unwrap();
        &self.mat[flat_idx]
    }

    fn step_multiple<S: Stimulus<R>>(&mut self, num_steps: u32, stim: &S) {
        let stim = self.eval_stimulus(stim);
        let (stim_e, stim_h) = (stim.e().into_inner(), stim.h().into_inner());
        self.backend.step_n(
            &self.diag,
            self.meta,
            self.mat.as_slice(),
            &*stim_e,
            &*stim_h,
            self.e.as_mut_slice(),
            self.h.as_mut_slice(),
            self.m.as_mut_slice(),
            num_steps,
        );
        self.step_no += num_steps as u64;
    }

    fn fields_at_index(&self, pos: Index) -> Fields<R> {
        let flat_idx = self.flat_index(pos);
        match flat_idx {
            Some(idx) => Fields::new(
                self.e[idx], self.h[idx], self.m[idx]
            ),
            None => Default::default(),
        }
    }

    fn to_generic(&self) -> GenericSim<Self::Real> {
        let mat = self.mat.iter().map(|m| m.clone().into()).collect();
        GenericSim {
            meta: self.meta,
            e: self.e.clone(),
            h: self.h.clone(),
            m: self.m.clone(),
            mat,
            step_no: self.step_no,
            backend: Default::default(),
            diag: self.diag.clone(),
        }
    }

    fn use_diagnostics(&mut self, diag: SyncDiagnostics) {
        self.diag = diag;
    }
}

impl<R: Real, M: Default, B> SpirvSim<R, M, B>
{
    pub fn new_with_backend(size: Index, feature_size: f32, backend: B) -> Self {
        info!("SpirvSim::new({:?}, {})", size, feature_size);
        let flat_size = size.volume() as usize;
        if flat_size * std::mem::size_of::<M>() >= 0x40000000 {
            warn!("simulation size ({:?}) is greater than what is tested for: anything could happen!", size);
        }
        let feature_size = feature_size.cast::<R>();
        let courant = 0.577.cast::<R>();
        let time_step = feature_size / R::c() * courant;
        let mut mat = Vec::new();
        mat.resize_with(flat_size, Default::default);
        Self {
            meta: SimMeta::new(*size, feature_size, time_step),
            e: vec![Vec3::zero(); flat_size],
            h: vec![Vec3::zero(); flat_size],
            m: vec![Vec3::zero(); flat_size],
            mat,
            step_no: 0,
            backend,
            diag: SyncDiagnostics::new(),
        }
    }

    /// test helper for when stimuli can't directly be used for one reason or another
    #[cfg(test)]
    pub(crate) fn set_fields_at_index(&mut self, pos: Index, f: Fields<R>) {
        let idx = self.flat_index(pos).unwrap();
        self.e[idx] = f.e();
        self.h[idx] = f.h();
        self.m[idx] = f.m();
    }
}

impl<R: Real, M: Default, B: Default> SpirvSim<R, M, B>
{
    pub fn new(size: Index, feature_size: f32) -> Self {
        Self::new_with_backend(size, feature_size, B::default())
    }
}

impl<R, M, B> SpirvSim<R, M, B> {
    fn flat_index(&self, idx: Index) -> Option<usize> {
        // TODO: port to DimSlice
        let dim = self.meta.dim();
        if idx.x() < dim.x() && idx.y() < dim.y() && idx.z() < dim.z() {
            Some(((idx.z()*dim.y() + idx.y())*dim.x() + idx.x()) as usize)
        } else {
            None
        }
    }
}

impl<R, M, B> SpirvSim<R, M, B>
where
    R: Real,
    M: Send + Sync + Material<R> + Clone + Into<GenericMaterial<R>>,
    B: Send + Sync + SimBackend<R, M>,
{
    fn eval_stimulus<'a, S: Stimulus<R>>(&self, stim: &'a S)
        -> Cow<'a, RenderedStimulus<R>>
    {
        self.diag.instrument_stimuli_blocked(move || {
            let dim = self.size();
            let feature_size = self.feature_size();
            let t_sec = self.time();
            let timestep = self.meta().time_step();

            let stim = stim.rendered(timestep.cast(), t_sec.cast(), feature_size.cast(), dim.into());
            assert_eq!(stim.e().dim(), dim.into());
            assert_eq!(stim.h().dim(), dim.into());
            stim
        })
    }
}


#[cfg(test)]
mod test {
    use super::*;
    use crate::mat::{self, AnisomorphicConductor};
    use crate::meas::Energy;
    use crate::real::{R32, ToFloat as _};
    use crate::stim::{
        self,
        ModulatedVectorField,
        Sinusoid,
        TimeVaryingExt as _,
        VectorField,
    };
    fn mean_magnitude_e<S: AbstractSim>(sim: &S) -> R32 {
        let m = sim.map_sum_enumerated(|_pos: Index, cell| {
            cell.e().mag().to_r64()
        })/(sim.width() * sim.height() * sim.depth()).to_r64();
        m.cast()
    }
    fn mean_magnitude_h<S: AbstractSim>(sim: &S) -> R32 {
        let m = sim.map_sum_enumerated(|_pos: Index, cell| {
            cell.h().mag().to_r64()
        })/(sim.width() * sim.height() * sim.depth()).to_r64();
        m.cast()
    }
    fn mean_magnitude_m<S: AbstractSim>(sim: &S) -> R32 {
        let m = sim.map_sum_enumerated(|_pos: Index, cell| {
            cell.m().mag().to_r64()
        })/(sim.width() * sim.height() * sim.depth()).to_r64();
        m.cast()
    }

    fn assert_simstate_eq<S0: AbstractSim<Real=R32>, S1: AbstractSim<Real=R32>>(
        real_state: &S0, ref_state: &S1
    ) {
        let rel_threshold = 1e-2f32.to_r32();
        let mean_e = mean_magnitude_e(ref_state);
        let mean_h = mean_magnitude_h(ref_state);
        let mean_m = mean_magnitude_m(ref_state);
        let dim = real_state.size();

        for z in 0..dim.z() {
            for y in 0..dim.y() {
                for x in 0..dim.x() {
                    let idx = Index::new(x, y, z);
                    let real_cell = real_state.sample(idx);
                    let ref_cell = ref_state.sample(idx);

                    let diff_e = real_cell.e() - ref_cell.e();
                    let diff_h = real_cell.h() - ref_cell.h();
                    let diff_m = real_cell.m() - ref_cell.m();

                    // if x == 1 && y == 1 && z == 0 {
                    // if ref_cell.m() != Vec3::zero() {
                    //     println!("e {} v.s. {}", real_cell.e(), ref_cell.e());
                    //     println!("h {} v.s. {}", real_cell.h(), ref_cell.h());
                    //     println!("m {} v.s. {}", real_cell.m(), ref_cell.m());
                    //     println!("diff_e: {}; diff_h: {}, diff_m: {}", diff_e, diff_h, diff_m);
                    // }
                    assert!(
                        diff_e.mag() <= mean_e * rel_threshold,
                        "idx.e: {:?}. {:?} v.s. {:?} at step {}",
                        idx, real_cell.e(), ref_cell.e(), ref_state.step_no()
                    );

                    assert!(
                        diff_h.mag() <= mean_h * rel_threshold,
                        "idx.h: {:?}. {:?} v.s. {:?} at step {}",
                        idx, real_cell.h(), ref_cell.h(), ref_state.step_no()
                    );

                    assert!(
                        diff_m.mag() <= mean_m * rel_threshold || diff_m.mag() <= ref_cell.m().mag(),
                        "idx.m: {:?}. {:?} v.s. {:?} at step {}",
                        idx, real_cell.m(), ref_cell.m(), ref_state.step_no()
                    );
                }
            }
        }
    }


    /// used in a ModulatedVectorField to apply a modulation to a *specific* index
    struct IndexGatedEField {
        loc: Index,
    }

    impl<R: Real> VectorField<R> for IndexGatedEField {
        fn at(&self, _feat_size: R, loc: Index) -> stim::Fields<R> {
            if loc == self.loc {
                stim::Fields::new_e(Vec3::new(0.0, 0.0, 1.0).cast())
            } else {
                stim::Fields::default()
            }
        }
    }

    /// returns a Stimulus which will create a smooth pulse at t=0 through to t=<frames>,
    ///   at only the provided location.
    /// caller can expect that its energy content is relatively focused to just the wavelength
    /// `frames`.
    fn smooth_pulse_at(timestep: f32, feat_size: f32, loc: Index, frames: u32) -> impl Stimulus<R32> {
        let wl = (timestep * frames as f32).to_r32();
        let quarter_wl = wl * (0.25).to_r32();
        let amp = (timestep*feat_size*feat_size).inv().to_r32();
        let window = Sinusoid::from_wavelength(wl)
            .shifted(quarter_wl)
            .summed(R32::one());
        ModulatedVectorField::new(
            IndexGatedEField { loc },
            Sinusoid::from_wavelength(wl).scaled(amp).scaled(window),
        )
    }

    /// step by ones, applying the stimulus at each step, and return the energy at the end.
    fn step_n<S: AbstractSim, Stim: Stimulus<S::Real>>(state: &mut S, stim: &Stim, frames: u32) -> f32 {
        for _ in 0..frames {
            state.step_multiple(1, stim);
        }
        Energy::world().data(state)
    }


    macro_rules! test_backend {
        ($shortname:ident, $backend:ty) => {
            mod $shortname {
                use super::*;

                use crate::geom::WorldRegion;
                use crate::stim::{NoopStimulus, RngStimulus};
                use coremem_cross::mat::{AnisomorphicConductor, IsomorphicConductor};
                use float_eq::assert_float_eq;
                use more_asserts::{assert_gt, assert_lt};

                ////////// SMOKE/SANITY TESTS //////////
                // these test that everything is roughly wired together right, but don't ensure
                // that the simulations will particularly reflect reality.

                #[test]
                fn accessors() {
                    let size = Index::new(15, 17, 19);
                    let mut state = SpirvSim::<R32, GenericMaterial<R32>, $backend>::new(size, 1e-6);
                    assert_eq!(state.width(), 15);
                    assert_eq!(state.height(), 17);
                    assert_eq!(state.depth(), 19);
                    assert_eq!(state.time(), 0.0);
                    // feat_size / C * courant: 1e-6 / 3e8 * 0.577
                    assert_eq!(state.timestep(), 0.0000000000000019246651);
                    state.step();
                    assert_eq!(state.time(), state.timestep());
                }

                #[test]
                fn energy_conservation_over_time() {
                    let mut state = SpirvSim::<R32, GenericMaterial<R32>, $backend>::new(
                        Index((2001, 1, 1).into()), 1e-6
                    );
                    let stim = smooth_pulse_at(state.timestep(), state.feature_size(), Index::new(1000, 0, 0), 100);

                    let energy_0 = step_n(&mut state, &stim, 100);
                    let energy_1 = step_n(&mut state, &NoopStimulus, 800);

                    assert_float_eq!(energy_1, energy_0, r2nd <= 1e-4);
                    // TODO: assert that energy is something reasonable to begin with, and not 0.
                }

                #[test]
                fn sane_boundary_conditions() {
                    let mut state = SpirvSim::<R32, GenericMaterial<R32>, $backend>::new(
                        Index((21, 21, 21).into()), 1e-6
                    );
                    let stim = smooth_pulse_at(state.timestep(), state.feature_size(), Index::new(10, 10, 10), 40);

                    let energy_0 = step_n(&mut state, &stim, 40);
                    let energy_1 = step_n(&mut state, &NoopStimulus, 500);

                    // Default boundary conditions reflect waves.
                    // but maybe not *perfectly*
                    assert_lt!(energy_1, energy_0);
                    assert_gt!(energy_1, 0.5 * energy_0);
                }

                /// Fill the world with the provided material and a stimulus.
                /// Measure energy at the start, and then again after advancing many steps.
                /// Return these two measurements (energy(t=0), energy(t=~=1000))
                fn conductor_test<M: Into<GenericMaterial<R32>>>(mat: M) -> (f32, f32) {
                    let mut state = SpirvSim::<R32, GenericMaterial<R32>, $backend>::new(
                        Index::new(201, 1, 1), 1e-6
                    );
                    state.fill_region(&WorldRegion, mat.into());
                    let stim = smooth_pulse_at(state.timestep(), state.feature_size(), Index::new(100, 0, 0), 100);

                    let energy_0 = step_n(&mut state, &stim, 100);
                    let energy_1 = step_n(&mut state, &NoopStimulus, 1000);

                    (energy_0, energy_1)
                }

                #[test]
                fn conductor_dissipates_energy() {
                    let (energy_0, energy_1) = conductor_test(IsomorphicConductor::new(1e3.cast()));
                    assert_float_eq!(energy_1/energy_0, 0.0, abs <= 1e-6);
                }

                #[test]
                fn anisotropic_conductor_inactive_x() {
                    let (energy_0, energy_1) = conductor_test(AnisomorphicConductor::new(
                        Vec3::new_x(1e3).cast()
                    ));
                    assert_gt!(energy_1, 0.9*energy_0);
                    // XXX: I think this gains energy because we only set the E field and not the H field
                    assert_lt!(energy_1, 1.8*energy_0);
                }

                #[test]
                fn anisotropic_conductor_inactive_y() {
                    let (energy_0, energy_1) = conductor_test(AnisomorphicConductor::new(
                        Vec3::new_y(1e3).cast()
                    ));
                    assert_gt!(energy_1, 0.9*energy_0);
                    assert_lt!(energy_1, 1.8*energy_0);
                }

                #[test]
                fn anisotropic_conductor_active_z() {
                    let (energy_0, energy_1) = conductor_test(AnisomorphicConductor::new(
                        Vec3::new_z(1e3).cast()
                    ));
                    assert_float_eq!(energy_1/energy_0, 0.0, abs <= 1e-6);
                }

                #[test]
                fn smoke_small() {
                    let mut state = SpirvSim::<R32, GenericMaterial<R32>, $backend>::new(
                        Index::new(8, 8, 8), 1e-3
                    );
                    state.step();
                }
                #[test]
                fn smoke_med_7bit() {
                    let mut state = SpirvSim::<R32, GenericMaterial<R32>, $backend>::new(
                        Index::new(124, 124, 124), 1e-3
                    );
                    state.step();
                }
                #[test]
                fn smoke_med128() {
                    let mut state = SpirvSim::<R32, GenericMaterial<R32>, $backend>::new(
                        Index::new(128, 128, 128), 1e-3
                    );
                    state.step();
                }
                #[test]
                fn smoke_med_23bit() {
                    let mut state = SpirvSim::<R32, GenericMaterial<R32>, $backend>::new(
                        Index::new(127, 256, 256), 1e-3
                    );
                    state.step();
                }
                #[test]
                fn smoke_med_0x800000_indexing() {
                    let mut state = SpirvSim::<R32, GenericMaterial<R32>, $backend>::new(
                        Index::new(128, 256, 256), 1e-3
                    );
                    state.step();
                }
                #[test]
                fn smoke_med_0x800000_address_space() {
                    let mut state = SpirvSim::<R32, GenericMaterial<R32>, $backend>::new(
                        Index::new(170, 256, 256), 1e-3
                    );
                    state.step();
                }
                #[test]
                fn smoke_large() {
                    let mut state = SpirvSim::<R32, GenericMaterial<R32>, $backend>::new(
                        Index::new(326, 252, 160), 1e-3
                    );
                    state.step();
                }
                #[test]
                fn smoke_not_multiple_of_4() {
                    let mut state = SpirvSim::<R32, GenericMaterial<R32>, $backend>::new(
                        Index::new(3, 2, 5), 1e-3
                    );
                    state.step();
                }

                fn mh_ferromagnet_smoke_test(
                    seed: u64, steps: u64, size: Index
                ) {
                    // XXX This doesn't do anything, except make sure we don't crash!
                    use rand::{Rng as _, SeedableRng as _};
                    let mut rng = rand::rngs::StdRng::seed_from_u64(seed);
                    let mut dut_state = SpirvSim::<R32, GenericMaterial<R32>, $backend>::new(size, 1e-3);

                    for z in 0..size.z() {
                        for y in 0..size.y() {
                            for x in 0..size.x() {
                                if rng.gen_range(0.0..1.0) < 0.5 {
                                    let h_start = rng.gen_range(1.0..1000.0);
                                    let mu_r = rng.gen_range(1.1..100000.0);
                                    let m_max = rng.gen_range(0.0..1_000_000.0);
                                    dut_state.put_material(Index::new(x, y, z), mat::MHPgram::new(
                                        h_start.cast(), mu_r.cast(), m_max.cast()
                                    ));
                                } else if rng.gen_range(0.0..1.0) < 0.5 {
                                    dut_state.put_material(Index::new(x, y, z), mat::IsomorphicConductor::new(
                                        rng.gen_range(0.0..1e7).cast()
                                    ));
                                }
                            }
                        }
                    }

                    let stim = RngStimulus::new(seed);
                    for step in 0..steps {
                        println!("step {}", step);
                        dut_state.step_multiple(1, &stim);
                    }
                }

                #[test]
                fn mh_ferromagnet_smoke_1_step() {
                    mh_ferromagnet_smoke_test(0x1234, 1, Index::new(4, 4, 4));
                }
                #[test]
                fn mh_ferromagnet_smoke_100_steps_larger() {
                    mh_ferromagnet_smoke_test(0x1234, 100, Index::new(328, 252, 160));
                }

                ////////// APPLICATION TESTS //////////
                // these test that the simulation accurately models *specific* applications

                /// returns the (E, H) value of the first cycle of a rightward-traveling ray at a given point in space/time.
                /// E amplitude of the ray is assumed to be 1; H is scaled appropriately to ensure that.
                fn ray(z_origin: i32, wavelength_cells: u32, sim: &SpirvSim<R32, GenericMaterial<R32>, $backend>, z: u32, frame: u32) -> (R32, R32) {
                    let time_step = sim.timestep().cast::<R32>();
                    let feat_size = sim.feature_size().cast::<R32>();

                    let amp_e = R32::one();
                    let amp_h = R32::c_inv() * R32::mu0_inv();
                    let t_shift = R32::from_primitive(frame) * time_step/feat_size * R32::c();
                    let idx = (z as f32 - z_origin as f32 - t_shift.cast::<f32>()) / (wavelength_cells as f32);
                    if idx < 0.0 || idx > 1.0 {
                        return (R32::zero(), R32::zero());
                    }

                    let phase = idx.cast::<R32>() * R32::two_pi();
                    let a = phase.sin();
                    (amp_e * a, amp_h * a)
                }

                /// inject one cycle of a rightward-traveling ray with the provided parameters into
                /// the sim.
                fn inject_ray(z_origin: i32, wavelength_cells: u32, sim: &mut SpirvSim<R32, GenericMaterial<R32>, $backend>) {
                    // inject the wave at t=0
                    for z in 0..sim.size().z() {
                        let (e, h) = ray(z_origin, wavelength_cells, &sim, z, 0);
                        // due to the staggered nature of stepping (step_e, then step_h),
                        // populating the initial E *and* H fields through a Stimulus is
                        // nontrivial. we just explicitly set the fields instead.
                        sim.set_fields_at_index(Index::new(0, 0, z), Fields::new(
                            Vec3::new(e, R32::zero(), R32::zero()),
                            Vec3::new(R32::zero(), h, R32::zero()),
                            Vec3::default(),
                        ));
                    }
                }

                /// verify that an EM ray propagates through the vacuum as expected
                #[test]
                fn ray_propagation() {
                    let size = Index::new(1, 1, 1536);
                    let feat_size = 1e-3;
                    let mut sim = SpirvSim::<R32, GenericMaterial<R32>, $backend>::new(size, feat_size);

                    inject_ray(512, 512, &mut sim);

                    // advance the simulation N steps.
                    // at each step, the wave we injected should have been shifted by some known amount,
                    // and we assert that the wave is where we expect it (and with the right
                    // amplitude)
                    for t in 1..512 {
                        sim.step();
                        let eh: Vec<_> = (0..size.z()).into_iter().map(|z| {
                            let f = sim.fields_at_index(Index::new(0, 0, z));
                            (f.e().x().cast::<f32>(), f.h().y().cast::<f32>())
                        }).collect();
                        for (z, (e, h)) in eh.iter().enumerate() {
                            let (exp_e, exp_h) = ray(512, 512, &sim, z as u32, t);
                            assert_float_eq!(*e, exp_e.cast::<f32>(), abs <= 1e-2, "t={}, z={} ... {:?}", t, z, eh);
                            assert_float_eq!(*h, exp_h.cast::<f32>(), abs <= 1e-4, "t={}, z={} ... {:?}", t, z, eh);
                        }
                    }
                }

                /// reflect an EM ray off a conductor and make sure the result is as expected
                #[test]
                fn conductor_reflection() {
                    let size = Index::new(1, 1, 2048);
                    let feat_size = 1e-3;
                    let mut sim = SpirvSim::<R32, GenericMaterial<R32>, $backend>::new(size, feat_size);

                    inject_ray(512, 512, &mut sim);

                    sim.put_material(Index::new(0, 0, 1536), mat::IsomorphicConductor::new(1e9.cast()));

                    // 512 cells for the ray to propagate *toward* the conductor,
                    // 512 cells for it to reflect,
                    // 512 cells for it to propagate back from the conductor.
                    // the wave crosses 0.577 cells per step (Courant value), so roughly 2662 steps
                    // for the wave to return home.
                    for t in 0..2662 {
                        sim.step();
                    }

                    // now the wave should have returned to the starting point
                    let eh: Vec<_> = (0..size.z()).into_iter().map(|z| {
                        let f = sim.fields_at_index(Index::new(0, 0, z));
                        (f.e().x().cast::<f32>(), f.h().y().cast::<f32>())
                    }).collect();
                    for (z, (e, h)) in eh.iter().enumerate() {
                        let (ray_e, ray_h) = ray(512, 512, &sim, z as u32, 0);
                        // a conductor forces E=0, hence the head of the return E wave is the
                        // negative of the head of the origin E wave.
                        // because of the opposite orientation, this actually negates again the
                        // return E wave evaluates to the same as the origin E wave.
                        // however, the return H wave is only flipped once, because of the
                        // orientation.
                        let (exp_e, exp_h) = (ray_e, -ray_h);
                        assert_float_eq!(*e, exp_e.cast::<f32>(), abs <= 2e-2, "z={} ... {:?}", z, eh);
                        assert_float_eq!(*h, exp_h.cast::<f32>(), abs <= 2e-4, "z={} ... {:?}", z, eh);
                    }
                }
            }
        }
    }

    test_backend!(cpu, CpuBackend);
    test_backend!(wgpu, WgpuBackend);

    /// sanity checks that cpu and wgpu backends give *similar* results
    mod same {
        use super::*;
        use crate::stim::{NoopStimulus, RngStimulus};
        fn test_same_explicit<S: Stimulus<R32>>(
            mut cpu_state: SpirvSim<R32, GenericMaterial<R32>, CpuBackend>,
            mut wgpu_state: SpirvSim<R32, GenericMaterial<R32>, WgpuBackend>,
            stim: &S,
            step_iters: u64,
            steps_per_iter: u32
        ) {
            for _ in 0..step_iters {
                cpu_state.step_multiple(steps_per_iter, stim);
                wgpu_state.step_multiple(steps_per_iter, stim);
                assert_simstate_eq(&wgpu_state, &cpu_state);
            }
        }
        fn test_same(seed: u64, step_iters: u64, steps_per_iter: u32, size: Index) {
            let cpu_state = SpirvSim::new_with_backend(size, 1e-3, CpuBackend::default());

            let wgpu_state = SpirvSim::new_with_backend(size, 1e-3, WgpuBackend::default());
            let stim = RngStimulus::new(seed);

            test_same_explicit(cpu_state, wgpu_state, &stim, step_iters, steps_per_iter);
        }

        #[test]
        fn same_no_step_no_stim() {
            let cpu_state = SpirvSim::new_with_backend(Index::new(8, 8, 8), 1e-3, CpuBackend::default());
            let wgpu_state = SpirvSim::new_with_backend(Index::new(8, 8, 8), 1e-3, WgpuBackend::default());
            test_same_explicit(cpu_state, wgpu_state, &NoopStimulus, 1, 1);
        }

        #[test]
        fn do_same_1_step() {
            test_same(0x1234, 1, 1, Index::new(4, 4, 4));
        }

        #[test]
        fn same_1000_step() {
            test_same(0x1234, 1000, 1, Index::new(4, 4, 4));
        }

        #[test]
        fn same_not_multiple_of_4() {
            test_same(0x1234, 100, 1, Index::new(3, 2, 5));
        }

        #[test]
        fn same_100_steps_larger() {
            test_same(0x1234, 100, 1, Index::new(24, 20, 44));
        }

        #[test]
        fn same_100_steps_of_10() {
            test_same(0x1234, 100, 10, Index::new(24, 20, 44));
        }

        fn test_same_conductor(
            seed: u64, step_iters: u64, steps_per_iter: u32, size: Index
        ) {
            use rand::{Rng as _, SeedableRng as _};
            let mut rng = rand::rngs::StdRng::seed_from_u64(seed);
            let mut cpu_state = SpirvSim::new_with_backend(size, 1e-3, CpuBackend::default());
            let mut wgpu_state = SpirvSim::new_with_backend(size, 1e-3, WgpuBackend::default());

            for z in 0..size.z() {
                for y in 0..size.y() {
                    for x in 0..size.x() {
                        let cond = Vec3::new(
                            rng.gen_range(0.0..1e6),
                            rng.gen_range(0.0..1e6),
                            rng.gen_range(0.0..1e6),
                        );
                        cpu_state.put_material(Index::new(x, y, z), AnisomorphicConductor::new(cond.cast()));
                        wgpu_state.put_material(Index::new(x, y, z), AnisomorphicConductor::new(cond.cast()));
                    }
                }
            }

            let stim = RngStimulus::new(seed);

            test_same_explicit(cpu_state, wgpu_state, &stim, step_iters, steps_per_iter);
        }

        #[test]
        fn conductor_1_step() {
            test_same_conductor(0x1234, 1, 1, Index::new(4, 4, 4));
        }

        #[test]
        fn conductor_many_steps_larger() {
            test_same_conductor(0x1234, 100, 10, Index::new(96, 16, 8));
        }

        fn test_same_mb_ferromagnet(
            seed: u64, step_iters: u64, steps_per_iter: u32, size: Index
        ) {
            use rand::{Rng as _, SeedableRng as _};
            let mut rng = rand::rngs::StdRng::seed_from_u64(seed);
            let mut cpu_state = SpirvSim::new_with_backend(size, 1e-3, CpuBackend::default());
            let mut wgpu_state = SpirvSim::new_with_backend(size, 1e-3, WgpuBackend::default());

            for z in 0..size.z() {
                for y in 0..size.y() {
                    for x in 0..size.x() {
                        let b_start = rng.gen_range(-1e-4..1e-4);
                        // if b_start is negative, we need the endpoint to be > |b_start|:
                        let b_range = (-2.0f32*b_start).max(0.0) + rng.gen_range(1e-5..1e-2);
                        // It would be weird for an increase in B to trigger a DECREASE in H.
                        // We have mu0(M+H) = B, so
                        // mu0(dM/dB = dH/dB) = 1
                        //   -> dH/dB = mu0^-1 - dM/dB > 0
                        let slope = rng.gen_range(0.0..0.9*f32::mu0_inv());
                        let b_end = b_start + b_range;
                        let m_max = b_range * slope;
                        let mat = coremem_cross::mat::MBPgram::new(
                            b_start.cast(), b_end.cast(), m_max.cast()
                        );
                        cpu_state.put_material(Index::new(x, y, z), mat.clone());
                        wgpu_state.put_material(Index::new(x, y, z), mat.clone());
                    }
                }
            }

            let stim = RngStimulus::new(seed);

            test_same_explicit(cpu_state, wgpu_state, &stim, step_iters, steps_per_iter);
        }

        #[test]
        fn mb_ferromagnet_1_step() {
            test_same_mb_ferromagnet(0x1234, 1, 1, Index::new(4, 4, 4));
        }

        #[test]
        fn mb_ferromagnet_50_steps_larger() {
            test_same_mb_ferromagnet(0x1234, 10, 5, Index::new(96, 16, 8));
        }

        #[test]
        fn step_multiple_with_stim() {
            let size = Index::new(4, 12, 8);
            let mut cpu_state: SpirvSim<R32, GenericMaterial<R32>, CpuBackend> =
                SpirvSim::new(size, 1e-3);
            let mut wgpu_state: SpirvSim<R32, GenericMaterial<R32>, WgpuBackend> =
                SpirvSim::new(size, 1e-3);
            let stim = stim::Fields::new_e(Vec3::new(1.0e15, 2.0e15, -3.0e15).cast::<R32>());
            for _ in 0..5 {
                cpu_state.step_multiple(100, &stim);
                wgpu_state.step_multiple(100, &stim);
                assert_simstate_eq(&wgpu_state, &cpu_state);
            }
        }
    }
}
