//! consumer/producer primitives

use crossbeam::channel::{self, Receiver, Sender};

pub struct JobPool<C, R> {
    command_chan: Sender<C>,
    response_chan: Receiver<R>,
    worker_command_chan: Receiver<C>,
    worker_response_chan: Sender<R>,
    handles: Vec<std::thread::JoinHandle<()>>,
}

struct Worker<C, R, W> {
    command_chan: Receiver<C>,
    response_chan: Sender<R>,
    work_fn: W,
}

impl<C, R, W: Clone> Clone for Worker<C, R, W> {
    fn clone(&self) -> Self {
        Self {
            command_chan: self.command_chan.clone(),
            response_chan: self.response_chan.clone(),
            work_fn: self.work_fn.clone(),
        }
    }
}

impl<C, R: Send, W: Fn(C) -> R> Worker<C, R, W> {
    fn to_completion(self) {
        for cmd in &self.command_chan {
            let resp = (self.work_fn)(cmd);
            let _ = self.response_chan.send(resp);
        }
    }
}

impl<C, R> JobPool<C, R> {
    pub fn new(buffer: usize) -> Self {
        let (cmd_send, cmd_recv) = channel::bounded(buffer);
        let (resp_send, resp_recv) = channel::bounded(buffer);
        Self {
            command_chan: cmd_send,
            response_chan: resp_recv,
            worker_command_chan: cmd_recv,
            worker_response_chan: resp_send,
            handles: vec![],
        }
    }
    pub fn num_workers(&self) -> u32 {
        self.handles.len().try_into().unwrap()
    }
    pub fn recv(&self) -> R {
        self.response_chan.recv().unwrap()
    }
    /// `try_recv`. named `tend` because this is often used when we want to ensure no workers are
    /// blocked due to lack of space in the output queue.
    pub fn tend(&self) -> Option<R> {
        self.response_chan.try_recv().ok()
    }
    pub fn join_workers(&mut self) {
        // hang up the sender, to signal workers to exit.
        let cap = self.command_chan.capacity().unwrap_or(0);
        (self.command_chan, self.worker_command_chan) = channel::bounded(cap);
        (self.worker_response_chan, self.response_chan) = channel::bounded(cap);
        for h in self.handles.drain(..) {
            h.join().unwrap();
        }
    }
}

impl<C: Send + 'static, R: Send + 'static> JobPool<C, R> {
    pub fn spawn_workers<W: Fn(C) -> R + Send + Clone + 'static>(&mut self, n: u32, work_fn: W) {
        for _ in 0..n {
            self.spawn_worker(work_fn.clone());
        }
    }
}

impl<C: Send + 'static, R: Send + 'static> JobPool<C, R> {
    pub fn spawn_worker<W: Fn(C) -> R + Send + 'static>(&mut self, work_fn: W) {
        let worker = Worker {
            command_chan: self.worker_command_chan.clone(),
            response_chan: self.worker_response_chan.clone(),
            work_fn,
        };
        self.handles.push(std::thread::spawn(move || {
            worker.to_completion()
        }));
    }
}

impl<C, R> Drop for JobPool<C, R> {
    fn drop(&mut self) {
        self.join_workers();
    }
}

impl<C: Send + 'static, R> JobPool<C, R> {
    pub fn send(&self, cmd: C) {
        self.command_chan.send(cmd).unwrap();
    }
}

#[cfg(test)]
mod test {
    use super::*;
    #[test]
    fn lifecycle_no_workers() {
        let _pool: JobPool<(), ()> = JobPool::new(0);
    }

    #[test]
    fn lifecycle_some_workers() {
        let mut pool: JobPool<(), ()> = JobPool::new(0);
        pool.spawn_worker(|_| ());
        pool.spawn_workers(2, |_| ());
    }

    #[test]
    fn single_worker() {
        let mut pool: JobPool<u32, u32> = JobPool::new(0);
        pool.spawn_worker(|x| x*2);
        pool.send(5);
        assert_eq!(pool.recv(), 10);
        pool.send(4);
        assert_eq!(pool.recv(), 8);
    }

    #[test]
    fn multi_worker() {
        use std::sync::{Arc, Mutex};
        let mutex = Arc::new(Mutex::new(()));
        let worker_mutex = mutex.clone();

        let mut pool: JobPool<u32, u32> = JobPool::new(0);
        pool.spawn_workers(2, move |x| {
            // wait until caller unlocks us
            let _ = worker_mutex.lock().unwrap();
            x*2
        });
        pool.send(1);
        assert_eq!(pool.recv(), 2);

        {
            let _lock = mutex.lock().unwrap();
            pool.send(4);
            pool.send(5); // shouldn't block
        }
        let mut replies = [pool.recv(), pool.recv()];
        replies.sort();
        assert_eq!(replies, [8, 10]);
    }

    #[test]
    fn exit_with_unclaimed_responses() {
        let mut pool: JobPool<u32, u32> = JobPool::new(0);
        pool.spawn_workers(2, |x| x*2);
        pool.send(5);
        pool.send(6);
    }

    #[test]
    fn num_workers() {
        let mut pool: JobPool<u32, u32> = JobPool::new(0);
        assert_eq!(pool.num_workers(), 0);

        pool.spawn_workers(2, |x| x*2);
        assert_eq!(pool.num_workers(), 2);

        pool.spawn_workers(1, |x| x*2);
        assert_eq!(pool.num_workers(), 3);

        pool.send(5);
        pool.send(6);
        assert_eq!(pool.num_workers(), 3);

        pool.recv();
        pool.recv();
        assert_eq!(pool.num_workers(), 3);
    }

    #[test]
    fn test_bounded() {
        let pool: JobPool<u32, u32> = JobPool::new(2);
        // we can do this without blocking even when there are no consumers
        // because it just gets buffered
        pool.send(1);
        pool.send(2);
    }

    #[test]
    fn join_workers() {
        let mut pool: JobPool<u32, u32> = JobPool::new(1);

        pool.spawn_workers(2, |x| x*2);
        pool.send(5);
        pool.join_workers();

        pool.spawn_workers(2, |x| x*2);
        pool.send(4);
        // the earlier response to '5' should be lost in the channel
        assert_eq!(pool.recv(), 8);

        // one message in the response queue; one in the send queue, 2 in the worker threads
        pool.send(3); pool.send(2); pool.send(1); pool.send(0);
        // should still be able to join even though everyone's blocked.
        pool.join_workers();

        pool.spawn_workers(1, |x| x*2);
        pool.send(7);
        // the old '0' command should be lost in the channel
        assert_eq!(pool.recv(), 14);

    }
}
