use crate::compound::peano::{Peano, P0};

mod flat;
// mod linked;
// mod tuple_consumer;
// pub use tuple_consumer::*;
// pub use linked::*;
pub use flat::{IntoList, Appendable, Appended, Prependable, Prepended};
pub use flat::exports::*;
use flat::{Node, SplitHead};

pub type Empty = flat::Null;

/// something which can be indexed at `P`.
/// a List of length N is expected to implement `Indexable<P>` for all `P < N`.
pub trait Indexable<P: Peano> {
    type Element;
    fn get(&self) -> Self::Element where Self::Element: Copy;
    fn get_ref(&self) -> &Self::Element;
    fn get_mut(&mut self) -> &mut Self::Element;
    fn set(&mut self, v: Self::Element);
}

pub trait IndexableExplicit {
    fn get<P: Peano>(&self) -> <Self as Indexable<P>>::Element
    where
        Self: Indexable<P>,
        Self::Element: Copy,
    {
        Indexable::get(self)
    }
    fn get_ref<P: Peano>(&self) -> &<Self as Indexable<P>>::Element
    where
        Self: Indexable<P>,
    {
        Indexable::get_ref(self)
    }
    fn get_mut<P: Peano>(&mut self) -> &mut <Self as Indexable<P>>::Element
    where
        Self: Indexable<P>,
    {
        Indexable::get_mut(self)
    }
    fn set<P: Peano>(&mut self, v: Self::Element)
    where
        Self: Indexable<P>,
    {
        Indexable::set(self, v)
    }

    fn get_first(&self) -> <Self as Indexable<P0>>::Element
    where
        Self: Indexable<P0>,
        Self::Element: Copy,
    {
        Indexable::get(self)
    }
    fn get_first_ref(&self) -> &<Self as Indexable<P0>>::Element
    where
        Self: Indexable<P0>,
    {
        Indexable::get_ref(self)
    }
    fn get_first_mut(&mut self) -> &mut <Self as Indexable<P0>>::Element
    where
        Self: Indexable<P0>,
    {
        Indexable::get_mut(self)
    }
    fn set_first(&mut self, v: Self::Element)
    where
        Self: Indexable<P0>,
    {
        Indexable::set(self, v)
    }
}

impl<L> IndexableExplicit for L {}

/// convenience to lookup the type of the element at index `P` of list `L`.
pub type ElementAt<P, L> = <L as Indexable<P>>::Element;

/// implemented by any List (including the Null, empty list)
pub trait Meta {
    type Length: Peano;
    fn len(&self) -> u32 {
        Self::Length::VALUE
    }
}


pub trait ListConsumer<L> {
    type Output;
    fn consume(self, a: L) -> Self::Output;
}

/// implement on your own type to process one list value and return whatever state is necessary to
/// process the subsequent value (and by extension all values)
pub trait FoldOp<State, V> {
    type Output;
    fn feed(&mut self, prev: State, next: V) -> Self::Output;
}

pub struct FoldImpl<Op, State>(Op, State);

//////// fold by-value
impl<Op, State> ListConsumer<Empty> for FoldImpl<Op, State> {
    type Output = State;
    fn consume(self, _l: Empty) -> Self::Output {
        self.1
    }
}

impl<H, T, Op, State> ListConsumer<Node<H, T>> for FoldImpl<Op, State>
where
    Op: FoldOp<State, H>,
    FoldImpl<Op, Op::Output>: ListConsumer<T>,
{
    type Output = <FoldImpl<Op, Op::Output> as ListConsumer<T>>::Output;
    fn consume(self, l: Node<H, T>) -> Self::Output {
        let FoldImpl(mut op, state) = self;
        let (head, tail) = l.split();
        let next_state = op.feed(state, head);
        FoldImpl(op, next_state).consume(tail)
    }
}

//////// fold by-ref
impl<Op, State> ListConsumer<&Empty> for FoldImpl<Op, State> {
    type Output = State;
    fn consume(self, _l: &Empty) -> Self::Output {
        self.1
    }
}

impl<'a, H, T, Op, State> ListConsumer<&'a Node<H, T>> for FoldImpl<Op, State>
where
    Op: FoldOp<State, &'a H>,
    FoldImpl<Op, Op::Output>: ListConsumer<&'a T>,
{
    type Output = <FoldImpl<Op, Op::Output> as ListConsumer<&'a T>>::Output;
    fn consume(self, l: &'a Node<H, T>) -> Self::Output {
        let FoldImpl(mut op, state) = self;
        let (head, tail) = l.split_ref();
        let next_state = op.feed(state, head);
        FoldImpl(op, next_state).consume(tail)
    }
}

pub trait Fold<Op, Init> {
    type Output;
    fn fold(self, op: Op, init: Init) -> Self::Output;
}

impl<L, Op, Init> Fold<Op, Init> for L
where
    FoldImpl<Op, Init>: ListConsumer<L>
{
    type Output = <FoldImpl<Op, Init> as ListConsumer<L>>::Output;
    fn fold(self, op: Op, init: Init) -> Self::Output {
        FoldImpl(op, init).consume(self)
    }
}

pub trait Visitor<E> {
    fn visit(&mut self, v: E);
}

pub struct VisitOp<V>(V);
impl<V, Next> FoldOp<(), Next> for VisitOp<V>
where
    V: Visitor<Next>
{
    type Output = ();
    fn feed(&mut self, _prev: (), next: Next) {
        self.0.visit(next)
    }
}

/// invokes the Visitor `V` on every element of the list.
pub trait Visit<V> {
    fn visit(self, v: V);
}
impl<V, L> Visit<V> for L
where
    L: Fold<VisitOp<V>, (), Output=()>
{
    fn visit(self, v: V) {
        self.fold(VisitOp(v), ())
    }
}


pub struct ReverseOp;
impl<Prev, Next> FoldOp<Prev, Next> for ReverseOp {
    type Output = Node<Next, Prev>;
    fn feed(&mut self, prev: Prev, next: Next) -> Self::Output {
        Node::new(next, prev)
    }
}

pub trait Reverse {
    type Output;
    fn reverse(self) -> Self::Output;
}

impl<L> Reverse for L
where
    L: Fold<ReverseOp, Empty>
{
    type Output = L::Output;
    fn reverse(self) -> Self::Output {
        self.fold(ReverseOp, Empty::default())
    }
}

pub struct SumOp;
impl<Prev, Next> FoldOp<Prev, Next> for SumOp
where
    Prev: core::ops::Add<Next>,
{
    type Output = Prev::Output;
    fn feed(&mut self, prev: Prev, next: Next) -> Self::Output {
        prev + next
    }
}

pub trait Sum<Init> {
    type Output;
    fn sum(self, init: Init) -> Self::Output;
}

impl<Init, L> Sum<Init> for L
where
    L: Fold<SumOp, Init>
{
    type Output = L::Output;
    fn sum(self, init: Init) -> Self::Output {
        self.fold(SumOp, init)
    }
}


#[cfg(feature = "std")]
mod into_vec {
    use super::*;
    pub struct IntoVecOp;
    impl<Next> FoldOp<Vec<Next>, Next> for IntoVecOp {
        type Output = Vec<Next>;
        fn feed(&mut self, mut prev: Vec<Next>, next: Next) -> Self::Output {
            prev.push(next);
            prev
        }
    }
    pub trait IntoVec<T> {
        fn into_vec(self) -> Vec<T>;
    }
    impl<T, L> IntoVec<T> for L
    where
        L: Fold<IntoVecOp, Vec<T>, Output=Vec<T>>
    {
        fn into_vec(self) -> Vec<T> {
            self.fold(IntoVecOp, Vec::new())
        }
    }
}
#[cfg(feature = "std")]
pub use into_vec::{IntoVec, IntoVecOp};

pub trait MapVisitor<V> {
    type Output;
    fn map(&self, elem: V) -> Self::Output;
}

pub struct MapOp<F>(F);
impl<Prev, Next, F> FoldOp<Prev, Next> for MapOp<F>
where
    F: MapVisitor<Next>,
    Prev: Appendable<F::Output>,
{
    type Output = Appended<Prev, F::Output>;
    fn feed(&mut self, prev: Prev, next: Next) -> Self::Output {
        prev.append(self.0.map(next))
    }
}

pub trait Map<Visitor> {
    type Output;
    fn map(self, op: Visitor) -> Self::Output;
}

impl<L, Visitor> Map<Visitor> for L
where
    L: Fold<MapOp<Visitor>, Empty>
{
    type Output = L::Output;
    fn map(self, visitor: Visitor) -> Self::Output {
        self.fold(MapOp(visitor), Empty::default())
    }
}

pub struct IdentityMapVisitor;
impl<V> MapVisitor<V> for IdentityMapVisitor {
    type Output = V;
    fn map(&self, elem: V) -> Self::Output {
        elem
    }
}

pub trait Extend<L> {
    type Output;
    fn extend(self, l: L) -> Self::Output;
}
impl<L0, L1> Extend<L1> for L0
where
    L1: Fold<MapOp<IdentityMapVisitor>, L0>
{
    type Output = L1::Output;
    fn extend(self, l: L1) -> Self::Output {
        l.fold(MapOp(IdentityMapVisitor), self)
    }
}

pub struct FlattenOp;
impl<Prev, Next> FoldOp<Prev, Next> for FlattenOp
where
    Prev: Extend<Next>,
{
    type Output = Prev::Output;
    fn feed(&mut self, prev: Prev, next: Next) -> Self::Output {
        prev.extend(next)
    }
}

pub trait Flatten {
    type Output;
    fn flatten(self) -> Self::Output;
}
impl<L> Flatten for L
where
    L: Fold<FlattenOp, Empty>
{
    type Output = L::Output;
    fn flatten(self) -> Self::Output {
        self.fold(FlattenOp, Empty::default())
    }
}


#[derive(Copy, Clone, Default, PartialEq)]
pub struct Tagged<P: Peano, V> {
    inner: V,
    _p: P::Unit,
}
impl<P: Peano, V> Tagged<P, V> {
    pub fn new(inner: V) -> Self {
        Self { inner, _p: P::Unit::default() }
    }
    pub fn into_inner(self) -> V {
        self.inner
    }
}

impl<P: Peano, V> core::ops::Deref for Tagged<P, V> {
    type Target = V;
    fn deref(&self) -> &Self::Target {
        &self.inner
    }
}
impl<P: Peano, V> core::ops::DerefMut for Tagged<P, V> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.inner
    }
}

pub struct EnumerateOp;
impl<L, Next> FoldOp<L, Next> for EnumerateOp
where
    L: Meta,
    L: Appendable<Tagged<L::Length, Next>>,
{
    type Output = Appended<L, Tagged<L::Length, Next>>;
    fn feed(&mut self, prev: L, next: Next) -> Self::Output {
        prev.append(Tagged::new(next))
    }
}

pub trait Enumerate {
    type Output;
    fn enumerate(self) -> Self::Output;
}

impl<L> Enumerate for L
where
    L: Fold<EnumerateOp, Empty>
{
    type Output = L::Output;
    fn enumerate(self) -> Self::Output {
        self.fold(EnumerateOp, Empty::default())
    }
}

pub struct MapTagToValueOp<T>(T /* unused */);
impl<P: Peano, V, T: From<u32>> MapVisitor<Tagged<P, V>> for MapTagToValueOp<T> {
    type Output = (T, V);
    fn map(&self, elem: Tagged<P, V>) -> Self::Output {
        (P::VALUE.into(), elem.into_inner())
    }
}

pub trait EnumerateU32 {
    type Output;
    fn enumerate_u32(self) -> Self::Output;
}
impl<L> EnumerateU32 for L
where
    L: Enumerate,
    L::Output: Map<MapTagToValueOp<u32>>,
{
    type Output = <L::Output as Map<MapTagToValueOp<u32>>>::Output;
    fn enumerate_u32(self) -> Self::Output {
        self.enumerate().map(MapTagToValueOp(0u32 /* unused */))
    }
}


#[cfg(test)]
mod test {
    use super::*;
    use crate::compound::peano::{P0, P1, P2};

    struct SumVal;

    impl FoldOp<i32, i32> for SumVal {
        type Output = i32;
        fn feed(&mut self, prev: i32, next: i32) -> Self::Output {
            prev + next
        }
    }

    #[test]
    fn fold_prim() {
        let list = (3, 4, 5i32).into_list();
        assert_eq!(list.fold(SumVal, 2i32), 2+3+4+5);
    }

    impl FoldOp<i32, f32> for SumVal {
        type Output = f32;
        fn feed(&mut self, prev: i32, next: f32) -> Self::Output {
            prev as f32 + next
        }
    }
    impl FoldOp<f32, f32> for SumVal {
        type Output = f32;
        fn feed(&mut self, prev: f32, next: f32) -> Self::Output {
            prev + next
        }
    }
    impl FoldOp<f32, i32> for SumVal {
        type Output = f32;
        fn feed(&mut self, prev: f32, next: i32) -> Self::Output {
            prev + next as f32
        }
    }

    #[test]
    fn fold_mixed() {
        // we fold:
        // 2i32 + 3i32
        // 5i32 + 4i32
        // 9i32 + 5.5f32
        // 14.5f32 + 6.5f32
        // 21f32 + 7i32
        let list = (3i32, 4i32, 5.5f32, 6.5f32, 7i32).into_list();
        assert_eq!(list.fold(SumVal, 2i32), 28f32);
    }


    #[derive(PartialEq)]
    struct NotCopy(i32);

    struct SumRef;
    impl FoldOp<i32, &NotCopy> for SumRef {
        type Output = i32;
        fn feed(&mut self, prev: i32, next: &NotCopy) -> Self::Output {
            prev + next.0
        }
    }
    impl FoldOp<i32, &i32> for SumRef {
        type Output = i32;
        fn feed(&mut self, prev: i32, next: &i32) -> Self::Output {
            prev + *next
        }
    }

    #[test]
    fn fold_ref() {
        let list = &(3i32, NotCopy(4i32), 5i32).into_list();
        assert_eq!(list.fold(SumRef, 2i32), 14i32);
        assert!(list == list); // just check that it wasn't consumed
    }

    struct NoopVisitor;
    impl<V> Visitor<V> for NoopVisitor {
        fn visit(&mut self, _e: V) {}
    }

    #[test]
    fn visit_noop() {
        let list = (3f32, NotCopy(4i32), 5u32).into_list();
        list.visit(NoopVisitor);

        let list = &(3f32, NotCopy(4i32), 5u32).into_list();
        list.visit(NoopVisitor);
    }

    struct AccumVisitor(i32);
    impl Visitor<i32> for &mut AccumVisitor {
        fn visit(&mut self, e: i32) {
            self.0 += e * 2;
        }
    }

    #[test]
    fn visit_mut() {
        let list = (3i32, 4i32, 5i32).into_list();
        let mut v = AccumVisitor(0);
        list.visit(&mut v);
        assert_eq!(v.0, 24);
    }

    #[test]
    fn into_vec() {
        let list = (3i32, 4i32, 5i32).into_list();
        assert_eq!(list.into_vec(), vec![3i32, 4, 5]);
    }

    #[test]
    fn into_vec_ref() {
        let list = &(3i32, 4i32, 5i32).into_list();
        assert_eq!(list.into_vec(), vec![&3i32, &4, &5]);
    }

    #[test]
    fn into_vec_empty() {
        assert_eq!(IntoVec::<u32>::into_vec(Empty::default()), vec![]);
    }

    #[test]
    fn sum() {
        let list = (3i32, 4i32, 5i32).into_list();
        assert_eq!(list.sum(2i32), 14i32);
    }

    struct SumA(f32);
    #[derive(Debug, PartialEq)]
    struct SumB(f32);
    impl core::ops::Add<i32> for SumA {
        type Output = SumB;
        fn add(self, other: i32) -> Self::Output {
            SumB(self.0 + other as f32)
        }
    }
    impl core::ops::Add<f32> for SumB {
        type Output = SumA;
        fn add(self, other: f32) -> Self::Output {
            SumA(self.0 + other)
        }
    }

    #[test]
    fn sum_mixed() {
        let list = (3i32, 4f32, 5i32).into_list();
        assert_eq!(list.sum(SumA(2f32)), SumB(14f32));
    }


    #[test]
    fn reverse_empty() {
        assert!(Empty::default().reverse() == Empty::default());
    }

    #[test]
    fn reverse_non_empty() {
        let list = (3i32, 4f32, 5u32).into_list();
        let expected = (5u32, 4f32, 3i32).into_list();
        assert!(list.reverse() == expected);
    }

    struct Double;
    impl MapVisitor<i32> for Double {
        type Output = i32;
        fn map(&self, v: i32) -> Self::Output {
            v + v
        }
    }
    impl MapVisitor<f32> for Double {
        type Output = f32;
        fn map(&self, v: f32) -> Self::Output {
            v + v
        }
    }

    #[test]
    fn map_empty() {
        assert!(Empty::default().map(Double) == Empty::default());
    }

    #[test]
    fn map_mixed() {
        let list = (2i32, 3f32, 4i32).into_list();
        let expected = (4i32, 6f32, 8i32).into_list();
        assert!(list.map(Double) == expected);
    }

    #[test]
    fn extend_empty() {
        let l0 = Empty::default();
        let l1 = Empty::default();
        assert!(l0.extend(l1) == Empty::default());
    }

    #[test]
    fn extend_with_empty() {
        let l0 = (2u32, 3f32).into_list();
        let l1 = Empty::default();
        let expected = (2u32, 3f32).into_list();
        assert!(l0.extend(l1) == expected);
    }

    #[test]
    fn extend_from_empty() {
        let l0 = Empty::default();
        let l1 = (2u32, 3f32).into_list();
        let expected = (2u32, 3f32).into_list();
        assert!(l0.extend(l1) == expected);
    }

    #[test]
    fn extend_mixed() {
        let l0 = (2u32, 3f32).into_list();
        let l1 = ("hello",).into_list();
        let expected = (2u32, 3f32, "hello").into_list();
        assert!(l0.extend(l1) == expected);
    }

    #[test]
    fn flatten_empty() {
        assert!(Empty::default().flatten() == Empty::default());
    }
    #[test]
    fn flatten_inner_empty1() {
        let l = (Empty::default(),).into_list();
        assert!(l.flatten() == Empty::default());
    }
    #[test]
    fn flatten_inner_empty2() {
        let l = (Empty::default(), Empty::default()).into_list();
        assert!(l.flatten() == Empty::default());
    }
    #[test]
    fn flatten_mixed() {
        let l = (
            (2u32, 3f32).into_list(),
            (4u32,).into_list(),
        ).into_list();
        let expected = (
            2u32,
            3f32,
            4u32,
        ).into_list();
        assert!(l.flatten() == expected);
    }
    #[test]
    fn flatten_nested() {
        let l = (
            (2u32, 3f32).into_list(),
            ("hello", ("every", "one").into_list()).into_list(),
            (4u32,).into_list(),
        ).into_list();
        let expected = (
            2u32,
            3f32,
            "hello",
            ("every", "one").into_list(),
            4u32,
        ).into_list();
        assert!(l.flatten() == expected);
    }

    #[test]
    fn enumerate_empty() {
        assert!(Empty::default().enumerate() == Empty::default());
    }

    #[test]
    fn enumerate_one() {
        let list = (2i32,).into_list();
        let expected = (Tagged::<P0, _>::new(2i32),).into_list();
        assert!(list.enumerate() == expected);
    }

    #[test]
    fn enumerate_multiple() {
        let list = (2i32, (), 4f32).into_list();
        let expected = (
            Tagged::<P0, _>::new(2i32),
            Tagged::<P1, _>::new(()),
            Tagged::<P2, _>::new(4f32),
        ).into_list();
        assert!(list.enumerate() == expected);
    }

    #[test]
    fn enumerate_u32_multiple() {
        let list = (2i32, (), 4f32).into_list();
        let expected = (
            (0u32, 2i32),
            (1u32, ()),
            (2u32, 4f32)
        ).into_list();
        assert!(list.enumerate_u32() == expected);
    }
}
