use crate::compound::peano::{P0, PNext};

#[cfg(feature = "serde")]
use serde::{Serialize, Deserialize};

pub trait TuplePrims: Sized {
    type Head; // single element
    type Tail; // variably-sized tuple
    fn into_head(self) -> Self::Head;
    fn into_tail(self) -> Self::Tail;
}
impl<E0> TuplePrims for (E0,) {
    type Head = E0;
    type Tail = ();
    fn into_head(self) -> Self::Head {
        self.0
    }
    fn into_tail(self) -> Self::Tail {
        ()
    }
}
impl<E0, E1> TuplePrims for (E0, E1) {
    type Head = E0;
    type Tail = (E1,);
    fn into_head(self) -> Self::Head {
        self.0
    }
    fn into_tail(self) -> Self::Tail {
        (self.1,)
    }
}
impl<E0, E1, E2> TuplePrims for (E0, E1, E2) {
    type Head = E0;
    type Tail = (E1, E2);
    fn into_head(self) -> Self::Head {
        self.0
    }
    fn into_tail(self) -> Self::Tail {
        (self.1, self.2)
    }
}
impl<E0, E1, E2, E3> TuplePrims for (E0, E1, E2, E3) {
    type Head = E0;
    type Tail = (E1, E2, E3);
    fn into_head(self) -> Self::Head {
        self.0
    }
    fn into_tail(self) -> Self::Tail {
        (self.1, self.2, self.3)
    }
}


// note that this construction allows a zero-length list (Null),
// which is sort of interesting.
#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
#[cfg_attr(feature = "fmt", derive(Debug))]
#[derive(Copy, Clone, Default, PartialEq)]
pub struct List<Args>(Args);

impl<Args> List<Args> {
    pub fn new(args: Args) -> Self {
        Self(args)
    }
}

pub trait ListPrims: Sized {
    type Head; // single element
    type Tail; // variably-sized List
    fn into_head(self) -> Self::Head;
    fn into_tail(self) -> Self::Tail;
}

impl<Args: TuplePrims> ListPrims for List<Args> {
    type Head = Args::Head;
    type Tail = List<Args::Tail>;
    fn into_head(self) -> Self::Head {
        self.0.into_head()
    }
    fn into_tail(self) -> Self::Tail {
        List(self.0.into_tail())
    }
}

pub trait Consumable<P> {
    type Result: ListPrims;
    fn consume(self) -> Self::Result;
}

impl<Args: TuplePrims> Consumable<P0> for List<Args> {
    type Result = Self;
    fn consume(self) -> Self::Result {
        self
    }
}

impl<Args, P> Consumable<PNext<P>> for List<Args>
where
    Self: ListPrims,
    <Self as ListPrims>::Tail: Consumable<P>,
{
    type Result = <<Self as ListPrims>::Tail as Consumable<P>>::Result;
    fn consume(self) -> Self::Result {
        self.into_tail().consume()
    }
}

impl<Args: TuplePrims> List<Args> {
    pub fn consume<P>(self) -> <Self as Consumable<P>>::Result
        where Self: Consumable<P>
    {
        Consumable::<P>::consume(self)
    }
    pub fn get<P>(self) -> <<Self as Consumable<P>>::Result as ListPrims>::Head
        where Self: Consumable<P>
    {
        self.consume().into_head()
    }
}


#[cfg(test)]
mod test {
    use super::*;
    use crate::compound::peano::{P1, P2};

    #[test]
    fn get() {
        let list = List((5u32, 4i32, 3f32));

        assert_eq!(list.get::<P0>(), 5u32);
        assert_eq!(list.get::<P1>(), 4i32);
        assert_eq!(list.get::<P2>(), 3f32);
    }

    // #[test]
    // fn get() {
    //     let list = (5u32, 4i32, 3f32).into_list();
    //     assert_eq!(list.get::<P0>(), &5u32);
    //     assert_eq!(list.get::<P1>(), &4i32);
    //     assert_eq!(list.get::<P2>(), &3f32);
    // }

    // #[test]
    // fn set() {
    //     let mut list = List::<(u32, i32, f32)>::default();

    //     list.set::<P0>(5u32);

    //     assert_eq!(list.get::<P0>(), &5u32);
    //     assert_eq!(list.get::<P1>(), &0i32);
    //     assert_eq!(list.get::<P2>(), &0f32);

    //     list.set::<P2>(3f32);
    //     list.set::<P1>(4i32);

    //     assert_eq!(list.get::<P0>(), &5u32);
    //     assert_eq!(list.get::<P1>(), &4i32);
    //     assert_eq!(list.get::<P2>(), &3f32);
    // }
}
