#[cfg(feature = "serde")]
use serde::{Serialize, Deserialize};

/// This is a spirv-compatible option type.
/// The native rust Option type produces invalid spirv due to its enum nature; this custom option
/// type creates code which will actually compile.
#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
#[cfg_attr(feature = "fmt", derive(Debug))]
#[derive(Copy, Clone, PartialEq)]
pub struct Optional<T> {
    // XXX: not a bool, because: "entrypoint parameter cannot contain a boolean"
    present: u8,
    data: T,
}

impl<T> Optional<T> {
    pub fn some(data: T) -> Self {
        Self {
            present: 1,
            data,
        }
    }

    pub fn explicit_none(data: T) -> Self {
        Self {
            present: 0,
            data,
        }
    }

    pub fn is_some(self) -> bool {
        self.present != 0
    }

    pub fn unwrap(self) -> T {
        debug_assert!(self.present != 0);
        self.data
    }

    pub fn map<U: Default, F: FnOnce(T) -> U>(self, f: F) -> Optional<U> {
        self.and_then(|inner| Optional::some(f(inner)))
    }

    pub fn and_then<U: Default, F: FnOnce(T) -> Optional<U>>(self, f: F) -> Optional<U> {
        if self.present != 0 {
            f(self.data)
        } else {
            Optional::none()
        }
    }

    pub fn unwrap_or(self, default: T) -> T {
        if self.present != 0 {
            self.data
        } else {
            default
        }
    }
}

impl<T: Default> Optional<T> {
    pub fn none() -> Self {
        Self::explicit_none(Default::default())
    }

    pub fn unwrap_or_default(self) -> T {
        self.unwrap_or(Default::default())
    }
}

impl<T: Default> Default for Optional<T> {
    fn default() -> Self {
        Self::none()
    }
}

impl<T0: Default, T1: Default> Optional<(T0, T1)> {
    pub fn flatten((f0, f1): (Optional<T0>, Optional<T1>)) -> Self {
        if f0.present != 0 && f1.present != 0 {
            Optional::some((f0.data, f1.data))
        } else {
            Optional::none()
        }
    }
}

