//! Peano numbers (also known as Church numerals) are type-level natural numbers.
//! each non-zero Peano number is defined as the unique successor of a previous Peano number.
//!
//! - given some Peano number I, we can derive its successor by S=PNext<I>.
//! - given a Peano number PNext<I>, we can define its predecessor as I=P.
//! - the base Peano number, which represents 0, is `P0`.
//! - the `Peano` trait exposes alternative syntaxes for these: `P::Next` and `P::Prev`,
//!   however the type system can reason less about these (they're just a convenience).
//!
//! the primary use of Peano numbers is to allow types to specialize on a specific natural number
//! out of some larger set of natural numbers. e.g. one might have a `struct List<Length: Peano>`
//! to allow constructing a list of compile-time constant length.
//!
//! this whole module will hopefully be obsoleted as Rust's type-level integers become more
//! capable, but in 2022 Peano numbers enable more operations (arithmetic, specialization) than type-level integers.

#[derive(Copy, Clone, Default, PartialEq)]
pub struct PNext<P>(P);
#[derive(Copy, Clone, Default, PartialEq)]
pub struct P0;
pub type P1 = PNext<P0>;

/// these are exported for the convenience of potential consumers: not needed internally
mod exports {
    #![allow(dead_code)]
    use super::{P1, PNext};
    pub type P2 = PNext<P1>;
    pub type P3 = PNext<P2>;
    pub type P4 = PNext<P3>;
    pub type P5 = PNext<P4>;
    pub type P6 = PNext<P5>;
    pub type P7 = PNext<P6>;
    pub type P8 = PNext<P7>;
    pub type P9 = PNext<P8>;
    pub type P10 = PNext<P9>;
    pub type P11 = PNext<P10>;
    pub type P12 = PNext<P11>;
    pub type P13 = PNext<P12>;
    pub type P14 = PNext<P13>;
    pub type P15 = PNext<P14>;
}
pub use exports::*;

pub trait Peano: Copy + Clone + Default + PartialEq {
    type Next: PeanoNonZero;
    type PrevOrZero: Peano;
    /// always set to ()
    /// this exists to allow Peano numbers to be used as struct parameters without PhantomData
    type Unit: Copy + Clone + Default + PartialEq;
    const VALUE: u32;
}

pub trait PeanoNonZero: Peano {
    type Prev: Peano;
}

impl Peano for P0 {
    type Next = P1;
    type PrevOrZero = P0;
    type Unit = ();
    const VALUE: u32 = 0;
}
impl<P: Peano> Peano for PNext<P> {
    type Next = PNext<PNext<P>>;
    type PrevOrZero = P;
    type Unit = ();
    const VALUE: u32 = 1 + P::VALUE;
}
impl<P: Peano> PeanoNonZero for PNext<P> {
    type Prev = P;
}

// A: LessThan<B> is satisfied only if A is strictly less than B.
pub trait LessThan<P: Peano> { }
impl<P: Peano> LessThan<PNext<P>> for P { }

