use core::convert::{AsMut, AsRef};
use core::iter::Zip;
use core::ops::{Index, IndexMut};

use crate::vec::Vec3u;

/// use this to wrap a flat region of memory into something which can be indexed by coordinates in
/// 3d space.
#[cfg_attr(feature = "fmt", derive(Debug))]
#[derive(Clone, Default, PartialEq)]
pub struct DimSlice<T> {
    dim: Vec3u,
    items: T,
}

impl<T> DimSlice<T> {
    pub fn new(dim: Vec3u, items: T) -> Self {
        Self { dim, items }
    }
    pub fn dim(&self) -> Vec3u {
        self.dim
    }
    pub fn into_inner(self) -> T {
        self.items
    }

    pub fn indices(&self) -> DimIter {
        DimIter::new(self.dim)
    }

    /// re-borrow the slice with a different lifetime.
    pub fn as_ref<R: ?Sized>(&self) -> DimSlice<&R>
        where T: AsRef<R>
    {
        DimSlice::new(self.dim, self.items.as_ref())
    }
    /// re-borrow the slice with a different lifetime.
    pub fn as_mut<R: ?Sized>(&mut self) -> DimSlice<&mut R>
        where T: AsMut<R>
    {
        DimSlice::new(self.dim, self.items.as_mut())
    }
}

#[cfg(feature = "iter")]
impl<T: IntoIterator> DimSlice<T> {
    pub fn enumerated(self) -> Zip<DimIter, T::IntoIter> {
        self.indices().zip(self.into_iter())
    }
}

fn index(loc: Vec3u, dim: Vec3u) -> usize {
    ((loc.z()*dim.y() + loc.y())*dim.x() + loc.x()) as usize
}

impl<'a, T: Index<usize> + ?Sized> Index<Vec3u> for DimSlice<&'a T> {
    type Output=T::Output;

    fn index(&self, idx: Vec3u) -> &Self::Output {
        let idx = index(idx, self.dim);
        &self.items[idx]
    }
}

impl<'a, T: Index<usize> + ?Sized> Index<Vec3u> for DimSlice<&'a mut T> {
    type Output=T::Output;

    fn index(&self, idx: Vec3u) -> &Self::Output {
        let idx = index(idx, self.dim);
        &self.items[idx]
    }
}

impl<'a, T: IndexMut<usize> + ?Sized> IndexMut<Vec3u> for DimSlice<&'a mut T> {
    fn index_mut(&mut self, idx: Vec3u) -> &mut Self::Output {
        let idx = index(idx, self.dim);
        &mut self.items[idx]
    }
}

#[cfg(feature = "std")]
impl<T> Index<Vec3u> for DimSlice<Vec<T>> {
    type Output=T;

    fn index(&self, idx: Vec3u) -> &Self::Output {
        let idx = index(idx, self.dim);
        &self.items[idx]
    }
}

#[cfg(feature = "std")]
impl<T> IndexMut<Vec3u> for DimSlice<Vec<T>> {
    fn index_mut(&mut self, idx: Vec3u) -> &mut Self::Output {
        let idx = index(idx, self.dim);
        &mut self.items[idx]
    }
}

#[cfg(feature = "std")]
impl<T> Index<Vec3u> for DimSlice<Box<[T]>> {
    type Output=T;

    fn index(&self, idx: Vec3u) -> &Self::Output {
        let idx = index(idx, self.dim);
        &self.items[idx]
    }
}

#[cfg(feature = "std")]
impl<T> IndexMut<Vec3u> for DimSlice<Box<[T]>> {
    fn index_mut(&mut self, idx: Vec3u) -> &mut Self::Output {
        let idx = index(idx, self.dim);
        &mut self.items[idx]
    }
}

impl<T: IntoIterator> IntoIterator for DimSlice<T> {
    type Item = T::Item;
    type IntoIter = T::IntoIter;
    fn into_iter(self) -> Self::IntoIter {
        self.items.into_iter()
    }
}

pub struct DimIter {
    idx: Vec3u,
    dim: Vec3u,
}

impl DimIter {
    fn new(dim: Vec3u) -> Self {
        Self {idx: Vec3u::default(), dim }
    }
}

#[cfg(feature = "iter")]
impl Iterator for DimIter {
    type Item = Vec3u;
    fn next(&mut self) -> Option<Self::Item> {
        if self.dim.x() == 0 || self.dim.y() == 0 || self.dim.z() == 0 {
            // no items
            return None;
        }
        if self.idx.z() == self.dim.z() {
            // reached the end
            return None;
        }
        let cur = self.idx;
        self.idx = match cur.x()+1 {
            // need to increment y
            next_x if next_x == self.dim.x() => match cur.y() + 1 {
                // need to increment z
                next_y if next_y == self.dim.y() => Vec3u::new(0, 0, cur.z() + 1),
                next_y => Vec3u::new(0, next_y, cur.z()),
            },
            next_x => Vec3u::new(next_x, cur.y(), cur.z()),
        };
        Some(cur)
    }
}


#[cfg(test)]
mod test {
    use super::*;
    #[test]
    fn test_index() {
        let dim = Vec3u::new(2, 3, 7);
        assert_eq!(index(Vec3u::new(0, 0, 0), dim), 0);
        assert_eq!(index(Vec3u::new(1, 0, 0), dim), 1);
        assert_eq!(index(Vec3u::new(0, 1, 0), dim), 2);
        assert_eq!(index(Vec3u::new(1, 1, 0), dim), 3);
        assert_eq!(index(Vec3u::new(0, 2, 0), dim), 4);
        assert_eq!(index(Vec3u::new(0, 0, 1), dim), 6);
        assert_eq!(index(Vec3u::new(1, 0, 1), dim), 7);
        assert_eq!(index(Vec3u::new(0, 1, 1), dim), 8);
        assert_eq!(index(Vec3u::new(1, 2, 1), dim), 11);
        assert_eq!(index(Vec3u::new(1, 2, 2), dim), 17);
    }

    #[test]
    fn as_ref() {
        let data = [1, 2];
        let s = DimSlice::new(Vec3u::new(1, 2, 1), &data[..]);
        assert_eq!(s.as_ref(), DimSlice::new(Vec3u::new(1, 2, 1), &data[..]));
    }

    #[test]
    fn dim_slice_index() {
        let data = [
            0, 1, 2,
            3, 4, 5,

            0, 10,20,
            30,40,50,
        ];
        let s = DimSlice::new(Vec3u::new(3, 2, 2), &data);
        assert_eq!(s[Vec3u::new(0, 0, 0)], 0);
        assert_eq!(s[Vec3u::new(1, 0, 0)], 1);
        assert_eq!(s[Vec3u::new(1, 1, 0)], 4);
        assert_eq!(s[Vec3u::new(1, 1, 1)], 40);
        assert_eq!(s[Vec3u::new(2, 1, 1)], 50);
    }

    #[test]
    fn dim_slice_index_mut() {
        let mut data = [
            0, 1, 2,
            3, 4, 5,

            0, 10,20,
            30,40,50,
        ];
        let mut s = DimSlice::new(Vec3u::new(3, 2, 2), &mut data);
        s[Vec3u::new(0, 0, 0)] = 100;
        s[Vec3u::new(0, 1, 1)] = 300;

        assert_eq!(data, [
            100,1, 2,
            3,  4, 5,

            0, 10, 20,
            300,40,50,
        ]);
    }

    #[test]
    fn dim_slice_into_iter() {
        let data = [1, 2, 3, 4, 5, 6];
        let s = DimSlice::new(Vec3u::new(3, 1, 2), &data);
        let mut i = s.into_iter();
        assert_eq!(*i.next().unwrap(), 1);
        assert_eq!(*i.next().unwrap(), 2);
        assert_eq!(*i.next().unwrap(), 3);
        assert_eq!(*i.next().unwrap(), 4);
        assert_eq!(*i.next().unwrap(), 5);
        assert_eq!(*i.next().unwrap(), 6);
        assert_eq!(i.next(), None);
    }

    #[test]
    fn dim_slice_into_iter_mut() {
        let mut data = [1, 2, 3, 4, 5, 6];
        let s = DimSlice::new(Vec3u::new(3, 1, 2), &mut data);
        let mut i = s.into_iter();
        *i.next().unwrap() = 10;
        assert_eq!(*i.next().unwrap(), 2);
        *i.next().unwrap() += 27;
        assert_eq!(*i.next().unwrap(), 4);
        *i.next().unwrap() *= 10;
        assert_eq!(*i.next().unwrap(), 6);
        assert_eq!(i.next(), None);

        assert_eq!(data, [10,2,30,4,50,6]);
    }

    #[test]
    fn dim_slice_indices() {
        let s = DimSlice::new(Vec3u::new(4, 3, 2), &[()]);
        let mut i = s.indices();
        assert_eq!(i.next().unwrap(), Vec3u::new(0, 0, 0));
        assert_eq!(i.next().unwrap(), Vec3u::new(1, 0, 0));
        assert_eq!(i.next().unwrap(), Vec3u::new(2, 0, 0));
        assert_eq!(i.next().unwrap(), Vec3u::new(3, 0, 0));

        assert_eq!(i.next().unwrap(), Vec3u::new(0, 1, 0));
        assert_eq!(i.next().unwrap(), Vec3u::new(1, 1, 0));
        assert_eq!(i.next().unwrap(), Vec3u::new(2, 1, 0));
        assert_eq!(i.next().unwrap(), Vec3u::new(3, 1, 0));

        assert_eq!(i.next().unwrap(), Vec3u::new(0, 2, 0));
        assert_eq!(i.next().unwrap(), Vec3u::new(1, 2, 0));
        assert_eq!(i.next().unwrap(), Vec3u::new(2, 2, 0));
        assert_eq!(i.next().unwrap(), Vec3u::new(3, 2, 0));

        assert_eq!(i.next().unwrap(), Vec3u::new(0, 0, 1));
        assert_eq!(i.next().unwrap(), Vec3u::new(1, 0, 1));
        assert_eq!(i.next().unwrap(), Vec3u::new(2, 0, 1));
        assert_eq!(i.next().unwrap(), Vec3u::new(3, 0, 1));

        assert_eq!(i.next().unwrap(), Vec3u::new(0, 1, 1));
        assert_eq!(i.next().unwrap(), Vec3u::new(1, 1, 1));
        assert_eq!(i.next().unwrap(), Vec3u::new(2, 1, 1));
        assert_eq!(i.next().unwrap(), Vec3u::new(3, 1, 1));

        assert_eq!(i.next().unwrap(), Vec3u::new(0, 2, 1));
        assert_eq!(i.next().unwrap(), Vec3u::new(1, 2, 1));
        assert_eq!(i.next().unwrap(), Vec3u::new(2, 2, 1));
        assert_eq!(i.next().unwrap(), Vec3u::new(3, 2, 1));

        assert_eq!(i.next(), None);
        assert_eq!(i.next(), None);
    }
    #[test]
    fn dim_slice_indices_zero_dim() {
        let s = DimSlice::new(Vec3u::new(4, 3, 0), &[()]);
        assert_eq!(s.indices().next(), None);
        assert_eq!(s.indices().next(), None);

        let s = DimSlice::new(Vec3u::new(4, 0, 2), &[()]);
        assert_eq!(s.indices().next(), None);
        assert_eq!(s.indices().next(), None);

        let s = DimSlice::new(Vec3u::new(0, 3, 2), &[()]);
        assert_eq!(s.indices().next(), None);
        assert_eq!(s.indices().next(), None);

        let s = DimSlice::new(Vec3u::new(3, 0, 0), &[()]);
        assert_eq!(s.indices().next(), None);
        assert_eq!(s.indices().next(), None);

        let s = DimSlice::new(Vec3u::new(0, 1, 0), &[()]);
        assert_eq!(s.indices().next(), None);
        assert_eq!(s.indices().next(), None);

        let s = DimSlice::new(Vec3u::new(0, 0, 2), &[()]);
        assert_eq!(s.indices().next(), None);
        assert_eq!(s.indices().next(), None);
    }

    #[test]
    fn dim_slice_enumerated() {
        let data = [
            10, 11,
            20, 21,
            30, 31,
        ];
        let s = DimSlice::new(Vec3u::new(1, 2, 3), &data);
        let mut i = s.enumerated();
        assert_eq!(i.next().unwrap(), (Vec3u::new(0, 0, 0), &10));
        assert_eq!(i.next().unwrap(), (Vec3u::new(0, 1, 0), &11));
        assert_eq!(i.next().unwrap(), (Vec3u::new(0, 0, 1), &20));
        assert_eq!(i.next().unwrap(), (Vec3u::new(0, 1, 1), &21));
        assert_eq!(i.next().unwrap(), (Vec3u::new(0, 0, 2), &30));
        assert_eq!(i.next().unwrap(), (Vec3u::new(0, 1, 2), &31));
        assert_eq!(i.next(), None);
    }

    #[test]
    fn dim_slice_enumerated_mut() {
        let mut data = [
            10, 11,
            20, 21,
            30, 31,
        ];
        let s = DimSlice::new(Vec3u::new(2, 1, 3), &mut data);
        let mut i = s.enumerated();

        let (idx, v) = i.next().unwrap();
        assert_eq!(idx, Vec3u::new(0, 0, 0));
        *v = 100;

        let (idx, v) = i.next().unwrap();
        assert_eq!(idx, Vec3u::new(1, 0, 0));
        *v = 110;
        i.next().unwrap();

        let (idx, v) = i.next().unwrap();
        assert_eq!(idx, Vec3u::new(1, 0, 1));
        *v = 210;

        assert_eq!(data, [100, 110, 20, 210, 30, 31]);
    }
}

