use core::convert::{AsMut, AsRef};
use core::iter::Zip;
use core::ops::{Index, IndexMut};

use crate::dim::{DimIter, DimSlice};
use crate::vec::Vec3u;


#[cfg_attr(feature = "fmt", derive(Debug))]
#[derive(Clone, Default, PartialEq)]
pub struct OffsetDimSlice<T> {
    offset: Vec3u,
    inner: DimSlice<T>,
}

impl<T> OffsetDimSlice<T> {
    pub fn new(offset: Vec3u, dim: Vec3u, items: T) -> Self {
        Self { offset, inner: DimSlice::new(dim, items) }
    }
    pub fn dim(&self) -> Vec3u {
        self.inner.dim()
    }
    pub fn offset(&self) -> Vec3u {
        self.offset
    }
    pub fn into_inner(self) -> T {
        self.inner.into_inner()
    }

    pub fn indices(&self) -> OffsetDimIter {
        OffsetDimIter::new(self.offset, self.inner.indices())
    }

    /// re-borrow the slice with a different lifetime.
    pub fn as_ref<R: ?Sized>(&self) -> OffsetDimSlice<&R>
        where T: AsRef<R>
    {
        OffsetDimSlice { offset: self.offset, inner: self.inner.as_ref()}
    }
    /// re-borrow the slice with a different lifetime.
    pub fn as_mut<R: ?Sized>(&mut self) -> OffsetDimSlice<&mut R>
        where T: AsMut<R>
    {
        OffsetDimSlice { offset: self.offset, inner: self.inner.as_mut()}
    }
}

#[cfg(feature = "iter")]
impl<T: IntoIterator> OffsetDimSlice<T> {
    pub fn enumerated(self) -> Zip<OffsetDimIter, T::IntoIter> {
        self.indices().zip(self.into_iter())
    }
}

impl<T> Index<Vec3u> for OffsetDimSlice<T>
where
    DimSlice<T>: Index<Vec3u>
{
    type Output=<DimSlice<T> as Index<Vec3u>>::Output;

    fn index(&self, idx: Vec3u) -> &Self::Output {
        &self.inner[idx - self.offset]
    }
}

impl<T> IndexMut<Vec3u> for OffsetDimSlice<T>
where
    DimSlice<T>: IndexMut<Vec3u>
{
    fn index_mut(&mut self, idx: Vec3u) -> &mut Self::Output {
        &mut self.inner[idx - self.offset]
    }
}

impl<T: IntoIterator> IntoIterator for OffsetDimSlice<T> {
    type Item = T::Item;
    type IntoIter = T::IntoIter;
    fn into_iter(self) -> Self::IntoIter {
        self.inner.into_iter()
    }
}

pub struct OffsetDimIter {
    offset: Vec3u,
    inner: DimIter,
}

impl OffsetDimIter {
    fn new(offset: Vec3u, inner: DimIter) -> Self {
        Self { offset, inner }
    }
}

#[cfg(feature = "iter")]
impl Iterator for OffsetDimIter {
    type Item = Vec3u;
    fn next(&mut self) -> Option<Self::Item> {
        self.inner.next().map(|i| i + self.offset)
    }
}


#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn offset_dim_slice_index() {
        let data = [
            0, 1, 2,
            3, 4, 5,

            0, 10,20,
            30,40,50,
        ];
        let s = OffsetDimSlice::new(Vec3u::new(1, 2, 3), Vec3u::new(3, 2, 2), &data);
        assert_eq!(s[Vec3u::new(1, 2, 3)], 0);
        assert_eq!(s[Vec3u::new(2, 2, 3)], 1);
        assert_eq!(s[Vec3u::new(2, 3, 3)], 4);
        assert_eq!(s[Vec3u::new(2, 3, 4)], 40);
        assert_eq!(s[Vec3u::new(3, 3, 4)], 50);
    }

    #[test]
    fn offset_dim_slice_index_mut() {
        let mut data = [
            0, 1, 2,
            3, 4, 5,

            0, 10,20,
            30,40,50,
        ];
        let mut s = OffsetDimSlice::new(Vec3u::new(1, 2, 3), Vec3u::new(3, 2, 2), &mut data);
        s[Vec3u::new(1, 2, 3)] = 100;
        s[Vec3u::new(1, 3, 4)] = 300;

        assert_eq!(data, [
            100,1, 2,
            3,  4, 5,

            0, 10, 20,
            300,40,50,
        ]);
    }

    #[test]
    fn offset_dim_slice_into_iter() {
        let data = [1, 2, 3, 4];
        let s = OffsetDimSlice::new(Vec3u::new(1, 2, 3), Vec3u::new(2, 1, 2), &data);
        let mut i = s.into_iter();
        assert_eq!(*i.next().unwrap(), 1);
        assert_eq!(*i.next().unwrap(), 2);
        assert_eq!(*i.next().unwrap(), 3);
        assert_eq!(*i.next().unwrap(), 4);
        assert_eq!(i.next(), None);
    }

    #[test]
    fn offset_dim_slice_into_iter_mut() {
        let mut data = [1, 2, 3, 4];
        let s = OffsetDimSlice::new(Vec3u::new(1, 2, 3), Vec3u::new(2, 1, 2), &mut data);
        let mut i = s.into_iter();
        *i.next().unwrap() = 10;
        assert_eq!(*i.next().unwrap(), 2);
        *i.next().unwrap() += 27;
        assert_eq!(*i.next().unwrap(), 4);
        assert_eq!(i.next(), None);

        assert_eq!(data, [10,2,30,4]);
    }

    #[test]
    fn offset_dim_slice_indices() {
        let s = OffsetDimSlice::new(Vec3u::new(10, 20, 30), Vec3u::new(2, 1, 2), &[()]);
        let mut i = s.indices();
        assert_eq!(i.next().unwrap(), Vec3u::new(10, 20, 30));
        assert_eq!(i.next().unwrap(), Vec3u::new(11, 20, 30));
        assert_eq!(i.next().unwrap(), Vec3u::new(10, 20, 31));
        assert_eq!(i.next().unwrap(), Vec3u::new(11, 20, 31));
        assert_eq!(i.next(), None);
    }

    #[test]
    fn offset_dim_slice_enumerated() {
        let data = [
            10, 11,
            20, 21,
            30, 31,
        ];
        let s = OffsetDimSlice::new(Vec3u::new(10, 20, 30), Vec3u::new(1, 2, 2), &data);
        let mut i = s.enumerated();
        assert_eq!(i.next().unwrap(), (Vec3u::new(10, 20, 30), &10));
        assert_eq!(i.next().unwrap(), (Vec3u::new(10, 21, 30), &11));
        assert_eq!(i.next().unwrap(), (Vec3u::new(10, 20, 31), &20));
        assert_eq!(i.next().unwrap(), (Vec3u::new(10, 21, 31), &21));
        assert_eq!(i.next(), None);
    }

    #[test]
    fn offset_dim_slice_enumerated_mut() {
        let mut data = [
            10, 11,
            20, 21,
        ];
        let s = OffsetDimSlice::new(Vec3u::new(10, 20, 30), Vec3u::new(2, 1, 2), &mut data);
        let mut i = s.enumerated();

        let (idx, v) = i.next().unwrap();
        assert_eq!(idx, Vec3u::new(10, 20, 30));
        *v = 100;

        i.next().unwrap();
        i.next().unwrap();

        let (idx, v) = i.next().unwrap();
        assert_eq!(idx, Vec3u::new(11, 20, 31));
        *v = 210;

        assert_eq!(i.next(), None);

        assert_eq!(data, [100, 11, 20, 210]);
    }
}
