use core::ops::{Index, IndexMut};

use crate::dim::DimSlice;
use crate::mat::Material;
use crate::real::Real;
use crate::vec::{Vec3, Vec3u};

mod support;
pub use support::{
    SimMeta,
    VolumeSampleNeg,
    VolumeSamplePos,
};

// TODO: make the fields private and hide behind a constructor?
pub struct StepEContext<'a, R, M> {
    pub inv_feature_size: R,
    pub time_step: R,
    pub stim_e: Vec3<R>,
    pub mat: &'a M,
    /// Input field sampled near this location
    pub in_h: VolumeSampleNeg<R>,
    pub in_e: Vec3<R>,
}

impl<'a, R: Real, M: Material<R>> StepEContext<'a, R, M> {
    /// given the simulation resoures (material, stimulus, fields),
    /// advance the `e` field at the provided index.
    /// this is a good toplevel function to use if you don't care about any of the details.
    pub fn step_flat_view<RM, RF, WF>(
        meta: SimMeta<R>,
        mat: &RM,
        stim_e: &RF,
        e: &mut WF,
        h: &RF,
        idx: Vec3u,
    )
    where
        RM: Index<usize, Output=M> + ?Sized,
        RF: Index<usize, Output=Vec3<R>> + ?Sized,
        WF: Index<usize, Output=Vec3<R>> + IndexMut<usize> + ?Sized,
    {
        let dim = meta.dim();
        let stim_e_matrix = DimSlice::new(dim, stim_e);
        let mat_matrix = DimSlice::new(dim, mat);
        let mut e_matrix = DimSlice::new(dim, e);
        let h_matrix = DimSlice::new(dim, h);

        let stim_e = stim_e_matrix[idx];
        let mat = &mat_matrix[idx];
        let in_e = e_matrix[idx];
        let in_h = VolumeSampleNeg::from_indexable(&h_matrix, idx);

        let update_state = StepEContext {
            inv_feature_size: meta.inv_feature_size(),
            time_step: meta.time_step(),
            stim_e,
            mat,
            in_h,
            in_e,
        };
        let new_e = update_state.step_e();
        e_matrix[idx] = new_e;
    }

    /// return the e field for this cell at one time step from where it was previously evaluated.
    pub fn step_e(self) -> Vec3<R> {
        // ```tex
        // Ampere's circuital law with Maxwell's addition, in SI units ("macroscopic version"):
        //     $\nabla x H = J_f + dD/dt$  (1)
        // where $J_f$ = current density = $\sigma E$, $\sigma$ being a material parameter ("conductivity")
        // note that $D = \epsilon_0 E + P$, but we don't simulate any material where $P \ne 0$.
        //
        // substitute $D = \epsilon_0 E$ into (1):
        //     $\nabla x H = J_f + \epsilon_0 dE/dt$
        // expand with $J_f = \sigma E$:
        //     $\nabla x H = \sigma E + \epsilon_0 dE/dt$
        // rearrange:
        //     $dE/dt = 1/\epsilon_0 (\nabla x H - \sigma E)$  (2)
        //
        // let $E_p$ be $E$ at $T - \Delta\!t$ and $E_n$ be $E$ at $T+\Delta\!t$.
        // apply these substitutions into a linear expansion of (2):
        //     $(E_n - E_p)/(2\Delta\!t) = 1/\epsilon_0 (\nabla x H - \sigma (E_n + E_p)/2)$
        // normalize:
        //     $E_n - E_p = 2\Delta\!t/\epsilon_0 (\nabla x H - \sigma (E_n + E_p)/2)$
        // expand:
        //     $E_n - E_p = 2\Delta\!t/\epsilon_0 \nabla x H - \sigma \Delta\!t/\epsilon_0 (E_n + E_p)$
        // replace $E_n = E_p + \Delta\!E$
        //     $\Delta\!E = 2\Delta\!t/\epsilon_0 \nabla x H - \sigma \Delta\!t/\epsilon_0 (2E_p + \Delta\!E)$
        // rearrange:
        //     $\Delta\!E (1 + \sigma\Delta\!t/\epsilon_0) = 2\Delta\!t/\epsilon_0 (\nabla x H - \sigma E_p$
        // normalize:
        //     $\Delta\!E (\epsilon_0 + \sigma\Delta\!t) = 2\Delta\!t (\nabla x H - \sigma E_p)$  (3)
        // then $\Delta\!E$ is trivially solved, and $E_n = E_p + \Delta\!E$
        // ```
        let twice_eps0 = R::twice_eps0();
        let deltas = self.in_h.delta_h();
        // \nabla x H
        let nabla_h = deltas.nabla() * self.inv_feature_size;
        let sigma = self.mat.conductivity();
        let e_prev = self.in_e;

        // evaluate (3)
        // NB: the simulation uses $\Delta t$ to mean the time between E_p and E_n,
        // whereas the math above denotes that as $2\Delta t$.
        // hence, we actually compute:
        // ```tex
        //     $\Delta\!E (\epsilon_0 + \frac{1}{2}\sigma\Delta\!t) = \Delta\!t (\nabla x H - \sigma E_p)$
        // or:
        //     $\Delta\!E (2\epsilon_0 + \sigma\Delta\!t) = 2\Delta\!t (\nabla x H - \sigma E_p)$
        // ```
        // XXX: this evaluation is very prone to rounding error.
        // - eps0 is *very* small -- about 1e-11.
        // - combined with the division, i worry about error accumulation here.
        let delta_e = (nabla_h - e_prev.elem_mul(sigma)).elem_div(
            sigma*self.time_step + Vec3::uniform(twice_eps0)
        )*(R::two()*self.time_step);

        // E_n = E_p + \Delta E + user_injected_field
        e_prev + delta_e + self.stim_e
    }
}


pub struct StepHContext<'a, R, M> {
    pub inv_feature_size: R,
    pub time_step: R,
    pub stim_h: Vec3<R>,
    pub mat: &'a M,
    /// Input field sampled near this location
    pub in_e: VolumeSamplePos<R>,
    pub in_h: Vec3<R>,
    pub in_m: Vec3<R>,
}

impl<'a, R: Real, M: Material<R>> StepHContext<'a, R, M> {
    /// given the simulation resoures (material, stimulus, fields),
    /// advance the `h` field at the provided index.
    /// this is a good toplevel function to use if you don't care about any of the details.
    pub fn step_flat_view<RM, RF, WF>(
        meta: SimMeta<R>,
        mat: &RM,
        stim_h: &RF,
        e: &RF,
        h: &mut WF,
        m: &mut WF,
        idx: Vec3u,
    )
    where
        RM: Index<usize, Output=M> + ?Sized,
        RF: Index<usize, Output=Vec3<R>> + ?Sized,
        WF: Index<usize, Output=Vec3<R>> + IndexMut<usize> + ?Sized,
    {
        let dim = meta.dim();
        let stim_h_matrix = DimSlice::new(dim, stim_h);
        let mat_matrix = DimSlice::new(dim, mat);
        let e_matrix = DimSlice::new(dim, e);
        let mut h_matrix = DimSlice::new(dim, h);
        let mut m_matrix = DimSlice::new(dim, m);

        let stim_h = stim_h_matrix[idx];
        let mat = &mat_matrix[idx];
        let in_e = VolumeSamplePos::from_indexable(&e_matrix, dim, idx);
        let in_h = h_matrix[idx];
        let in_m = m_matrix[idx];

        let update_state = StepHContext {
            inv_feature_size: meta.inv_feature_size(),
            time_step: meta.time_step(),
            stim_h,
            mat,
            in_e,
            in_h,
            in_m,
        };
        let (new_h, new_m) = update_state.step_h();
        h_matrix[idx] = new_h;
        m_matrix[idx] = new_m;
    }

    /// return the `(h, m)` fields for this cell at one time-step from where it was previously
    /// evaluated.
    pub fn step_h(self) -> (Vec3<R>, Vec3<R>) {
        // ```tex
        // Maxwell-Faraday equation:
        //     $\nabla x E = -dB/dt$  (1)
        //
        // where
        //     $B = \mu_0 (H + M)$  (2)
        // ```
        let mu0 = R::mu0();
        let mu0_inv = R::mu0_inv();
        let deltas = self.in_e.delta_e();
        // \nabla x E
        // TODO: inv_feature_size and time_step could be folded
        let nabla_e = deltas.nabla() * self.inv_feature_size;
        let delta_b = nabla_e * (-self.time_step);

        // Relation between these is: B = mu0*(H + M)
        let old_h = self.in_h;
        let old_m = self.in_m;
        let old_b = (old_h + old_m) * mu0;

        // evaluate the next B value.
        // eq (1) enforces this relation. H and M can vary freely within that context,
        // so long as they sum to B as per (2).
        // we let the Material dictate M, and from that compute H.
        // TODO(opt): fold stim_h * mu0 into the old_b = ... * mu0 calculation above
        let new_b = old_b + delta_b + self.stim_h * mu0;
        let mat = self.mat;  // XXX: copy off of self to avoid a rust-gpu bug around ZSTs.
        let new_m = mat.move_b_vec(old_m, new_b);
        let new_h = new_b * mu0_inv - new_m;

        (new_h, new_m)
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::compound::Optional;
    use crate::mat::{AnisomorphicConductor, Vacuum};
    use float_eq::assert_float_eq;
    use std::cell::Cell;

    fn assert_vec_eq(a: Vec3<f64>, b: Vec3<f64>) {
        assert_float_eq!(a.x(), b.x(), r2nd <= 1e-9, "{:?} != {:?}", a, b);
        assert_float_eq!(a.y(), b.y(), r2nd <= 1e-9, "{:?} != {:?}", a, b);
        assert_float_eq!(a.z(), b.z(), r2nd <= 1e-9, "{:?} != {:?}", a, b);
    }

    #[test]
    fn step_e_trivial() {
        let ctx = StepEContext {
            inv_feature_size: 1.0,
            time_step: 1.0,
            stim_e: Vec3::zero(),
            mat: &Vacuum,
            in_h: VolumeSampleNeg::default(),
            in_e: Vec3::new(1.0, 2.0, 3.0),
        };

        let new_e = ctx.step_e();
        // when the h field has zero spatial derivative, E has zero time derivative.
        assert_vec_eq(new_e, Vec3::new(1.0, 2.0, 3.0));
    }
    #[test]
    fn step_e_applies_stimulus() {
        let ctx = StepEContext {
            inv_feature_size: 1.0,
            time_step: 1.0,
            stim_e: Vec3::new(-3.0, -2.0, -1.0),
            mat: &Vacuum,
            in_h: VolumeSampleNeg::default(),
            in_e: Vec3::new(1.0, 2.0, 3.0),
        };

        let new_e = ctx.step_e();
        // when the h field has zero spatial derivative, E has zero time derivative.
        assert_vec_eq(new_e, Vec3::new(-2.0, 0.0, 2.0));
    }
    #[test]
    fn step_e_understands_nabla_h() {
        let mid = Vec3::new(4.0, 8.0, 12.0);
        let ctx = StepEContext {
            inv_feature_size: 1.0,
            time_step: 1.0,
            stim_e: Vec3::zero(),
            mat: &Vacuum,
            in_h: VolumeSampleNeg {
                mid,
                xm1: Optional::some(mid - Vec3::new(1.0, 2.0, 3.0)),
                ym1: Optional::some(mid - Vec3::new(4.0, 5.0, 6.0)),
                zm1: Optional::some(mid - Vec3::new(7.0, 8.0, 9.0)),
            },
            in_e: Vec3::zero(),
        };
        //```tex
        // $\nabla x H$:
        //     $[ \Delta\!H_z/\Delta\!y - \Delta\!H_y/\Delta\!z$
        //     $  \Delta\!H_x/\Delta\!z - \Delta\!H_z/\Delta\!x$
        //     $  \Delta\!H_y/\Delta\!x - \Delta\!H_x/\Delta\!y ]$
        // we take these values from `in_h`:
        //    $[ 6.0 - 8.0 $
        //    $  7.0 - 3.0$
        //    $  2.0 - 4.0 ]$
        // this gives $dD/dt$, and then we scale by $\epsilon_0^{-1}$ to get $dE/dt$.
        //```
        let delta_d = Vec3::new(-2.0, 4.0, -2.0);
        let delta_e = delta_d * f64::eps0_inv();

        let new_e = ctx.step_e();
        // when the h field has zero spatial derivative, E has zero time derivative.
        assert_vec_eq(new_e, delta_e);
    }

    #[test]
    fn step_e_respects_material() {
        let cond = Vec3::new(4.0e-10, 5.0e-10, 6.0e-10);
        let mat = AnisomorphicConductor::new(cond);
        let in_e = Vec3::new(1.0, 2.0, 3.0);
        let ctx = StepEContext {
            inv_feature_size: 1.0,
            time_step: 1.0,
            stim_e: Vec3::zero(),
            mat: &mat,
            in_h: VolumeSampleNeg::default(),
            in_e,
        };
        //```tex
        // guiding Maxwell equation:
        //     $\nabla x H = J_f + dD/dt$
        // we've forced $\Nabla x H = 0$, so
        //     $dD/dt = -J_f$
        // simplified:
        //     $\epsilon_0 dE/dt = -\sigma E$
        //```
        let new_e = ctx.step_e();
        let mid_e = (in_e + new_e)*0.5;
        let delta_e = new_e - in_e;
        assert_vec_eq(delta_e, cond.elem_mul(mid_e) * -f64::eps0_inv());
    }
    #[test]
    fn step_e_understands_time_scale_feature_size() {
        let cond = Vec3::new(4.0, 5.0, 6.0);
        let mat = AnisomorphicConductor::new(cond);
        let in_e = Vec3::new(1.0, 2.0, 3.0);
        let ctx = StepEContext {
            inv_feature_size: 1e6,
            time_step: 1e-10,
            stim_e: Vec3::zero(),
            mat: &mat,
            in_h: VolumeSampleNeg {
                mid: Vec3::zero(),
                xm1: Optional::some(-Vec3::new(1e-6, 2e-6, 3e-6)),
                ym1: Optional::some(-Vec3::new(4e-6, 5e-6, 6e-6)),
                zm1: Optional::some(-Vec3::new(7e-6, 8e-6, 9e-6)),
            },
            in_e,
        };
        //```tex
        // guiding Maxwell equation:
        //     $\nabla x H = J_f + dD/dt$
        // where $D = \epsilon_0 E$
        // and $J_f = \sigma E$
        //```
        let new_e = ctx.step_e();
        let mid_e = (in_e + new_e)*0.5;
        let delta_e = new_e - in_e;
        // see step_e_understands_nabla_h
        // we pre-scaled the in_h values by feature_size in order to get the same nabla_h as before
        let nabla_h_component = Vec3::new(-2.0, 4.0, -2.0);
        let current_component = cond.elem_mul(mid_e);
        assert_vec_eq(delta_e, (nabla_h_component - current_component) * f64::eps0_inv() * 1e-10);
    }

    #[test]
    fn step_h_trivial() {
        let ctx = StepHContext {
            inv_feature_size: 1.0,
            time_step: 1.0,
            stim_h: Vec3::zero(),
            mat: &Vacuum,
            in_e: VolumeSamplePos::default(),
            in_h: Vec3::new(1.0, 2.0, 3.0),
            in_m: Vec3::zero(),
        };

        let (new_h, new_m) = ctx.step_h();
        // when the e field has zero spatial derivative, B has zero time derivative.
        assert_vec_eq(new_h, Vec3::new(1.0, 2.0, 3.0));
        assert_vec_eq(new_m, Vec3::zero());
    }
    #[test]
    fn step_h_applies_stimulus() {
        let ctx = StepHContext {
            inv_feature_size: 1.0,
            time_step: 1.0,
            stim_h: Vec3::new(-3.0, -2.0, -1.0),
            mat: &Vacuum,
            in_e: VolumeSamplePos::default(),
            in_h: Vec3::new(1.0, 2.0, 3.0),
            in_m: Vec3::zero(),
        };

        let (new_h, new_m) = ctx.step_h();
        // when the e field has zero spatial derivative, B has zero time derivative.
        assert_vec_eq(new_h, Vec3::new(-2.0, 0.0, 2.0));
        assert_vec_eq(new_m, Vec3::zero());
    }
    #[test]
    fn step_h_understands_nabla_e() {
        let mid = Vec3::new(4.0, 8.0, 12.0);
        let ctx = StepHContext {
            inv_feature_size: 1.0,
            time_step: 1.0,
            stim_h: Vec3::zero(),
            mat: &Vacuum,
            in_e: VolumeSamplePos {
                mid,
                xp1: Optional::some(mid + Vec3::new(1.0, 2.0, 3.0)),
                yp1: Optional::some(mid + Vec3::new(4.0, 5.0, 6.0)),
                zp1: Optional::some(mid + Vec3::new(7.0, 8.0, 9.0)),
            },
            in_h: Vec3::zero(),
            in_m: Vec3::zero(),
        };

        //```tex
        // $\nabla x E$:
        //     $[ \Delta\!E_z/\Delta\!y - \Delta\!E_y/\Delta\!z$
        //     $  \Delta\!E_x/\Delta\!z - \Delta\!E_z/\Delta\!x$
        //     $  \Delta\!E_y/\Delta\!x - \Delta\!E_x/\Delta\!y ]$
        // we take these values from `in_e`:
        //    $[ 6.0 - 8.0 $
        //    $  7.0 - 3.0$
        //    $  2.0 - 4.0 ]$
        // this gets negated, and then scaled by $\mu_0^{-1}$
        //```
        let delta_b = -Vec3::new(-2.0, 4.0, -2.0);

        let (new_h, new_m) = ctx.step_h();
        assert_vec_eq(new_h, delta_b * f64::mu0_inv());
        assert_vec_eq(new_m, Vec3::zero());
    }
    #[test]
    fn step_h_understands_time_scale_feature_size() {
        let mid = Vec3::new(4.0, 8.0, 12.0);
        let ctx = StepHContext {
            inv_feature_size: 5.0, // $\Delta x = 0.2$
            time_step: 1e-3,
            stim_h: Vec3::zero(),
            mat: &Vacuum,
            in_e: VolumeSamplePos {
                mid,
                xp1: Optional::some(mid + Vec3::new(1.0, 2.0, 3.0)),
                yp1: Optional::some(mid + Vec3::new(4.0, 5.0, 6.0)),
                zp1: Optional::some(mid + Vec3::new(7.0, 8.0, 9.0)),
            },
            in_h: Vec3::zero(),
            in_m: Vec3::zero(),
        };

        // see step_h_understands_nabla_e
        // the nabla is dE/dx, so divide by \Delta x.
        // nabla relates to dB/dt, so multiply by \Delta t to get \Delta B
        let delta_b = -Vec3::new(-2.0, 4.0, -2.0) * 1e-3 / 0.2;

        let (new_h, new_m) = ctx.step_h();
        assert_vec_eq(new_h, delta_b * f64::mu0_inv());
        assert_vec_eq(new_m, Vec3::zero());
    }

    struct MockHMaterial {
        response: Vec3<f64>,
        called_with: Cell<(Vec3<f64>, Vec3<f64>)>,
    }
    impl Material<f64> for MockHMaterial {
        fn move_b_vec(&self, m: Vec3<f64>, target_b: Vec3<f64>) -> Vec3<f64> {
            self.called_with.set((m, target_b));
            self.response
        }
    }
    #[test]
    fn step_h_respects_material() {
        let mock_mat = MockHMaterial {
            response: Vec3::new(1e-6, 2e-6, 3e-6),
            called_with: Default::default(),
        };
        let ctx = StepHContext {
            inv_feature_size: 1.0,
            time_step: 1.0,
            stim_h: Vec3::zero(),
            mat: &mock_mat,
            in_e: VolumeSamplePos {
                mid: Vec3::zero(),
                xp1: Optional::some(Vec3::new(1.0, 2.0, 3.0)),
                yp1: Optional::some(Vec3::new(4.0, 5.0, 6.0)),
                zp1: Optional::some(Vec3::new(7.0, 8.0, 9.0)),
            },
            in_h: Vec3::zero(),
            in_m: Vec3::zero(),
        };
        // see step_h_understands_nabla_e
        let delta_b = -Vec3::new(-2.0, 4.0, -2.0);

        let (new_h, new_m) = ctx.step_h();
        assert_vec_eq(mock_mat.called_with.get().0, Vec3::zero()); // not magnetized
        assert_vec_eq(mock_mat.called_with.get().1, delta_b);
        assert_vec_eq(new_m, mock_mat.response);
        assert_vec_eq(new_h, delta_b * f64::mu0_inv() - new_m);
    }
    #[test]
    fn step_h_understands_previous_m_h() {
        let mock_mat = MockHMaterial {
            response: Vec3::new(1e-6, 2e-6, 3e-6),
            called_with: Default::default(),
        };
        let in_h = Vec3::new(4e-6, 5e-6, 6e-6);
        let in_m = Vec3::new(7e-6, 8e-6, 9e-6);
        let ctx = StepHContext {
            inv_feature_size: 1.0,
            time_step: 1.0,
            stim_h: Vec3::zero(),
            mat: &mock_mat,
            in_e: VolumeSamplePos {
                mid: Vec3::zero(),
                xp1: Optional::some(Vec3::new(1.0, 2.0, 3.0)),
                yp1: Optional::some(Vec3::new(4.0, 5.0, 6.0)),
                zp1: Optional::some(Vec3::new(7.0, 8.0, 9.0)),
            },
            in_h,
            in_m,
        };
        // see step_h_understands_nabla_e
        let delta_b = -Vec3::new(-2.0, 4.0, -2.0);
        let prev_b = (in_h + in_m) * f64::mu0();
        let new_b = prev_b + delta_b;

        let (new_h, new_m) = ctx.step_h();
        assert_vec_eq(mock_mat.called_with.get().0, in_m);
        assert_vec_eq(mock_mat.called_with.get().1, new_b);
        assert_vec_eq(new_m, mock_mat.response);
        assert_vec_eq(new_h, new_b * f64::mu0_inv() - new_m);
    }
}
