use core::ops::Index;

use crate::compound::Optional;
use crate::real::Real;
use crate::vec::{Vec3, Vec3u};

#[cfg(feature = "serde")]
use serde::{Serialize, Deserialize};


#[cfg_attr(feature = "serde", derive(Deserialize, Serialize))]
#[cfg_attr(feature = "fmt", derive(Debug))]
#[derive(Copy, Clone, Default, PartialEq)]
pub struct SimMeta<R> {
    dim: Vec3u,
    inv_feature_size: R,
    time_step: R,
    feature_size: R,
}

impl<R: Real> SimMeta<R> {
    pub fn new(dim: Vec3u, feature_size: R, time_step: R) -> Self {
        Self {
            dim,
            inv_feature_size: feature_size.inv(),
            time_step,
            feature_size
        }
    }
}

impl<R> SimMeta<R> {
    pub fn dim(&self) -> Vec3u {
        self.dim
    }
}

impl<R: Copy> SimMeta<R> {
    pub fn inv_feature_size(&self) -> R {
        self.inv_feature_size
    }
    pub fn time_step(&self) -> R {
        self.time_step
    }
    pub fn feature_size(&self) -> R {
        self.feature_size
    }
}

impl<R: Real> SimMeta<R> {
    pub fn cast<R2: Real>(self) -> SimMeta<R2> {
        SimMeta {
            dim: self.dim,
            inv_feature_size: self.inv_feature_size.cast(),
            time_step: self.time_step.cast(),
            feature_size: self.feature_size.cast(),
        }
    }
}

/// Package the field vectors adjacent to some particular location.
/// Particular those at negative offsets from the midpoint.
/// This is used in step_e when looking at the H field deltas.
#[derive(Copy, Clone, Default)]
pub struct VolumeSampleNeg<R> {
    pub mid: Vec3<R>,
    // TODO(optimization): if we just force the boundary cells to be always E=0, H=0,
    // we can save a lot of bounds checks and special cases, at the slight expense that the
    // boundaries now reflect all energy and we use slightly more memory (6*N^2 extra cells).
    // having the edges be perfectly reflective might actually simplify testing though (e.g. energy
    // would actually be preserved?)
    pub xm1: Optional<Vec3<R>>,
    pub ym1: Optional<Vec3<R>>,
    pub zm1: Optional<Vec3<R>>,
}

impl<R: Copy + Default> VolumeSampleNeg<R> {
    pub fn from_indexable<I: Index<Vec3u, Output=Vec3<R>>>(i: &I, idx: Vec3u) -> Self {
        VolumeSampleNeg {
            mid: i[idx],
            xm1: prev_x(idx).map(|idx| i[idx]),
            ym1: prev_y(idx).map(|idx| i[idx]),
            zm1: prev_z(idx).map(|idx| i[idx]),
        }
    }
}

fn prev_x(idx: Vec3u) -> Optional<Vec3u> {
    match idx.into() {
        (0, _, _) => Optional::none(),
        (x, y, z) => Optional::some(Vec3u::new(x-1, y, z)),
    }
}
fn prev_y(idx: Vec3u) -> Optional<Vec3u> {
    match idx.into() {
        (_, 0, _) => Optional::none(),
        (x, y, z) => Optional::some(Vec3u::new(x, y-1, z)),
    }
}
fn prev_z(idx: Vec3u) -> Optional<Vec3u> {
    match idx.into() {
        (_, _, 0) => Optional::none(),
        (x, y, z) => Optional::some(Vec3u::new(x, y, z-1)),
    }
}

impl<R: Real> VolumeSampleNeg<R> {
    /// Calculate the delta in H values amongst this cell and its neighbors (left/up/out)
    pub fn delta_h(self) -> FieldDeltas<R> {
        let mid = self.mid;
        // let (dfy_dx, dfz_dx) = self.xm1.map(|xm1| {
        //     (mid.y() - xm1.y(), mid.z() - xm1.z())
        // }).unwrap_or_default();

        // let (dfx_dy, dfz_dy) = self.ym1.map(|ym1| {
        //     (mid.x() - ym1.x(), mid.z() - ym1.z())
        // }).unwrap_or_default();

        // let (dfx_dz, dfy_dz) = self.zm1.map(|zm1| {
        //     (mid.x() - zm1.x(), mid.y() - zm1.y())
        // }).unwrap_or_default();

        let (dfy_dx, dfz_dx) = if self.xm1.is_some() {
            (mid.y() - self.xm1.unwrap().y(), mid.z() - self.xm1.unwrap().z())
        } else {
            (R::zero(), R::zero())
        };

        let (dfx_dy, dfz_dy) = if self.ym1.is_some() {
            (mid.x() - self.ym1.unwrap().x(), mid.z() - self.ym1.unwrap().z())
        } else {
            (R::zero(), R::zero())
        };

        let (dfx_dz, dfy_dz) = if self.zm1.is_some() {
            (mid.x() - self.zm1.unwrap().x(), mid.y() - self.zm1.unwrap().y())
        } else {
            (R::zero(), R::zero())
        };
        
        FieldDeltas {
            dfy_dx,
            dfz_dx,
            dfx_dy,
            dfz_dy,
            dfx_dz,
            dfy_dz,
        }
    }
}


/// Package the field vectors adjacent to some particular location.
/// Particular those at positive offsets from the midpoint.
/// This is used in step_h when looking at the E field deltas.
#[derive(Copy, Clone, Default)]
pub struct VolumeSamplePos<R> {
    pub mid: Vec3<R>,
    pub xp1: Optional<Vec3<R>>,
    pub yp1: Optional<Vec3<R>>,
    pub zp1: Optional<Vec3<R>>
}

impl<R: Copy + Default> VolumeSamplePos<R> {
    pub fn from_indexable<I: Index<Vec3u, Output=Vec3<R>>>(i: &I, dim: Vec3u, idx: Vec3u) -> Self {
        VolumeSamplePos {
            mid: i[idx],
            xp1: next_x(dim, idx).map(|idx| i[idx]),
            yp1: next_y(dim, idx).map(|idx| i[idx]),
            zp1: next_z(dim, idx).map(|idx| i[idx]),
        }
    }
}

fn next_x(dim: Vec3u, idx: Vec3u) -> Optional<Vec3u> {
    match idx.into() {
        (x, y, z) if x + 1 < dim.x() => Optional::some(Vec3u::new(x+1, y, z)),
        _ => Optional::none(),
    }
}
fn next_y(dim: Vec3u, idx: Vec3u) -> Optional<Vec3u> {
    match idx.into() {
        (x, y, z) if y + 1 < dim.y() => Optional::some(Vec3u::new(x, y+1, z)),
        _ => Optional::none(),
    }
}
fn next_z(dim: Vec3u, idx: Vec3u) -> Optional<Vec3u> {
    match idx.into() {
        (x, y, z) if z + 1 < dim.z() => Optional::some(Vec3u::new(x, y, z+1)),
        _ => Optional::none(),
    }
}

impl<R: Real> VolumeSamplePos<R> {
    /// Calculate the delta in E values amongst this cell and its neighbors (right/down/in)
    pub fn delta_e(self) -> FieldDeltas<R> {
        let mid = self.mid;
        // let (dfy_dx, dfz_dx) = self.xp1.map(|xp1| {
        //     (xp1.y() - mid.y(), xp1.z() - mid.z())
        // }).unwrap_or_default();

        // let (dfx_dy, dfz_dy) = self.yp1.map(|yp1| {
        //     (yp1.x() - mid.x(), yp1.z() - mid.z())
        // }).unwrap_or_default();

        // let (dfx_dz, dfy_dz) = self.zp1.map(|zp1| {
        //     (zp1.x() - mid.x(), zp1.y() - mid.y())
        // }).unwrap_or_default();

        let (dfy_dx, dfz_dx) = if self.xp1.is_some() {
            (self.xp1.unwrap().y() - mid.y(), self.xp1.unwrap().z() - mid.z())
        } else {
            (R::zero(), R::zero())
        };

        let (dfx_dy, dfz_dy) = if self.yp1.is_some() {
            (self.yp1.unwrap().x() - mid.x(), self.yp1.unwrap().z() - mid.z())
        } else {
            (R::zero(), R::zero())
        };

        let (dfx_dz, dfy_dz) = if self.zp1.is_some() {
            (self.zp1.unwrap().x() - mid.x(), self.zp1.unwrap().y() - mid.y())
        } else {
            (R::zero(), R::zero())
        };
        
        FieldDeltas {
            dfy_dx,
            dfz_dx,
            dfx_dy,
            dfz_dy,
            dfx_dz,
            dfy_dz,
        }
    }
}

pub struct FieldDeltas<R> {
    dfy_dx: R,
    dfz_dx: R,
    dfx_dy: R,
    dfz_dy: R,
    dfx_dz: R,
    dfy_dz: R,
}

impl<R: Real> FieldDeltas<R> {
    pub fn nabla(self) -> Vec3<R> {
        Vec3::new(
            self.dfz_dy - self.dfy_dz,
            self.dfx_dz - self.dfz_dx,
            self.dfy_dx - self.dfx_dy,
        )
    }
}

