#!/usr/bin/env python3

from inverter_characteristics import Piecewise

import subprocess
import sys

# these floats can't be represented as f32, so stringify bad on the rust side
_15e9  =  "15000000512e0"
_3e10  =  "30000001024e0"
_5e10  =  "49999998976e0"
_15e10 = "149999992832e0"

class SimParams:
    def __init__(self, couplings: int, wrappings_spec: int, um: int, drive: str) -> str:
        self.couplings = couplings
        self.wrappings_spec = wrappings_spec
        self.um = um
        self.drive_str = drive

    def __eq__(self, other: 'SimParams') -> bool:
        return self.tuple == other.tuple

    def __repr__(self) -> str:
        return f"SimParams({self.human_name})"

    @property
    def run(self) -> str:
        raise NotImplementedError

    @property
    def drive(self) -> int:
        return int(float(self.drive_str))

    @property
    def wrappings(self) -> int:
        return self.wrappings_spec * 2 + 1

    @property
    def um_str(self) -> str:
        return {
            400:  "0.0004",
            600:  "0.00059999997",
            800:  "0.0008",
            1200: "0.0011999999",
        }[self.um]

    @property
    def is_inverter(self) -> bool:
        raise NotImplementedError

    @property
    def is_discharge_high(self) -> bool:
        return False  # see 41-xx

    @property
    def t_init(self) -> float:
        return 2e-9

    @property
    def t_last(self) -> float:
        return 3e-9

    @property
    def t_mid(self) -> float:
        return None

    @property
    def machine_name(self) -> str:
        return f"{self.run}-{self.um_str}rad-{self.couplings}coupling-{self.wrappings}_1_winding-{self.drive_str}-drive"

    @property
    def human_name(self) -> str:
        return f"{self.run} {self.couplings}x {self.wrappings}:1 ({self.um}um, {self.drive_str} I)"

    @property
    def tuple(self):
        return (self.run, self.um, self.couplings, self.wrappings, self.is_inverter, self.drive)

    def ensure_inverter(self, curve):
        if not self.is_inverter:
            return curve.logically_inverted()
        return curve

class SimParams40(SimParams):
    @property
    def run(self) -> str:
        return "40"

    @property
    def is_inverter(self) -> bool:
        return False

class SimParams41(SimParams):
    @property
    def run(self) -> str:
        return "41"

    @property
    def is_inverter(self) -> bool:
        return False

    @property
    def is_discharge_high(self) -> bool:
        # 41-xx has the unfortunate case that we encoded the CLEAR operation
        # to charge core 0 HIGH, and push core 1 towards LOW on successful transfer
        return True

class SimParams42(SimParams):
    @property
    def run(self) -> str:
        return "42"

    @property
    def is_inverter(self) -> bool:
        return True

class SimParams43(SimParams):
    @property
    def run(self) -> str:
        return "43"

    @property
    def is_inverter(self) -> bool:
        return False

    @property
    def t_last(self) -> float:
        return 5e-9

    @property
    def t_mid(self) -> float:
        return 4e-9

class SimParams46(SimParams):
    @property
    def run(self) -> str:
        return "46"

    @property
    def is_inverter(self) -> bool:
        return True

    @property
    def t_last(self) -> float:
        return 5e-9

    @property
    def t_mid(self) -> float:
        return 4e-9

class SimParams46Float(SimParams46):
    # 46, but with the final core left floating for the full sim
    @property
    def run(self) -> str:
        return "46-float"

class SimParams47(SimParams):
    @property
    def run(self) -> str:
        return "47"

    @property
    def is_inverter(self) -> bool:
        return False

    @property
    def t_last(self) -> float:
        return 5e-9

    @property
    def t_mid(self) -> float:
        return 4e-9

class SimParamsV3(SimParams):
    """
    SimParams where the conductivities and clock length are configurable
    """
    def __init__(
        self,
        control_cond: int,
        coupling_cond: int,
        clock_length_ps: int,
        clock_decay_ps: int,
        couplings: int,
        wrappings_spec: int,
        um: int,
        drive: str
    ):
        super().__init__(couplings, wrappings_spec, um, drive)
        self.control_cond = int(control_cond)
        self.coupling_cond = int(coupling_cond)
        self.clock_length_ps = int(clock_length_ps)
        self.clock_decay_ps = int(clock_decay_ps)

    @property
    def t_init(self) -> float:
        return 2e-12 * self.clock_length_ps

    @property
    def t_last(self) -> float:
        return 3e-12 * self.clock_length_ps

    @property
    def machine_name(self) -> str:
        return f"{self.run}-{self.um_str}rad-{self.control_cond}ctl_cond-{self.coupling_cond}coupling_cond-{self.clock_length_ps}ps-{self.clock_decay_ps}ps-{self.couplings}ctl-{self.couplings}coupling-{self.wrappings}_1_winding-{self.drive_str}-drive"

    @property
    def human_name(self) -> str:
        name = super().human_name
        return f"{name} ({self.clock_length_ps}ps {self.control_cond}/{self.coupling_cond} cond)"

class SimParams48(SimParamsV3):
    @property
    def run(self) -> str:
        return "48"

    @property
    def is_inverter(self) -> bool:
        return False

class SimParams50(SimParamsV3):
    @property
    def run(self) -> str:
        return "50"

    @property
    def is_inverter(self) -> bool:
        return False

    @property
    def t_last(self) -> float:
        return 5e-12 * self.clock_length_ps

    @property
    def t_mid(self) -> float:
        return 4e-12 * self.clock_length_ps

class SimParams51(SimParamsV3):
    @property
    def run(self) -> str:
        return "51"

    @property
    def is_inverter(self) -> bool:
        return True

    @property
    def t_init(self) -> float:
        return 1e-12 * self.clock_length_ps

    @property
    def t_mid(self) -> float:
        return 2e-12 * self.clock_length_ps

    @property
    def t_last(self) -> float:
        return 2e-12 * self.clock_length_ps

sims = [
    # params, human friendly db name, normalization
    (SimParams40(6,  1,  400, "5e10"), "fwd_40_6_3_1_5e10",        17000),
    (SimParams40(12, 1,  400, "5e10"), "fwd_40_12_3_1_5e10",       17000),
    (SimParams40(6,  3,  400, "5e10"), "fwd_40_6_7_1_5e10",        15000),

    (SimParams40(20, 1,  600, "3e10"), "fwd_40_600um_20_3_1_3e10", 10000),
    (SimParams40(20, 1,  600, "5e10"), "fwd_40_600um_20_3_1_5e10", 19000),
    (SimParams40(20, 1,  600, "1e11"), "fwd_40_600um_20_3_1_1e11", 15000),
    (SimParams40(20, 1,  600, "2e11"), "fwd_40_600um_20_3_1_2e11", 15000),
    (SimParams40(6,  2,  600, "5e10"), "fwd_40_600um_6_5_1_5e10",  20000),
    (SimParams40(8,  2,  600, "5e10"), "fwd_40_600um_8_5_1_5e10",  20000),
    (SimParams40(10, 2,  600, "5e10"), "fwd_40_600um_10_5_1_5e10", 20000),
    (SimParams40(12, 2,  600, "5e10"), "fwd_40_600um_12_5_1_5e10", 12000),
    (SimParams40(12, 2,  600, "1e11"), "fwd_40_600um_12_5_1_1e11", 14000),
    (SimParams40(12, 2,  600, "2e11"), "fwd_40_600um_12_5_1_2e11", 14000),
    (SimParams40(6,  3,  600, "5e10"), "fwd_40_600um_6_7_1_5e10",  20000),
    (SimParams40(8,  3,  600, "5e10"), "fwd_40_600um_8_7_1_5e10",  12000),
    (SimParams40(8,  3,  600, "1e11"), "fwd_40_600um_8_7_1_1e11",  12000),
    # (SimParams40(10, 3, 600, "5e10"), "fwd_40_600um_10_7_1_5e10", 20000),
    (SimParams40(6,  4,  600, "5e10"), "fwd_40_600um_6_9_1_5e10",  12000),
    (SimParams40(6,  4,  600, "1e11"), "fwd_40_600um_6_9_1_1e11",  12000),
    (SimParams40(6,  5,  600, "5e10"), "fwd_40_600um_6_11_1_5e10", 10000),

    (SimParams40(24, 1,  800, "1e11"), "fwd_40_800um_24_3_1_1e11", 12000),
    (SimParams40(12, 2,  800, "5e10"), "fwd_40_800um_12_5_1_5e10",  8000),
    (SimParams40(12, 2,  800, "1e11"), "fwd_40_800um_12_5_1_1e11",  8000),
    (SimParams40(18, 2,  800, "5e10"), "fwd_40_800um_18_5_1_5e10", 10000),
    (SimParams40(18, 2,  800, "1e11"), "fwd_40_800um_18_5_1_1e11", 10000),
    (SimParams40(18, 2,  800, "2e11"), "fwd_40_800um_18_5_1_1e11", 20000),
    (SimParams40(12, 3,  800, "5e10"), "fwd_40_800um_12_7_1_5e10",  8000),
    (SimParams40(12, 3,  800, "1e11"), "fwd_40_800um_12_7_1_1e11",  8000),
    (SimParams40(12, 3,  800, "2e11"), "fwd_40_800um_12_7_1_2e11",  8000),
    (SimParams40(12, 3,  800, "4e11"), "fwd_40_800um_12_7_1_4e11",  8000),
    (SimParams40(10, 4,  800, "5e10"), "fwd_40_800um_10_9_1_5e10",  8000),
    (SimParams40(10, 4,  800, "1e11"), "fwd_40_800um_10_9_1_1e11",  8000),
    (SimParams40(10, 4,  800, "2e11"), "fwd_40_800um_10_9_1_2e11",  8000),
    (SimParams40(8,  5,  800, "5e10"), "fwd_40_800um_8_11_1_5e10",  8000),
    (SimParams40(8,  5,  800, "1e11"), "fwd_40_800um_8_11_1_1e11",  8000),

    (SimParams40(12, 4, 1200, "1e11"), "fwd_40_1200um_12_9_1_1e11",  8000),
    (SimParams40(12, 4, 1200, "2e11"), "fwd_40_1200um_12_9_1_2e11",  8000),
    (SimParams40(10, 5, 1200, "1e11"), "fwd_40_1200um_10_11_1_1e11", 8000),

    (SimParams41(18, 0, 400,  "5e9"),  None, 20000),
    (SimParams41(18, 0, 400,  "1e10"), None, 20000),
    (SimParams41(9,  1, 400,  "2e9"),  None, 20000),
    (SimParams41(9,  1, 400,  "3e9"),  None, 20000),
    (SimParams41(9,  1, 400,  "5e9"),  None, 20000),
    (SimParams41(9,  1, 400,  "8e9"),  None, 20000),
    (SimParams41(9,  1, 400,  "1e10"), None, 20000),
    (SimParams41(9,  1, 400,  "12e9"), None, 20000),
    (SimParams41(9,  1, 400,  _15e9),  None, 20000),
    (SimParams41(9,  1, 400,  "2e10"), None, 20000),
    (SimParams41(6,  2, 400,  "5e9"),  None, 20000),
    (SimParams41(6,  2, 400,  "1e10"), None, 20000),
    (SimParams41(6,  2, 400,  "2e10"), None, 20000),
    (SimParams41(4,  3, 400,  "1e10"), None, 20000),
    (SimParams41(4,  3, 400,  "2e10"), None, 20000),
    (SimParams41(4,  3, 400,  "4e10"), None, 20000),

    (SimParams41(18, 1, 600,  "3e9"),  None, 20000),
    (SimParams41(18, 1, 600,  "5e9"),  None, 20000),
    (SimParams41(18, 1, 600,  "1e10"), None, 20000),
    (SimParams41(18, 1, 600,  "2e10"), None, 20000),
    (SimParams41(12, 2, 600,  "5e9"),  None, 20000),
    (SimParams41(12, 2, 600,  "1e10"), None, 20000),
    (SimParams41(9,  3, 600,  "3e9"),  None, 20000),
    (SimParams41(9,  3, 600,  "1e10"), None, 20000),
    (SimParams41(9,  3, 600,  "2e10"), None, 20000),

    (SimParams41(24, 1, 800,  "3e9"),  None, 20000),
    (SimParams41(24, 1, 800,  "5e9"),  None, 20000),
    (SimParams41(16, 2, 800,  "5e9"),  None, 20000),
    (SimParams41(16, 2, 800,  "1e10"), None, 20000),
    (SimParams41(16, 2, 800,  "2e10"), None, 20000),
    (SimParams41(10, 3, 800,  "25e8"), None, 20000),
    (SimParams41(10, 3, 800,  "3e9"),  None, 20000),
    (SimParams41(10, 3, 800,  "1e10"), None, 20000),
    (SimParams41(10, 3, 800,  "2e10"), None, 20000),
    (SimParams41(36, 1, 1200, "3e9"),  None, 20000),
    (SimParams41(36, 1, 1200, "4e9"),  None, 20000),
    (SimParams41(36, 1, 1200, "5e9"),  None, 20000),
    (SimParams41(24, 2, 1200, "5e9"),  None, 20000),
    (SimParams41(24, 2, 1200, "1e10"), None, 20000),

    (SimParams42(9,  1, 400,  "5e9"),  None, 20000),
    (SimParams42(9,  1, 400,  "8e9"),  None, 20000),
    (SimParams42(9,  1, 400,  "1e10"), None, 20000),
    (SimParams42(9,  1, 400,  "12e9"), None, 20000),
    (SimParams42(9,  1, 400,  _15e9),  None, 20000),
    (SimParams42(9,  1, 400,  "2e10"), None, 20000),
    (SimParams42(6,  2, 400,  "1e10"), None, 20000),
    (SimParams42(6,  2, 400,  _15e9),  None, 20000),
    (SimParams42(6,  2, 400,  "2e10"), None, 20000),
    (SimParams42(6,  2, 400,  _3e10),  None, 20000),
    (SimParams42(4,  3, 400,  "1e10"), None, 20000),
    (SimParams42(4,  3, 400,  _15e9),  None, 20000),
    (SimParams42(4,  3, 400,  "2e10"), None, 20000),
    (SimParams42(4,  3, 400,  _3e10),  None, 20000),
    (SimParams42(2,  6, 400,  "2e10"), None, 20000),
    (SimParams42(2,  6, 400,  _3e10),  None, 20000),
    (SimParams42(2,  6, 400,  "4e10"), None, 20000),
    (SimParams42(2,  6, 400,  "8e10"), None, 20000),
    (SimParams42(2,  6, 400,  _15e10), None, 20000),
    (SimParams42(15, 1, 600,  "5e9"),  None, 20000),
    (SimParams42(10, 2, 600,  "1e10"), None, 20000),
    (SimParams42(20, 1, 800,  "5e9"), None, 20000),
    (SimParams42(13, 2, 800,  "1e10"), None, 20000),

    (SimParams43(2,  3, 400,  "1e10"), None, 20000),
    (SimParams43(2,  3, 400,  "2e10"), None, 20000),
    (SimParams43(3,  2, 400,  "2e9"),  None, 20000),
    (SimParams43(3,  2, 400,  "3e9"),  None, 20000),
    (SimParams43(3,  2, 400,  "5e9"),  None, 20000),
    (SimParams43(3,  2, 400,  "1e10"), None, 20000),
    (SimParams43(3,  2, 400,  "2e10"), None, 20000),
    (SimParams43(3,  2, 400,  _3e10),  None, 20000),
    (SimParams43(5,  1, 400,  "2e9"),  None, 20000),
    (SimParams43(5,  1, 400,  "3e9"),  None, 20000),
    (SimParams43(5,  1, 400,  "5e9"),  None, 20000),
    (SimParams43(5,  1, 400,  "8e9"),  None, 20000),
    (SimParams43(5,  1, 400,  "1e10"), None, 20000),
    (SimParams43(5,  1, 400,  _15e10), None, 20000),
    (SimParams43(5,  1, 400,  "2e10"), None, 20000),
    (SimParams43(5,  1, 400,  "4e10"), None, 20000),
    (SimParams43(7,  1, 600,  "3e9"),  None, 20000),
    (SimParams43(7,  1, 600,  "5e9"),  None, 20000),
    (SimParams43(7,  1, 600,  "1e10"), None, 20000),
    (SimParams43(7,  1, 600,  "2e10"), None, 20000),
    (SimParams43(4,  2, 600,  "1e10"), None, 20000),
    (SimParams43(4,  2, 600,  "2e10"), None, 20000),
    (SimParams43(4,  2, 600,  _3e10),  None, 20000),

    (SimParams46(5,  1, 400,  "5e9"),  None, 20000),
    (SimParams46(5,  1, 400,  "8e9"),  None, 20000),
    (SimParams46(5,  1, 400,  "1e10"), None, 20000),
    (SimParams46(5,  1, 400,  "2e10"), None, 20000),
    (SimParams46(5,  1, 400,  "4e10"), None, 20000),

    (SimParams46(7,  1, 600,  "5e9"),  None, 20000),
    (SimParams46(7,  1, 600,  "8e9"),  None, 20000),
    (SimParams46(7,  1, 600,  "1e10"), None, 20000),
    (SimParams46(7,  1, 600,  "2e10"), None, 20000),
    (SimParams46(7,  1, 600,  _3e10),  None, 20000),
    (SimParams46(4,  2, 600,  "2e10"), None, 20000),
    (SimParams46(4,  2, 600,  "5e10"), None, 20000),
    (SimParams46(11, 1, 800,  "2e10"), None, 20000),
    (SimParams46(7,  2, 800,  _3e10),  None, 20000),

    (SimParams46Float(5,  1, 400,  "2e10"), None, 20000),
    (SimParams46Float(5,  1, 400,  "4e10"), None, 20000),
    (SimParams46Float(7,  1, 600,  _3e10),  None, 20000),

    (SimParams47(5,  1, 400,  "5e9"),  None, 20000),
    (SimParams47(5,  1, 400,  "8e9"),  None, 20000),
    (SimParams47(5,  1, 400,  "1e10"), None, 20000),
    (SimParams47(5,  1, 400,  "2e10"), None, 20000),
    (SimParams47(5,  1, 400,  "4e10"), None, 20000),
    (SimParams47(7,  1, 600,  "5e9"),  None, 20000),
    (SimParams47(7,  1, 600,  "8e9"),  None, 20000),
    (SimParams47(7,  1, 600,  "1e10"), None, 20000),
    (SimParams47(7,  1, 600,  "2e10"), None, 20000),
    (SimParams47(7,  1, 600,  _3e10),  None, 20000),
    (SimParams47(4,  2, 600,  "2e10"), None, 20000),
    (SimParams47(4,  2, 600,  "5e10"), None, 20000),

    # control conductivity, coupling cond, clock decay (ps), clock duration, coupling loops, asymmetric s0 loops, radius, drive current
    (SimParams48(5e2, 5e3, 1000, 50,  9, 1, 400, "1e10"), None, 20000),
    (SimParams48(5e2, 5e3, 1000, 50,  9, 1, 400, "2e10"), None, 20000),
    (SimParams48(2e2, 1e4, 1000, 50,  9, 1, 400, "1e10"), None, 20000),
    (SimParams48(5e2, 1e4, 1000, 50,  9, 1, 400, "3e9"),  None, 20000),
    (SimParams48(5e2, 1e4, 1000, 50,  9, 1, 400, "5e9"),  None, 20000),
    (SimParams48(5e2, 1e4, 1000, 50,  9, 1, 400, "1e10"), None, 20000),
    (SimParams48(5e2, 1e4, 1000, 50,  9, 1, 400, "2e10"), None, 20000),
    (SimParams48(5e2, 2e4, 2000, 100, 9, 1, 400, "5e9"),  None, 20000),
    (SimParams48(5e2, 2e4, 2000, 100, 9, 1, 400, "1e10"), None, 20000),
    (SimParams48(1e3, 1e4, 2000, 100, 9, 1, 400, "3e9"),  None, 20000),
    (SimParams48(1e3, 1e4, 2000, 100, 9, 1, 400, "5e9"),  None, 20000),
    (SimParams48(1e3, 1e4, 2000, 100, 9, 1, 400, "1e10"), None, 20000),
    (SimParams48(1e3, 1e4, 2000, 100, 9, 1, 400, "2e10"), None, 20000),
    (SimParams48(5e2, 2e4, 4000, 200, 9, 1, 400, "1e10"), None, 20000),
    (SimParams48(2e3, 2e4, 4000, 200, 9, 1, 400, "1e9"),  None, 20000),
    (SimParams48(2e3, 2e4, 4000, 200, 9, 1, 400, "2e9"),  None, 20000),
    (SimParams48(5e2, 4e4, 2000, 100, 9, 1, 400, "5e9"),  None, 20000),
    (SimParams48(5e2, 4e4, 4000, 200, 9, 1, 400, "1e10"), None, 20000),
    (SimParams48(5e2, 1e5,10000, 500, 9, 1, 400, "5e9"),  None, 20000),
    (SimParams48(5e2, 1e5,10000, 500, 9, 1, 400, "1e10"), None, 20000),
    (SimParams48(5e3, 5e4,10000, 500, 9, 1, 400, "5e8"),  None, 20000),
    (SimParams48(5e3, 5e4,10000, 500, 9, 1, 400, "1e9"),  None, 20000),

    (SimParams48(5e2, 2e4, 2000, 100, 6, 2, 400, "2e10"), None, 20000),

    (SimParams50(5e2, 1e4, 1000, 50,  5, 1, 400, "1e10"), None, 20000),
    (SimParams50(5e2, 2e4, 2000, 100, 5, 1, 400, "1e10"), None, 20000),
    (SimParams50(1e3, 2e4, 2000, 100, 5, 1, 400, "5e9"),  None, 20000),
    (SimParams50(1e3, 2e4, 2000, 100, 5, 1, 400, "1e10"), None, 20000),
    (SimParams50(1e3, 2e4, 2000, 100, 5, 1, 400, "2e10"), None, 20000),
    (SimParams50(2e3, 2e4, 2000, 100, 5, 1, 400, "5e9"),  None, 20000),
    (SimParams50(2e3, 2e4, 2000, 100, 5, 1, 400, "1e10"), None, 20000),
    (SimParams50(2e3, 2e4, 2000, 100, 5, 1, 400, "2e10"), None, 20000),
    (SimParams50(5e3, 2e4, 2000, 100, 5, 1, 400, "1e10"), None, 20000),
    (SimParams50(5e3, 2e4, 2000, 100, 5, 1, 400, "2e10"), None, 20000),
    (SimParams50(1e4, 2e4, 2000, 100, 5, 1, 400, "1e10"), None, 20000),
    (SimParams50(1e4, 2e4, 2000, 100, 5, 1, 400, "2e10"), None, 20000),
    (SimParams50(2e4, 2e4, 2000, 100, 5, 1, 400, "1e10"), None, 20000),
    (SimParams50(5e2, 4e4, 4000, 200, 5, 1, 400, "2e10"), None, 20000),
    (SimParams50(5e3, 4e4, 4000, 200, 5, 1, 400, "1e10"), None, 20000),
    (SimParams50(5e3, 4e4, 4000, 200, 5, 1, 400, "2e10"), None, 20000),

    (SimParams51(5e2, 2e4, 2000, 100, 5, 1, 400, _5e10),  None, 20000),
    (SimParams51(1e3, 2e4, 2000, 100, 5, 1, 400, "2e10"), None, 20000),
    (SimParams51(2e3, 2e4, 2000, 100, 5, 1, 400, "1e10"), None, 20000),
    (SimParams51(2e3, 2e4, 2000, 100, 5, 1, 400, "2e10"), None, 20000),
    (SimParams51(2e3, 2e4, 2000, 100, 5, 1, 400, _3e10),  None, 20000),
    (SimParams51(5e3, 2e4, 2000, 100, 5, 1, 400, _3e10),  None, 20000),
    (SimParams51(5e3, 2e4, 2000, 100, 5, 1, 400, _5e10),  None, 20000),
    (SimParams51(2e3, 2e4, 2000, 100, 3, 2, 400, "2e10"), None, 20000),
    (SimParams51(2e3, 2e4, 2000, 100, 3, 2, 400, _5e10),  None, 20000),
    (SimParams51(2e3, 2e4, 2000, 100, 2, 3, 400, _5e10),  None, 20000),
    (SimParams51(2e3, 2e4, 2000, 100, 2, 3, 400, 1e11),   None, 20000),
]

measurements = { real.machine_name: [human, norm, None, None] for (real, human, norm) in sims }

def matches(actual, expected) -> bool:
    return expected == None or actual == expected

def matches_attr(obj, attr: str, expected) -> bool:
    """
    this is `matches`, but handles the case where the object lacks the queried attribute.
    returns None if the object does not have this attribute.
    """
    return expected is None or (hasattr(obj, attr) and getattr(obj, attr) == expected)

def filter_meas(
    run: str = None,
    rad_um: int = None,
    drive: float = None,
    couplings: int = None,
    wrappings: int = None,
    viable_inverter: bool = None,
    is_inverter: bool = None,
    control_cond: int = None,
    coupling_cond: int = None,
    clock_length_ps: int = None,
    clock_decay_ps: int = None,
) -> list:
    """
    return only that set of (SimParams, Piecewise) which matches the given criteria.
    leave any option as `None` to not filter on it.
    the results are sorted.
    """
    return sorted(
        [
            (p, get_meas(p)) for (p, _, _) in sims
                if matches(p.run, run) \
                    and matches(p.um, rad_um) \
                    and matches(p.drive, drive) \
                    and matches(p.couplings, couplings) \
                    and matches(p.wrappings, wrappings) \
                    and matches(p.is_inverter, is_inverter) \
                    and matches_attr(p, "control_cond", control_cond) \
                    and matches_attr(p, "coupling_cond", coupling_cond) \
                    and matches_attr(p, "clock_length_ps", clock_length_ps) \
                    and matches_attr(p, "clock_decay_ps", clock_decay_ps) \
                    and get_meas(p) \
                    and get_meas(p).num_pieces > 0 \
                    and matches(p.ensure_inverter(get_meas(p)).is_viable_inverter(), viable_inverter)
        ],
        key = lambda v: v[0].tuple
    )

def get_meas(params: SimParams) -> Piecewise:
    pw = measurements[params.machine_name][3]
    if params.is_discharge_high:
        pw = pw.logically_inverted_xy()
    return pw

def set_meas(real: str, expr: str):
    pw = eval(expr)
    measurements[real][2] = expr
    human, norm, _, _ = measurements[real]
    # measurements[real][3] = pw = pw.normalized(norm)
    measurements[real][3] = pw = pw.normalized(17000)
    if human:
        globals()[human] = pw

def try_measure(real: str, t_init: float, t_last: float, t_mid: float = None):
    try:
        new_value = subprocess.check_output(
            [
                "./stacked_cores_39xx.py",
                f"../../../../out/applications/stacked_cores/{real}",
                str(t_init),
                str(t_last),
            ] + ([str(t_mid)] if t_mid is not None else [])
        ).decode()
    except subprocess.CalledProcessError as e:
        print(f"failed to extract measurements for {real}: {e}")
    else:
        if new_value.strip():
            set_meas(real, "\n" + new_value)

def update_all():
    for (params, _human, _norm) in sims:
        try_measure(params.machine_name, t_init=params.t_init, t_last=params.t_last, t_mid=params.t_mid)

def dump(path: str):
    orig = open(path).read()
    marker = "## " + "AUTOGENERATED"
    prelude = orig[:orig.index(marker)]
    with open(path, "w") as f:
        f.write(prelude)
        f.write(marker)
        for (real, _human, _norm) in sims:
            encoded = measurements[real.machine_name][2]
            if encoded:
                f.write(f"\n\nset_meas(\"{real.machine_name}\", \"\"\"{encoded}\"\"\")")

        f.write("\n\nif __name__ == '__main__': main()\n")

def main():
    update_all()
    dump(__file__)


## AUTOGENERATED

set_meas("40-0.0004rad-6coupling-3_1_winding-5e10-drive", """
Piecewise(
    [
        [ -16689,  -6116 ],  # -2.000
        [ -16510,  -5954 ],  # -1.500
        [ -16046,  -5674 ],  # -1.000
        [ -15523,  -5367 ],  # -0.800
        [ -14807,  -4923 ],  # -0.600
        [ -14350,  -4632 ],  # -0.500
        [ -13668,  -4326 ],  # -0.400
        [ -12591,  -3719 ],  # -0.300
        [  -9752,  -1873 ],  # -0.250
        [  -5310,   1484 ],  # -0.200
        [  -2675,   3634 ],  # -0.170
        [  -1048,   4991 ],  # -0.150
        [    474,   6252 ],  # -0.130
        [   2517,   7918 ],  # -0.100
        [   4274,   9344 ],  # -0.070
        [   5284,  10134 ],  # -0.050
        [   7297,  11623 ],  #  0.000
        [   8629,  12439 ],  #  0.050
        [   9611,  12955 ],  #  0.100
        [  11360,  13845 ],  #  0.200
        [  12188,  14174 ],  #  0.300
        [  12831,  14537 ],  #  0.500
        [  13573,  14838 ],  #  0.800
        [  14082,  14999 ],  #  1.000
        [  15309,  15457 ],  #  1.500
        [  16329,  15683 ],  #  2.000
    ]
)
""")

set_meas("40-0.0004rad-12coupling-3_1_winding-5e10-drive", """
Piecewise(
    [
        [ -16453,  -5877 ],  # -2.000
        [ -16171,  -5444 ],  # -1.500
        [ -15617,  -4900 ],  # -1.000
        [ -15086,  -4405 ],  # -0.800
        [ -14357,  -3941 ],  # -0.600
        [ -13828,  -3800 ],  # -0.500
        [ -13111,  -3294 ],  # -0.400
        [ -12203,  -2582 ],  # -0.300
        [  -9321,   -169 ],  # -0.250
        [  -4645,   4342 ],  # -0.200
        [  -1619,   7434 ],  # -0.170
        [    422,   9543 ],  # -0.150
        [   2471,  11641 ],  # -0.130
        [   5472,  14584 ],  # -0.100
        [   7482,  15921 ],  # -0.070
        [   8233,  16078 ],  # -0.050
        [   9249,  16224 ],  #  0.000
        [   9700,  16263 ],  #  0.050
        [  10044,  16286 ],  #  0.100
        [  10643,  16323 ],  #  0.200
        [  11168,  16348 ],  #  0.300
        [  11605,  16367 ],  #  0.500
        [  12054,  16387 ],  #  0.800
        [  12411,  16402 ],  #  1.000
        [  13301,  16442 ],  #  1.500
        [  14146,  16474 ],  #  2.000
    ]
)
""")

set_meas("40-0.0004rad-6coupling-7_1_winding-5e10-drive", """
Piecewise(
    [
        [ -16244,  -8159 ],  # -2.000
        [ -15756,  -7835 ],  # -1.500
        [ -14687,  -7326 ],  # -1.000
        [ -14108,  -7030 ],  # -0.800
        [ -13435,  -6690 ],  # -0.600
        [ -13049,  -6503 ],  # -0.500
        [ -12522,  -6223 ],  # -0.400
        [ -11785,  -5833 ],  # -0.300
        [  -9190,  -4410 ],  # -0.250
        [  -4649,  -1447 ],  # -0.200
        [  -1703,    727 ],  # -0.170
        [    250,   2263 ],  # -0.150
        [   1821,   3566 ],  # -0.130
        [   3132,   4799 ],  # -0.100
        [   3928,   5689 ],  # -0.070
        [   4306,   6177 ],  # -0.050
        [   4961,   7059 ],  #  0.000
        [   5320,   7400 ],  #  0.050
        [   5631,   7670 ],  #  0.100
        [   6516,   8370 ],  #  0.200
        [   8605,   9674 ],  #  0.300
        [   9764,  10342 ],  #  0.500
        [  10753,  10874 ],  #  0.800
        [  11283,  11147 ],  #  1.000
        [  12329,  11657 ],  #  1.500
        [  13223,  12045 ],  #  2.000
    ]
)
""")

set_meas("40-0.00059999997rad-20coupling-3_1_winding-3e10-drive", """
Piecewise(
    [
        [  -9940,  -2886 ],  # -2.000
        [  -9292,  -2432 ],  # -1.500
        [  -8364,  -1816 ],  # -1.000
        [  -7849,  -1487 ],  # -0.800
        [  -6885,  -1141 ],  # -0.600
        [  -4088,    260 ],  # -0.500
        [   -408,   2478 ],  # -0.400
        [   2176,   4614 ],  # -0.300
        [   3050,   5423 ],  # -0.250
        [   3816,   6050 ],  # -0.200
        [   4229,   6382 ],  # -0.170
        [   4472,   6579 ],  # -0.150
        [   4680,   6751 ],  # -0.130
        [   4927,   6955 ],  # -0.100
        [   5107,   7052 ],  # -0.070
        [   5185,   7056 ],  # -0.050
        [   5354,   7045 ],  #  0.000
        [   5513,   7025 ],  #  0.050
        [   5663,   7003 ],  #  0.100
        [   5948,   6957 ],  #  0.200
        [   6214,   6917 ],  #  0.300
        [   6696,   6857 ],  #  0.500
        [   7144,   6766 ],  #  0.800
        [   7327,   6732 ],  #  1.000
        [   7699,   6755 ],  #  1.500
        [   8113,   6906 ],  #  2.000
    ]
)
""")

set_meas("40-0.00059999997rad-20coupling-3_1_winding-5e10-drive", """
Piecewise(
    [
        [ -11623,  -2886 ],  # -2.000
        [ -10847,  -2125 ],  # -1.500
        [  -9878,  -1348 ],  # -1.000
        [  -9392,   -985 ],  # -0.800
        [  -8694,   -508 ],  # -0.600
        [  -8256,   -360 ],  # -0.500
        [  -7609,   -221 ],  # -0.400
        [  -4128,   1930 ],  # -0.300
        [   -974,   4269 ],  # -0.250
        [   2142,   6978 ],  # -0.200
        [   3339,   8247 ],  # -0.170
        [   3895,   8940 ],  # -0.150
        [   4394,   9598 ],  # -0.130
        [   5049,  10512 ],  # -0.100
        [   5564,  11276 ],  # -0.070
        [   5838,  11677 ],  # -0.050
        [   6192,  11925 ],  #  0.000
        [   6415,  12017 ],  #  0.050
        [   6613,  12086 ],  #  0.100
        [   6960,  12202 ],  #  0.200
        [   7282,  12311 ],  #  0.300
        [   7658,  12407 ],  #  0.500
        [   8037,  12562 ],  #  0.800
        [   8272,  12695 ],  #  1.000
        [   8790,  13075 ],  #  1.500
        [   9248,  13503 ],  #  2.000
    ]
)
""")

set_meas("40-0.00059999997rad-20coupling-3_1_winding-1e11-drive", """
Piecewise(
    [
        [ -14284,  -3414 ],  # -2.000
        [ -13332,  -2406 ],  # -1.500
        [ -12167,  -1233 ],  # -1.000
        [ -11552,   -655 ],  # -0.800
        [ -10820,    -93 ],  # -0.600
        [ -10370,    199 ],  # -0.500
        [  -9825,    434 ],  # -0.400
        [  -9043,    844 ],  # -0.300
        [  -8573,   1101 ],  # -0.250
        [  -7905,   1565 ],  # -0.200
        [  -6613,   2609 ],  # -0.170
        [  -4197,   4541 ],  # -0.150
        [  -1670,   6569 ],  # -0.130
        [   1886,   9344 ],  # -0.100
        [   4013,  11023 ],  # -0.070
        [   5011,  11888 ],  # -0.050
        [   6604,  12899 ],  #  0.000
        [   7358,  13302 ],  #  0.050
        [   7932,  13602 ],  #  0.100
        [   8423,  13730 ],  #  0.200
        [   8645,  13888 ],  #  0.300
        [   8942,  13962 ],  #  0.500
        [   9318,  14166 ],  #  0.800
        [   9554,  14252 ],  #  1.000
        [  10166,  14547 ],  #  1.500
        [  10774,  14789 ],  #  2.000
    ]
)
""")

set_meas("40-0.00059999997rad-20coupling-3_1_winding-2e11-drive", """
Piecewise(
    [
        [ -17130,  -2396 ],  # -2.000
        [ -16555,  -1969 ],  # -1.500
        [ -15189,  -1044 ],  # -1.000
        [ -14486,   -584 ],  # -0.800
        [ -13641,    -20 ],  # -0.600
        [ -13147,    302 ],  # -0.500
        [ -12539,    723 ],  # -0.400
        [ -11816,   1218 ],  # -0.300
        [ -11388,   1501 ],  # -0.250
        [ -10882,   1826 ],  # -0.200
        [ -10473,   2112 ],  # -0.170
        [ -10095,   2383 ],  # -0.150
        [  -9674,   2701 ],  # -0.130
        [  -8747,   3406 ],  # -0.100
        [  -3356,   6322 ],  # -0.070
        [    747,   8265 ],  # -0.050
        [   6051,   9708 ],  #  0.000
        [   8022,  10412 ],  #  0.050
        [   8421,  10382 ],  #  0.100
        [   8817,  10592 ],  #  0.200
        [   9116,  10658 ],  #  0.300
        [   9620,  10803 ],  #  0.500
        [  10324,  11056 ],  #  0.800
        [  10785,  11303 ],  #  1.000
        [  11938,  11738 ],  #  1.500
        [  13090,  12062 ],  #  2.000
    ]
)
""")

set_meas("40-0.00059999997rad-6coupling-5_1_winding-5e10-drive", """
Piecewise(
    [
        [ -13402,  -7952 ],  # -2.000
        [ -12000,  -7336 ],  # -1.500
        [ -10508,  -6529 ],  # -1.000
        [  -9877,  -6187 ],  # -0.800
        [  -9170,  -5941 ],  # -0.600
        [  -8751,  -5783 ],  # -0.500
        [  -8105,  -5550 ],  # -0.400
        [  -5403,  -4604 ],  # -0.300
        [  -3123,  -3697 ],  # -0.250
        [  -1245,  -2867 ],  # -0.200
        [   -253,  -2312 ],  # -0.170
        [    321,  -1964 ],  # -0.150
        [    854,  -1637 ],  # -0.130
        [   1567,  -1089 ],  # -0.100
        [   2213,   -547 ],  # -0.070
        [   2597,   -173 ],  # -0.050
        [   3420,    483 ],  #  0.000
        [   4021,    893 ],  #  0.050
        [   4541,   1235 ],  #  0.100
        [   5638,   1906 ],  #  0.200
        [   6442,   2343 ],  #  0.300
        [   7066,   2622 ],  #  0.500
        [   7795,   2977 ],  #  0.800
        [   8296,   3245 ],  #  1.000
        [   9478,   3935 ],  #  1.500
        [  10530,   4558 ],  #  2.000
    ]
)
""")

set_meas("40-0.00059999997rad-8coupling-5_1_winding-5e10-drive", """
Piecewise(
    [
        [ -12065,  -7853 ],  # -2.000
        [ -10986,  -6995 ],  # -1.500
        [  -9808,  -6083 ],  # -1.000
        [  -9285,  -5788 ],  # -0.800
        [  -8592,  -5673 ],  # -0.600
        [  -8153,  -5422 ],  # -0.500
        [  -7507,  -5218 ],  # -0.400
        [  -4424,  -4031 ],  # -0.300
        [  -1908,  -2828 ],  # -0.250
        [    -23,  -1673 ],  # -0.200
        [    866,  -1035 ],  # -0.170
        [   1359,   -631 ],  # -0.150
        [   1820,   -212 ],  # -0.130
        [   2402,    397 ],  # -0.100
        [   2911,    962 ],  # -0.070
        [   3198,   1342 ],  # -0.050
        [   3755,   1922 ],  #  0.000
        [   4135,   2227 ],  #  0.050
        [   4461,   2471 ],  #  0.100
        [   5195,   3005 ],  #  0.200
        [   5896,   3442 ],  #  0.300
        [   6480,   3700 ],  #  0.500
        [   7084,   3989 ],  #  0.800
        [   7497,   4234 ],  #  1.000
        [   8504,   4899 ],  #  1.500
        [   9421,   5547 ],  #  2.000
    ]
)
""")

set_meas("40-0.00059999997rad-10coupling-5_1_winding-5e10-drive", """
Piecewise(
    [
        [ -11455,  -7219 ],  # -2.000
        [ -10621,  -6290 ],  # -1.500
        [  -9634,  -5367 ],  # -1.000
        [  -9129,  -5200 ],  # -0.800
        [  -8406,  -5069 ],  # -0.600
        [  -7984,  -4925 ],  # -0.500
        [  -7384,  -4682 ],  # -0.400
        [  -4194,  -3342 ],  # -0.300
        [  -1242,  -1795 ],  # -0.250
        [    745,   -450 ],  # -0.200
        [   2015,    707 ],  # -0.150
        [   2946,   1814 ],  # -0.100
        [   3364,   2409 ],  # -0.070
        [   3589,   2765 ],  # -0.050
        [   3984,   3233 ],  #  0.000
        [   4239,   3447 ],  #  0.050
        [   4463,   3620 ],  #  0.100
        [   5011,   4049 ],  #  0.200
        [   5629,   4468 ],  #  0.300
        [   6134,   4705 ],  #  0.500
        [   6638,   4932 ],  #  0.800
        [   6956,   5107 ],  #  1.000
        [   7769,   5635 ],  #  1.500
        [   8569,   6226 ],  #  2.000
    ]
)
""")

set_meas("40-0.00059999997rad-12coupling-5_1_winding-5e10-drive", """
Piecewise(
    [
        [ -11347,  -6303 ],  # -2.000
        [ -10626,  -5304 ],  # -1.500
        [  -9680,  -4520 ],  # -1.000
        [  -9156,  -4477 ],  # -0.800
        [  -8432,  -4378 ],  # -0.600
        [  -8023,  -4242 ],  # -0.500
        [  -7423,  -4024 ],  # -0.400
        [  -4149,  -2537 ],  # -0.300
        [  -1033,   -850 ],  # -0.250
        [   1286,    777 ],  # -0.200
        [   2067,   1524 ],  # -0.170
        [   2496,   2006 ],  # -0.150
        [   2874,   2482 ],  # -0.130
        [   3338,   3142 ],  # -0.100
        [   3686,   3723 ],  # -0.070
        [   3865,   4057 ],  # -0.050
        [   4147,   4400 ],  #  0.000
        [   4326,   4538 ],  #  0.050
        [   4486,   4651 ],  #  0.100
        [   4891,   4972 ],  #  0.200
        [   5437,   5353 ],  #  0.300
        [   5897,   5559 ],  #  0.500
        [   6370,   5741 ],  #  0.800
        [   6616,   5858 ],  #  1.000
        [   7278,   6265 ],  #  1.500
        [   7937,   6759 ],  #  2.000
    ]
)
""")

set_meas("40-0.00059999997rad-12coupling-5_1_winding-1e11-drive", """
Piecewise(
    [
        [ -14026,  -4901 ],  # -2.000
        [ -13014,  -4368 ],  # -1.500
        [ -11814,  -3658 ],  # -1.000
        [ -11192,  -3265 ],  # -0.800
        [ -10456,  -2801 ],  # -0.600
        [ -10025,  -2587 ],  # -0.500
        [  -9522,  -2256 ],  # -0.400
        [  -8805,  -1836 ],  # -0.300
        [  -8362,  -1556 ],  # -0.250
        [  -7702,  -1105 ],  # -0.200
        [  -6613,   -274 ],  # -0.170
        [  -4268,   1258 ],  # -0.150
        [  -2026,   2809 ],  # -0.130
        [    315,   4617 ],  # -0.100
        [   1954,   6243 ],  # -0.070
        [   2810,   7246 ],  # -0.050
        [   4359,   9016 ],  #  0.000
        [   5225,   9707 ],  #  0.050
        [   5981,  10366 ],  #  0.100
        [   6782,  10836 ],  #  0.200
        [   7084,  11037 ],  #  0.300
        [   7501,  11328 ],  #  0.500
        [   8045,  11716 ],  #  0.800
        [   8395,  12070 ],  #  1.000
        [   9248,  12388 ],  #  1.500
        [  10140,  13087 ],  #  2.000
    ]
)
""")

set_meas("40-0.00059999997rad-12coupling-5_1_winding-2e11-drive", """
Piecewise(
    [
        [ -17015,  -1518 ],  # -2.000
        [ -16471,  -1243 ],  # -1.500
        [ -15036,   -499 ],  # -1.000
        [ -14316,   -134 ],  # -0.800
        [ -13477,    290 ],  # -0.600
        [ -12998,    524 ],  # -0.500
        [ -12426,    801 ],  # -0.400
        [ -11761,   1134 ],  # -0.300
        [ -11372,   1315 ],  # -0.250
        [ -10905,   1547 ],  # -0.200
        [ -10526,   1731 ],  # -0.170
        [ -10172,   1926 ],  # -0.150
        [  -9747,   2188 ],  # -0.130
        [  -8733,   2765 ],  # -0.100
        [  -4346,   5002 ],  # -0.070
        [  -1335,   6369 ],  # -0.050
        [   3759,   8707 ],  #  0.000
        [   6392,   9509 ],  #  0.050
        [   7090,   9733 ],  #  0.100
        [   7637,   9681 ],  #  0.200
        [   8041,   9839 ],  #  0.300
        [   8724,   9855 ],  #  0.500
        [   9635,  10142 ],  #  0.800
        [  10214,  10435 ],  #  1.000
        [  11594,  10686 ],  #  1.500
        [  12900,  10836 ],  #  2.000
    ]
)
""")

set_meas("40-0.00059999997rad-6coupling-7_1_winding-5e10-drive", """
Piecewise(
    [
        [ -11955,  -8858 ],  # -2.000
        [ -10844,  -8392 ],  # -1.500
        [  -9653,  -7760 ],  # -1.000
        [  -9148,  -7480 ],  # -0.800
        [  -8477,  -7343 ],  # -0.600
        [  -8058,  -7101 ],  # -0.500
        [  -7437,  -6885 ],  # -0.400
        [  -4518,  -6025 ],  # -0.300
        [  -2176,  -5225 ],  # -0.250
        [   -490,  -4531 ],  # -0.200
        [    302,  -4102 ],  # -0.170
        [    756,  -3831 ],  # -0.150
        [   1149,  -3593 ],  # -0.130
        [   1675,  -3170 ],  # -0.100
        [   2104,  -2811 ],  # -0.070
        [   2352,  -2551 ],  # -0.050
        [   2854,  -2097 ],  #  0.000
        [   3197,  -1839 ],  #  0.050
        [   3517,  -1617 ],  #  0.100
        [   4341,  -1140 ],  #  0.200
        [   5178,   -709 ],  #  0.300
        [   5774,   -435 ],  #  0.500
        [   6369,   -180 ],  #  0.800
        [   6782,      1 ],  #  1.000
        [   7828,    503 ],  #  1.500
        [   8887,   1051 ],  #  2.000
    ]
)
""")

set_meas("40-0.00059999997rad-8coupling-7_1_winding-5e10-drive", """
Piecewise(
    [
        [ -11191,  -8744 ],  # -2.000
        [ -10386,  -8104 ],  # -1.500
        [  -9433,  -7450 ],  # -1.000
        [  -8946,  -7292 ],  # -0.800
        [  -8265,  -7035 ],  # -0.600
        [  -7873,  -6805 ],  # -0.500
        [  -7292,  -6588 ],  # -0.400
        [  -4183,  -5572 ],  # -0.300
        [  -1326,  -4520 ],  # -0.250
        [    385,  -3634 ],  # -0.200
        [   1083,  -3152 ],  # -0.170
        [   1463,  -2837 ],  # -0.150
        [   1801,  -2531 ],  # -0.130
        [   2207,  -2095 ],  # -0.100
        [   2521,  -1693 ],  # -0.070
        [   2685,  -1451 ],  # -0.050
        [   2988,  -1084 ],  #  0.000
        [   3186,   -916 ],  #  0.050
        [   3382,   -764 ],  #  0.100
        [   3998,   -338 ],  #  0.200
        [   4755,     99 ],  #  0.300
        [   5265,    344 ],  #  0.500
        [   5782,    539 ],  #  0.800
        [   6062,    652 ],  #  1.000
        [   6868,   1020 ],  #  1.500
        [   7684,   1464 ],  #  2.000
    ]
)
""")

set_meas("40-0.00059999997rad-8coupling-7_1_winding-1e11-drive", """
Piecewise(
    [
        [ -14253,  -5453 ],  # -2.000
        [ -13129,  -5047 ],  # -1.500
        [ -11874,  -4604 ],  # -1.000
        [ -11257,  -4275 ],  # -0.800
        [ -10540,  -4074 ],  # -0.600
        [ -10145,  -3874 ],  # -0.500
        [  -9684,  -3722 ],  # -0.400
        [  -9016,  -3401 ],  # -0.300
        [  -8575,  -3221 ],  # -0.250
        [  -7816,  -2849 ],  # -0.200
        [  -6723,  -2182 ],  # -0.170
        [  -4777,  -1380 ],  # -0.150
        [  -3087,   -580 ],  # -0.130
        [  -1072,    696 ],  # -0.100
        [    487,   1724 ],  # -0.070
        [   1344,   2397 ],  # -0.050
        [   3043,   3893 ],  #  0.000
        [   4117,   4567 ],  #  0.050
        [   5132,   5187 ],  #  0.100
        [   6206,   5762 ],  #  0.200
        [   6580,   5966 ],  #  0.300
        [   7096,   6239 ],  #  0.500
        [   7783,   6638 ],  #  0.800
        [   8223,   6861 ],  #  1.000
        [   9254,   7438 ],  #  1.500
        [  10288,   7992 ],  #  2.000
    ]
)
""")

set_meas("40-0.00059999997rad-6coupling-9_1_winding-5e10-drive", """
Piecewise(
    [
        [ -11138,  -9779 ],  # -2.000
        [ -10280,  -9491 ],  # -1.500
        [  -9298,  -9070 ],  # -1.000
        [  -8827,  -8906 ],  # -0.800
        [  -8171,  -8585 ],  # -0.600
        [  -7786,  -8410 ],  # -0.500
        [  -7209,  -8166 ],  # -0.400
        [  -4205,  -7466 ],  # -0.300
        [  -1652,  -6796 ],  # -0.250
        [    -89,  -6232 ],  # -0.200
        [    580,  -5919 ],  # -0.170
        [    940,  -5712 ],  # -0.150
        [   1260,  -5491 ],  # -0.130
        [   1644,  -5196 ],  # -0.100
        [   1941,  -4928 ],  # -0.070
        [   2099,  -4755 ],  # -0.050
        [   2412,  -4431 ],  #  0.000
        [   2629,  -4251 ],  #  0.050
        [   2853,  -4088 ],  #  0.100
        [   3581,  -3693 ],  #  0.200
        [   4449,  -3306 ],  #  0.300
        [   4986,  -3048 ],  #  0.500
        [   5509,  -2843 ],  #  0.800
        [   5803,  -2733 ],  #  1.000
        [   6678,  -2391 ],  #  1.500
        [   7561,  -1986 ],  #  2.000
    ]
)
""")

set_meas("40-0.00059999997rad-6coupling-9_1_winding-1e11-drive", """
Piecewise(
    [
        [ -14471,  -5820 ],  # -2.000
        [ -13297,  -5501 ],  # -1.500
        [ -12009,  -5126 ],  # -1.000
        [ -11389,  -4911 ],  # -0.800
        [ -10680,  -4725 ],  # -0.600
        [ -10302,  -4603 ],  # -0.500
        [  -9865,  -4457 ],  # -0.400
        [  -9237,  -4234 ],  # -0.300
        [  -8811,  -4089 ],  # -0.250
        [  -8049,  -3862 ],  # -0.200
        [  -7004,  -3391 ],  # -0.170
        [  -5300,  -2848 ],  # -0.150
        [  -3750,  -2294 ],  # -0.130
        [  -1836,  -1408 ],  # -0.100
        [   -309,   -584 ],  # -0.070
        [    551,    -36 ],  # -0.050
        [   2324,   1083 ],  #  0.000
        [   3513,   1623 ],  #  0.050
        [   4680,   2080 ],  #  0.100
        [   5918,   2563 ],  #  0.200
        [   6333,   2691 ],  #  0.300
        [   6896,   2970 ],  #  0.500
        [   7657,   3297 ],  #  0.800
        [   8136,   3399 ],  #  1.000
        [   9255,   3962 ],  #  1.500
        [  10362,   4340 ],  #  2.000
    ]
)
""")

set_meas("40-0.00059999997rad-6coupling-11_1_winding-5e10-drive", """
Piecewise(
    [
        [ -10841, -10684 ],  # -2.000
        [ -10118, -10416 ],  # -1.500
        [  -9209, -10033 ],  # -1.000
        [  -8755,  -9809 ],  # -0.800
        [  -8113,  -9520 ],  # -0.600
        [  -7741,  -9312 ],  # -0.500
        [  -7180,  -9051 ],  # -0.400
        [  -4152,  -8422 ],  # -0.300
        [  -1328,  -7790 ],  # -0.250
        [    117,  -7317 ],  # -0.200
        [    720,  -7057 ],  # -0.170
        [   1037,  -6894 ],  # -0.150
        [   1291,  -6744 ],  # -0.130
        [   1580,  -6541 ],  # -0.100
        [   1773,  -6366 ],  # -0.070
        [   1870,  -6257 ],  # -0.050
        [   2055,  -6046 ],  #  0.000
        [   2191,  -5921 ],  #  0.050
        [   2366,  -5783 ],  #  0.100
        [   3097,  -5430 ],  #  0.200
        [   4014,  -5084 ],  #  0.300
        [   4496,  -4826 ],  #  0.500
        [   4942,  -4641 ],  #  0.800
        [   5175,  -4562 ],  #  1.000
        [   5884,  -4342 ],  #  1.500
        [   6611,  -4071 ],  #  2.000
    ]
)
""")

set_meas("40-0.0008rad-24coupling-3_1_winding-1e11-drive", """
Piecewise(
    [
        [  -9820,  -2521 ],  # -2.000
        [  -8879,  -1658 ],  # -1.500
        [  -7734,   -623 ],  # -1.000
        [  -7195,   -152 ],  # -0.800
        [  -6536,    353 ],  # -0.600
        [  -6131,    578 ],  # -0.500
        [  -5591,    852 ],  # -0.400
        [  -4914,   1143 ],  # -0.300
        [  -4439,   1401 ],  # -0.250
        [  -3452,   2107 ],  # -0.200
        [  -1471,   3443 ],  # -0.170
        [   -351,   4282 ],  # -0.150
        [    599,   5057 ],  # -0.130
        [   1827,   6231 ],  # -0.100
        [   2866,   7402 ],  # -0.070
        [   3456,   8065 ],  # -0.050
        [   4483,   9100 ],  #  0.000
        [   5059,   9621 ],  #  0.050
        [   5529,   9869 ],  #  0.100
        [   5926,  10058 ],  #  0.200
        [   6124,  10161 ],  #  0.300
        [   6417,  10457 ],  #  0.500
        [   6818,  10623 ],  #  0.800
        [   7070,  10703 ],  #  1.000
        [   7665,  11010 ],  #  1.500
        [   8196,  11153 ],  #  2.000
    ]
)
""")

set_meas("40-0.0008rad-12coupling-5_1_winding-5e10-drive", """
Piecewise(
    [
        [  -5684,  -3702 ],  # -1.000
        [   1295,    -89 ],  # -0.170
        [   1525,    164 ],  # -0.150
        [   1720,    388 ],  # -0.130
        [   1982,    740 ],  # -0.100
        [   2192,   1060 ],  # -0.070
        [   2311,   1260 ],  # -0.050
        [   2549,   1545 ],  #  0.000
        [   2729,   1725 ],  #  0.050
        [   2906,   1875 ],  #  0.100
        [   4790,   2570 ],  #  1.000
    ]
)
""")

set_meas("40-0.0008rad-18coupling-5_1_winding-5e10-drive", """
Piecewise(
    [
        [  -6973,  -3314 ],  # -2.000
        [  -6359,  -2380 ],  # -1.500
        [  -5476,  -1830 ],  # -1.000
        [  -4932,  -1774 ],  # -0.800
        [  -4300,  -1686 ],  # -0.600
        [  -3906,  -1549 ],  # -0.500
        [  -3118,  -1134 ],  # -0.400
        [    -92,    305 ],  # -0.300
        [    868,   1000 ],  # -0.250
        [   1629,   1735 ],  # -0.200
        [   1966,   2160 ],  # -0.170
        [   2142,   2430 ],  # -0.150
        [   2285,   2676 ],  # -0.130
        [   2453,   3016 ],  # -0.100
        [   2574,   3302 ],  # -0.070
        [   2638,   3460 ],  # -0.050
        [   2757,   3600 ],  #  0.000
        [   2851,   3678 ],  #  0.050
        [   2946,   3762 ],  #  0.100
        [   3235,   4021 ],  #  0.200
        [   3559,   4245 ],  #  0.300
        [   3856,   4339 ],  #  0.500
        [   4154,   4342 ],  #  0.800
        [   4308,   4322 ],  #  1.000
        [   4707,   4173 ],  #  1.500
        [   5138,   4065 ],  #  2.000
    ]
)
""")

set_meas("40-0.0008rad-18coupling-5_1_winding-1e11-drive", """
Piecewise(
    [
        [  -9383,  -4435 ],  # -2.000
        [  -8496,  -3627 ],  # -1.500
        [  -7402,  -2697 ],  # -1.000
        [  -6879,  -2366 ],  # -0.800
        [  -6221,  -1934 ],  # -0.600
        [  -5813,  -1708 ],  # -0.500
        [  -5278,  -1432 ],  # -0.400
        [  -4635,  -1141 ],  # -0.300
        [  -4198,   -879 ],  # -0.250
        [  -3311,   -301 ],  # -0.200
        [  -1487,    831 ],  # -0.170
        [   -503,   1567 ],  # -0.150
        [    312,   2266 ],  # -0.130
        [   1320,   3288 ],  # -0.100
        [   2111,   4289 ],  # -0.070
        [   2532,   4939 ],  # -0.050
        [   3242,   5979 ],  #  0.000
        [   3652,   6352 ],  #  0.050
        [   4067,   6713 ],  #  0.100
        [   4566,   7101 ],  #  0.200
        [   4783,   7244 ],  #  0.300
        [   5077,   7430 ],  #  0.500
        [   5448,   7668 ],  #  0.800
        [   5691,   7818 ],  #  1.000
        [   6284,   8237 ],  #  1.500
        [   6840,   8656 ],  #  2.000
    ]
)
""")

set_meas("40-0.0008rad-18coupling-5_1_winding-2e11-drive", """
Piecewise(
    [
        [ -10488,  -1071 ],  # -1.000
        [  -5841,   1751 ],  # -0.150
        [  -4203,   2883 ],  # -0.100
        [    412,   5826 ],  # -0.050
        [   3125,   7681 ],  #  0.000
        [   4595,   8487 ],  #  0.050
        [   5050,   8482 ],  #  0.100
        [   7042,   8931 ],  #  1.000
    ]
)
""")

set_meas("40-0.0008rad-12coupling-7_1_winding-5e10-drive", """
Piecewise(
    [
        [  -6872,  -6278 ],  # -2.000
        [  -6226,  -5292 ],  # -1.500
        [  -5361,  -4584 ],  # -1.000
        [  -4837,  -4405 ],  # -0.800
        [  -4222,  -4180 ],  # -0.600
        [  -3833,  -3965 ],  # -0.500
        [  -3070,  -3613 ],  # -0.400
        [   -510,  -2854 ],  # -0.300
        [    339,  -2479 ],  # -0.250
        [    967,  -2083 ],  # -0.200
        [   1237,  -1853 ],  # -0.170
        [   1381,  -1696 ],  # -0.150
        [   1499,  -1549 ],  # -0.130
        [   1639,  -1346 ],  # -0.100
        [   1745,  -1152 ],  # -0.070
        [   1804,  -1036 ],  # -0.050
        [   1924,   -868 ],  #  0.000
        [   2022,   -749 ],  #  0.050
        [   2143,   -638 ],  #  0.100
        [   2562,   -344 ],  #  0.200
        [   3034,    -95 ],  #  0.300
        [   3367,     34 ],  #  0.500
        [   3693,     20 ],  #  0.800
        [   3870,    -18 ],  #  1.000
        [   4347,   -159 ],  #  1.500
        [   4843,   -206 ],  #  2.000
    ]
)
""")

set_meas("40-0.0008rad-12coupling-7_1_winding-1e11-drive", """
Piecewise(
    [
        [  -9694,  -6356 ],  # -2.000
        [  -7577,  -5115 ],  # -1.000
        [  -7056,  -4782 ],  # -0.800
        [  -6077,  -4361 ],  # -0.500
        [  -5057,  -3730 ],  # -0.300
        [  -3765,  -3111 ],  # -0.200
        [  -2393,  -2551 ],  # -0.170
        [  -1520,  -2077 ],  # -0.150
        [   -755,  -1672 ],  # -0.130
        [    214,   -946 ],  # -0.100
        [    995,   -345 ],  # -0.070
        [   1434,    168 ],  # -0.050
        [   2253,   1101 ],  #  0.000
        [   2806,   1547 ],  #  0.050
        [   3407,   1949 ],  #  0.100
        [   4095,   2422 ],  #  0.200
        [   4350,   2604 ],  #  0.300
        [   4703,   2801 ],  #  0.500
        [   5149,   3072 ],  #  0.800
        [   5431,   3247 ],  #  1.000
        [   6796,   4233 ],  #  2.000
    ]
)
""")

set_meas("40-0.0008rad-12coupling-7_1_winding-2e11-drive", """
Piecewise(
    [
        [   7009,   6247 ],  #  1.000
    ]
)
""")

set_meas("40-0.0008rad-10coupling-9_1_winding-5e10-drive", """
Piecewise(
    [
        [  -5306,  -6061 ],  # -1.000
        [   1379,  -3250 ],  # -0.050
        [   1475,  -3098 ],  #  0.000
        [   3081,  -2179 ],  #  0.500
        [   3561,  -2147 ],  #  1.000
    ]
)
""")

set_meas("40-0.0008rad-10coupling-9_1_winding-1e11-drive", """
Piecewise(
    [
        [  -7431,  -6530 ],  # -1.000
        [  -2427,  -4298 ],  # -0.170
        [  -1596,  -3970 ],  # -0.150
        [   -874,  -3700 ],  # -0.130
        [     14,  -3172 ],  # -0.100
        [    712,  -2699 ],  # -0.070
        [   1091,  -2385 ],  # -0.050
        [   1789,  -1764 ],  #  0.000
        [   2292,  -1392 ],  #  0.050
        [   2891,  -1020 ],  #  0.100
        [   4991,    223 ],  #  1.000
    ]
)
""")

set_meas("40-0.0008rad-10coupling-9_1_winding-2e11-drive", """
Piecewise(
    [
        [ -10851,  -3027 ],  # -1.000
        [  -9108,  -2492 ],  # -0.500
        [  -7329,  -1814 ],  # -0.200
        [  -6967,  -1564 ],  # -0.170
        [  -6658,  -1477 ],  # -0.150
        [  -6280,  -1214 ],  # -0.130
        [  -5205,   -679 ],  # -0.100
        [  -2706,    389 ],  # -0.070
        [  -1244,   1084 ],  # -0.050
        [   1602,   2786 ],  #  0.000
        [   3466,   3593 ],  #  0.050
        [   4152,   3872 ],  #  0.100
        [   4626,   4063 ],  #  0.200
        [   5467,   4427 ],  #  0.500
        [   6597,   4912 ],  #  1.000
    ]
)
""")

set_meas("40-0.0008rad-8coupling-11_1_winding-5e10-drive", """
Piecewise(
    [
        [  -5239,  -7279 ],  # -1.000
        [    754,  -5300 ],  # -0.170
        [    865,  -5224 ],  # -0.150
        [    955,  -5150 ],  # -0.130
        [   1050,  -5053 ],  # -0.100
        [   1121,  -4957 ],  # -0.070
        [   1158,  -4899 ],  # -0.050
        [   1257,  -4763 ],  #  0.000
        [   1356,  -4640 ],  #  0.050
        [   1499,  -4516 ],  #  0.100
        [   3477,  -3796 ],  #  1.000
    ]
)
""")

set_meas("40-0.0008rad-8coupling-11_1_winding-1e11-drive", """
Piecewise(
    [
        [  -7556,  -7056 ],  # -1.000
        [  -2779,  -5200 ],  # -0.170
        [  -1966,  -4870 ],  # -0.150
        [  -1253,  -4699 ],  # -0.130
        [   -374,  -4370 ],  # -0.100
        [    324,  -4011 ],  # -0.070
        [    709,  -3720 ],  # -0.050
        [   1442,  -3226 ],  #  0.000
        [   1990,  -2876 ],  #  0.050
        [   2653,  -2497 ],  #  0.100
        [   4887,  -1305 ],  #  1.000
    ]
)
""")

set_meas("40-0.0011999999rad-12coupling-9_1_winding-1e11-drive", """
Piecewise(
    [
        [  -4389,  -5228 ],  # -1.000
        [  -1417,  -3560 ],  # -0.170
        [  -1057,  -3486 ],  # -0.150
        [   -717,  -3395 ],  # -0.130
        [   -290,  -3094 ],  # -0.100
        [     73,  -2847 ],  # -0.070
        [    289,  -2706 ],  # -0.050
        [    752,  -2285 ],  #  0.000
        [   1151,  -1909 ],  #  0.050
        [   1607,  -1739 ],  #  0.100
        [   2989,   -992 ],  #  1.000
    ]
)
""")

set_meas("40-0.0011999999rad-12coupling-9_1_winding-2e11-drive", """
Piecewise(
    [
        [  -7871,  -3446 ],  # -1.000
        [  -4722,  -2357 ],  # -0.150
        [  -3688,  -1830 ],  # -0.100
        [  -1468,   -971 ],  # -0.050
        [    506,     -5 ],  #  0.000
        [   2001,    524 ],  #  0.050
        [   2409,    671 ],  #  0.100
        [   4291,   1349 ],  #  1.000
    ]
)
""")

set_meas("40-0.0011999999rad-10coupling-11_1_winding-1e11-drive", """
Piecewise(
    [
        [  -4432,  -5765 ],  # -1.000
        [  -1539,  -4158 ],  # -0.170
        [  -1180,  -4162 ],  # -0.150
        [   -865,  -3992 ],  # -0.130
        [   -448,  -3763 ],  # -0.100
        [    -98,  -3617 ],  # -0.070
        [    104,  -3547 ],  # -0.050
        [    561,  -3136 ],  #  0.000
        [    967,  -2799 ],  #  0.050
        [   1444,  -2615 ],  #  0.100
        [   2850,  -1806 ],  #  1.000
    ]
)
""")

set_meas("41-0.0004rad-18coupling-1_1_winding-5e9-drive", """
Piecewise(
    [
        [ -16843, -16108 ],  # -1.000
        [ -13196, -15734 ],  #  0.000
        [ -12712, -15657 ],  #  0.010
        [ -11747, -15471 ],  #  0.020
        [  -9794, -14646 ],  #  0.030
        [  -7750, -13292 ],  #  0.040
        [  -5665, -11792 ],  #  0.050
        [  -3558, -10236 ],  #  0.060
        [  -1436,  -8654 ],  #  0.070
        [    696,  -7042 ],  #  0.080
        [   2835,  -5423 ],  #  0.090
        [   4975,  -3806 ],  #  0.100
        [   7114,  -2174 ],  #  0.110
        [   9241,   -538 ],  #  0.120
        [  11357,   1101 ],  #  0.130
        [  13451,   2718 ],  #  0.140
        [  15294,   4164 ],  #  0.150
        [  16591,   5294 ],  #  0.170
        [  17147,   5761 ],  #  0.200
        [  17439,   5999 ],  #  0.300
        [  17639,   6255 ],  #  1.000
    ]
)
""")

set_meas("41-0.0004rad-18coupling-1_1_winding-1e10-drive", """
Piecewise(
    [
        [ -16866, -16415 ],  # -1.000
        [ -13335, -15908 ],  #  0.000
        [ -11545, -15502 ],  #  0.010
        [  -7643, -13782 ],  #  0.020
        [  -3518, -11128 ],  #  0.030
        [    700,  -8171 ],  #  0.040
        [   4956,  -5072 ],  #  0.050
        [   9211,  -1898 ],  #  0.060
        [  13411,   1261 ],  #  0.070
        [  16152,   3463 ],  #  0.080
        [  16892,   4104 ],  #  0.090
        [  17203,   4355 ],  #  0.100
        [  17370,   4481 ],  #  0.110
        [  17436,   4534 ],  #  0.120
        [  17469,   4563 ],  #  0.130
        [  17491,   4585 ],  #  0.140
        [  17505,   4601 ],  #  0.150
        [  17524,   4624 ],  #  0.170
        [  17544,   4653 ],  #  0.200
        [  17610,   4740 ],  #  0.300
        [  17802,   5001 ],  #  1.000
    ]
)
""")

set_meas("41-0.0004rad-9coupling-3_1_winding-2e9-drive", """
Piecewise(
    [
        [ -16571,   2372 ],  # -1.000
        [  -5986,   5641 ],  # -0.170
        [  -5733,   5725 ],  # -0.150
        [  -5481,   5809 ],  # -0.130
        [  -5105,   5933 ],  # -0.100
        [  -4734,   6057 ],  # -0.070
        [  -4491,   6139 ],  # -0.050
        [  -3909,   6334 ],  #  0.000
        [  -3352,   6521 ],  #  0.050
        [  -2818,   6702 ],  #  0.100
        [  16714,  12996 ],  #  1.000
    ]
)
""")

set_meas("41-0.0004rad-9coupling-3_1_winding-3e9-drive", """
Piecewise(
    [
        [ -16887,  -4873 ],  # -2.000
        [ -16802,  -4812 ],  # -1.000
        [ -16749,  -4794 ],  # -0.800
        [ -16343,  -4696 ],  # -0.600
        [ -14391,  -4098 ],  # -0.500
        [ -13065,  -3673 ],  # -0.400
        [ -12414,  -3462 ],  # -0.300
        [ -12066,  -3349 ],  # -0.250
        [ -11687,  -3228 ],  # -0.200
        [ -11442,  -3150 ],  # -0.170
        [ -11270,  -3094 ],  # -0.150
        [ -11088,  -3036 ],  # -0.130
        [ -10791,  -2940 ],  # -0.100
        [ -10471,  -2836 ],  # -0.070
        [ -10248,  -2763 ],  # -0.050
        [  -9667,  -2573 ],  #  0.000
        [  -9056,  -2372 ],  #  0.050
        [  -8283,  -2113 ],  #  0.100
        [  -3257,   -395 ],  #  0.200
        [   2914,   1681 ],  #  0.300
        [  15191,   5995 ],  #  0.500
        [  16935,   6397 ],  #  0.800
        [  16994,   6381 ],  #  1.000
        [  17039,   6389 ],  #  1.500
        [  17079,   6399 ],  #  2.000
    ]
)
""")

set_meas("41-0.0004rad-9coupling-3_1_winding-5e9-drive", """
Piecewise(
    [
        [ -16935, -16004 ],  # -2.000
        [ -16914, -16001 ],  # -1.500
        [ -16858, -15999 ],  # -1.000
        [ -16830, -15997 ],  # -0.800
        [ -16798, -15996 ],  # -0.600
        [ -16759, -15994 ],  # -0.500
        [ -16562, -15988 ],  # -0.400
        [ -14276, -15873 ],  # -0.300
        [ -13550, -15831 ],  # -0.250
        [ -13075, -15803 ],  # -0.200
        [ -12812, -15788 ],  # -0.170
        [ -12636, -15779 ],  # -0.150
        [ -12458, -15768 ],  # -0.130
        [ -12368, -15763 ],  # -0.120
        [ -12187, -15753 ],  # -0.100
        [ -11911, -15737 ],  # -0.070
        [ -11721, -15725 ],  # -0.050
        [ -11192, -15692 ],  #  0.000
        [ -10706, -15656 ],  #  0.030
        [ -10148, -15605 ],  #  0.050
        [  -9496, -15538 ],  #  0.060
        [  -8574, -15440 ],  #  0.070
        [  -6588, -15189 ],  #  0.090
        [  -5572, -15000 ],  #  0.100
        [  -4537, -14720 ],  #  0.110
        [  -3490, -14372 ],  #  0.120
        [  -2439, -13953 ],  #  0.130
        [  -1384, -13473 ],  #  0.140
        [   -327, -12941 ],  #  0.150
        [   1794, -11684 ],  #  0.170
        [   4975,  -9524 ],  #  0.200
        [   7093,  -7868 ],  #  0.220
        [  10263,  -5050 ],  #  0.250
        [  12345,  -3038 ],  #  0.270
        [  15184,    -54 ],  #  0.300
        [  16971,   1553 ],  #  0.500
        [  17049,   1613 ],  #  0.800
        [  17079,   1636 ],  #  1.000
        [  17136,   1671 ],  #  1.500
        [  17169,   1702 ],  #  2.000
    ]
)
""")

set_meas("41-0.0004rad-9coupling-3_1_winding-8e9-drive", """
Piecewise(
    [
        [ -16909, -16768 ],  # -1.000
        [ -11255, -16761 ],  #  0.000
        [ -10944, -16761 ],  #  0.010
        [ -10540, -16759 ],  #  0.020
        [ -10007, -16757 ],  #  0.030
        [  -9161, -16751 ],  #  0.040
        [  -7631, -16729 ],  #  0.050
        [  -6014, -16643 ],  #  0.060
        [  -4357, -16490 ],  #  0.070
        [  -2680, -16347 ],  #  0.080
        [   -994, -16199 ],  #  0.090
        [    698, -15974 ],  #  0.100
        [   2394, -15487 ],  #  0.110
        [   4090, -14493 ],  #  0.120
        [   5788, -13198 ],  #  0.130
        [   7484, -11632 ],  #  0.140
        [   9177,  -9839 ],  #  0.150
        [  10856,  -7906 ],  #  0.160
        [  12518,  -5948 ],  #  0.170
        [  14139,  -4023 ],  #  0.180
        [  15362,  -2585 ],  #  0.190
        [  15900,  -1982 ],  #  0.200
        [  16428,  -1406 ],  #  0.220
        [  16775,  -1056 ],  #  0.250
        [  16899,   -930 ],  #  0.270
        [  16974,   -854 ],  #  0.300
        [  17040,   -776 ],  #  0.400
        [  17067,   -739 ],  #  0.500
        [  17139,   -663 ],  #  0.800
        [  17171,   -627 ],  #  1.000
    ]
)
""")

set_meas("41-0.0004rad-9coupling-3_1_winding-1e10-drive", """
Piecewise(
    [
        [ -17010, -16789 ],  # -2.000
        [ -16931, -16791 ],  # -1.000
        [ -16907, -16791 ],  # -0.800
        [ -16880, -16792 ],  # -0.600
        [ -16858, -16793 ],  # -0.500
        [ -16830, -16793 ],  # -0.400
        [ -16798, -16793 ],  # -0.300
        [ -16758, -16793 ],  # -0.250
        [ -16500, -16793 ],  # -0.200
        [ -15511, -16791 ],  # -0.170
        [ -14631, -16790 ],  # -0.150
        [ -14115, -16790 ],  # -0.130
        [ -13476, -16790 ],  # -0.100
        [ -12908, -16788 ],  # -0.070
        [ -12523, -16787 ],  # -0.050
        [ -11224, -16783 ],  #  0.000
        [ -10776, -16780 ],  #  0.010
        [ -10182, -16778 ],  #  0.020
        [  -9321, -16772 ],  #  0.030
        [  -7616, -16763 ],  #  0.040
        [  -5604, -16743 ],  #  0.050
        [  -3533, -16622 ],  #  0.060
        [  -1432, -16422 ],  #  0.070
        [    683, -16220 ],  #  0.080
        [   2804, -15888 ],  #  0.090
        [   4926, -14787 ],  #  0.100
        [   7048, -12924 ],  #  0.110
        [   9165, -10620 ],  #  0.120
        [  11262,  -8235 ],  #  0.130
        [  13331,  -5802 ],  #  0.140
        [  15157,  -3634 ],  #  0.150
        [  15902,  -2784 ],  #  0.160
        [  16289,  -2360 ],  #  0.170
        [  16515,  -2120 ],  #  0.180
        [  16668,  -1965 ],  #  0.190
        [  16783,  -1850 ],  #  0.200
        [  16927,  -1705 ],  #  0.220
        [  16998,  -1629 ],  #  0.250
        [  17021,  -1604 ],  #  0.270
        [  17042,  -1577 ],  #  0.300
        [  17078,  -1523 ],  #  0.400
        [  17111,  -1482 ],  #  0.500
        [  17183,  -1405 ],  #  0.800
        [  17210,  -1376 ],  #  1.000
        [  17283,  -1286 ],  #  2.000
    ]
)
""")

set_meas("41-0.0004rad-9coupling-3_1_winding-12e9-drive", """
Piecewise(
    [
        [ -17026, -16808 ],  # -2.000
        [ -16957, -16812 ],  # -1.000
        [ -16928, -16813 ],  # -0.800
        [ -16878, -16813 ],  # -0.500
        [ -16816, -16814 ],  # -0.300
        [ -16735, -16814 ],  # -0.200
        [ -11301, -16801 ],  #  0.000
        [ -10718, -16797 ],  #  0.010
        [  -9875, -16793 ],  #  0.020
        [  -8375, -16784 ],  #  0.030
        [  -6008, -16763 ],  #  0.040
        [  -3540, -16732 ],  #  0.050
        [  -1019, -16516 ],  #  0.060
        [   1521, -16262 ],  #  0.070
        [   4067, -15678 ],  #  0.080
        [   6615, -13688 ],  #  0.090
        [   9157, -10970 ],  #  0.100
        [  11669,  -8122 ],  #  0.110
        [  14111,  -5332 ],  #  0.120
        [  15660,  -3571 ],  #  0.130
        [  16228,  -2955 ],  #  0.140
        [  16517,  -2653 ],  #  0.150
        [  16697,  -2474 ],  #  0.160
        [  16829,  -2343 ],  #  0.170
        [  16914,  -2258 ],  #  0.180
        [  16961,  -2211 ],  #  0.190
        [  16989,  -2181 ],  #  0.200
        [  17022,  -2144 ],  #  0.220
        [  17050,  -2109 ],  #  0.250
        [  17059,  -2093 ],  #  0.270
        [  17072,  -2073 ],  #  0.300
        [  17114,  -2018 ],  #  0.400
        [  17150,  -1974 ],  #  0.500
        [  17218,  -1896 ],  #  0.800
        [  17243,  -1869 ],  #  1.000
    ]
)
""")

set_meas("41-0.0004rad-9coupling-3_1_winding-15000000512e0-drive", """
Piecewise(
    [
        [ -16968, -16803 ],  # -1.000
        [ -11397, -16776 ],  #  0.000
        [ -10554, -16773 ],  #  0.010
        [  -9244, -16770 ],  #  0.020
        [  -6613, -16764 ],  #  0.030
        [  -3543, -16748 ],  #  0.040
        [   -387, -16629 ],  #  0.050
        [   2790, -16137 ],  #  0.060
        [   5974, -14065 ],  #  0.070
        [   9153, -11018 ],  #  0.080
        [  12290,  -7733 ],  #  0.090
        [  15144,  -4604 ],  #  0.100
        [  16120,  -3555 ],  #  0.110
        [  16523,  -3141 ],  #  0.120
        [  16740,  -2935 ],  #  0.130
        [  16889,  -2790 ],  #  0.140
        [  16962,  -2721 ],  #  0.150
        [  17000,  -2682 ],  #  0.160
        [  17023,  -2657 ],  #  0.170
        [  17039,  -2639 ],  #  0.180
        [  17052,  -2624 ],  #  0.190
        [  17061,  -2612 ],  #  0.200
        [  17072,  -2594 ],  #  0.220
        [  17088,  -2569 ],  #  0.250
        [  17099,  -2555 ],  #  0.270
        [  17114,  -2534 ],  #  0.300
        [  17162,  -2478 ],  #  0.400
        [  17198,  -2436 ],  #  0.500
        [  17260,  -2363 ],  #  0.800
        [  17284,  -2336 ],  #  1.000
    ]
)
""")

set_meas("41-0.0004rad-9coupling-3_1_winding-2e10-drive", """
Piecewise(
    [
        [ -17063, -16715 ],  # -2.000
        [ -16990, -16708 ],  # -1.000
        [ -16965, -16710 ],  # -0.800
        [ -16931, -16706 ],  # -0.500
        [ -16870, -16704 ],  # -0.300
        [ -16815, -16699 ],  # -0.200
        [ -16776, -16697 ],  # -0.170
        [ -16722, -16697 ],  # -0.150
        [ -16629, -16696 ],  # -0.130
        [ -16189, -16686 ],  # -0.100
        [ -14738, -16659 ],  # -0.070
        [ -13985, -16644 ],  # -0.050
        [ -11222, -16598 ],  #  0.000
        [  -9830, -16591 ],  #  0.010
        [  -7550, -16593 ],  #  0.020
        [  -3518, -16650 ],  #  0.030
        [    684, -16552 ],  #  0.040
        [   4919, -14379 ],  #  0.050
        [   9157, -10822 ],  #  0.060
        [  13320,  -6893 ],  #  0.070
        [  15906,  -4409 ],  #  0.080
        [  16531,  -3841 ],  #  0.090
        [  16811,  -3604 ],  #  0.100
        [  16959,  -3478 ],  #  0.110
        [  17016,  -3427 ],  #  0.120
        [  17044,  -3400 ],  #  0.130
        [  17077,  -3365 ],  #  0.150
        [  17093,  -3341 ],  #  0.170
        [  17114,  -3315 ],  #  0.200
        [  17180,  -3242 ],  #  0.300
        [  17257,  -3159 ],  #  0.500
        [  17313,  -3098 ],  #  0.800
        [  17335,  -3072 ],  #  1.000
        [  17422,  -2979 ],  #  2.000
    ]
)
""")

set_meas("41-0.0004rad-6coupling-5_1_winding-5e9-drive", """
Piecewise(
    [
        [ -16822,  -3557 ],  # -1.000
        [  -7287,  -1010 ],  #  0.000
        [  16912,   5750 ],  #  1.000
    ]
)
""")

set_meas("41-0.0004rad-6coupling-5_1_winding-1e10-drive", """
Piecewise(
    [
        [ -16983, -16406 ],  # -2.000
        [ -16916, -16409 ],  # -1.000
        [ -16891, -16410 ],  # -0.800
        [ -16827, -16410 ],  # -0.500
        [ -16586, -16407 ],  # -0.300
        [ -12412, -16329 ],  # -0.200
        [  -7736, -16227 ],  #  0.000
        [  -7511, -16221 ],  #  0.010
        [  -7269, -16214 ],  #  0.020
        [  -7004, -16207 ],  #  0.030
        [  -6705, -16199 ],  #  0.040
        [  -6356, -16189 ],  #  0.050
        [  -5931, -16175 ],  #  0.060
        [  -5397, -16158 ],  #  0.070
        [  -4652, -16132 ],  #  0.080
        [  -3451, -16081 ],  #  0.090
        [  -2119, -16004 ],  #  0.100
        [   -758, -15882 ],  #  0.110
        [    622, -15703 ],  #  0.120
        [   2009, -15454 ],  #  0.130
        [   3405, -15041 ],  #  0.140
        [   4803, -14314 ],  #  0.150
        [   6200, -13287 ],  #  0.160
        [   7596, -12082 ],  #  0.170
        [   8989, -10772 ],  #  0.180
        [  10380,  -9352 ],  #  0.190
        [  11755,  -7850 ],  #  0.200
        [  13097,  -6349 ],  #  0.210
        [  14397,  -4854 ],  #  0.220
        [  15349,  -3736 ],  #  0.230
        [  15800,  -3231 ],  #  0.240
        [  16075,  -2930 ],  #  0.250
        [  16407,  -2583 ],  #  0.270
        [  16671,  -2335 ],  #  0.300
        [  16896,  -2115 ],  #  0.400
        [  16930,  -2077 ],  #  0.500
        [  16998,  -2016 ],  #  0.800
        [  17025,  -1993 ],  #  1.000
        [  17095,  -1929 ],  #  2.000
    ]
)
""")

set_meas("41-0.0004rad-6coupling-5_1_winding-2e10-drive", """
Piecewise(
    [
        [ -16979, -16643 ],  # -1.000
        [  17128,  -5314 ],  #  1.000
    ]
)
""")

set_meas("41-0.0004rad-4coupling-7_1_winding-1e10-drive", """
Piecewise(
    [
        [ -16750,  -5443 ],  # -1.000
        [  -4968,  -2684 ],  #  0.000
        [  -4300,  -2513 ],  #  0.050
        [  -3998,  -2437 ],  #  0.070
        [  -3466,  -2303 ],  #  0.100
        [  -3025,  -2190 ],  #  0.120
        [  -2764,  -2124 ],  #  0.130
        [  -2113,  -1957 ],  #  0.150
        [  -1117,  -1704 ],  #  0.170
        [   1249,  -1106 ],  #  0.200
        [  16760,   4392 ],  #  1.000
    ]
)
""")

set_meas("41-0.0004rad-4coupling-7_1_winding-2e10-drive", """
Piecewise(
    [
        [ -16917, -16479 ],  # -2.000
        [ -16886, -16481 ],  # -1.500
        [ -16847, -16482 ],  # -1.000
        [ -16827, -16483 ],  # -0.800
        [ -16798, -16484 ],  # -0.600
        [ -16777, -16484 ],  # -0.500
        [ -16752, -16485 ],  # -0.400
        [ -16720, -16486 ],  # -0.300
        [ -16684, -16485 ],  # -0.250
        [ -16085, -16483 ],  # -0.200
        [ -13344, -16450 ],  # -0.170
        [ -11376, -16393 ],  # -0.150
        [  -9503, -16316 ],  # -0.120
        [  -8593, -16271 ],  # -0.100
        [  -7469, -16214 ],  # -0.070
        [  -6800, -16179 ],  # -0.050
        [  -5121, -16087 ],  #  0.000
        [  -4740, -16066 ],  #  0.010
        [  -4338, -16041 ],  #  0.020
        [  -3910, -16010 ],  #  0.030
        [  -3443, -15974 ],  #  0.040
        [  -2932, -15932 ],  #  0.050
        [  -2356, -15877 ],  #  0.060
        [  -1695, -15804 ],  #  0.070
        [   -920, -15703 ],  #  0.080
        [     66, -15524 ],  #  0.090
        [   1339, -15181 ],  #  0.100
        [   2954, -14493 ],  #  0.110
        [   4648, -13420 ],  #  0.120
        [   6350, -12059 ],  #  0.130
        [   8050, -10421 ],  #  0.140
        [   9746,  -8522 ],  #  0.150
        [  11429,  -6493 ],  #  0.160
        [  13104,  -4410 ],  #  0.170
        [  14755,  -2333 ],  #  0.180
        [  15750,  -1106 ],  #  0.190
        [  16093,   -720 ],  #  0.200
        [  16442,   -360 ],  #  0.220
        [  16683,   -139 ],  #  0.250
        [  16716,   -106 ],  #  0.270
        [  16726,    -88 ],  #  0.300
        [  16759,    -55 ],  #  0.400
        [  16783,    -32 ],  #  0.500
        [  16832,     19 ],  #  0.800
        [  16855,     41 ],  #  1.000
        [  16897,     85 ],  #  1.500
        [  16934,    121 ],  #  2.000
    ]
)
""")

set_meas("41-0.0004rad-4coupling-7_1_winding-4e10-drive", """
Piecewise(
    [
        [ -16924, -15335 ],  # -1.000
        [  -4821, -15100 ],  #  0.000
        [  -3651, -15119 ],  #  0.010
        [  -2371, -15139 ],  #  0.020
        [   -942, -15142 ],  #  0.030
        [    701, -15007 ],  #  0.040
        [   2704, -14469 ],  #  0.050
        [   5175, -13372 ],  #  0.060
        [   8247, -11347 ],  #  0.070
        [  11601,  -8649 ],  #  0.080
        [  14917,  -5376 ],  #  0.090
        [  16153,  -4116 ],  #  0.100
        [  16489,  -3790 ],  #  0.110
        [  16668,  -3620 ],  #  0.120
        [  16745,  -3538 ],  #  0.130
        [  16761,  -3510 ],  #  0.140
        [  16770,  -3493 ],  #  0.150
        [  16779,  -3495 ],  #  0.160
        [  16812,  -3457 ],  #  0.200
        [  16976,  -3316 ],  #  1.000
    ]
)
""")

set_meas("41-0.00059999997rad-18coupling-3_1_winding-3e9-drive", """
Piecewise(
    [
        [ -17109, -11798 ],  # -2.000
        [ -16987, -11712 ],  # -1.000
        [ -16952, -11691 ],  # -0.800
        [ -16679, -11590 ],  # -0.500
        [ -13120, -10459 ],  # -0.300
        [ -12423, -10236 ],  # -0.200
        [ -11018,  -9787 ],  #  0.000
        [ -10533,  -9630 ],  #  0.050
        [  -9926,  -9430 ],  #  0.070
        [  -9242,  -9204 ],  #  0.080
        [  -7737,  -8706 ],  #  0.100
        [  -6135,  -8179 ],  #  0.120
        [  -5321,  -7911 ],  #  0.130
        [  -3667,  -7366 ],  #  0.150
        [  -1998,  -6817 ],  #  0.170
        [    524,  -5982 ],  #  0.200
        [   2208,  -5415 ],  #  0.220
        [   4736,  -4526 ],  #  0.250
        [   6420,  -3880 ],  #  0.270
        [   8935,  -2733 ],  #  0.300
        [  17208,   3349 ],  #  0.500
        [  17520,   3549 ],  #  0.800
        [  17553,   3554 ],  #  1.000
        [  17653,   3624 ],  #  2.000
    ]
)
""")

set_meas("41-0.00059999997rad-18coupling-3_1_winding-5e9-drive", """
Piecewise(
    [
        [ -17163, -16780 ],  # -2.000
        [ -17051, -16783 ],  # -1.000
        [ -17012, -16784 ],  # -0.800
        [ -16938, -16785 ],  # -0.500
        [ -16653, -16784 ],  # -0.300
        [ -13537, -16769 ],  # -0.200
        [ -11597, -16745 ],  #  0.000
        [ -11467, -16743 ],  #  0.010
        [ -11307, -16739 ],  #  0.020
        [ -11081, -16734 ],  #  0.030
        [ -10454, -16714 ],  #  0.040
        [  -9203, -16662 ],  #  0.050
        [  -7883, -16598 ],  #  0.060
        [  -6518, -16527 ],  #  0.070
        [  -5131, -16458 ],  #  0.080
        [  -3735, -16402 ],  #  0.090
        [  -2332, -16326 ],  #  0.100
        [   -928, -16229 ],  #  0.110
        [    477, -16080 ],  #  0.120
        [   1883, -15801 ],  #  0.130
        [   3288, -15035 ],  #  0.140
        [   4694, -13753 ],  #  0.150
        [   6097, -12354 ],  #  0.160
        [   7494, -10892 ],  #  0.170
        [   8888,  -9370 ],  #  0.180
        [  10279,  -7832 ],  #  0.190
        [  11658,  -6311 ],  #  0.200
        [  14345,  -3304 ],  #  0.220
        [  16512,   -760 ],  #  0.250
        [  16917,   -263 ],  #  0.270
        [  17246,    128 ],  #  0.300
        [  17511,    420 ],  #  0.400
        [  17569,    480 ],  #  0.500
        [  17650,    573 ],  #  0.800
        [  17687,    620 ],  #  1.000
        [  17779,    725 ],  #  2.000
    ]
)
""")

set_meas("41-0.00059999997rad-18coupling-3_1_winding-1e10-drive", """
Piecewise(
    [
        [ -17264, -16835 ],  # -2.000
        [ -17122, -16837 ],  # -1.000
        [ -17086, -16837 ],  # -0.800
        [ -17014, -16838 ],  # -0.500
        [ -16929, -16838 ],  # -0.300
        [ -16840, -16838 ],  # -0.200
        [ -11615, -16836 ],  #  0.000
        [ -11128, -16835 ],  #  0.010
        [ -10132, -16834 ],  #  0.020
        [  -7824, -16829 ],  #  0.030
        [  -5109, -16820 ],  #  0.040
        [  -2336, -16805 ],  #  0.050
        [    458, -16594 ],  #  0.060
        [   3260, -16363 ],  #  0.070
        [   6056, -15675 ],  #  0.080
        [   8838, -12956 ],  #  0.090
        [  11594,  -9700 ],  #  0.100
        [  14261,  -6408 ],  #  0.110
        [  16182,  -4014 ],  #  0.120
        [  16781,  -3252 ],  #  0.130
        [  17096,  -2850 ],  #  0.140
        [  17300,  -2603 ],  #  0.150
        [  17416,  -2469 ],  #  0.160
        [  17473,  -2401 ],  #  0.170
        [  17512,  -2355 ],  #  0.180
        [  17543,  -2319 ],  #  0.190
        [  17566,  -2290 ],  #  0.200
        [  17595,  -2254 ],  #  0.220
        [  17623,  -2217 ],  #  0.250
        [  17636,  -2199 ],  #  0.270
        [  17652,  -2173 ],  #  0.300
        [  17708,  -2095 ],  #  0.400
        [  17752,  -2034 ],  #  0.500
        [  17836,  -1920 ],  #  0.800
        [  17871,  -1870 ],  #  1.000
        [  17965,  -1736 ],  #  2.000
    ]
)
""")

set_meas("41-0.00059999997rad-18coupling-3_1_winding-2e10-drive", """
Piecewise(
    [
        [ -17372, -16784 ],  # -2.000
        [ -17204, -16778 ],  # -1.000
        [ -17160, -16777 ],  # -0.800
        [ -17085, -16773 ],  # -0.500
        [ -16995, -16769 ],  # -0.300
        [ -16923, -16767 ],  # -0.200
        [ -11520, -16670 ],  #  0.000
        [    533, -16752 ],  #  0.030
        [   6106, -14832 ],  #  0.040
        [  11623, -10301 ],  #  0.050
        [  16206,  -5773 ],  #  0.060
        [  17156,  -4738 ],  #  0.070
        [  17487,  -4394 ],  #  0.080
        [  17638,  -4238 ],  #  0.100
        [  17686,  -4184 ],  #  0.120
        [  17700,  -4166 ],  #  0.130
        [  17722,  -4137 ],  #  0.150
        [  17746,  -4107 ],  #  0.170
        [  17779,  -4067 ],  #  0.200
        [  17798,  -4044 ],  #  0.220
        [  17823,  -4014 ],  #  0.250
        [  17837,  -3995 ],  #  0.270
        [  17857,  -3971 ],  #  0.300
        [  17950,  -3856 ],  #  0.500
        [  18036,  -3745 ],  #  0.800
        [  18074,  -3691 ],  #  1.000
        [  18200,  -3530 ],  #  2.000
    ]
)
""")

set_meas("41-0.00059999997rad-12coupling-5_1_winding-5e9-drive", """
Piecewise(
    [
        [ -17072,  -9628 ],  # -1.000
        [  -7871,  -7153 ],  #  0.000
        [  17393,   1587 ],  #  1.000
    ]
)
""")

set_meas("41-0.00059999997rad-12coupling-5_1_winding-1e10-drive", """
Piecewise(
    [
        [ -17278, -16787 ],  # -2.000
        [ -17152, -16791 ],  # -1.000
        [ -17122, -16791 ],  # -0.800
        [ -17049, -16793 ],  # -0.500
        [ -16939, -16793 ],  # -0.300
        [ -16309, -16793 ],  # -0.200
        [  -8145, -16790 ],  #  0.000
        [  -7861, -16789 ],  #  0.010
        [  -7546, -16789 ],  #  0.020
        [  -7185, -16789 ],  #  0.030
        [  -6761, -16788 ],  #  0.040
        [  -6218, -16787 ],  #  0.050
        [  -5198, -16784 ],  #  0.060
        [  -3481, -16771 ],  #  0.070
        [  -1700, -16721 ],  #  0.080
        [    100, -16605 ],  #  0.090
        [   1910, -16485 ],  #  0.100
        [   3723, -16376 ],  #  0.110
        [   5537, -16217 ],  #  0.120
        [   7350, -15846 ],  #  0.130
        [   9159, -14427 ],  #  0.140
        [  10962, -12489 ],  #  0.150
        [  12734, -10349 ],  #  0.160
        [  14431,  -8186 ],  #  0.170
        [  15823,  -6330 ],  #  0.180
        [  16337,  -5661 ],  #  0.190
        [  16625,  -5291 ],  #  0.200
        [  16964,  -4863 ],  #  0.220
        [  17209,  -4567 ],  #  0.250
        [  17270,  -4494 ],  #  0.270
        [  17325,  -4428 ],  #  0.300
        [  17398,  -4332 ],  #  0.400
        [  17436,  -4284 ],  #  0.500
        [  17508,  -4191 ],  #  0.800
        [  17532,  -4162 ],  #  1.000
        [  17624,  -4054 ],  #  2.000
    ]
)
""")

set_meas("41-0.00059999997rad-9coupling-7_1_winding-3e9-drive", """
Piecewise(
    [
        [ -17102,   6449 ],  # -2.000
        [ -17031,   6440 ],  # -1.500
        [ -16783,   6426 ],  # -1.000
        [ -16089,   6493 ],  # -0.800
        [  -9028,   8147 ],  # -0.600
        [  -6768,   8679 ],  # -0.500
        [  -6195,   8807 ],  # -0.400
        [  -5589,   8931 ],  # -0.300
        [  -5268,   8992 ],  # -0.250
        [  -4937,   9056 ],  # -0.200
        [  -4735,   9094 ],  # -0.170
        [  -4600,   9118 ],  # -0.150
        [  -4463,   9142 ],  # -0.130
        [  -4257,   9181 ],  # -0.100
        [  -4052,   9219 ],  # -0.070
        [  -3915,   9246 ],  # -0.050
        [  -3573,   9323 ],  #  0.000
        [  -3236,   9403 ],  #  0.050
        [  -2906,   9484 ],  #  0.100
        [  -2270,   9641 ],  #  0.200
        [  -1639,   9803 ],  #  0.300
        [   4763,  11378 ],  #  0.500
        [  16092,  13915 ],  #  0.800
        [  16937,  13981 ],  #  1.000
        [  17192,  13998 ],  #  1.500
        [  17253,  14001 ],  #  2.000
    ]
)
""")

set_meas("41-0.00059999997rad-9coupling-7_1_winding-1e10-drive", """
Piecewise(
    [
        [ -17161, -16039 ],  # -1.000
        [  -9246, -15387 ],  # -0.170
        [  -8663, -15290 ],  # -0.150
        [  -8204, -15211 ],  # -0.130
        [  -7628, -15110 ],  # -0.100
        [  -7141, -15021 ],  # -0.070
        [  -6826, -14962 ],  # -0.050
        [  -5969, -14789 ],  #  0.000
        [  -4881, -14560 ],  #  0.050
        [  -4279, -14432 ],  #  0.070
        [  -2118, -13958 ],  #  0.100
        [   4725, -12075 ],  #  0.150
        [   7491, -10729 ],  #  0.170
        [  11600,  -7816 ],  #  0.200
        [  17401,  -2236 ],  #  1.000
    ]
)
""")

set_meas("41-0.00059999997rad-9coupling-7_1_winding-2e10-drive", """
Piecewise(
    [
        [ -17416, -16677 ],  # -2.000
        [ -17259, -16681 ],  # -1.000
        [ -17217, -16681 ],  # -0.800
        [ -17145, -16681 ],  # -0.500
        [ -17059, -16682 ],  # -0.300
        [ -16957, -16681 ],  # -0.200
        [  -6085, -16652 ],  #  0.000
        [  -5524, -16650 ],  #  0.010
        [  -4908, -16646 ],  #  0.020
        [  -4208, -16643 ],  #  0.030
        [  -3340, -16640 ],  #  0.040
        [  -2039, -16633 ],  #  0.050
        [    585, -16623 ],  #  0.060
        [   3351, -16583 ],  #  0.070
        [   6122, -16343 ],  #  0.080
        [   8884, -15936 ],  #  0.090
        [  11604, -14517 ],  #  0.100
        [  14241, -11570 ],  #  0.110
        [  16127,  -9006 ],  #  0.120
        [  16616,  -8357 ],  #  0.130
        [  16865,  -8033 ],  #  0.140
        [  17033,  -7821 ],  #  0.150
        [  17125,  -7704 ],  #  0.160
        [  17179,  -7636 ],  #  0.170
        [  17216,  -7592 ],  #  0.180
        [  17240,  -7562 ],  #  0.190
        [  17257,  -7538 ],  #  0.200
        [  17283,  -7500 ],  #  0.220
        [  17316,  -7453 ],  #  0.250
        [  17333,  -7429 ],  #  0.270
        [  17356,  -7398 ],  #  0.300
        [  17409,  -7323 ],  #  0.400
        [  17445,  -7270 ],  #  0.500
        [  17518,  -7168 ],  #  0.800
        [  17554,  -7118 ],  #  1.000
        [  17669,  -6958 ],  #  2.000
    ]
)
""")

set_meas("41-0.0008rad-24coupling-3_1_winding-3e9-drive", """
Piecewise(
    [
        [ -17069, -11730 ],  # -1.000
        [ -11176,  -9819 ],  #  0.000
        [ -10918,  -9736 ],  #  0.030
        [ -10655,  -9650 ],  #  0.050
        [ -10046,  -9448 ],  #  0.070
        [  -9370,  -9224 ],  #  0.080
        [  -8625,  -8977 ],  #  0.090
        [  -7846,  -8720 ],  #  0.100
        [  -6237,  -8190 ],  #  0.120
        [  -3752,  -7373 ],  #  0.150
        [  -2076,  -6821 ],  #  0.170
        [    448,  -5988 ],  #  0.200
        [   4659,  -4545 ],  #  0.250
        [   8859,  -2793 ],  #  0.300
        [  17259,   3366 ],  #  0.500
        [  17595,   3583 ],  #  0.800
        [  17643,   3591 ],  #  1.000
    ]
)
""")

set_meas("41-0.0008rad-24coupling-3_1_winding-5e9-drive", """
Piecewise(
    [
        [ -17239, -16835 ],  # -2.000
        [ -17135, -16838 ],  # -1.000
        [ -17096, -16838 ],  # -0.800
        [ -17001, -16839 ],  # -0.500
        [ -16679, -16838 ],  # -0.300
        [ -13843, -16820 ],  # -0.200
        [ -11603, -16793 ],  #  0.000
        [ -11435, -16791 ],  #  0.010
        [ -11219, -16786 ],  #  0.020
        [ -10872, -16778 ],  #  0.030
        [ -10323, -16761 ],  #  0.040
        [  -9251, -16718 ],  #  0.050
        [  -7929, -16659 ],  #  0.060
        [  -6572, -16587 ],  #  0.070
        [  -5193, -16522 ],  #  0.080
        [  -3800, -16474 ],  #  0.090
        [  -2401, -16416 ],  #  0.100
        [   -999, -16338 ],  #  0.110
        [    405, -16220 ],  #  0.120
        [   1809, -15952 ],  #  0.130
        [   3214, -15087 ],  #  0.140
        [   4618, -13905 ],  #  0.150
        [   6021, -12563 ],  #  0.160
        [   7414, -11146 ],  #  0.170
        [   8801,  -9695 ],  #  0.180
        [  10186,  -8221 ],  #  0.190
        [  11558,  -6738 ],  #  0.200
        [  14214,  -3745 ],  #  0.220
        [  16615,   -864 ],  #  0.250
        [  16980,   -390 ],  #  0.270
        [  17299,     14 ],  #  0.300
        [  17579,    344 ],  #  0.400
        [  17647,    423 ],  #  0.500
        [  17747,    536 ],  #  0.800
        [  17781,    584 ],  #  1.000
        [  17874,    697 ],  #  2.000
    ]
)
""")

set_meas("41-0.0008rad-16coupling-5_1_winding-5e9-drive", """
Piecewise(
    [
        [ -17134,  -9658 ],  # -1.000
        [  17450,   1684 ],  #  1.000
    ]
)
""")

set_meas("41-0.0008rad-16coupling-5_1_winding-1e10-drive", """
Piecewise(
    [
        [ -17238, -16871 ],  # -1.000
        [  -7946, -16859 ],  #  0.000
        [  -7589, -16858 ],  #  0.010
        [  -7194, -16857 ],  #  0.020
        [  -6752, -16855 ],  #  0.030
        [  -6247, -16854 ],  #  0.040
        [  -5653, -16850 ],  #  0.050
        [  -4907, -16844 ],  #  0.060
        [  -3543, -16829 ],  #  0.070
        [  -1763, -16776 ],  #  0.080
        [     40, -16662 ],  #  0.090
        [   1848, -16536 ],  #  0.100
        [   3660, -16436 ],  #  0.110
        [   5471, -16307 ],  #  0.120
        [   7280, -15964 ],  #  0.130
        [   9074, -14841 ],  #  0.140
        [  10831, -13055 ],  #  0.150
        [  12556, -10985 ],  #  0.160
        [  14254,  -8775 ],  #  0.170
        [  15669,  -6906 ],  #  0.180
        [  16448,  -5866 ],  #  0.190
        [  16701,  -5533 ],  #  0.200
        [  17008,  -5132 ],  #  0.220
        [  17246,  -4830 ],  #  0.250
        [  17306,  -4752 ],  #  0.270
        [  17359,  -4686 ],  #  0.300
        [  17445,  -4573 ],  #  0.400
        [  17495,  -4508 ],  #  0.500
        [  17587,  -4394 ],  #  0.800
        [  17623,  -4347 ],  #  1.000
        [  17715,  -4225 ],  #  2.000
    ]
)
""")

set_meas("41-0.0008rad-16coupling-5_1_winding-2e10-drive", """
Piecewise(
    [
        [ -17467, -16810 ],  # -2.000
        [ -17320, -16811 ],  # -1.000
        [ -17281, -16811 ],  # -0.800
        [ -17205, -16810 ],  # -0.500
        [ -17123, -16810 ],  # -0.300
        [ -17041, -16810 ],  # -0.200
        [ -16796, -16808 ],  # -0.120
        [ -16135, -16803 ],  # -0.100
        [ -13294, -16767 ],  # -0.070
        [ -11735, -16745 ],  # -0.050
        [  -7651, -16679 ],  #  0.000
        [  -4083, -16647 ],  #  0.030
        [   1874, -16746 ],  #  0.050
        [   5490, -16708 ],  #  0.060
        [   9077, -15721 ],  #  0.070
        [  12544, -12704 ],  #  0.080
        [  15642,  -9367 ],  #  0.090
        [  16738,  -8083 ],  #  0.100
        [  17053,  -7710 ],  #  0.110
        [  17241,  -7488 ],  #  0.120
        [  17325,  -7388 ],  #  0.130
        [  17378,  -7325 ],  #  0.140
        [  17406,  -7290 ],  #  0.150
        [  17446,  -7238 ],  #  0.170
        [  17492,  -7179 ],  #  0.200
        [  17514,  -7149 ],  #  0.220
        [  17543,  -7111 ],  #  0.250
        [  17560,  -7090 ],  #  0.270
        [  17581,  -7061 ],  #  0.300
        [  17639,  -6986 ],  #  0.400
        [  17679,  -6933 ],  #  0.500
        [  17748,  -6840 ],  #  0.800
        [  17777,  -6800 ],  #  1.000
        [  17871,  -6673 ],  #  2.000
    ]
)
""")

set_meas("41-0.0008rad-10coupling-7_1_winding-25e8-drive", """
Piecewise(
    [
        [ -12586,  10173 ],  # -1.000
        [  -1330,  11989 ],  # -0.170
        [  -1243,  12002 ],  # -0.150
        [  -1157,  12014 ],  # -0.130
        [  -1034,  12033 ],  # -0.100
        [   -916,  12051 ],  # -0.070
        [   -841,  12062 ],  # -0.050
        [   -662,  12087 ],  #  0.000
        [   -506,  12107 ],  #  0.050
        [   -378,  12124 ],  #  0.100
        [  12563,  12412 ],  #  1.000
    ]
)
""")

set_meas("41-0.0008rad-10coupling-7_1_winding-3e9-drive", """
Piecewise(
    [
        [ -16374,   8010 ],  # -1.0
    ]
)
""")

set_meas("41-0.0008rad-10coupling-7_1_winding-1e10-drive", """
Piecewise(
    [
        [ -17135, -11905 ],  # -2.000
        [ -17045, -11918 ],  # -1.000
        [ -17010, -11920 ],  # -0.800
        [ -16906, -11920 ],  # -0.500
        [ -16353, -11863 ],  # -0.300
        [  -9810, -10355 ],  # -0.200
        [  -8957, -10158 ],  # -0.170
        [  -8473, -10047 ],  # -0.150
        [  -8028,  -9944 ],  # -0.130
        [  -7819,  -9896 ],  # -0.120
        [  -7424,  -9805 ],  # -0.100
        [  -6870,  -9676 ],  # -0.070
        [  -6507,  -9592 ],  # -0.050
        [  -5569,  -9373 ],  #  0.000
        [  -4937,  -9226 ],  #  0.030
        [  -4459,  -9115 ],  #  0.050
        [  -4197,  -9055 ],  #  0.060
        [  -3917,  -8989 ],  #  0.070
        [  -3287,  -8843 ],  #  0.090
        [  -2919,  -8757 ],  #  0.100
        [  -2493,  -8657 ],  #  0.110
        [  -1952,  -8530 ],  #  0.120
        [  -1124,  -8334 ],  #  0.130
        [      3,  -8067 ],  #  0.140
        [   1148,  -7794 ],  #  0.150
        [   3449,  -7232 ],  #  0.170
        [   6900,  -6215 ],  #  0.200
        [   9172,  -5277 ],  #  0.220
        [  12488,  -3370 ],  #  0.250
        [  14637,  -1772 ],  #  0.270
        [  16415,   -254 ],  #  0.300
        [  17009,    181 ],  #  0.500
        [  17111,    230 ],  #  0.800
        [  17145,    245 ],  #  1.000
        [  17230,    295 ],  #  2.000
    ]
)
""")

set_meas("41-0.0008rad-10coupling-7_1_winding-2e10-drive", """
Piecewise(
    [
        [ -17212, -16635 ],  # -2.000
        [ -17123, -16636 ],  # -1.000
        [ -17101, -16636 ],  # -0.800
        [ -17050, -16637 ],  # -0.500
        [ -16961, -16636 ],  # -0.300
        [ -16825, -16637 ],  # -0.200
        [  -5610, -16583 ],  #  0.000
        [  -5008, -16578 ],  #  0.010
        [  -4367, -16573 ],  #  0.020
        [  -3670, -16565 ],  #  0.030
        [  -2906, -16552 ],  #  0.040
        [  -2039, -16533 ],  #  0.050
        [  -1016, -16500 ],  #  0.060
        [    294, -16430 ],  #  0.070
        [   2312, -16256 ],  #  0.080
        [   4612, -16021 ],  #  0.090
        [   6911, -15567 ],  #  0.100
        [   9168, -14383 ],  #  0.110
        [  11386, -12285 ],  #  0.120
        [  13563,  -9740 ],  #  0.130
        [  15518,  -7287 ],  #  0.140
        [  16433,  -6133 ],  #  0.150
        [  16626,  -5903 ],  #  0.160
        [  16743,  -5767 ],  #  0.170
        [  16836,  -5660 ],  #  0.180
        [  16903,  -5585 ],  #  0.190
        [  16943,  -5537 ],  #  0.200
        [  16990,  -5480 ],  #  0.220
        [  17032,  -5427 ],  #  0.250
        [  17053,  -5400 ],  #  0.270
        [  17079,  -5367 ],  #  0.300
        [  17140,  -5288 ],  #  0.400
        [  17178,  -5240 ],  #  0.500
        [  17242,  -5162 ],  #  0.800
        [  17262,  -5136 ],  #  1.000
        [  17330,  -5056 ],  #  2.000
    ]
)
""")

set_meas("41-0.0011999999rad-36coupling-3_1_winding-3e9-drive", """
Piecewise(
    [
        [ -17043, -11870 ],  # -1.000
        [ -11293, -10023 ],  #  0.000
        [  17606,   3293 ],  #  1.000
    ]
)
""")

set_meas("41-0.0011999999rad-36coupling-3_1_winding-4e9-drive", """
Piecewise(
    [
        [ -17066, -16520 ],  # -1.000
        [ -11344, -16473 ],  #  0.000
        [  -9659, -16451 ],  #  0.050
        [  -5101, -16241 ],  #  0.100
        [  -3978, -15995 ],  #  0.110
        [  -2849, -15692 ],  #  0.120
        [  -1714, -15297 ],  #  0.130
        [   -579, -14795 ],  #  0.140
        [    557, -14176 ],  #  0.150
        [   6237, -10026 ],  #  0.200
        [   9598,  -7124 ],  #  0.230
        [  10704,  -6102 ],  #  0.240
        [  11794,  -5059 ],  #  0.250
        [  16366,   -304 ],  #  0.300
        [  17531,   1117 ],  #  0.500
        [  17676,   1267 ],  #  1.000
    ]
)
""")

set_meas("41-0.0011999999rad-36coupling-3_1_winding-5e9-drive", """
Piecewise(
    [
        [ -17088, -16823 ],  # -1.000
        [ -17053, -16823 ],  # -0.800
        [ -16982, -16824 ],  # -0.500
        [ -16705, -16824 ],  # -0.300
        [ -10940, -16779 ],  #  0.000
        [ -10565, -16771 ],  #  0.010
        [ -10120, -16759 ],  #  0.020
        [  -9595, -16742 ],  #  0.030
        [  -8973, -16719 ],  #  0.040
        [  -8226, -16688 ],  #  0.050
        [  -7348, -16647 ],  #  0.060
        [  -6310, -16600 ],  #  0.070
        [  -5057, -16560 ],  #  0.080
        [  -3673, -16528 ],  #  0.090
        [  -2271, -16494 ],  #  0.100
        [   -859, -16444 ],  #  0.110
        [    557, -16351 ],  #  0.120
        [   1974, -15945 ],  #  0.130
        [   3392, -15196 ],  #  0.140
        [   4810, -14082 ],  #  0.150
        [   6226, -12725 ],  #  0.160
        [   7628, -11301 ],  #  0.170
        [   9022,  -9850 ],  #  0.180
        [  10405,  -8349 ],  #  0.190
        [  11764,  -6827 ],  #  0.200
        [  13080,  -5311 ],  #  0.210
        [  14348,  -3822 ],  #  0.220
        [  15465,  -2501 ],  #  0.230
        [  16321,  -1492 ],  #  0.240
        [  16745,   -952 ],  #  0.250
        [  16921,   -711 ],  #  0.260
        [  17054,   -528 ],  #  0.270
        [  17323,   -169 ],  #  0.300
        [  17549,    109 ],  #  0.400
        [  17612,    182 ],  #  0.500
        [  17698,    284 ],  #  0.800
        [  17729,    327 ],  #  1.000
    ]
)
""")

set_meas("41-0.0011999999rad-24coupling-5_1_winding-5e9-drive", """
Piecewise(
    [
        [ -17116, -10370 ],  # -1.000
        [  -7768,  -7873 ],  #  0.000
        [  -7261,  -7732 ],  #  0.030
        [  -6838,  -7615 ],  #  0.050
        [  -6591,  -7546 ],  #  0.060
        [  -6314,  -7469 ],  #  0.070
        [  -6006,  -7383 ],  #  0.080
        [  -5661,  -7287 ],  #  0.090
        [  -5274,  -7179 ],  #  0.100
        [  -4838,  -7058 ],  #  0.110
        [  -4342,  -6920 ],  #  0.120
        [  -3767,  -6760 ],  #  0.130
        [  -3073,  -6567 ],  #  0.140
        [  -2185,  -6320 ],  #  0.150
        [   -317,  -5799 ],  #  0.170
        [   2499,  -5012 ],  #  0.200
        [   4378,  -4485 ],  #  0.220
        [   7194,  -3685 ],  #  0.250
        [   9051,  -3122 ],  #  0.270
        [  11784,  -2091 ],  #  0.300
        [  16846,    999 ],  #  0.400
        [  17218,   1235 ],  #  0.500
        [  17390,   1269 ],  #  0.800
        [  17433,   1280 ],  #  1.000
    ]
)
""")

set_meas("41-0.0011999999rad-24coupling-5_1_winding-1e10-drive", """
Piecewise(
    [
        [ -17194, -16847 ],  # -1.000
        [  -7141, -16835 ],  #  0.000
        [  -6549, -16832 ],  #  0.010
        [  -5905, -16829 ],  #  0.020
        [  -5197, -16824 ],  #  0.030
        [  -4420, -16819 ],  #  0.040
        [  -3567, -16811 ],  #  0.050
        [  -2629, -16795 ],  #  0.060
        [  -1590, -16759 ],  #  0.070
        [   -423, -16696 ],  #  0.080
        [    917, -16613 ],  #  0.090
        [   2558, -16529 ],  #  0.100
        [   4412, -16454 ],  #  0.110
        [   6276, -16325 ],  #  0.120
        [   8125, -15601 ],  #  0.130
        [   9951, -13826 ],  #  0.140
        [  11746, -11695 ],  #  0.150
        [  13446,  -9541 ],  #  0.160
        [  15025,  -7458 ],  #  0.170
        [  16182,  -5954 ],  #  0.180
        [  16707,  -5254 ],  #  0.190
        [  16881,  -5016 ],  #  0.200
        [  17010,  -4842 ],  #  0.210
        [  17111,  -4706 ],  #  0.220
        [  17258,  -4514 ],  #  0.250
        [  17309,  -4447 ],  #  0.270
        [  17357,  -4383 ],  #  0.300
        [  17437,  -4280 ],  #  0.400
        [  17482,  -4221 ],  #  0.500
        [  17553,  -4130 ],  #  0.800
        [  17582,  -4093 ],  #  1.000
    ]
)
""")

set_meas("42-0.0004rad-9coupling-3_1_winding-5e9-drive", """
Piecewise(
    [
        [ -16953,   1967 ],  # -1.000
        [ -15202,    409 ],  # -0.300
        [ -12404,  -2369 ],  # -0.270
        [ -10309,  -4357 ],  # -0.250
        [  -5015,  -9040 ],  # -0.200
        [  -2894, -10740 ],  # -0.180
        [   -775, -12375 ],  # -0.160
        [    280, -13126 ],  # -0.150
        [   5462, -15543 ],  # -0.100
        [   9315, -15973 ],  # -0.050
        [  10424, -16039 ],  #  0.000
        [  13125, -16147 ],  #  0.200
        [  16821, -16252 ],  #  1.000
    ]
)
""")

set_meas("42-0.0004rad-9coupling-3_1_winding-8e9-drive", """
Piecewise(
    [
        [ -17028,    -32 ],  # -1.000
        [ -16860,   -213 ],  # -0.300
        [ -16671,   -383 ],  # -0.250
        [ -15880,  -1121 ],  # -0.200
        [ -15375,  -1631 ],  # -0.190
        [ -14252,  -2821 ],  # -0.180
        [ -12607,  -4558 ],  # -0.170
        [ -10931,  -6288 ],  # -0.160
        [  -9238,  -8004 ],  # -0.150
        [  -7540,  -9692 ],  # -0.140
        [  -5842, -11279 ],  # -0.130
        [  -4142, -12833 ],  # -0.120
        [  -2442, -14325 ],  # -0.110
        [   -742, -15569 ],  # -0.100
        [    956, -16090 ],  # -0.090
        [   2649, -16242 ],  # -0.080
        [   4334, -16371 ],  # -0.070
        [   6003, -16518 ],  # -0.060
        [   7627, -16656 ],  # -0.050
        [   9154, -16725 ],  # -0.040
        [   9894, -16735 ],  # -0.030
        [  10353, -16739 ],  # -0.020
        [  10726, -16742 ],  # -0.010
        [  11029, -16744 ],  #  0.000
        [  11290, -16744 ],  #  0.010
        [  11529, -16745 ],  #  0.020
        [  12176, -16749 ],  #  0.050
        [  12523, -16752 ],  #  0.070
        [  13010, -16755 ],  #  0.100
        [  13809, -16759 ],  #  0.150
        [  15012, -16765 ],  #  0.200
        [  16724, -16771 ],  #  0.300
        [  16868, -16769 ],  #  1.000
    ]
)
""")

set_meas("42-0.0004rad-9coupling-3_1_winding-1e10-drive", """
Piecewise(
    [
        [ -17058,   -701 ],  # -1.000
        [ -16918,   -872 ],  # -0.300
        [ -16899,   -893 ],  # -0.270
        [ -16878,   -915 ],  # -0.250
        [ -16813,   -979 ],  # -0.220
        [ -16674,  -1109 ],  # -0.200
        [ -16569,  -1203 ],  # -0.190
        [ -16433,  -1326 ],  # -0.180
        [ -16229,  -1523 ],  # -0.170
        [ -15873,  -1887 ],  # -0.160
        [ -15180,  -2619 ],  # -0.150
        [ -13427,  -4501 ],  # -0.140
        [ -11336,  -6619 ],  # -0.130
        [  -9221,  -8727 ],  # -0.120
        [  -7097, -10810 ],  # -0.110
        [  -4971, -12812 ],  # -0.100
        [  -2844, -14563 ],  # -0.090
        [   -717, -15913 ],  # -0.080
        [   1408, -16328 ],  # -0.070
        [   3525, -16512 ],  # -0.060
        [   5626, -16686 ],  # -0.050
        [   7675, -16734 ],  # -0.040
        [   9344, -16748 ],  # -0.030
        [  10132, -16756 ],  # -0.020
        [  10681, -16761 ],  # -0.010
        [  11121, -16764 ],  #  0.000
        [  13525, -16776 ],  #  0.100
        [  16456, -16784 ],  #  0.200
        [  16883, -16784 ],  #  1.000
    ]
)
""")

set_meas("42-0.0004rad-9coupling-3_1_winding-12e9-drive", """
Piecewise(
    [
        [ -17082,  -1246 ],  # -1.000
        [ -16942,  -1399 ],  # -0.300
        [ -16872,  -1480 ],  # -0.200
        [ -16845,  -1504 ],  # -0.190
        [ -16802,  -1541 ],  # -0.180
        [ -16720,  -1612 ],  # -0.170
        [ -16598,  -1715 ],  # -0.160
        [ -16441,  -1849 ],  # -0.150
        [ -16184,  -2083 ],  # -0.140
        [ -15653,  -2601 ],  # -0.130
        [ -14247,  -4101 ],  # -0.120
        [ -11762,  -6794 ],  # -0.110
        [  -9226,  -9334 ],  # -0.100
        [  -6677, -11728 ],  # -0.090
        [  -4125, -13936 ],  # -0.080
        [  -1573, -15721 ],  # -0.070
        [    977, -16397 ],  # -0.060
        [   3514, -16641 ],  # -0.050
        [  11078, -16764 ],  #  0.000
        [  16686, -16774 ],  #  0.200
        [  16893, -16773 ],  #  1.000
    ]
)
""")

set_meas("42-0.0004rad-9coupling-3_1_winding-15000000512e0-drive", """
Piecewise(
    [
        [ -17110,  -2110 ],  # -1.000
        [ -16976,  -2260 ],  # -0.300
        [ -16931,  -2322 ],  # -0.200
        [ -16924,  -2331 ],  # -0.190
        [ -16913,  -2343 ],  # -0.180
        [ -16898,  -2357 ],  # -0.170
        [ -16877,  -2378 ],  # -0.160
        [ -16843,  -2410 ],  # -0.150
        [ -16775,  -2470 ],  # -0.140
        [ -16635,  -2593 ],  # -0.130
        [ -16442,  -2757 ],  # -0.120
        [ -16078,  -3091 ],  # -0.110
        [ -15175,  -3969 ],  # -0.100
        [ -12385,  -6776 ],  # -0.090
        [  -9216,  -9896 ],  # -0.080
        [  -6025, -12920 ],  # -0.070
        [  -2832, -15395 ],  # -0.060
        [    360, -16474 ],  # -0.050
        [  11090, -16322 ],  #  0.000
        [  16743, -16310 ],  #  0.200
        [  16908, -16342 ],  #  1.000
    ]
)
""")

set_meas("42-0.0004rad-9coupling-3_1_winding-2e10-drive", """
Piecewise(
    [
        [ -17150,  -2787 ],  # -1.000
        [ -16914,  -3077 ],  # -0.130
        [ -16694,  -3287 ],  # -0.100
        [ -16449,  -3506 ],  # -0.090
        [ -15875,  -4060 ],  # -0.080
        [ -13426,  -6648 ],  # -0.070
        [  -9210, -10853 ],  # -0.060
        [  -4953, -14268 ],  # -0.050
        [   -695, -16397 ],  # -0.040
        [  11016, -15700 ],  #  0.000
        [  16936, -15322 ],  #  1.000
    ]
)
""")

set_meas("42-0.0004rad-6coupling-5_1_winding-1e10-drive", """
Piecewise(
    [
        [ -16953,  -1236 ],  # -1.000
        [ -16867,  -1305 ],  # -0.500
        [ -16838,  -1341 ],  # -0.400
        [ -16620,  -1540 ],  # -0.300
        [ -16476,  -1662 ],  # -0.280
        [ -16376,  -1751 ],  # -0.270
        [ -16250,  -1872 ],  # -0.260
        [ -16062,  -2061 ],  # -0.250
        [ -15792,  -2343 ],  # -0.240
        [ -15351,  -2824 ],  # -0.230
        [ -14424,  -3913 ],  # -0.220
        [ -13123,  -5448 ],  # -0.210
        [ -11787,  -6960 ],  # -0.200
        [ -10417,  -8395 ],  # -0.190
        [  -9028,  -9761 ],  # -0.180
        [  -7638, -11057 ],  # -0.170
        [  -6245, -12251 ],  # -0.160
        [  -4851, -13364 ],  # -0.150
        [  -3455, -14384 ],  # -0.140
        [  -2060, -15176 ],  # -0.130
        [   -671, -15606 ],  # -0.120
        [    710, -15827 ],  # -0.110
        [   2081, -15977 ],  # -0.100
        [   3420, -16083 ],  # -0.090
        [   4582, -16158 ],  # -0.080
        [   5305, -16198 ],  # -0.070
        [   5835, -16224 ],  # -0.060
        [   6243, -16242 ],  # -0.050
        [   6854, -16266 ],  # -0.030
        [   7099, -16275 ],  # -0.020
        [   7533, -16291 ],  #  0.000
        [   8456, -16327 ],  #  0.050
        [   9359, -16363 ],  #  0.100
        [  10373, -16399 ],  #  0.150
        [  12271, -16458 ],  #  0.200
        [  16568, -16554 ],  #  0.300
        [  16878, -16551 ],  #  1.000
    ]
)
""")

set_meas("42-0.0004rad-6coupling-5_1_winding-15000000512e0-drive", """
Piecewise(
    [
        [ -17005,  -3590 ],  # -1.000
        [ -16930,  -3676 ],  # -0.500
        [ -16902,  -3711 ],  # -0.400
        [ -16862,  -3764 ],  # -0.300
        [ -16859,  -3770 ],  # -0.290
        [ -16853,  -3779 ],  # -0.280
        [ -16848,  -3788 ],  # -0.270
        [ -16842,  -3799 ],  # -0.260
        [ -16834,  -3810 ],  # -0.250
        [ -16822,  -3823 ],  # -0.240
        [ -16802,  -3845 ],  # -0.230
        [ -16774,  -3878 ],  # -0.220
        [ -16716,  -3939 ],  # -0.210
        [ -16626,  -4032 ],  # -0.200
        [ -16521,  -4135 ],  # -0.190
        [ -16380,  -4274 ],  # -0.180
        [ -16171,  -4499 ],  # -0.170
        [ -15792,  -4930 ],  # -0.160
        [ -15005,  -5867 ],  # -0.150
        [ -13126,  -8145 ],  # -0.140
        [ -11099, -10303 ],  # -0.130
        [  -9021, -12245 ],  # -0.120
        [  -6931, -13987 ],  # -0.110
        [  -4837, -15331 ],  # -0.100
        [  -2741, -16042 ],  # -0.090
        [   -650, -16321 ],  # -0.080
        [   1424, -16524 ],  # -0.070
        [   3452, -16593 ],  # -0.060
        [   5174, -16594 ],  # -0.050
        [   6088, -16593 ],  # -0.040
        [   6667, -16594 ],  # -0.030
        [   7108, -16595 ],  # -0.020
        [   7479, -16595 ],  # -0.010
        [   7812, -16597 ],  #  0.000
        [   8410, -16600 ],  #  0.020
        [   9254, -16603 ],  #  0.050
        [  10724, -16610 ],  #  0.100
        [  11520, -16616 ],  #  0.120
        [  14354, -16637 ],  #  0.150
        [  15962, -16647 ],  #  0.170
        [  16585, -16650 ],  #  0.200
        [  16790, -16651 ],  #  0.300
        [  16910, -16650 ],  #  1.000
    ]
)
""")

set_meas("42-0.0004rad-6coupling-5_1_winding-2e10-drive", """
Piecewise(
    [
        [ -17036,  -4747 ],  # -1.000
        [ -16973,  -4815 ],  # -0.500
        [ -16947,  -4844 ],  # -0.400
        [ -16910,  -4887 ],  # -0.300
        [ -16895,  -4904 ],  # -0.270
        [ -16889,  -4911 ],  # -0.260
        [ -16884,  -4918 ],  # -0.250
        [ -16878,  -4926 ],  # -0.240
        [ -16872,  -4933 ],  # -0.230
        [ -16867,  -4942 ],  # -0.220
        [ -16860,  -4952 ],  # -0.210
        [ -16853,  -4965 ],  # -0.200
        [ -16843,  -4979 ],  # -0.190
        [ -16828,  -4994 ],  # -0.180
        [ -16800,  -5025 ],  # -0.170
        [ -16745,  -5081 ],  # -0.160
        [ -16634,  -5191 ],  # -0.150
        [ -16486,  -5328 ],  # -0.140
        [ -16257,  -5557 ],  # -0.130
        [ -15795,  -6052 ],  # -0.120
        [ -14433,  -7655 ],  # -0.110
        [ -11792, -10704 ],  # -0.100
        [  -9029, -13408 ],  # -0.090
        [  -6242, -15302 ],  # -0.080
        [  -3449, -16151 ],  # -0.070
        [   -661, -16459 ],  # -0.060
        [   2088, -16422 ],  # -0.050
        [   4534, -16328 ],  # -0.040
        [   5864, -16297 ],  # -0.030
        [   6705, -16287 ],  # -0.020
        [   7334, -16284 ],  # -0.010
        [   7857, -16284 ],  #  0.000
        [   8323, -16286 ],  #  0.010
        [   8752, -16290 ],  #  0.020
        [   9162, -16297 ],  #  0.030
        [   9564, -16302 ],  #  0.040
        [   9967, -16308 ],  #  0.050
        [  10385, -16317 ],  #  0.060
        [  10832, -16326 ],  #  0.070
        [  11340, -16337 ],  #  0.080
        [  11960, -16356 ],  #  0.090
        [  12796, -16374 ],  #  0.100
        [  16586, -16472 ],  #  0.150
        [  16776, -16478 ],  #  0.200
        [  16830, -16480 ],  #  0.300
        [  16928, -16487 ],  #  1.000
    ]
)
""")

set_meas("42-0.0004rad-6coupling-5_1_winding-30000001024e0-drive", """
Piecewise(
    [
        [ -17094,  -6205 ],  # -1.000
        [ -16986,  -6322 ],  # -0.300
        [ -16962,  -6347 ],  # -0.250
        [ -16931,  -6386 ],  # -0.200
        [ -16924,  -6398 ],  # -0.190
        [ -16916,  -6409 ],  # -0.180
        [ -16906,  -6421 ],  # -0.170
        [ -16896,  -6431 ],  # -0.160
        [ -16886,  -6443 ],  # -0.150
        [ -16876,  -6459 ],  # -0.140
        [ -16864,  -6481 ],  # -0.130
        [ -16842,  -6512 ],  # -0.120
        [ -16793,  -6563 ],  # -0.110
        [ -16647,  -6707 ],  # -0.100
        [ -16395,  -6930 ],  # -0.090
        [ -15804,  -7526 ],  # -0.080
        [ -13172, -10328 ],  # -0.070
        [  -9055, -13889 ],  # -0.060
        [  -4874, -15886 ],  # -0.050
        [   7617, -14837 ],  #  0.000
        [  12117, -14761 ],  #  0.050
        [  16555, -15029 ],  #  0.100
        [  16803, -15048 ],  #  0.150
        [  16841, -15055 ],  #  0.200
        [  16879, -15063 ],  #  0.300
        [  16968, -15089 ],  #  1.000
    ]
)
""")

set_meas("42-0.0004rad-4coupling-7_1_winding-1e10-drive", """
Piecewise(
    [
        [ -16737,   4481 ],  # -1.000
        [  -9660,    519 ],  # -0.300
        [  -5425,   -748 ],  # -0.250
        [  -1200,  -1850 ],  # -0.200
        [   -379,  -2063 ],  # -0.190
        [    431,  -2274 ],  # -0.180
        [   1202,  -2476 ],  # -0.170
        [   1849,  -2646 ],  # -0.160
        [   2332,  -2774 ],  # -0.150
        [   2705,  -2873 ],  # -0.140
        [   3005,  -2953 ],  # -0.130
        [   3256,  -3019 ],  # -0.120
        [   3471,  -3077 ],  # -0.110
        [   3658,  -3126 ],  # -0.100
        [   3827,  -3172 ],  # -0.090
        [   3983,  -3213 ],  # -0.080
        [   4129,  -3252 ],  # -0.070
        [   4269,  -3289 ],  # -0.060
        [   4404,  -3325 ],  # -0.050
        [   5052,  -3500 ],  #  0.000
        [   5659,  -3664 ],  #  0.050
        [   6244,  -3818 ],  #  0.100
        [   6827,  -3967 ],  #  0.150
        [   7456,  -4125 ],  #  0.200
        [   9934,  -4753 ],  #  0.300
        [  16734,  -6322 ],  #  1.000
    ]
)
""")

set_meas("42-0.0004rad-4coupling-7_1_winding-15000000512e0-drive", """
Piecewise(
    [
        [ -16782,   1586 ],  # -1.000
        [ -16485,   1362 ],  # -0.300
        [ -15944,    873 ],  # -0.260
        [ -15620,    548 ],  # -0.250
        [ -14665,   -476 ],  # -0.240
        [ -13436,  -1754 ],  # -0.230
        [ -12188,  -2989 ],  # -0.220
        [ -10931,  -4145 ],  # -0.210
        [  -9670,  -5213 ],  # -0.200
        [  -8404,  -6223 ],  # -0.190
        [  -7134,  -7190 ],  # -0.180
        [  -5862,  -8123 ],  # -0.170
        [  -4587,  -9035 ],  # -0.160
        [  -3314,  -9916 ],  # -0.150
        [  -2049, -10749 ],  # -0.140
        [   -790, -11528 ],  # -0.130
        [    437, -12232 ],  # -0.120
        [   1489, -12796 ],  # -0.110
        [   2238, -13167 ],  # -0.100
        [   2785, -13422 ],  # -0.090
        [   3203, -13608 ],  # -0.080
        [   5245, -14402 ],  #  0.000
        [  16786, -15767 ],  #  1.000
    ]
)
""")

set_meas("42-0.0004rad-4coupling-7_1_winding-2e10-drive", """
Piecewise(
    [
        [ -16820,     25 ],  # -1.000
        [ -16704,    -82 ],  # -0.300
        [ -16668,   -124 ],  # -0.250
        [ -16094,   -669 ],  # -0.200
        [ -15755,  -1038 ],  # -0.190
        [ -14674,  -2338 ],  # -0.180
        [ -13031,  -4309 ],  # -0.170
        [ -11360,  -6215 ],  # -0.160
        [  -9680,  -8020 ],  # -0.150
        [  -7991,  -9723 ],  # -0.140
        [  -6295, -11288 ],  # -0.130
        [  -4596, -12674 ],  # -0.120
        [  -2902, -13763 ],  # -0.110
        [  -1220, -14570 ],  # -0.100
        [    316, -15098 ],  # -0.090
        [   1443, -15369 ],  # -0.080
        [   2269, -15522 ],  # -0.070
        [   2903, -15616 ],  # -0.060
        [   3414, -15687 ],  # -0.050
        [   3853, -15740 ],  # -0.040
        [   4246, -15783 ],  # -0.030
        [   4609, -15819 ],  # -0.020
        [   4954, -15852 ],  # -0.010
        [   5285, -15884 ],  #  0.000
        [   5604, -15914 ],  #  0.010
        [   5914, -15942 ],  #  0.020
        [   6218, -15969 ],  #  0.030
        [   6515, -15995 ],  #  0.040
        [   6808, -16022 ],  #  0.050
        [   7102, -16048 ],  #  0.060
        [   7399, -16075 ],  #  0.070
        [   7707, -16102 ],  #  0.080
        [   8032, -16130 ],  #  0.090
        [   8377, -16158 ],  #  0.100
        [  10922, -16339 ],  #  0.150
        [  16062, -16479 ],  #  0.200
        [  16702, -16483 ],  #  0.300
        [  16818, -16479 ],  #  1.000
    ]
)
""")

set_meas("42-0.0004rad-4coupling-7_1_winding-30000001024e0-drive", """
Piecewise(
    [
        [ -16887,  -2079 ],  # -1.000
        [ -16768,  -2208 ],  # -0.300
        [ -16756,  -2221 ],  # -0.270
        [ -16748,  -2235 ],  # -0.250
        [ -16732,  -2252 ],  # -0.220
        [ -16721,  -2268 ],  # -0.200
        [ -16717,  -2275 ],  # -0.190
        [ -16710,  -2284 ],  # -0.180
        [ -16693,  -2305 ],  # -0.170
        [ -16635,  -2362 ],  # -0.160
        [ -16504,  -2479 ],  # -0.150
        [ -16311,  -2667 ],  # -0.140
        [ -15964,  -3047 ],  # -0.130
        [ -14712,  -4573 ],  # -0.120
        [ -12241,  -7566 ],  # -0.110
        [  -9726, -10234 ],  # -0.100
        [  -7193, -12344 ],  # -0.090
        [  -4666, -13727 ],  # -0.080
        [  -2384, -14509 ],  # -0.070
        [   -598, -14942 ],  # -0.060
        [    808, -15236 ],  # -0.050
        [   1913, -15438 ],  # -0.040
        [   2839, -15536 ],  # -0.030
        [   3642, -15562 ],  # -0.020
        [   4365, -15570 ],  # -0.010
        [   5045, -15582 ],  #  0.000
        [   6297, -15611 ],  #  0.020
        [   8100, -15663 ],  #  0.050
        [  12162, -15750 ],  #  0.100
        [  16534, -15882 ],  #  0.150
        [  16726, -15884 ],  #  0.200
        [  16772, -15890 ],  #  0.300
        [  16872, -15895 ],  #  1.000
    ]
)
""")

set_meas("42-0.0004rad-2coupling-13_1_winding-2e10-drive", """
Piecewise(
    [
        [ -16512,   9715 ],  # -1.000
        [  -8781,   7383 ],  # -0.300
        [  -1664,   5754 ],  # -0.200
        [  -1198,   5633 ],  # -0.190
        [   -779,   5523 ],  # -0.180
        [   -414,   5428 ],  # -0.170
        [    -68,   5340 ],  # -0.160
        [    236,   5266 ],  # -0.150
        [    509,   5199 ],  # -0.140
        [    768,   5137 ],  # -0.130
        [    995,   5084 ],  # -0.120
        [   1205,   5034 ],  # -0.110
        [   1394,   4990 ],  # -0.100
        [   1575,   4947 ],  # -0.090
        [   1744,   4907 ],  # -0.080
        [   1903,   4869 ],  # -0.070
        [   2056,   4832 ],  # -0.060
        [   2205,   4795 ],  # -0.050
        [   2900,   4600 ],  #  0.000
        [   3565,   4378 ],  #  0.050
        [   4265,   4185 ],  #  0.100
        [   5042,   4009 ],  #  0.150
        [   6035,   3838 ],  #  0.200
        [   9904,   3347 ],  #  0.300
        [  16623,   2579 ],  #  1.000
    ]
)
""")

set_meas("42-0.0004rad-2coupling-13_1_winding-30000001024e0-drive", """
Piecewise(
    [
        [ -16625,   7458 ],  # -1.000
        [ -16342,   7411 ],  # -0.300
        [  -8904,   4907 ],  # -0.200
        [  -7755,   4493 ],  # -0.190
        [  -6648,   4091 ],  # -0.180
        [  -5607,   3713 ],  # -0.170
        [  -4627,   3351 ],  # -0.160
        [  -3734,   3017 ],  # -0.150
        [  -2933,   2714 ],  # -0.140
        [  -2192,   2429 ],  # -0.130
        [  -1538,   2175 ],  # -0.120
        [   -946,   1943 ],  # -0.110
        [   -415,   1736 ],  # -0.100
        [     62,   1547 ],  # -0.090
        [    501,   1373 ],  # -0.080
        [    890,   1214 ],  # -0.070
        [   1259,   1063 ],  # -0.060
        [   1591,    924 ],  # -0.050
        [   3009,    235 ],  #  0.000
        [   4284,   -403 ],  #  0.050
        [   5707,   -934 ],  #  0.100
        [   7644,  -1482 ],  #  0.150
        [  10706,  -2127 ],  #  0.200
        [  16478,  -3591 ],  #  0.300
        [  16719,  -3647 ],  #  1.000
    ]
)
""")

set_meas("42-0.0004rad-2coupling-13_1_winding-4e10-drive", """
Piecewise(
    [
        [ -16705,   5261 ],  # -1.000
        [ -16502,   5142 ],  # -0.300
        [ -16463,   5153 ],  # -0.270
        [ -16425,   5121 ],  # -0.250
        [ -16351,   5167 ],  # -0.220
        [ -15985,   5113 ],  # -0.200
        [ -15106,   4844 ],  # -0.190
        [ -13658,   4344 ],  # -0.180
        [ -12155,   3781 ],  # -0.170
        [ -10638,   3248 ],  # -0.160
        [  -9157,   2707 ],  # -0.150
        [  -7746,   2166 ],  # -0.140
        [  -6419,   1670 ],  # -0.130
        [  -5201,   1210 ],  # -0.120
        [  -4066,    771 ],  # -0.110
        [  -3056,    394 ],  # -0.100
        [  -2149,     37 ],  # -0.090
        [  -1339,   -293 ],  # -0.080
        [   -629,   -579 ],  # -0.070
        [     14,   -853 ],  # -0.060
        [    615,  -1107 ],  # -0.050
        [   1152,  -1342 ],  # -0.040
        [   1657,  -1575 ],  # -0.030
        [   2128,  -1797 ],  # -0.020
        [   2574,  -2025 ],  # -0.010
        [   3007,  -2268 ],  #  0.000
        [   3841,  -2752 ],  #  0.020
        [   5110,  -3408 ],  #  0.050
        [   7640,  -4525 ],  #  0.100
        [  11518,  -5999 ],  #  0.150
        [  16197,  -7489 ],  #  0.200
        [  16627,  -7601 ],  #  0.300
        [  16779,  -7638 ],  #  1.000
    ]
)
""")

set_meas("42-0.0004rad-2coupling-13_1_winding-8e10-drive", """
Piecewise(
    [
        [ -16858,    108 ],  # -1.000
        [ -16740,     53 ],  # -0.200
        [ -16340,    -85 ],  # -0.100
        [ -14990,   -563 ],  # -0.090
        [ -12507,  -1430 ],  # -0.080
        [ -10107,  -2238 ],  # -0.070
        [  -7856,  -3003 ],  # -0.060
        [  -5763,  -3667 ],  # -0.050
        [  -3845,  -4323 ],  # -0.040
        [  -2095,  -4921 ],  # -0.030
        [   -467,  -5477 ],  # -0.020
        [   1074,  -5995 ],  # -0.010
        [   2575,  -6544 ],  #  0.000
        [   4045,  -7105 ],  #  0.010
        [   5521,  -7587 ],  #  0.020
        [   7014,  -8245 ],  #  0.030
        [   8551,  -8602 ],  #  0.040
        [  10143,  -9100 ],  #  0.050
        [  13576, -10098 ],  #  0.070
        [  16612, -11125 ],  #  0.100
        [  16754, -11172 ],  #  0.150
        [  16783, -11134 ],  #  0.200
        [  16810, -11166 ],  #  0.300
        [  16861, -11171 ],  #  1.000
    ]
)
""")

set_meas("42-0.0004rad-2coupling-13_1_winding-149999992832e0-drive", """
Piecewise(
    [
        [ -16883,  -1609 ],  # -1.000
        [ -16756,  -1653 ],  # -0.100
        [ -16749,  -1646 ],  # -0.090
        [ -16741,  -1664 ],  # -0.080
        [ -16728,  -1676 ],  # -0.070
        [ -16693,  -1698 ],  # -0.060
        [ -16560,  -1752 ],  # -0.050
        [ -13841,  -2098 ],  # -0.040
        [  -9580,  -2504 ],  # -0.030
        [  -5486,  -2864 ],  # -0.020
        [  -1570,  -3191 ],  # -0.010
        [   2214,  -3372 ],  #  0.000
        [   6023,  -3590 ],  #  0.010
        [   9862,  -3706 ],  #  0.020
        [  13696,  -4160 ],  #  0.030
        [  16173,  -4111 ],  #  0.040
        [  16671,  -4168 ],  #  0.050
        [  16741,  -4228 ],  #  0.060
        [  16763,  -4267 ],  #  0.070
        [  16773,  -4245 ],  #  0.080
        [  16781,  -4310 ],  #  0.090
        [  16788,  -4321 ],  #  0.100
        [  16863,  -4281 ],  #  1.000
    ]
)
""")

set_meas("42-0.00059999997rad-15coupling-3_1_winding-5e9-drive", """
Piecewise(
    [
        [ -17098,   1499 ],  # -1.000
        [  10615, -16424 ],  #  0.000
        [  16885, -16633 ],  #  1.000
    ]
)
""")

set_meas("42-0.00059999997rad-10coupling-5_1_winding-1e10-drive", """
Piecewise(
    [
        [ -17078,  -2357 ],  # -1.000
        [  16945, -16736 ],  #  1.000
    ]
)
""")

set_meas("43-0.0004rad-2coupling-7_1_winding-1e10-drive", """
Piecewise(
    [
        [ -16501,   4203 ],  # -2.000, M1=14319
        [ -16325,   4200 ],  # -1.000, M1=14299
        [   1164,   4497 ],  #  0.000, M1=13190
        [  16471,   4871 ],  #  1.000, M1=11807
        [  16605,   4877 ],  #  2.000, M1=11828
    ]
)
""")

set_meas("43-0.0004rad-2coupling-7_1_winding-2e10-drive", """
Piecewise(
    [
        [ -16668,  -7942 ],  # -2.000, M1=12893
        [ -16532,  -7937 ],  # -1.000, M1=12820
        [ -16350,  -7936 ],  # -0.500, M1=12801
        [  -8796,  -7712 ],  # -0.300, M1=11704
        [  -4996,  -7592 ],  # -0.250, M1=11218
        [  -1907,  -7497 ],  # -0.200, M1=10821
        [     35,  -7432 ],  # -0.150, M1=10556
        [   1341,  -7388 ],  # -0.100, M1=10375
        [   2408,  -7348 ],  # -0.050, M1=10213
        [   3386,  -7305 ],  #  0.000, M1=10047
        [   4351,  -7257 ],  #  0.050, M1= 9865
        [   5386,  -7214 ],  #  0.100, M1= 9698
        [   6513,  -7176 ],  #  0.150, M1= 9548
        [   7732,  -7140 ],  #  0.200, M1= 9405
        [   9157,  -7101 ],  #  0.250, M1= 9247
        [  11134,  -7049 ],  #  0.300, M1= 9023
        [  16529,  -6901 ],  #  0.500, M1= 8402
        [  16653,  -6902 ],  #  1.000, M1= 8421
        [  16760,  -6900 ],  #  2.000, M1= 8470
    ]
)
""")

set_meas("43-0.0004rad-3coupling-5_1_winding-2e9-drive", """
Piecewise(
    [
        [   -362,    570 ],  # -1.000, M1=  380
        [      0,    565 ],  #  0.000, M1=  371
        [    358,    612 ],  #  1.000, M1=  401
        [  11296,    801 ],  #  2.000, M1=  912
    ]
)
""")

set_meas("43-0.0004rad-3coupling-5_1_winding-3e9-drive", """
Piecewise(
    [
        [ -16518,   6297 ],  # -2.000, M1= 4568
        [  -4652,   7064 ],  # -1.000, M1= 4598
        [   -172,   7240 ],  # -0.500, M1= 4591
        [    -45,   7267 ],  # -0.300, M1= 4587
        [    -27,   7271 ],  # -0.250, M1= 4586
        [    -14,   7274 ],  # -0.200, M1= 4585
        [     -5,   7276 ],  # -0.150, M1= 4585
        [      0,   7277 ],  # -0.100, M1= 4585
        [      1,   7277 ],  # -0.050, M1= 4586
        [      5,   7281 ],  #  0.150, M1= 4587
        [     16,   7285 ],  #  0.200, M1= 4587
        [     35,   7293 ],  #  0.250, M1= 4586
        [     62,   7303 ],  #  0.300, M1= 4585
        [    271,   7382 ],  #  0.500, M1= 4577
        [   4683,   7975 ],  #  1.000, M1= 4633
        [  16522,   9525 ],  #  2.000, M1= 6758
    ]
)
""")

set_meas("43-0.0004rad-3coupling-5_1_winding-5e9-drive", """
Piecewise(
    [
        [ -16645,   5703 ],  # -2.000, M1=15028
        [ -16040,   5792 ],  # -1.000, M1=14902
        [  -1445,   6076 ],  # -0.500, M1=14012
        [    -89,   6124 ],  # -0.300, M1=13868
        [     13,   6128 ],  # -0.250, M1=13856
        [    103,   6132 ],  # -0.200, M1=13844
        [    182,   6136 ],  # -0.150, M1=13834
        [    252,   6139 ],  # -0.100, M1=13826
        [    303,   6142 ],  # -0.050, M1=13819
        [    351,   6144 ],  #  0.000, M1=13812
        [    484,   6149 ],  #  0.050, M1=13796
        [    712,   6158 ],  #  0.100, M1=13769
        [   1024,   6171 ],  #  0.150, M1=13733
        [   1399,   6186 ],  #  0.200, M1=13690
        [   1823,   6202 ],  #  0.250, M1=13643
        [   2280,   6220 ],  #  0.300, M1=13592
        [   4334,   6297 ],  #  0.500, M1=13372
        [  16046,   6691 ],  #  1.000, M1=12186
        [  16666,   6738 ],  #  2.000, M1=12153
    ]
)
""")

set_meas("43-0.0004rad-3coupling-5_1_winding-1e10-drive", """
Piecewise(
    [
        [ -16760,  -3662 ],  # -2.000, M1=13584
        [ -16662,  -3712 ],  # -1.000, M1=13565
        [ -16048,  -3716 ],  # -0.500, M1=13514
        [  -4607,  -3242 ],  # -0.300, M1=11797
        [  -1438,  -3105 ],  # -0.250, M1=11305
        [    862,  -3004 ],  # -0.200, M1=10948
        [   2005,  -2956 ],  # -0.150, M1=10775
        [   2751,  -2924 ],  # -0.100, M1=10663
        [   3422,  -2895 ],  # -0.050, M1=10561
        [   3812,  -2879 ],  # -0.020, M1=10502
        [   3940,  -2873 ],  # -0.010, M1=10483
        [   4069,  -2868 ],  #  0.000, M1=10463
        [   4199,  -2862 ],  #  0.010, M1=10443
        [   4333,  -2857 ],  #  0.020, M1=10422
        [   4469,  -2851 ],  #  0.030, M1=10401
        [   4610,  -2844 ],  #  0.040, M1=10380
        [   4756,  -2839 ],  #  0.050, M1=10358
        [   5530,  -2806 ],  #  0.100, M1=10241
        [   6348,  -2773 ],  #  0.150, M1=10120
        [   7199,  -2738 ],  #  0.200, M1= 9997
        [   8104,  -2701 ],  #  0.250, M1= 9867
        [   9110,  -2662 ],  #  0.300, M1= 9726
        [  16015,  -2413 ],  #  0.500, M1= 8846
        [  16684,  -2401 ],  #  1.000, M1= 8804
        [  16776,  -2409 ],  #  2.000, M1= 8836
    ]
)
""")

set_meas("43-0.0004rad-3coupling-5_1_winding-2e10-drive", """
Piecewise(
    [
        [ -16832, -15970 ],  # -2.000, M1=12049
        [ -16778, -15971 ],  # -1.000, M1=12015
        [ -16685, -15969 ],  # -0.500, M1=11997
        [ -16537, -15970 ],  # -0.300, M1=11974
        [ -16059, -15965 ],  # -0.250, M1=11901
        [ -11196, -15897 ],  # -0.200, M1=10885
        [  -4720, -15790 ],  # -0.150, M1= 9490
        [    797, -15677 ],  # -0.100, M1= 8349
        [   1601, -15658 ],  # -0.090, M1= 8193
        [   2309, -15643 ],  # -0.080, M1= 8057
        [   2926, -15628 ],  # -0.070, M1= 7941
        [   3468, -15616 ],  # -0.060, M1= 7841
        [   3968, -15603 ],  # -0.050, M1= 7751
        [   4424, -15592 ],  # -0.040, M1= 7670
        [   4848, -15583 ],  # -0.030, M1= 7596
        [   5253, -15572 ],  # -0.020, M1= 7525
        [   5645, -15562 ],  # -0.010, M1= 7458
        [   6024, -15554 ],  #  0.000, M1= 7393
        [   6392, -15544 ],  #  0.010, M1= 7330
        [   6753, -15535 ],  #  0.020, M1= 7269
        [   7107, -15525 ],  #  0.030, M1= 7209
        [   7458, -15515 ],  #  0.040, M1= 7151
        [   7806, -15506 ],  #  0.050, M1= 7093
        [   8151, -15497 ],  #  0.060, M1= 7036
        [   8496, -15489 ],  #  0.070, M1= 6980
        [   8841, -15479 ],  #  0.080, M1= 6923
        [   9190, -15469 ],  #  0.090, M1= 6868
        [   9540, -15458 ],  #  0.100, M1= 6812
        [  11407, -15404 ],  #  0.150, M1= 6508
        [  13924, -15320 ],  #  0.200, M1= 6090
        [  16013, -15247 ],  #  0.250, M1= 5753
        [  16489, -15231 ],  #  0.300, M1= 5689
        [  16700, -15224 ],  #  0.500, M1= 5664
        [  16789, -15223 ],  #  1.000, M1= 5657
        [  16836, -15223 ],  #  2.000, M1= 5668
    ]
)
""")

set_meas("43-0.0004rad-3coupling-5_1_winding-30000001024e0-drive", """
Piecewise(
    [
        [ -16855, -15745 ],  # -2.000, M1=10885
        [ -16818, -15742 ],  # -1.000, M1=10865
        [ -16765, -15743 ],  # -0.500, M1=10870
        [ -16684, -15744 ],  # -0.300, M1=10853
        [ -16638, -15742 ],  # -0.250, M1=10834
        [ -16550, -15742 ],  # -0.200, M1=10834
        [ -14280, -15740 ],  # -0.150, M1=10466
        [  -5019, -15722 ],  # -0.100, M1= 8601
        [   2131, -15645 ],  # -0.050, M1= 7044
        [   6131, -15554 ],  #  0.000, M1= 6150
        [   9158, -15469 ],  #  0.050, M1= 5485
        [  12202, -15376 ],  #  0.100, M1= 4820
        [  15674, -15265 ],  #  0.150, M1= 4090
        [  16459, -15243 ],  #  0.200, M1= 3954
        [  16584, -15239 ],  #  0.250, M1= 3930
        [  16655, -15237 ],  #  0.300, M1= 3917
        [  16766, -15233 ],  #  0.500, M1= 3900
        [  16826, -15232 ],  #  1.000, M1= 3890
        [  16848, -15229 ],  #  2.000, M1= 3887
    ]
)
""")

set_meas("43-0.0004rad-5coupling-3_1_winding-2e9-drive", """
Piecewise(
    [
        [ -16661,   8062 ],  # -2.000, M1= 7350
        [  -7384,   8571 ],  # -1.000, M1= 7355
        [     15,   8710 ],  #  0.000, M1= 7307
        [   7396,   9645 ],  #  1.000, M1= 7479
        [  16661,   9468 ],  #  2.000, M1=10184
    ]
)
""")

set_meas("43-0.0004rad-5coupling-3_1_winding-3e9-drive", """
Piecewise(
    [
        [ -16754,   3880 ],  # -2.000, M1=15073
        [ -16113,   4023 ],  # -1.000, M1=14879
        [   1027,   4728 ],  #  0.000, M1=13015
        [  16108,   5683 ],  #  1.000, M1=10787
        [  16750,   5716 ],  #  2.000, M1=10859
    ]
)
""")

set_meas("43-0.0004rad-5coupling-3_1_winding-5e9-drive", """
Piecewise(
    [
        [ -16811,   -499 ],  # -2.000, M1=13429
        [ -16744,   -546 ],  # -1.000, M1=13419
        [   5584,    685 ],  #  0.000, M1= 9205
        [  16716,   1188 ],  #  1.000, M1= 7537
        [  16801,   1183 ],  #  2.000, M1= 7577
    ]
)
""")

set_meas("43-0.0004rad-5coupling-3_1_winding-8e9-drive", """
Piecewise(
    [
        [ -16848, -10669 ],  # -2.000, M1=12257
        [ -16802, -10679 ],  # -1.000, M1=12248
        [ -16691, -10678 ],  # -0.500, M1=12243
        [ -11982, -10359 ],  # -0.300, M1=11128
        [  -7374, -10013 ],  # -0.250, M1= 9922
        [  -2727,  -9654 ],  # -0.200, M1= 8688
        [   1840,  -9307 ],  # -0.150, M1= 7503
        [   5760,  -9021 ],  # -0.100, M1= 6532
        [   7616,  -8890 ],  # -0.050, M1= 6090
        [   8766,  -8811 ],  #  0.000, M1= 5826
        [   9753,  -8742 ],  #  0.050, M1= 5598
        [  10718,  -8675 ],  #  0.100, M1= 5375
        [  11693,  -8607 ],  #  0.150, M1= 5154
        [  12682,  -8538 ],  #  0.200, M1= 4933
        [  13713,  -8467 ],  #  0.250, M1= 4703
        [  14679,  -8401 ],  #  0.300, M1= 4492
        [  16540,  -8290 ],  #  0.500, M1= 4147
        [  16770,  -8284 ],  #  1.000, M1= 4141
        [  16826,  -8280 ],  #  2.000, M1= 4153
    ]
)
""")

set_meas("43-0.0004rad-5coupling-3_1_winding-1e10-drive", """
Piecewise(
    [
        [ -16861, -15654 ],  # -2.000, M1=11767
        [ -16824, -15656 ],  # -1.000, M1=11756
        [   9512, -13473 ],  #  0.000, M1= 4787
        [  16791, -12547 ],  #  1.000, M1= 3045
        [  16839, -12547 ],  #  2.000, M1= 3056
    ]
)
""")

set_meas("43-0.0004rad-5coupling-3_1_winding-149999992832e0-drive", """
Piecewise(
    [
        [ -17018,  -3558 ],  # -2.000, M1= 4805
        [ -16995,  -3535 ],  # -1.000, M1= 4798
        [   4265,  -3990 ],  #  0.000, M1=  795
        [  16771,  -4213 ],  #  1.000, M1=-1158
        [  16801,  -4217 ],  #  2.000, M1=-1195
    ]
)
""")

set_meas("43-0.0004rad-5coupling-3_1_winding-2e10-drive", """
Piecewise(
    [
        [ -16879, -16390 ],  # -1.000, M1=10687
        [   9727, -15280 ],  #  0.000, M1= 2570
        [  16837, -14284 ],  #  1.000, M1=  635
    ]
)
""")

set_meas("43-0.0004rad-5coupling-3_1_winding-4e10-drive", """
Piecewise(
    [
        [ -16945, -13758 ],  # -2.000, M1= 8911
        [ -16930, -13762 ],  # -1.000, M1= 8906
        [   8866, -14925 ],  #  0.000, M1= 1780
        [  16596, -14097 ],  #  0.300, M1= -431
        [  16841, -14067 ],  #  1.000, M1= -510
        [  16855, -14050 ],  #  2.000, M1= -545
    ]
)
""")

set_meas("43-0.00059999997rad-7coupling-3_1_winding-3e9-drive", """
Piecewise(
    [
        [ -16766,   4631 ],  # -2.000, M1=15398
        [ -15848,   4681 ],  # -1.000, M1=15222
        [   -492,   5288 ],  # -0.500, M1=13568
        [     12,   5319 ],  # -0.300, M1=13489
        [     71,   5323 ],  # -0.250, M1=13479
        [    126,   5327 ],  # -0.200, M1=13469
        [    174,   5331 ],  # -0.150, M1=13460
        [    220,   5335 ],  # -0.100, M1=13451
        [    297,   5340 ],  # -0.050, M1=13438
        [    472,   5351 ],  #  0.000, M1=13412
        [    734,   5368 ],  #  0.050, M1=13373
        [   1061,   5389 ],  #  0.100, M1=13325
        [   1445,   5413 ],  #  0.150, M1=13268
        [   1882,   5441 ],  #  0.200, M1=13204
        [   2363,   5471 ],  #  0.250, M1=13134
        [   2885,   5504 ],  #  0.300, M1=13058
        [   5324,   5662 ],  #  0.500, M1=12699
        [  15868,   6306 ],  #  1.000, M1=11234
        [  16768,   6393 ],  #  2.000, M1=11290
    ]
)
""")

set_meas("43-0.00059999997rad-7coupling-3_1_winding-5e9-drive", """
Piecewise(
    [
        [ -16823,    394 ],  # -2.000, M1=13717
        [ -16750,    340 ],  # -1.000, M1=13708
        [   5257,   1485 ],  #  0.000, M1= 9856
        [  16738,   2003 ],  #  1.000, M1= 8168
        [  16818,   1978 ],  #  2.000, M1= 8227
    ]
)
""")

set_meas("43-0.00059999997rad-7coupling-3_1_winding-1e10-drive", """
Piecewise(
    [
        [ -16863, -14903 ],  # -2.000, M1=11784
        [ -16834, -14911 ],  # -1.000, M1=11771
        [   9322, -12390 ],  #  0.000, M1= 5092
        [  16817, -11570 ],  #  1.000, M1= 3347
        [  16847, -11558 ],  #  2.000, M1= 3352
    ]
)
""")

set_meas("43-0.00059999997rad-7coupling-3_1_winding-2e10-drive", """
Piecewise(
    [
        [ -16899, -16098 ],  # -2.000, M1=10402
        [ -16874, -16099 ],  # -1.000, M1=10381
        [   8624, -14621 ],  #  0.000, M1= 2849
        [  16849, -13622 ],  #  1.000, M1=  559
        [  16867, -13615 ],  #  2.000, M1=  545
    ]
)
""")

set_meas("43-0.00059999997rad-4coupling-5_1_winding-1e10-drive", """
Piecewise(
    [
        [ -16755,  -1427 ],  # -2.000, M1=14094
        [ -16699,  -1435 ],  # -1.000, M1=14064
        [   3637,   -627 ],  #  0.000, M1=11242
        [  16714,   -168 ],  #  1.000, M1= 9706
        [  16778,   -183 ],  #  2.000, M1= 9805
    ]
)
""")

set_meas("43-0.00059999997rad-4coupling-5_1_winding-2e10-drive", """
Piecewise(
    [
        [ -16821, -15570 ],  # -2.000, M1=12422
        [ -16771, -15569 ],  # -1.000, M1=12390
        [   5377, -14463 ],  #  0.000, M1= 7930
        [  16802, -13682 ],  #  1.000, M1= 6052
        [  16851, -13693 ],  #  2.000, M1= 6091
    ]
)
""")

set_meas("43-0.00059999997rad-4coupling-5_1_winding-30000001024e0-drive", """
Piecewise(
    [
        [ -16857, -15327 ],  # -2.000, M1=11068
        [ -16807, -15323 ],  # -1.000, M1=11024
        [   5536, -14712 ],  #  0.000, M1= 6705
        [  16840, -14200 ],  #  1.000, M1= 4505
        [  16860, -14200 ],  #  2.000, M1= 4510
    ]
)
""")

set_meas("46-0.0004rad-5coupling-3_1_winding-5e9-drive", """
Piecewise(
    [
        [ -16747,   -259 ],  # -1.000, M1=13716
        [  16702,  -1813 ],  #  1.000, M1= 8145
    ]
)
""")

set_meas("46-0.0004rad-5coupling-3_1_winding-8e9-drive", """
Piecewise(
    [
        [ -16810,   9403 ],  # -1.000, M1=12599
        [  16766,   7366 ],  #  1.000, M1= 5175
    ]
)
""")

set_meas("46-0.0004rad-5coupling-3_1_winding-1e10-drive", """
Piecewise(
    [
        [ -16834,  15248 ],  # -1.000, M1=12135
        [  16776,  12945 ],  #  1.000, M1= 4207
    ]
)
""")

set_meas("46-0.0004rad-5coupling-3_1_winding-2e10-drive", """
Piecewise(
    [
        [ -16884,  16555 ],  # -1.000, M1=11155
        [  16820,  16003 ],  #  1.000, M1= 2124
    ]
)
""")

set_meas("46-0.0004rad-5coupling-3_1_winding-4e10-drive", """
Piecewise(
    [
        [ -16938,  14685 ],  # -1.000, M1= 9414
        [  16829,  14366 ],  #  1.000, M1= 1070
    ]
)
""")

set_meas("46-0.00059999997rad-7coupling-3_1_winding-5e9-drive", """
Piecewise(
    [
        [ -16751,  -1189 ],  # -1.000, M1=13954
        [  16733,  -2656 ],  #  1.000, M1= 8726
    ]
)
""")

set_meas("46-0.00059999997rad-7coupling-3_1_winding-8e9-drive", """
Piecewise(
    [
        [ -16818,   7615 ],  # -1.000, M1=12663
        [  16798,   5659 ],  #  1.000, M1= 5518
    ]
)
""")

set_meas("46-0.00059999997rad-7coupling-3_1_winding-1e10-drive", """
Piecewise(
    [
        [ -16841,  13439 ],  # -1.000, M1=12161
        [  16808,  11127 ],  #  1.000, M1= 4429
    ]
)
""")

set_meas("46-0.00059999997rad-7coupling-3_1_winding-2e10-drive", """
Piecewise(
    [
        [ -16879,  16505 ],  # -1.000, M1=10842
        [  16840,  15033 ],  #  1.000, M1= 2034
    ]
)
""")

set_meas("46-0.00059999997rad-7coupling-3_1_winding-30000001024e0-drive", """
Piecewise(
    [
        [ -16904,  14734 ],  # -1.000, M1= 9857
        [ -16844,  14733 ],  # -0.200, M1= 9836
        [   7423,  13889 ],  #  0.000, M1= 3718
        [  16063,  12993 ],  #  0.100, M1= 1582
        [  16460,  12949 ],  #  0.200, M1= 1486
        [  16844,  12906 ],  #  1.000, M1= 1387
    ]
)
""")

set_meas("46-0.00059999997rad-4coupling-5_1_winding-2e10-drive", """
Piecewise(
    [
        [ -16778,  15097 ],  # -1.000, M1=12485
        [  16802,  12503 ],  #  1.000, M1= 6611
    ]
)
""")

set_meas("46-0.0008rad-11coupling-3_1_winding-2e10-drive", """
Piecewise(
    [
        [ -17000,  16570 ],  # -1.000, M1=10636
        [  16889,  14758 ],  #  1.000, M1= 1867
    ]
)
""")

set_meas("46-0.0008rad-7coupling-5_1_winding-30000001024e0-drive", """
Piecewise(
    [
        [ -16961,  14970 ],  # -1.000, M1=11300
        [  16917,  15104 ],  #  1.000, M1= 4724
    ]
)
""")

set_meas("46-float-0.0004rad-5coupling-3_1_winding-2e10-drive", """
Piecewise(
    [
        [ -16877,  16626 ],  # -1.000, M1= 8560
        [  16816,  16640 ],  #  1.000, M1=-6641
    ]
)
""")

set_meas("46-float-0.0004rad-5coupling-3_1_winding-4e10-drive", """
Piecewise(
    [
        [ -16931,  15314 ],  # -1.000, M1= 6216
        [ -16882,  15295 ],  # -0.350, M1= 6103
        [ -16848,  15295 ],  # -0.200, M1= 6095
        [ -16719,  15302 ],  # -0.100, M1= 6059
        [   9914,  15423 ],  #  0.000, M1=-5270
        [  16114,  15344 ],  #  0.100, M1=-7671
        [  16419,  15335 ],  #  0.200, M1=-7757
        [  16608,  15336 ],  #  0.350, M1=-7777
        [  16831,  15382 ],  #  1.000, M1=-7595
    ]
)
""")

set_meas("46-float-0.00059999997rad-7coupling-3_1_winding-30000001024e0-drive", """
Piecewise(
    [
        [ -16892,  15421 ],  # -1.000, M1= 7146
        [ -16803,  15419 ],  # -0.200, M1= 7091
        [   9678,  15665 ],  #  0.000, M1=-4257
        [  15797,  15652 ],  #  0.100, M1=-6726
        [  16423,  15654 ],  #  0.200, M1=-6947
        [  16838,  15733 ],  #  1.000, M1=-6481
    ]
)
""")

set_meas("47-0.0004rad-5coupling-3_1_winding-5e9-drive", """
Piecewise(
    [
        [ -16759,   1164 ],  # -1.000, M1=-13497
        [  16712,   2756 ],  #  1.000, M1=-7528
    ]
)
""")

set_meas("47-0.0004rad-5coupling-3_1_winding-8e9-drive", """
Piecewise(
    [
        [ -16828,  -8107 ],  # -1.000, M1=-12624
        [  16772,  -5884 ],  #  1.000, M1=-4215
    ]
)
""")

set_meas("47-0.0004rad-5coupling-3_1_winding-1e10-drive", """
Piecewise(
    [
        [ -16854, -14076 ],  # -1.000, M1=-11998
        [ -16545, -14068 ],  # -0.350, M1=-11961
        [  -7375, -13452 ],  # -0.200, M1=-9609
        [   4140, -12512 ],  # -0.100, M1=-6400
        [   9313, -12046 ],  #  0.000, M1=-5027
        [  11841, -11810 ],  #  0.100, M1=-4377
        [  14247, -11582 ],  #  0.200, M1=-3766
        [  16258, -11398 ],  #  0.350, M1=-3299
        [  16786, -11358 ],  #  1.000, M1=-3214
    ]
)
""")

set_meas("47-0.0004rad-5coupling-3_1_winding-2e10-drive", """
Piecewise(
    [
        [ -16920, -16586 ],  # -1.000, M1=-10963
        [  16827, -15737 ],  #  1.000, M1=-1168
    ]
)
""")

set_meas("47-0.0004rad-5coupling-3_1_winding-4e10-drive", """
Piecewise(
    [
        [ -16987, -14442 ],  # -1.000, M1=-8976
        [  16829, -13443 ],  #  1.000, M1=  421
    ]
)
""")

set_meas("47-0.00059999997rad-7coupling-3_1_winding-5e9-drive", """
Piecewise(
    [
        [ -16763,   2041 ],  # -1.000, M1=-13759
        [  16734,   3545 ],  #  1.000, M1=-8164
    ]
)
""")

set_meas("47-0.00059999997rad-7coupling-3_1_winding-8e9-drive", """
Piecewise(
    [
        [ -16829,  -6479 ],  # -1.000, M1=-12766
        [ -14123,  -6313 ],  # -0.350, M1=-12124
        [  -1242,  -5418 ],  # -0.200, M1=-8708
        [   5339,  -4994 ],  # -0.100, M1=-7104
        [   8141,  -4824 ],  #  0.000, M1=-6461
        [  10321,  -4693 ],  #  0.100, M1=-5972
        [  12538,  -4563 ],  #  0.200, M1=-5485
        [  15644,  -4382 ],  #  0.350, M1=-4806
        [  16799,  -4325 ],  #  1.000, M1=-4619
    ]
)
""")

set_meas("47-0.00059999997rad-7coupling-3_1_winding-1e10-drive", """
Piecewise(
    [
        [ -16852, -12074 ],  # -1.000, M1=-12070
        [ -16512, -12067 ],  # -0.350, M1=-12033
        [  -5580, -11331 ],  # -0.200, M1=-9189
        [   4228, -10606 ],  # -0.100, M1=-6584
        [   8721, -10256 ],  #  0.000, M1=-5425
        [  11561, -10031 ],  #  0.100, M1=-4723
        [  14334,  -9799 ],  #  0.200, M1=-4037
        [  16256,  -9638 ],  #  0.350, M1=-3580
        [  16813,  -9600 ],  #  1.000, M1=-3486
    ]
)
""")

set_meas("47-0.00059999997rad-7coupling-3_1_winding-2e10-drive", """
Piecewise(
    [
        [ -16902, -16514 ],  # -1.000, M1=-10719
        [ -16701, -16513 ],  # -0.200, M1=-10675
        [  -5648, -16369 ],  # -0.100, M1=-7360
        [   8575, -15470 ],  #  0.000, M1=-3263
        [  15217, -14725 ],  #  0.100, M1=-1597
        [  16344, -14580 ],  #  0.200, M1=-1314
        [  16843, -14510 ],  #  1.000, M1=-1188
    ]
)
""")

set_meas("47-0.00059999997rad-7coupling-3_1_winding-30000001024e0-drive", """
Piecewise(
    [
        [ -16932, -14525 ],  # -1.000, M1=-9579
        [  16845, -12288 ],  #  1.000, M1=   63
    ]
)
""")

set_meas("47-0.00059999997rad-4coupling-5_1_winding-2e10-drive", """
Piecewise(
    [
        [ -16785, -15047 ],  # -1.000, M1=-12562
        [  16793, -12591 ],  #  1.000, M1=-6601
    ]
)
""")

set_meas("48-0.0004rad-500ctl_cond-5000coupling_cond-1000ps-50ps-9ctl-9coupling-3_1_winding-1e10-drive", """
Piecewise(
    [
        [ -17210,   1376 ],  # -1.000
        [ -17061,   1548 ],  # -0.350
        [ -16998,   1629 ],  # -0.250
        [ -16783,   1850 ],  # -0.200
        [ -16289,   2360 ],  # -0.170
        [ -15902,   2784 ],  # -0.160
        [ -15157,   3634 ],  # -0.150
        [ -13331,   5802 ],  # -0.140
        [ -11262,   8235 ],  # -0.130
        [  -9165,  10620 ],  # -0.120
        [  -7048,  12924 ],  # -0.110
        [  -4926,  14787 ],  # -0.100
        [  -2804,  15888 ],  # -0.090
        [   -683,  16220 ],  # -0.080
        [   1432,  16422 ],  # -0.070
        [   5604,  16743 ],  # -0.050
        [  10182,  16778 ],  # -0.020
        [  11224,  16783 ],  #  0.000
        [  12523,  16787 ],  #  0.050
        [  12908,  16788 ],  #  0.070
        [  13476,  16790 ],  #  0.100
        [  13889,  16790 ],  #  0.120
        [  16500,  16793 ],  #  0.200
        [  16816,  16793 ],  #  0.350
        [  16931,  16791 ],  #  1.000
    ]
)
""")

set_meas("48-0.0004rad-500ctl_cond-5000coupling_cond-1000ps-50ps-9ctl-9coupling-3_1_winding-2e10-drive", """
Piecewise(
    [
        [ -17335,   3072 ],  # -1.000
        [ -17206,   3217 ],  # -0.350
        [ -17186,   3237 ],  # -0.310
        [ -17180,   3242 ],  # -0.300
        [ -17175,   3249 ],  # -0.290
        [ -17169,   3254 ],  # -0.280
        [ -17163,   3260 ],  # -0.270
        [ -17156,   3267 ],  # -0.260
        [ -17149,   3275 ],  # -0.250
        [ -17114,   3315 ],  # -0.200
        [ -17077,   3365 ],  # -0.150
        [ -16811,   3604 ],  # -0.100
        [ -16531,   3841 ],  # -0.090
        [ -15906,   4409 ],  # -0.080
        [ -13320,   6893 ],  # -0.070
        [  -9157,  10822 ],  # -0.060
        [  -4919,  14379 ],  # -0.050
        [   -684,  16552 ],  # -0.040
        [   3518,  16650 ],  # -0.030
        [   7550,  16593 ],  # -0.020
        [   9830,  16591 ],  # -0.010
        [  11222,  16598 ],  #  0.000
        [  13985,  16644 ],  #  0.050
        [  16990,  16708 ],  #  1.000
    ]
)
""")

set_meas("48-0.0004rad-200ctl_cond-10000coupling_cond-1000ps-50ps-9ctl-9coupling-3_1_winding-1e10-drive", """
Piecewise(
    [
        [ -18665,   2331 ],  # -1.000
        [ -18367,   2808 ],  # -0.350
        [ -18262,   2964 ],  # -0.300
        [ -18225,   3017 ],  # -0.290
        [ -18177,   3086 ],  # -0.280
        [ -18103,   3188 ],  # -0.270
        [ -17962,   3383 ],  # -0.260
        [ -17758,   3673 ],  # -0.250
        [ -17482,   4070 ],  # -0.240
        [ -17147,   4571 ],  # -0.230
        [ -16723,   5206 ],  # -0.220
        [ -16212,   5981 ],  # -0.210
        [ -15523,   7019 ],  # -0.200
        [ -14381,   8713 ],  # -0.190
        [ -12852,  10876 ],  # -0.180
        [ -11271,  12966 ],  # -0.170
        [  -9673,  14779 ],  # -0.160
        [  -8061,  16111 ],  # -0.150
        [  -6445,  16527 ],  # -0.140
        [  -4827,  16712 ],  # -0.130
        [  -3209,  16747 ],  # -0.120
        [  -1592,  16771 ],  # -0.110
        [     23,  16792 ],  # -0.100
        [   7948,  16834 ],  # -0.050
        [  11334,  16838 ],  #  0.000
        [  12855,  16839 ],  #  0.100
        [  17046,  16842 ],  #  0.350
        [  17352,  16841 ],  #  1.000
    ]
)
""")

set_meas("48-0.0004rad-500ctl_cond-10000coupling_cond-1000ps-50ps-9ctl-9coupling-3_1_winding-3e9-drive", """
Piecewise(
    [
        [ -18331,  -6085 ],  # -1.000
        [  17160,   6840 ],  #  1.000
    ]
)
""")

set_meas("48-0.0004rad-500ctl_cond-10000coupling_cond-1000ps-50ps-9ctl-9coupling-3_1_winding-5e9-drive", """
Piecewise(
    [
        [ -18553,  -1270 ],  # -1.000
        [ -16802,    913 ],  # -0.350
        [ -16502,   1314 ],  # -0.340
        [ -16155,   1781 ],  # -0.330
        [ -15738,   2335 ],  # -0.320
        [ -15218,   3018 ],  # -0.310
        [ -14402,   4035 ],  # -0.300
        [ -13438,   5163 ],  # -0.290
        [ -12446,   6280 ],  # -0.280
        [ -11448,   7404 ],  # -0.270
        [ -10441,   8484 ],  # -0.260
        [  -9426,   9490 ],  # -0.250
        [  -8409,  10396 ],  # -0.240
        [  -7390,  11313 ],  # -0.230
        [  -6370,  12220 ],  # -0.220
        [  -5350,  12984 ],  # -0.210
        [  -4329,  13677 ],  # -0.200
        [  -3308,  14279 ],  # -0.190
        [  -2286,  14744 ],  # -0.180
        [  -1266,  15074 ],  # -0.170
        [   -245,  15280 ],  # -0.160
        [    775,  15429 ],  # -0.150
        [   1793,  15555 ],  # -0.140
        [   2810,  15666 ],  # -0.130
        [   3824,  15748 ],  # -0.120
        [   4834,  15816 ],  # -0.110
        [   5838,  15881 ],  # -0.100
        [   6835,  15943 ],  # -0.090
        [   7813,  16000 ],  # -0.080
        [   8767,  16044 ],  # -0.070
        [   9589,  16078 ],  # -0.060
        [  10211,  16103 ],  # -0.050
        [  10689,  16120 ],  # -0.040
        [  10887,  16127 ],  # -0.030
        [  11026,  16132 ],  # -0.020
        [  11140,  16135 ],  # -0.010
        [  11237,  16138 ],  #  0.000
        [  11323,  16141 ],  #  0.010
        [  11403,  16144 ],  #  0.020
        [  11478,  16146 ],  #  0.030
        [  11550,  16148 ],  #  0.040
        [  11620,  16151 ],  #  0.050
        [  11758,  16154 ],  #  0.070
        [  11959,  16160 ],  #  0.100
        [  12092,  16164 ],  #  0.120
        [  12621,  16179 ],  #  0.200
        [  15897,  16276 ],  #  0.350
        [  17271,  16313 ],  #  1.000
    ]
)
""")

set_meas("48-0.0004rad-500ctl_cond-10000coupling_cond-1000ps-50ps-9ctl-9coupling-3_1_winding-1e10-drive", """
Piecewise(
    [
        [ -18828,   1410 ],  # -1.000
        [ -18549,   1845 ],  # -0.350
        [ -18403,   2082 ],  # -0.250
        [ -18372,   2127 ],  # -0.240
        [ -18326,   2191 ],  # -0.230
        [ -18265,   2280 ],  # -0.220
        [ -18149,   2439 ],  # -0.210
        [ -17921,   2757 ],  # -0.200
        [ -17580,   3240 ],  # -0.190
        [ -17135,   3888 ],  # -0.180
        [ -16541,   4753 ],  # -0.170
        [ -15735,   5927 ],  # -0.160
        [ -14337,   7931 ],  # -0.150
        [ -12397,  10649 ],  # -0.140
        [ -10392,  13366 ],  # -0.130
        [  -8360,  15457 ],  # -0.120
        [  -6319,  16168 ],  # -0.110
        [  -4275,  16419 ],  # -0.100
        [  -2232,  16650 ],  # -0.090
        [   -192,  16765 ],  # -0.080
        [   1843,  16788 ],  # -0.070
        [   3866,  16805 ],  # -0.060
        [   5867,  16812 ],  # -0.050
        [   7818,  16815 ],  # -0.040
        [   9607,  16817 ],  # -0.030
        [  10547,  16819 ],  # -0.020
        [  10983,  16820 ],  # -0.010
        [  11309,  16820 ],  #  0.000
        [  11573,  16820 ],  #  0.010
        [  11784,  16820 ],  #  0.020
        [  11968,  16821 ],  #  0.030
        [  12137,  16821 ],  #  0.040
        [  12296,  16821 ],  #  0.050
        [  12594,  16822 ],  #  0.070
        [  13053,  16822 ],  #  0.100
        [  13407,  16823 ],  #  0.120
        [  16668,  16825 ],  #  0.200
        [  17106,  16825 ],  #  0.350
        [  17405,  16824 ],  #  1.000
    ]
)
""")

set_meas("48-0.0004rad-500ctl_cond-10000coupling_cond-1000ps-50ps-9ctl-9coupling-3_1_winding-2e10-drive", """
Piecewise(
    [
        [ -19108,   2815 ],  # -1.000
        [ -18809,   3240 ],  # -0.300
        [ -18753,   3318 ],  # -0.250
        [ -18686,   3414 ],  # -0.200
        [ -18637,   3489 ],  # -0.170
        [ -18597,   3554 ],  # -0.150
        [ -18470,   3726 ],  # -0.120
        [ -18362,   3859 ],  # -0.110
        [ -18008,   4281 ],  # -0.100
        [ -17204,   5257 ],  # -0.090
        [ -15766,   6970 ],  # -0.080
        [ -12417,  10676 ],  # -0.070
        [  -8389,  14672 ],  # -0.060
        [  -4320,  16641 ],  # -0.050
        [   7649,  16852 ],  # -0.020
        [  11247,  16858 ],  #  0.000
        [  13466,  16861 ],  #  0.050
        [  14225,  16862 ],  #  0.070
        [  16856,  16862 ],  #  0.120
        [  17516,  16859 ],  #  1.000
    ]
)
""")

set_meas("48-0.0004rad-500ctl_cond-20000coupling_cond-2000ps-100ps-9ctl-9coupling-3_1_winding-5e9-drive", """
Piecewise(
    [
        [ -18385,    377 ],  # -1.000
        [ -17714,   1264 ],  # -0.400
        [ -16717,   2796 ],  # -0.350
        [ -16429,   3229 ],  # -0.340
        [ -16098,   3715 ],  # -0.330
        [ -15695,   4277 ],  # -0.320
        [ -15192,   4953 ],  # -0.310
        [ -14391,   5983 ],  # -0.300
        [ -13426,   7231 ],  # -0.290
        [ -12434,   8475 ],  # -0.280
        [ -11434,   9620 ],  # -0.270
        [ -10426,  10668 ],  # -0.260
        [  -9409,  11790 ],  # -0.250
        [  -8390,  12770 ],  # -0.240
        [  -7369,  13639 ],  # -0.230
        [  -6347,  14376 ],  # -0.220
        [  -4301,  15224 ],  # -0.200
        [  -1233,  15693 ],  # -0.170
        [    811,  15853 ],  # -0.150
        [   3872,  16043 ],  # -0.120
        [   5900,  16128 ],  # -0.100
        [   8860,  16230 ],  # -0.070
        [  10290,  16275 ],  # -0.050
        [  11468,  16310 ],  # -0.020
        [  11672,  16316 ],  #  0.000
        [  11988,  16323 ],  #  0.050
        [  12093,  16326 ],  #  0.070
        [  12246,  16329 ],  #  0.100
        [  12346,  16331 ],  #  0.120
        [  12750,  16340 ],  #  0.200
        [  15902,  16419 ],  #  0.350
        [  17215,  16453 ],  #  1.000
    ]
)
""")

set_meas("48-0.0004rad-500ctl_cond-20000coupling_cond-2000ps-100ps-9ctl-9coupling-3_1_winding-1e10-drive", """
Piecewise(
    [
        [ -18644,   2982 ],  # -1.000
        [ -18484,   3261 ],  # -0.500
        [ -18419,   3377 ],  # -0.400
        [ -18380,   3451 ],  # -0.350
        [ -18362,   3488 ],  # -0.330
        [ -18343,   3528 ],  # -0.310
        [ -18332,   3550 ],  # -0.300
        [ -18319,   3573 ],  # -0.290
        [ -18286,   3631 ],  # -0.270
        [ -18264,   3667 ],  # -0.260
        [ -18239,   3710 ],  # -0.250
        [ -18209,   3760 ],  # -0.240
        [ -18164,   3831 ],  # -0.230
        [ -18103,   3929 ],  # -0.220
        [ -17989,   4104 ],  # -0.210
        [ -17769,   4450 ],  # -0.200
        [ -17442,   4975 ],  # -0.190
        [ -17018,   5680 ],  # -0.180
        [ -16449,   6621 ],  # -0.170
        [ -15674,   7898 ],  # -0.160
        [ -14299,  10123 ],  # -0.150
        [ -12360,  13165 ],  # -0.140
        [ -10352,  15693 ],  # -0.130
        [  -8317,  16251 ],  # -0.120
        [  -6271,  16512 ],  # -0.110
        [  -4221,  16752 ],  # -0.100
        [   1920,  16832 ],  # -0.070
        [   5972,  16841 ],  # -0.050
        [  10908,  16845 ],  # -0.020
        [  11760,  16845 ],  #  0.000
        [  12557,  16845 ],  #  0.050
        [  12782,  16845 ],  #  0.070
        [  13133,  16845 ],  #  0.100
        [  13402,  16846 ],  #  0.120
        [  16662,  16848 ],  #  0.200
        [  17065,  16847 ],  #  0.350
        [  17348,  16846 ],  #  1.000
    ]
)
""")

set_meas("48-0.0004rad-1000ctl_cond-10000coupling_cond-2000ps-100ps-9ctl-9coupling-3_1_winding-3e9-drive", """
Piecewise(
    [
        [ -16999,  -3777 ],  # -1.000
        [ -16742,  -3787 ],  # -0.500
        [ -13768,  -1435 ],  # -0.350
        [  -9559,   1353 ],  # -0.300
        [  -5278,   3430 ],  # -0.250
        [  -4420,   3777 ],  # -0.240
        [  -3562,   4109 ],  # -0.230
        [  -2705,   4423 ],  # -0.220
        [  -1847,   4721 ],  # -0.210
        [   -992,   5010 ],  # -0.200
        [   -139,   5293 ],  # -0.190
        [    711,   5569 ],  # -0.180
        [   1560,   5843 ],  # -0.170
        [   2407,   6113 ],  # -0.160
        [   3249,   6381 ],  # -0.150
        [   4077,   6643 ],  # -0.140
        [   4900,   6904 ],  # -0.130
        [   5719,   7162 ],  # -0.120
        [   6526,   7417 ],  # -0.110
        [   7296,   7660 ],  # -0.100
        [   8044,   7897 ],  # -0.090
        [   8771,   8127 ],  # -0.080
        [   9422,   8335 ],  # -0.070
        [  10024,   8527 ],  # -0.060
        [  10446,   8662 ],  # -0.050
        [  10580,   8704 ],  # -0.040
        [  10674,   8733 ],  # -0.030
        [  10762,   8759 ],  # -0.020
        [  10845,   8784 ],  # -0.010
        [  10926,   8809 ],  #  0.000
        [  11004,   8832 ],  #  0.010
        [  11081,   8855 ],  #  0.020
        [  11156,   8878 ],  #  0.030
        [  11230,   8900 ],  #  0.040
        [  11304,   8922 ],  #  0.050
        [  11449,   8966 ],  #  0.070
        [  11660,   9029 ],  #  0.100
        [  11799,   9071 ],  #  0.120
        [  12332,   9229 ],  #  0.200
        [  13386,   9543 ],  #  0.350
        [  16812,  10556 ],  #  1.000
    ]
)
""")

set_meas("48-0.0004rad-1000ctl_cond-10000coupling_cond-2000ps-100ps-9ctl-9coupling-3_1_winding-5e9-drive", """
Piecewise(
    [
        [ -17090,  -1037 ],  # -1.000
        [ -16920,   -886 ],  # -0.350
        [ -16229,   -202 ],  # -0.250
        [ -15980,     74 ],  # -0.240
        [ -15616,    480 ],  # -0.230
        [ -14998,   1181 ],  # -0.220
        [ -13739,   2608 ],  # -0.210
        [ -12356,   4146 ],  # -0.200
        [ -10956,   5720 ],  # -0.190
        [  -9537,   7231 ],  # -0.180
        [  -8109,   8659 ],  # -0.170
        [  -6678,  10153 ],  # -0.160
        [  -5245,  11575 ],  # -0.150
        [  -3811,  12962 ],  # -0.140
        [  -2377,  14221 ],  # -0.130
        [   -942,  15144 ],  # -0.120
        [    492,  15592 ],  # -0.110
        [   1925,  15850 ],  # -0.100
        [   3354,  16030 ],  # -0.090
        [   4775,  16145 ],  # -0.080
        [   6187,  16239 ],  # -0.070
        [   7566,  16321 ],  # -0.060
        [   8917,  16391 ],  # -0.050
        [  10193,  16454 ],  # -0.040
        [  11267,  16505 ],  # -0.030
        [  11557,  16519 ],  # -0.020
        [  11712,  16525 ],  # -0.010
        [  11831,  16530 ],  #  0.000
        [  11931,  16534 ],  #  0.010
        [  12020,  16537 ],  #  0.020
        [  12104,  16541 ],  #  0.030
        [  12186,  16544 ],  #  0.040
        [  12267,  16547 ],  #  0.050
        [  12425,  16552 ],  #  0.070
        [  12659,  16560 ],  #  0.100
        [  12814,  16565 ],  #  0.120
        [  13516,  16586 ],  #  0.200
        [  16720,  16652 ],  #  0.350
        [  16874,  16650 ],  #  1.000
    ]
)
""")

set_meas("48-0.0004rad-1000ctl_cond-10000coupling_cond-2000ps-100ps-9ctl-9coupling-3_1_winding-1e10-drive", """
Piecewise(
    [
        [ -17206,   2148 ],  # -1.000
        [ -17066,   2328 ],  # -0.350
        [ -17024,   2390 ],  # -0.250
        [ -16990,   2439 ],  # -0.200
        [ -16976,   2457 ],  # -0.190
        [ -16958,   2478 ],  # -0.180
        [ -16928,   2512 ],  # -0.170
        [ -16880,   2565 ],  # -0.160
        [ -16782,   2671 ],  # -0.150
        [ -16632,   2832 ],  # -0.140
        [ -16401,   3087 ],  # -0.130
        [ -15978,   3579 ],  # -0.120
        [ -14925,   4854 ],  # -0.110
        [ -12303,   8135 ],  # -0.100
        [  -9481,  11557 ],  # -0.090
        [  -6618,  14604 ],  # -0.080
        [  -3747,  15942 ],  # -0.070
        [   1991,  16581 ],  # -0.050
        [   7631,  16791 ],  # -0.030
        [  10231,  16806 ],  # -0.020
        [  11417,  16810 ],  # -0.010
        [  11896,  16811 ],  #  0.000
        [  13064,  16813 ],  #  0.050
        [  13432,  16814 ],  #  0.070
        [  14095,  16814 ],  #  0.100
        [  14946,  16815 ],  #  0.120
        [  16756,  16816 ],  #  0.200
        [  16829,  16817 ],  #  0.350
        [  16938,  16814 ],  #  1.000
    ]
)
""")

set_meas("48-0.0004rad-1000ctl_cond-10000coupling_cond-2000ps-100ps-9ctl-9coupling-3_1_winding-2e10-drive", """
Piecewise(
    [
        [ -17334,   4023 ],  # -1.000
        [ -17188,   4205 ],  # -0.300
        [ -17159,   4240 ],  # -0.250
        [ -17124,   4279 ],  # -0.200
        [ -17116,   4289 ],  # -0.190
        [ -17108,   4300 ],  # -0.180
        [ -17098,   4310 ],  # -0.170
        [ -17090,   4322 ],  # -0.160
        [ -17081,   4333 ],  # -0.150
        [ -17071,   4344 ],  # -0.140
        [ -17062,   4357 ],  # -0.130
        [ -17054,   4372 ],  # -0.120
        [ -17044,   4388 ],  # -0.110
        [ -17025,   4410 ],  # -0.100
        [ -16991,   4445 ],  # -0.090
        [ -16910,   4521 ],  # -0.080
        [ -16651,   4757 ],  # -0.070
        [ -15978,   5402 ],  # -0.060
        [ -15495,   5886 ],  # -0.057
        [ -14912,   6487 ],  # -0.055
        [ -14438,   6980 ],  # -0.054
        [ -13925,   7513 ],  # -0.053
        [ -13398,   8059 ],  # -0.052
        [ -12848,   8628 ],  # -0.051
        [ -12290,   9198 ],  # -0.050
        [ -11729,   9766 ],  # -0.049
        [ -10601,  10885 ],  # -0.047
        [  -9466,  11979 ],  # -0.045
        [  -8327,  13033 ],  # -0.043
        [  -7755,  13544 ],  # -0.042
        [  -7182,  14043 ],  # -0.041
        [  -6608,  14526 ],  # -0.040
        [  -5461,  15432 ],  # -0.038
        [  -3740,  16411 ],  # -0.035
        [   -875,  16724 ],  # -0.030
        [   4807,  16835 ],  # -0.020
        [   9949,  16844 ],  # -0.010
        [  11921,  16847 ],  #  0.000
        [  14540,  16851 ],  #  0.050
        [  15809,  16852 ],  #  0.070
        [  16620,  16852 ],  #  0.120
        [  16986,  16850 ],  #  1.000
    ]
)
""")

set_meas("48-0.0004rad-2000ctl_cond-20000coupling_cond-4000ps-200ps-9ctl-9coupling-3_1_winding-1e9-drive", """
Piecewise(
    [
        [ -16695, -12962 ],  # -1.000
        [    189,  -7932 ],  # -0.350
        [    893,  -7705 ],  # -0.300
        [   1302,  -7574 ],  # -0.250
        [   1728,  -7439 ],  # -0.200
        [   1816,  -7411 ],  # -0.190
        [   1905,  -7383 ],  # -0.180
        [   1996,  -7355 ],  # -0.170
        [   2087,  -7326 ],  # -0.160
        [   2181,  -7297 ],  # -0.150
        [   2276,  -7267 ],  # -0.140
        [   2371,  -7238 ],  # -0.130
        [   2468,  -7207 ],  # -0.120
        [   2566,  -7177 ],  # -0.110
        [   2665,  -7146 ],  # -0.100
        [   2765,  -7114 ],  # -0.090
        [   2867,  -7083 ],  # -0.080
        [   2970,  -7051 ],  # -0.070
        [   3074,  -7018 ],  # -0.060
        [   3179,  -6986 ],  # -0.050
        [   3286,  -6952 ],  # -0.040
        [   3393,  -6919 ],  # -0.030
        [   3502,  -6885 ],  # -0.020
        [   3611,  -6851 ],  # -0.010
        [   3722,  -6817 ],  #  0.000
        [   4293,  -6640 ],  #  0.050
        [   4528,  -6567 ],  #  0.070
        [   4888,  -6456 ],  #  0.100
        [   5132,  -6380 ],  #  0.120
        [   6146,  -6068 ],  #  0.200
        [   8166,  -5448 ],  #  0.350
        [  16600,  -2984 ],  #  1.000
    ]
)
""")

set_meas("48-0.0004rad-2000ctl_cond-20000coupling_cond-4000ps-200ps-9ctl-9coupling-3_1_winding-2e9-drive", """
Piecewise(
    [
        [ -16955,  -3998 ],  # -1.000
        [ -14123,  -1880 ],  # -0.350
        [  -5549,   3367 ],  # -0.250
        [  -1211,   5059 ],  # -0.200
        [   1372,   5916 ],  # -0.170
        [   2229,   6190 ],  # -0.160
        [   3082,   6462 ],  # -0.150
        [   3921,   6727 ],  # -0.140
        [   4753,   6988 ],  # -0.130
        [   5580,   7248 ],  # -0.120
        [   6394,   7502 ],  # -0.110
        [   7168,   7745 ],  # -0.100
        [   7922,   7982 ],  # -0.090
        [   8641,   8208 ],  # -0.080
        [   9288,   8413 ],  # -0.070
        [   9863,   8595 ],  # -0.060
        [  10239,   8714 ],  # -0.050
        [  10344,   8747 ],  # -0.040
        [  10433,   8774 ],  # -0.030
        [  10519,   8800 ],  # -0.020
        [  10603,   8826 ],  # -0.010
        [  10685,   8851 ],  #  0.000
        [  11066,   8967 ],  #  0.050
        [  11208,   9010 ],  #  0.070
        [  11411,   9072 ],  #  0.100
        [  11542,   9111 ],  #  0.120
        [  12021,   9255 ],  #  0.200
        [  13494,   9696 ],  #  0.350
        [  16794,  10663 ],  #  1.000
    ]
)
""")

set_meas("48-0.0004rad-500ctl_cond-40000coupling_cond-2000ps-100ps-9ctl-9coupling-3_1_winding-5e9-drive", """
Piecewise(
    [
        [ -21111,  -4121 ],  # -1.000
        [ -20681,  -3380 ],  # -0.500
        [ -19361,  -1091 ],  # -0.450
        [ -18997,   -468 ],  # -0.440
        [ -18209,    881 ],  # -0.420
        [ -17350,   2421 ],  # -0.400
        [ -16373,   4089 ],  # -0.380
        [ -15169,   5937 ],  # -0.360
        [ -14393,   7054 ],  # -0.350
        [ -13568,   8200 ],  # -0.340
        [ -10183,  12278 ],  # -0.300
        [  -5878,  15315 ],  # -0.250
        [  -1569,  15892 ],  # -0.200
        [   -708,  15955 ],  # -0.190
        [    152,  16015 ],  # -0.180
        [   1013,  16070 ],  # -0.170
        [   1873,  16116 ],  # -0.160
        [   2732,  16157 ],  # -0.150
        [   3590,  16193 ],  # -0.140
        [   4445,  16227 ],  # -0.130
        [   5293,  16260 ],  # -0.120
        [   6125,  16291 ],  # -0.110
        [   6858,  16317 ],  # -0.100
        [   7485,  16339 ],  # -0.090
        [   8086,  16360 ],  # -0.080
        [   8666,  16380 ],  # -0.070
        [   9223,  16400 ],  # -0.060
        [   9385,  16405 ],  # -0.057
        [   9492,  16408 ],  # -0.055
        [   9545,  16410 ],  # -0.054
        [   9598,  16412 ],  # -0.053
        [   9651,  16413 ],  # -0.052
        [   9703,  16415 ],  # -0.051
        [   9755,  16417 ],  # -0.050
        [   9807,  16418 ],  # -0.049
        [   9910,  16422 ],  # -0.047
        [  10012,  16424 ],  # -0.045
        [  10112,  16428 ],  # -0.043
        [  10162,  16429 ],  # -0.042
        [  10211,  16431 ],  # -0.041
        [  10260,  16432 ],  # -0.040
        [  10356,  16435 ],  # -0.038
        [  10498,  16440 ],  # -0.035
        [  10723,  16446 ],  # -0.030
        [  11003,  16455 ],  # -0.020
        [  11120,  16458 ],  # -0.010
        [  11211,  16460 ],  #  0.000
        [  11518,  16469 ],  #  0.050
        [  11606,  16471 ],  #  0.070
        [  11805,  16476 ],  #  0.120
        [  18116,  16679 ],  #  1.000
    ]
)
""")

set_meas("48-0.0004rad-500ctl_cond-40000coupling_cond-4000ps-200ps-9ctl-9coupling-3_1_winding-1e10-drive", """
Piecewise(
    [
        [ -18503,   3956 ],  # -1.000
        [ -18244,   4450 ],  # -0.350
        [ -18105,   4724 ],  # -0.250
        [ -17635,   5520 ],  # -0.200
        [ -16347,   7816 ],  # -0.170
        [ -15588,   9160 ],  # -0.160
        [ -14219,  11518 ],  # -0.150
        [ -12285,  14698 ],  # -0.140
        [ -10278,  16161 ],  # -0.130
        [  -4150,  16834 ],  # -0.100
        [   6054,  16875 ],  # -0.050
        [  12063,  16878 ],  #  0.000
        [  12697,  16879 ],  #  0.050
        [  13136,  16879 ],  #  0.100
        [  16644,  16881 ],  #  0.200
        [  17028,  16882 ],  #  0.350
        [  17304,  16881 ],  #  1.000
    ]
)
""")

set_meas("48-0.0004rad-500ctl_cond-100000coupling_cond-10000ps-500ps-9ctl-9coupling-3_1_winding-5e9-drive", """
Piecewise(
    [
        [ -18009,   2478 ],  # -1.000
        [ -17799,   2787 ],  # -0.500
        [ -17330,   3494 ],  # -0.400
        [ -17188,   3724 ],  # -0.390
        [ -17022,   3998 ],  # -0.380
        [ -16837,   4301 ],  # -0.370
        [ -16630,   4643 ],  # -0.360
        [ -16392,   5027 ],  # -0.350
        [ -16119,   5448 ],  # -0.340
        [ -15804,   5917 ],  # -0.330
        [ -15419,   6482 ],  # -0.320
        [ -14933,   7221 ],  # -0.310
        [ -14121,   8389 ],  # -0.300
        [ -13163,   9633 ],  # -0.290
        [ -12182,  10842 ],  # -0.280
        [ -11191,  12100 ],  # -0.270
        [ -10188,  13226 ],  # -0.260
        [  -9179,  14230 ],  # -0.250
        [  -8166,  14951 ],  # -0.240
        [  -7152,  15383 ],  # -0.230
        [  -6137,  15611 ],  # -0.220
        [  -4104,  15895 ],  # -0.200
        [    -43,  16175 ],  # -0.160
        [   1987,  16260 ],  # -0.140
        [   3002,  16298 ],  # -0.130
        [   4015,  16332 ],  # -0.120
        [  12095,  16557 ],  #  0.000
        [  17086,  16676 ],  #  1.000
    ]
)
""")

set_meas("48-0.0004rad-500ctl_cond-100000coupling_cond-10000ps-500ps-9ctl-9coupling-3_1_winding-1e10-drive", """
Piecewise(
    [
        [ -18257,   4896 ],  # -1.000
        [ -18098,   5209 ],  # -0.500
        [ -18035,   5337 ],  # -0.400
        [ -17996,   5420 ],  # -0.350
        [ -17988,   5441 ],  # -0.340
        [ -17980,   5461 ],  # -0.330
        [ -17970,   5482 ],  # -0.320
        [ -17961,   5505 ],  # -0.310
        [ -17950,   5529 ],  # -0.300
        [ -17937,   5555 ],  # -0.290
        [ -17922,   5584 ],  # -0.280
        [ -17857,   5709 ],  # -0.250
        [ -17368,   6578 ],  # -0.200
        [ -16114,   8952 ],  # -0.170
        [ -15371,  10306 ],  # -0.160
        [ -13994,  12730 ],  # -0.150
        [ -12074,  15831 ],  # -0.140
        [ -10078,  16409 ],  # -0.130
        [  -8057,  16687 ],  # -0.120
        [  -6025,  16905 ],  # -0.110
        [  -3987,  16940 ],  # -0.100
        [   6168,  16967 ],  # -0.050
        [  12301,  16969 ],  #  0.000
        [  17221,  16973 ],  #  1.000
    ]
)
""")

set_meas("48-0.0004rad-5000ctl_cond-50000coupling_cond-10000ps-500ps-9ctl-9coupling-3_1_winding-5e8-drive", """
Piecewise(
    [
        [ -16608, -12761 ],  # -1.000
        [    107,  -7801 ],  # -0.350
        [    923,  -7540 ],  # -0.300
        [   1350,  -7404 ],  # -0.250
        [   1793,  -7265 ],  # -0.200
        [   1975,  -7208 ],  # -0.180
        [   2069,  -7179 ],  # -0.170
        [   2164,  -7149 ],  # -0.160
        [   2260,  -7119 ],  # -0.150
        [   2357,  -7089 ],  # -0.140
        [   2456,  -7058 ],  # -0.130
        [   2556,  -7027 ],  # -0.120
        [   2657,  -6996 ],  # -0.110
        [   2760,  -6964 ],  # -0.100
        [   3073,  -6866 ],  # -0.070
        [   3288,  -6800 ],  # -0.050
        [   3620,  -6698 ],  # -0.020
        [   3847,  -6628 ],  #  0.000
        [   4432,  -6448 ],  #  0.050
        [   4672,  -6373 ],  #  0.070
        [   5041,  -6260 ],  #  0.100
        [   5292,  -6183 ],  #  0.120
        [   6325,  -5867 ],  #  0.200
        [   8379,  -5240 ],  #  0.350
        [  16585,  -2845 ],  #  1.000
    ]
)
""")

set_meas("48-0.0004rad-5000ctl_cond-50000coupling_cond-10000ps-500ps-9ctl-9coupling-3_1_winding-1e9-drive", """
Piecewise(
    [
        [ -16845,  -3852 ],  # -1.000
        [ -14379,  -2012 ],  # -0.350
        [ -10136,   1101 ],  # -0.300
        [  -5768,   3505 ],  # -0.250
        [  -1389,   5290 ],  # -0.200
        [   -517,   5596 ],  # -0.190
        [    352,   5890 ],  # -0.180
        [   1219,   6176 ],  # -0.170
        [   2085,   6455 ],  # -0.160
        [   2947,   6730 ],  # -0.150
        [   3796,   6999 ],  # -0.140
        [   4636,   7263 ],  # -0.130
        [   5471,   7524 ],  # -0.120
        [   6294,   7781 ],  # -0.110
        [   7077,   8026 ],  # -0.100
        [   7840,   8264 ],  # -0.090
        [   8565,   8491 ],  # -0.080
        [   9217,   8696 ],  # -0.070
        [   9792,   8878 ],  # -0.060
        [  10163,   8994 ],  # -0.050
        [  10267,   9026 ],  # -0.040
        [  10356,   9053 ],  # -0.030
        [  10440,   9079 ],  # -0.020
        [  10523,   9104 ],  # -0.010
        [  10603,   9129 ],  #  0.000
        [  10982,   9244 ],  #  0.050
        [  11124,   9287 ],  #  0.070
        [  11326,   9348 ],  #  0.100
        [  11455,   9387 ],  #  0.120
        [  11934,   9530 ],  #  0.200
        [  13718,  10061 ],  #  0.350
        [  16758,  10948 ],  #  1.000
    ]
)
""")

set_meas("48-0.0004rad-500ctl_cond-20000coupling_cond-2000ps-100ps-6ctl-6coupling-5_1_winding-2e10-drive", """
Piecewise(
    [
        [ -18414,  11031 ],  # -1.000
        [ -18111,  11626 ],  # -0.300
        [ -18048,  11747 ],  # -0.250
        [ -17962,  11929 ],  # -0.200
        [ -17937,  11980 ],  # -0.190
        [ -17905,  12038 ],  # -0.180
        [ -17847,  12138 ],  # -0.170
        [ -17743,  12312 ],  # -0.160
        [ -17486,  12730 ],  # -0.150
        [ -17062,  13404 ],  # -0.140
        [ -16465,  14325 ],  # -0.130
        [ -15561,  15520 ],  # -0.120
        [ -13589,  16528 ],  # -0.110
        [ -11006,  16749 ],  # -0.100
        [  -8330,  16800 ],  # -0.090
        [  -5637,  16837 ],  # -0.080
        [  -2938,  16844 ],  # -0.070
        [   -243,  16846 ],  # -0.060
        [   2441,  16844 ],  # -0.050
        [   5074,  16843 ],  # -0.040
        [   7375,  16844 ],  # -0.030
        [   8177,  16844 ],  # -0.020
        [   8671,  16844 ],  # -0.010
        [   9037,  16844 ],  #  0.000
        [  10339,  16844 ],  #  0.050
        [  10842,  16844 ],  #  0.070
        [  14825,  16845 ],  #  0.120
        [  17592,  16842 ],  #  1.000
    ]
)
""")

set_meas("50-0.0004rad-500ctl_cond-10000coupling_cond-1000ps-50ps-5ctl-5coupling-3_1_winding-1e10-drive", """
Piecewise(
    [
        [ -17425,   3923 ],  # -1.000, M1=-12660
        [ -17001,   4090 ],  # -0.350, M1=-12447
        [ -16429,   4261 ],  # -0.300, M1=-12228
        [ -13120,   5011 ],  # -0.250, M1=-11235
        [  -9690,   5770 ],  # -0.220, M1=-10232
        [  -7385,   6255 ],  # -0.200, M1=-9575
        [  -1872,   7376 ],  # -0.150, M1=-7948
        [    802,   7871 ],  # -0.100, M1=-7164
        [   5084,   8592 ],  #  0.000, M1=-5944
        [   9278,   9209 ],  #  0.100, M1=-4806
        [  13045,   9699 ],  #  0.200, M1=-3809
        [  16741,  10142 ],  #  1.000, M1=-2780
    ]
)
""")

set_meas("50-0.0004rad-500ctl_cond-20000coupling_cond-2000ps-100ps-5ctl-5coupling-3_1_winding-1e10-drive", """
Piecewise(
    [
        [ -17349,   4910 ],  # -1.000, M1=-12580
        [   4848,   8967 ],  #  0.000, M1=-5823
        [  16730,  10287 ],  #  1.000, M1=-2565
    ]
)
""")

set_meas("50-0.0004rad-1000ctl_cond-20000coupling_cond-2000ps-100ps-5ctl-5coupling-3_1_winding-5e9-drive", """
Piecewise(
    [
        [ -17260,  -3784 ],  # -1.000, M1=-13550
        [   1294,  -1671 ],  #  0.000, M1=-8291
        [  16709,   -676 ],  #  1.000, M1=-4803
    ]
)
""")

set_meas("50-0.0004rad-1000ctl_cond-20000coupling_cond-2000ps-100ps-5ctl-5coupling-3_1_winding-1e10-drive", """
Piecewise(
    [
        [ -17427,   4252 ],  # -1.000, M1=-12305
        [ -15269,   5104 ],  # -0.200, M1=-11411
        [   -110,   9208 ],  # -0.100, M1=-6920
        [   3269,  10079 ],  # -0.070, M1=-5895
        [   6604,  10894 ],  #  0.000, M1=-4913
        [  10959,  11919 ],  #  0.100, M1=-3666
        [  16740,  13291 ],  #  1.000, M1=-1934
    ]
)
""")

set_meas("50-0.0004rad-1000ctl_cond-20000coupling_cond-2000ps-100ps-5ctl-5coupling-3_1_winding-2e10-drive", """
Piecewise(
    [
        [ -17587,  10360 ],  # -1.000, M1=-11417
        [ -17437,  10523 ],  # -0.350, M1=-11275
        [ -17413,  10547 ],  # -0.300, M1=-11253
        [ -17379,  10577 ],  # -0.250, M1=-11227
        [ -17331,  10616 ],  # -0.200, M1=-11192
        [ -17227,  10685 ],  # -0.150, M1=-11129
        [ -17045,  10777 ],  # -0.130, M1=-11046
        [ -16825,  10886 ],  # -0.120, M1=-10946
        [ -16402,  11093 ],  # -0.110, M1=-10756
        [ -15273,  11624 ],  # -0.100, M1=-10256
        [ -12326,  12652 ],  # -0.090, M1=-9237
        [  -9209,  13601 ],  # -0.080, M1=-8208
        [  -6063,  14439 ],  # -0.070, M1=-7208
        [  -2898,  15175 ],  # -0.060, M1=-6231
        [    274,  15787 ],  # -0.050, M1=-5270
        [   3417,  16146 ],  # -0.040, M1=-4324
        [   6421,  16260 ],  # -0.030, M1=-3416
        [   8677,  16324 ],  # -0.020, M1=-2721
        [   9570,  16350 ],  # -0.010, M1=-2443
        [  10229,  16369 ],  #  0.000, M1=-2240
        [  14759,  16483 ],  #  0.100, M1= -875
        [  16335,  16527 ],  #  0.200, M1= -354
        [  16771,  16545 ],  #  1.000, M1= -120
    ]
)
""")

set_meas("50-0.0004rad-2000ctl_cond-20000coupling_cond-2000ps-100ps-5ctl-5coupling-3_1_winding-5e9-drive", """
Piecewise(
    [
        [ -17402,  -1620 ],  # -1.000, M1=-13338
        [   4620,   3431 ],  #  0.000, M1=-6623
        [  16729,   5549 ],  #  1.000, M1=-3374
    ]
)
""")

set_meas("50-0.0004rad-2000ctl_cond-20000coupling_cond-2000ps-100ps-5ctl-5coupling-3_1_winding-1e10-drive", """
Piecewise(
    [
        [ -17555,   2961 ],  # -1.000, M1=-12179
        [ -17243,   3193 ],  # -0.200, M1=-11947
        [ -14966,   4108 ],  # -0.130, M1=-11033
        [ -12700,   4794 ],  # -0.120, M1=-10348
        [ -10348,   5479 ],  # -0.110, M1=-9664
        [  -7974,   6152 ],  # -0.100, M1=-8984
        [  -5593,   6821 ],  # -0.090, M1=-8300
        [  -3210,   7517 ],  # -0.080, M1=-7588
        [   -846,   8215 ],  # -0.070, M1=-6872
        [   1466,   8878 ],  # -0.060, M1=-6186
        [   7967,  10735 ],  #  0.000, M1=-4212
        [  12706,  11973 ],  #  0.100, M1=-2840
        [  16755,  13090 ],  #  1.000, M1=-1516
    ]
)
""")

set_meas("50-0.0004rad-2000ctl_cond-20000coupling_cond-2000ps-100ps-5ctl-5coupling-3_1_winding-2e10-drive", """
Piecewise(
    [
        [ -17730,   7307 ],  # -1.000, M1=-11351
        [ -17492,   7545 ],  # -0.200, M1=-11122
        [ -17311,   7667 ],  # -0.100, M1=-11002
        [ -17214,   7709 ],  # -0.090, M1=-10960
        [ -16931,   7834 ],  # -0.080, M1=-10838
        [ -16177,   8165 ],  # -0.070, M1=-10510
        [ -14947,   8667 ],  # -0.065, M1=-10009
        [ -12688,   9434 ],  # -0.060, M1=-9233
        [  -7938,  10941 ],  # -0.050, M1=-7656
        [  -3128,  12304 ],  # -0.040, M1=-6135
        [   1707,  13513 ],  # -0.030, M1=-4660
        [   6447,  14550 ],  # -0.020, M1=-3240
        [   9775,  15181 ],  # -0.010, M1=-2241
        [  10966,  15386 ],  #  0.000, M1=-1880
        [  14177,  15832 ],  #  0.050, M1= -937
        [  16107,  16041 ],  #  0.100, M1= -310
        [  16789,  16128 ],  #  1.000, M1=   65
    ]
)
""")

set_meas("50-0.0004rad-5000ctl_cond-20000coupling_cond-2000ps-100ps-5ctl-5coupling-3_1_winding-1e10-drive", """
Piecewise(
    [
        [ -17874,   -336 ],  # -1.000, M1=-12353
        [ -17709,   -216 ],  # -0.350, M1=-12199
        [ -17685,   -199 ],  # -0.300, M1=-12177
        [ -17658,   -180 ],  # -0.250, M1=-12153
        [ -17621,   -156 ],  # -0.200, M1=-12122
        [ -17563,   -121 ],  # -0.150, M1=-12078
        [ -17436,    -58 ],  # -0.100, M1=-11997
        [  -8261,   2237 ],  # -0.050, M1=-9027
        [   8890,   6091 ],  #  0.000, M1=-4131
        [  16207,   7754 ],  #  0.100, M1=-2005
        [  16491,   7861 ],  #  0.200, M1=-1869
        [  16625,   7929 ],  #  0.350, M1=-1781
        [  16811,   8050 ],  #  1.000, M1=-1626
    ]
)
""")

set_meas("50-0.0004rad-5000ctl_cond-20000coupling_cond-2000ps-100ps-5ctl-5coupling-3_1_winding-2e10-drive", """
Piecewise(
    [
        [ -18082,   3583 ],  # -1.000, M1=-11466
        [ -17810,   3787 ],  # -0.200, M1=-11214
        [ -17731,   3848 ],  # -0.120, M1=-11140
        [ -17701,   3874 ],  # -0.100, M1=-11108
        [ -17683,   3884 ],  # -0.090, M1=-11097
        [ -17630,   3909 ],  # -0.070, M1=-11064
        [ -17522,   3957 ],  # -0.050, M1=-11005
        [ -13081,   5195 ],  # -0.030, M1=-9465
        [  -3399,   7497 ],  # -0.020, M1=-6549
        [   6241,   9671 ],  # -0.010, M1=-3725
        [  10894,  10687 ],  #  0.000, M1=-2354
        [  16083,  11810 ],  #  0.050, M1= -775
        [  16377,  11907 ],  #  0.100, M1= -635
        [  16869,  12160 ],  #  1.000, M1= -258
    ]
)
""")

set_meas("50-0.0004rad-10000ctl_cond-20000coupling_cond-2000ps-100ps-5ctl-5coupling-3_1_winding-1e10-drive", """
Piecewise(
    [
        [ -18342,  -3636 ],  # -1.000, M1=-12916
        [ -18115,  -3533 ],  # -0.350, M1=-12746
        [ -18084,  -3519 ],  # -0.300, M1=-12723
        [ -18051,  -3503 ],  # -0.250, M1=-12695
        [ -18014,  -3484 ],  # -0.200, M1=-12662
        [ -17961,  -3460 ],  # -0.150, M1=-12623
        [ -17887,  -3424 ],  # -0.100, M1=-12563
        [ -17632,  -3334 ],  # -0.050, M1=-12415
        [   8753,    735 ],  #  0.000, M1=-5331
        [  16488,   1959 ],  #  0.100, M1=-3107
        [  16640,   2024 ],  #  0.200, M1=-2992
        [  16754,   2075 ],  #  0.350, M1=-2899
        [  17009,   2174 ],  #  1.000, M1=-2717
    ]
)
""")

set_meas("50-0.0004rad-10000ctl_cond-20000coupling_cond-2000ps-100ps-5ctl-5coupling-3_1_winding-2e10-drive", """
Piecewise(
    [
        [ -18603,    -61 ],  # -1.000, M1=-11774
        [  10164,   4652 ],  #  0.000, M1=-3814
        [  17128,   6016 ],  #  1.000, M1=-1449
    ]
)
""")

set_meas("50-0.0004rad-20000ctl_cond-20000coupling_cond-2000ps-100ps-5ctl-5coupling-3_1_winding-1e10-drive", """
Piecewise(
    [
        [ -19216,  -6910 ],  # -1.000, M1=-13886
        [   7843,  -4486 ],  #  0.000, M1=-7707
        [  17714,  -3467 ],  #  1.000, M1=-5023
    ]
)
""")

set_meas("50-0.0004rad-500ctl_cond-40000coupling_cond-4000ps-200ps-5ctl-5coupling-3_1_winding-2e10-drive", """
Piecewise(
    [
        [ -17439,  13243 ],  # -1.000, M1=-11394
        [ -17115,  13551 ],  # -0.200, M1=-11152
        [  -7315,  16318 ],  # -0.100, M1=-7850
        [   8831,  16749 ],  #  0.000, M1=-2973
        [  13401,  16778 ],  #  0.100, M1=-1604
        [  16756,  16787 ],  #  1.000, M1= -484
    ]
)
""")

set_meas("50-0.0004rad-5000ctl_cond-40000coupling_cond-4000ps-200ps-5ctl-5coupling-3_1_winding-1e10-drive", """
Piecewise(
    [
        [ -17679,   4637 ],  # -1.000, M1=-11683
        [ -17453,   4861 ],  # -0.200, M1=-11461
        [ -17354,   4940 ],  # -0.120, M1=-11382
        [ -17328,   4957 ],  # -0.110, M1=-11365
        [ -17284,   4980 ],  # -0.100, M1=-11342
        [ -17200,   5019 ],  # -0.090, M1=-11303
        [ -16274,   5424 ],  # -0.070, M1=-10900
        [ -13045,   6569 ],  # -0.060, M1=-9769
        [  -8261,   8048 ],  # -0.050, M1=-8297
        [  -3416,   9500 ],  # -0.040, M1=-6828
        [   1429,  10929 ],  # -0.030, M1=-5370
        [   6039,  12284 ],  # -0.020, M1=-3955
        [   8583,  13014 ],  # -0.010, M1=-3163
        [   9463,  13253 ],  #  0.000, M1=-2891
        [  12891,  14121 ],  #  0.050, M1=-1877
        [  16088,  14947 ],  #  0.100, M1= -877
        [  16761,  15235 ],  #  1.000, M1= -496
    ]
)
""")

set_meas("50-0.0004rad-5000ctl_cond-40000coupling_cond-4000ps-200ps-5ctl-5coupling-3_1_winding-2e10-drive", """
Piecewise(
    [
        [ -17870,   9437 ],  # -1.000, M1=-10755
        [ -17705,   9613 ],  # -0.350, M1=-10579
        [ -17656,   9663 ],  # -0.250, M1=-10529
        [ -17627,   9694 ],  # -0.200, M1=-10498
        [ -17585,   9733 ],  # -0.150, M1=-10459
        [ -17554,   9763 ],  # -0.120, M1=-10430
        [ -17527,   9788 ],  # -0.100, M1=-10405
        [ -17511,   9801 ],  # -0.090, M1=-10391
        [ -17464,   9834 ],  # -0.070, M1=-10357
        [ -17366,   9896 ],  # -0.050, M1=-10295
        [ -13069,  11459 ],  # -0.030, M1=-8663
        [  -3376,  14111 ],  # -0.020, M1=-5523
        [   6319,  15855 ],  # -0.010, M1=-2530
        [  11902,  16301 ],  #  0.000, M1= -813
        [  15944,  16406 ],  #  0.050, M1=  459
        [  16258,  16415 ],  #  0.100, M1=  598
        [  16799,  16436 ],  #  1.000, M1=  973
    ]
)
""")

set_meas("51-0.0004rad-500ctl_cond-20000coupling_cond-2000ps-100ps-5ctl-5coupling-3_1_winding-49999998976e0-drive", """
Piecewise(
    [
        [ -17584,  -6603 ],  # -1.000, M1=-2536
        [ -17456,  -6667 ],  # -0.350, M1=-2377
        [ -17393,  -6696 ],  # -0.300, M1=-2309
        [ -17002,  -6914 ],  # -0.250, M1=-1975
        [ -14787,  -8326 ],  # -0.200, M1= -491
        [  -8494, -11989 ],  # -0.150, M1= 3186
        [  -2111, -14126 ],  # -0.100, M1= 6269
        [   4214, -15241 ],  # -0.050, M1= 8808
        [  10359, -15908 ],  #  0.000, M1=11011
        [  14576, -16159 ],  #  0.050, M1=12412
        [  16005, -16197 ],  #  0.100, M1=12901
        [  16645, -16197 ],  #  0.200, M1=13173
        [  16909, -16185 ],  #  1.000, M1=13354
    ]
)
""")

set_meas("51-0.0004rad-1000ctl_cond-20000coupling_cond-2000ps-100ps-5ctl-5coupling-3_1_winding-2e10-drive", """
Piecewise(
    [
        [ -17462,    955 ],  # -1.000, M1=-6544
        [ -11848,  -3516 ],  # -0.350, M1=-2673
        [  -8384,  -6316 ],  # -0.300, M1= -516
        [  -4922,  -8983 ],  # -0.250, M1= 1553
        [  -1482, -11476 ],  # -0.200, M1= 3520
        [   1903, -13698 ],  # -0.150, M1= 5384
        [   5191, -15268 ],  # -0.100, M1= 7122
        [   8322, -15915 ],  # -0.050, M1= 8725
        [  11174, -16156 ],  #  0.000, M1=10122
        [  13214, -16263 ],  #  0.050, M1=11025
        [  14526, -16312 ],  #  0.100, M1=11538
        [  16070, -16357 ],  #  0.200, M1=12096
        [  16910, -16375 ],  #  1.000, M1=12480
    ]
)
""")

set_meas("51-0.0004rad-2000ctl_cond-20000coupling_cond-2000ps-100ps-5ctl-5coupling-3_1_winding-1e10-drive", """
Piecewise(
    [
        [ -17412,   7351 ],  # -1.000, M1=-9661
        [  -5507,   1206 ],  # -0.350, M1=-3475
        [  -2926,   -132 ],  # -0.300, M1=-2233
        [   -374,  -1375 ],  # -0.250, M1=-1051
        [   2145,  -2528 ],  # -0.200, M1=   93
        [   4593,  -3593 ],  # -0.150, M1= 1189
        [   6922,  -4577 ],  # -0.100, M1= 2205
        [   9061,  -5432 ],  # -0.050, M1= 3096
        [  10849,  -6119 ],  #  0.000, M1= 3826
        [  12083,  -6573 ],  #  0.050, M1= 4335
        [  13169,  -6962 ],  #  0.100, M1= 4781
        [  14949,  -7537 ],  #  0.200, M1= 5498
        [  16881,  -8016 ],  #  1.000, M1= 6376
    ]
)
""")

set_meas("51-0.0004rad-2000ctl_cond-20000coupling_cond-2000ps-100ps-5ctl-5coupling-3_1_winding-2e10-drive", """
Piecewise(
    [
        [ -17596,   4048 ],  # -1.000, M1=-7725
        [ -17162,   3838 ],  # -0.350, M1=-7354
        [ -16533,   3548 ],  # -0.300, M1=-6886
        [ -13668,   1824 ],  # -0.250, M1=-5211
        [  -8574,  -1224 ],  # -0.200, M1=-2500
        [  -3482,  -4166 ],  # -0.150, M1=  127
        [   1525,  -6940 ],  # -0.100, M1= 2643
        [   6355,  -9461 ],  # -0.050, M1= 4983
        [  10815, -11658 ],  #  0.000, M1= 7056
        [  13837, -13075 ],  #  0.050, M1= 8459
        [  15352, -13707 ],  #  0.100, M1= 9179
        [  16596, -14137 ],  #  0.200, M1= 9832
        [  16938, -14287 ],  #  1.000, M1=10139
    ]
)
""")

set_meas("51-0.0004rad-2000ctl_cond-20000coupling_cond-2000ps-100ps-5ctl-5coupling-3_1_winding-30000001024e0-drive", """
Piecewise(
    [
        [ -17699,   1454 ],  # -1.000, M1=-6236
        [ -17551,   1379 ],  # -0.350, M1=-6069
        [ -17522,   1350 ],  # -0.300, M1=-6025
        [ -17425,   1317 ],  # -0.250, M1=-5948
        [ -16719,   1033 ],  # -0.200, M1=-5451
        [ -11661,  -2116 ],  # -0.150, M1=-2565
        [  -4081,  -6784 ],  # -0.100, M1= 1528
        [   3400, -10964 ],  # -0.050, M1= 5393
        [  10508, -14107 ],  #  0.000, M1= 8877
        [  16171, -15692 ],  #  0.100, M1=11522
        [  16692, -15776 ],  #  0.200, M1=11822
        [  16959, -15828 ],  #  1.000, M1=12067
    ]
)
""")

set_meas("51-0.0004rad-5000ctl_cond-20000coupling_cond-2000ps-100ps-5ctl-5coupling-3_1_winding-30000001024e0-drive", """
Piecewise(
    [
        [ -18072,   5436 ],  # -1.000, M1=-7703
        [ -17883,   5390 ],  # -0.350, M1=-7524
        [ -17859,   5365 ],  # -0.300, M1=-7486
        [ -17831,   5381 ],  # -0.250, M1=-7471
        [ -17796,   5380 ],  # -0.200, M1=-7440
        [ -17738,   5369 ],  # -0.150, M1=-7391
        [ -16762,   5145 ],  # -0.100, M1=-6795
        [  -4256,    769 ],  # -0.050, M1=-1544
        [   9643,  -3986 ],  #  0.000, M1= 4057
        [  15927,  -5962 ],  #  0.050, M1= 6589
        [  16601,  -6128 ],  #  0.100, M1= 6942
        [  16747,  -6163 ],  #  0.200, M1= 7068
        [  17073,  -6302 ],  #  1.000, M1= 7390
    ]
)
""")

set_meas("51-0.0004rad-5000ctl_cond-20000coupling_cond-2000ps-100ps-5ctl-5coupling-3_1_winding-49999998976e0-drive", """
Piecewise(
    [
        [ -18258,   2488 ],  # -1.000, M1=-5473
        [ -18044,   2435 ],  # -0.350, M1=-5284
        [ -18016,   2439 ],  # -0.300, M1=-5261
        [ -17985,   2424 ],  # -0.250, M1=-5225
        [ -17949,   2414 ],  # -0.200, M1=-5187
        [ -17906,   2405 ],  # -0.150, M1=-5142
        [ -17849,   2419 ],  # -0.100, M1=-5102
        [ -15109,   1672 ],  # -0.050, M1=-3758
        [   8689,  -4897 ],  #  0.000, M1= 4806
        [  16381,  -7002 ],  #  0.050, M1= 7767
        [  16601,  -7024 ],  #  0.100, M1= 7888
        [  16749,  -7082 ],  #  0.200, M1= 8032
        [  17125,  -7188 ],  #  1.000, M1= 8350
    ]
)
""")

set_meas("51-0.0004rad-2000ctl_cond-20000coupling_cond-2000ps-100ps-3ctl-3coupling-5_1_winding-2e10-drive", """
Piecewise(
    [
        [ -17109,   7082 ],  # -1.000, M1=-10230
        [ -13678,   6078 ],  # -0.350, M1=-9021
        [ -10605,   5162 ],  # -0.300, M1=-8102
        [  -7507,   4334 ],  # -0.250, M1=-7225
        [  -4413,   3616 ],  # -0.200, M1=-6377
        [  -1375,   3013 ],  # -0.150, M1=-5588
        [   1551,   2490 ],  # -0.100, M1=-4862
        [   4314,   2033 ],  # -0.050, M1=-4207
        [   6844,   1617 ],  #  0.000, M1=-3613
        [   8990,   1226 ],  #  0.050, M1=-3115
        [  11075,    819 ],  #  0.100, M1=-2636
        [  15249,    -43 ],  #  0.200, M1=-1660
        [  16849,   -309 ],  #  1.000, M1=-1196
    ]
)
""")

set_meas("51-0.0004rad-2000ctl_cond-20000coupling_cond-2000ps-100ps-3ctl-3coupling-5_1_winding-49999998976e0-drive", """
Piecewise(
    [
        [ -17321,    427 ],  # -1.000, M1=-6935
        [ -17218,    384 ],  # -0.350, M1=-6841
        [ -17204,    379 ],  # -0.300, M1=-6827
        [ -17173,    385 ],  # -0.250, M1=-6817
        [ -17096,    388 ],  # -0.200, M1=-6792
        [ -16357,    152 ],  # -0.150, M1=-6472
        [  -9419,  -2287 ],  # -0.100, M1=-4280
        [  -1497,  -4809 ],  # -0.050, M1=-1873
        [   6332,  -7139 ],  #  0.000, M1=  473
        [  13416,  -9136 ],  #  0.050, M1= 2536
        [  16310,  -9904 ],  #  0.100, M1= 3368
        [  16709,  -9989 ],  #  0.200, M1= 3522
        [  16912, -10051 ],  #  1.000, M1= 3670
    ]
)
""")

set_meas("51-0.0004rad-2000ctl_cond-20000coupling_cond-2000ps-100ps-2ctl-2coupling-7_1_winding-49999998976e0-drive", """
Piecewise(
    [
        [ -16983,   3748 ],  # -1.000, M1=-9308
        [ -16836,   3708 ],  # -0.350, M1=-9223
        [ -16816,   3724 ],  # -0.300, M1=-9230
        [ -16760,   3729 ],  # -0.250, M1=-9220
        [ -15731,   3548 ],  # -0.200, M1=-9013
        [ -10811,   2528 ],  # -0.150, M1=-8083
        [  -5747,   1384 ],  # -0.100, M1=-7037
        [   -756,    388 ],  # -0.050, M1=-6076
        [   4031,   -524 ],  #  0.000, M1=-5170
        [   8551,  -1241 ],  #  0.050, M1=-4402
        [  12996,  -1856 ],  #  0.100, M1=-3697
        [  16579,  -2292 ],  #  0.200, M1=-3156
        [  16819,  -2309 ],  #  1.000, M1=-3077
    ]
)
""")

if __name__ == '__main__': main()
